/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

   ludovic.apvrille AT telecom-paristech.fr

   This software is a computer program whose purpose is to allow malware analysis

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class AlligatorClient
   * Client of Alligator to connect to the alligator daemon server
   * Creation: 15/01/2013
   * @version 1.0 15/01/2013
   * @author Ludovic APVRILLE
   * @see
   */


import java.util.*;
//import java.io.*;

import analyzer.*;
import info.*;
import myutil.*;
//import cluster.*;
//import java.awt.*;
import java.io.*;

import java.net.*;



public class AlligatorClient  {
    private static boolean allsamples = false;
    private static boolean hasNbOfSamples = false;
    private static int nbOfSamples = 1;

    private static boolean fullReport = false;
    private static boolean debug = false;
    private static boolean help = true;
    private static String guessFile;
    private static String serverHost;
    private static int port = AlligatorDaemonAnalyzer.ALLIGATOR_PORT;

    private static int nbOfAnalyzed = 0;
    private static int nbOfClean = 0;
    private static int nbOfMalware = 0;


    public static void printCopyright() {
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator: (C) Institut Mines Telecom / Telecom ParisTech, Ludovic APVRILLE, ludovic.apvrille@telecom-paristech.fr");
        System.out.println(AlligatorLanguageReader.icontext[0] + " http://perso.telecom-paristech.fr/~apvrille/alligator.html");
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator is released under a CECILL License. See http://www.cecill.info/index.en.html");
        System.out.println(AlligatorLanguageReader.icontext[0] + " Enjoy!!!\n");
    }

    public static void printUsage() {
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator client: usage");
        System.out.println("Alligator <options> -guess <file>");
        System.out.println("<options> are optional.  Might be : -debug, -help, -host <hostname or ip address>, -port <port value>; -nbofsamples <int>, -allsamples, -fullreport");
        System.out.println("-nbofsamples precise the n first samples to be analyzed in the file provided as argument. By default, only the first sample is analyzed");
        System.out.println("-allsamples analyzes all samples provided in the input file. It that option is specified, -nbofsamples is ignored");
        System.out.println("-fullreport returns a more complete report about the analysis of each sample, and not only a regular and malware score");

        //System.out.println("-learning specifies that all weight intervals will be studied in order to compute the most accurate weights to apply");
        System.out.println("<file> must be a reference to a file containing the samples to analyze");
    }

    public static boolean checkArgs(String [] args) {
        return !(args.length < 3);
    }

    public static boolean hasArg(String [] args, String argument) {
        for(int i=0; i<args.length; i++) {
            if (args[i].equals(argument)) {
                return true;
            }
        }

        return false;
    }

    public static String analyzeFileArgument(String[] args, String argument) {
        for(int i=0; i<args.length-1; i++) {
            if (args[i].equals(argument)) {
                return args[i+1];
            }
        }
        return null;
    }

    public static boolean analyseArgs(String [] args) {

        fullReport = hasArg(args, "-fullreport");
        debug = hasArg(args, "-debug");
        help = hasArg(args, "-help");
        allsamples = hasArg(args, "-allsamples");
        hasNbOfSamples = hasArg(args, "-nbofsamples");

        if (help) {
            return false;
        }

        if (debug) {
            TraceManager.devPolicy = TraceManager.TO_CONSOLE;
        } else {
            TraceManager.devPolicy = TraceManager.TO_DEVNULL;
        }

        guessFile = analyzeFileArgument(args, "-guess");
        if ((guessFile == null) && (help == false)) {
            TraceManager.addUser("Error: guess file not specified");
            return false;
        }

        serverHost = analyzeFileArgument(args, "-host");
        if (serverHost == null) {
            serverHost = "localhost";
        }

        String s = analyzeFileArgument(args, "-port");
        if (s != null) {
            //System.out.println("Has port argument");
            try {
                port = new Integer(s).intValue();
            } catch (Exception e) {
                System.out.println("Exception on port attribute: " + e.getMessage());
                port = AlligatorDaemonAnalyzer.ALLIGATOR_PORT;
            }

            if (port < 1 || port>65535) {
                port = AlligatorDaemonAnalyzer.ALLIGATOR_PORT;
            }
        }

        if (hasNbOfSamples) {
            if (allsamples) {
                System.out.println("Option \"nbofsamples\" has been ignored since \"allsamples\" is specified");
                hasNbOfSamples = false;
            } else {
                s = analyzeFileArgument(args, "-nbofsamples");
                try {
                    nbOfSamples = new Integer(s).intValue();
                } catch (Exception e) {
                    System.out.println("Exception on nbofsamples attribute: " + e.getMessage());
                }

                if (nbOfSamples < 1) {
                    nbOfSamples = 1;
                }

            }
        }


        return true;
    }


    public static void connectAnalyzeLine(String line) throws Exception {
        Socket clientSocket = new Socket(serverHost, port);
        DataOutputStream outToServer = new DataOutputStream(clientSocket.getOutputStream());
        BufferedReader inFromServer = new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));
        //TraceManager.addDev("Analyzing line:" + line);
        if (fullReport) {
            outToServer.writeBytes(AlligatorDaemonAnalyzer.FULL_REPORT_REQUEST + line + '\n');
        } else {
            outToServer.writeBytes(AlligatorDaemonAnalyzer.GUESS_REQUEST + line + '\n');
        }
        boolean done = false;
        while(!done) {
            done = analyzeServerAnswer(inFromServer);
        }
        clientSocket.close();

    }


    public static void basicAnalyzeLine(String line, DataOutputStream outToServer, BufferedReader inFromServer) throws Exception {
        //TraceManager.addDev("Analyzing line:" + line);
        outToServer.writeBytes(AlligatorDaemonAnalyzer.GUESS_REQUEST + line + '\n');
        boolean done = false;
        while(!done) {
            done = analyzeServerAnswer(inFromServer);
        }

    }


    public static void main(String[] args) {
        TraceManager.addUser("Alligator client: AnaLyzing maLware wIth partitioninG and probAbiliTy-based algORithms client");
        printCopyright();
        TraceManager.addUser("*** Your Alligator version is: " + DefaultText.getFullVersion() + " ***\n");

        if (!analyseArgs(args)) {
            printUsage();
            System.exit(-1);
        }

        int nbOfAnalyzedSamples = 0;

        String line = null;

        if (allsamples) {
            nbOfSamples = Integer.MAX_VALUE;
        }


        try {

            String contentOfGuessFile = FileUtils.loadFile(guessFile);
            Scanner scanner = new Scanner(contentOfGuessFile);

            while (scanner.hasNextLine() && (nbOfSamples>0)) {
                line = scanner.nextLine().trim();
                if (!(line.startsWith("!") || line.startsWith("#") || line.length()==0)) {
                    // Found a new line to analyze!
                    //basicAnalyzeLine(line, outToServer, inFromServer);
                    connectAnalyzeLine(line);
                    nbOfSamples --;
                    nbOfAnalyzedSamples++;
                }
                TraceManager.addDev("Remaining samples:" + nbOfSamples);
            }

            scanner.close();

            /*if (line != null) {
            //Sending line to server

            TraceManager.addDev("Connecting to server");
            Socket clientSocket = new Socket(serverHost, port);
            DataOutputStream outToServer = new DataOutputStream(clientSocket.getOutputStream());
            BufferedReader inFromServer = new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));

            outToServer.writeBytes(AlligatorDaemonAnalyzer.GUESS_REQUEST + line + '\n');

            // Waiting for response
            boolean done = false;
            while(!done) {
            done = analyzeServerAnswer(inFromServer);

            }*/

            //}
        } catch (Exception e) {
            System.out.println("ERROR: " + e.getMessage());
        }

        TraceManager.addUser("\nClient terminated, " + nbOfAnalyzedSamples + " sample(s) analyzed");
	double perc1 = ((double)((int)(((double)nbOfClean)/(nbOfAnalyzed)*10000)))/100;
	double perc2 = ((double)((int)(((double)nbOfMalware)/(nbOfAnalyzed)*10000)))/100;
	TraceManager.addUser(" Clean:" + nbOfClean + " (" + perc1 + "%)   Malware:" + nbOfMalware + " (" + perc2 + "%)\n");


        System.exit(0);
    }

    // returns true if an end-of-transaction response has been obtained
    public static boolean analyzeServerAnswer(BufferedReader br) throws Exception {
        boolean ret = false;

        String response = br.readLine();
        response = Conversion.replaceAllString(response, "@", "\n");


        TraceManager.addDev("Got from server: " + response);

        // information
        if (response.startsWith(AlligatorDaemonAnalyzer.INFO_FROM_SERVER)) {
            TraceManager.addDev("Information from server: " + response.substring(1, response.length()));

            // error
        } else if (response.startsWith(AlligatorDaemonAnalyzer.ERROR_FROM_SERVER)) {
            TraceManager.addDev("ERROR: " + response.substring(1, response.length()));
            ret = true;

        } else if (response.startsWith(AlligatorDaemonAnalyzer.DONE)) {
            System.out.println("DONE: " + response.substring(2, response.length()));
	    TraceManager.addDev("Class: " + response.substring(1, 2));
	    if (response.substring(1, 2).compareTo("1") == 0) {
		nbOfClean ++;
	    } else {
		nbOfMalware ++;
	    }
	    nbOfAnalyzed ++;

            ret = true;

        } else {
            TraceManager.addDev("Unknown response from server: " + response.substring(1, response.length()));
        }

        return ret;

    }

} // Class Main
