/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

   ludovic.apvrille AT telecom-paristech.fr

   This software is a computer program whose purpose is to allow malware analysis

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class AlligatorDaemon
   * Starts the daemon server of the alligator system
   * Creation: 14/01/2013
   * @version 1.0 14/01/2013
   * @author Ludovic APVRILLE
   * @see
   */


//import java.util.*;
//import java.io.*;

import analyzer.*;
import info.*;
import myutil.*;
import cluster.*;
//import java.awt.*;
//import java.io.*;

public class AlligatorDaemon  {
    private static boolean hasMaxPercentageOfInvalidProperties = false;
    private static int maxPercentageOfInvalidProperties = 1;
    private static boolean debug = false;
    private static boolean help = true;
    //    private static boolean report = false;
    //    private static boolean properties = false;
    private static String regularFile;
    private static String malwareFile;
    //    private static String toAnalyzeFile;
    private static String toExecuteFile;
    //    private static String reportFile;
    private static int port = AlligatorDaemonAnalyzer.ALLIGATOR_PORT;
    private static boolean precompute = false;
    private static boolean automaticRestart = false;

    private static boolean maxClusterSize = false;
    private static int maxNbOfElementsPerCluster = Integer.MAX_VALUE;

    private static boolean nocsvname = false;


    public static void printCopyright() {
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator: (C) Institut Mines Telecom / Telecom ParisTech, Ludovic APVRILLE, ludovic.apvrille@telecom-paristech.fr");
        System.out.println(AlligatorLanguageReader.icontext[0] + " http://perso.telecom-paristech.fr/~apvrille/alligator.html");
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator is released under a CECILL License. See http://www.cecill.info/index.en.html");
        System.out.println(AlligatorLanguageReader.icontext[0] + " Enjoy!!!\n");
    }

    public static void printUsage() {
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator daemon: usage");
        System.out.println("Alligator <options> -regular <file> -malware <file> -execute <file>");
        System.out.println("<options> are optional.  Might be : -debug, -help, properties, -maxclustersize <int>, -maxpercentageofinvalidproperties <int in [1..100]>, -port, -precompute");
        System.out.println("----");
        System.out.println("-execute must refer to a file in alligator language");
        System.out.println("-maxclustersize means the maximum number of elements a cluster may contain. If the option is not provided, the cluster may contain up to MAX_INT (system-dependent) elements");
        System.out.println("-maxpercentageofinvalidproperties <int in [1..100]>: cluster elements having more than the provided percentage of invalid properties are not taken into account");
        System.out.println("-port: default port is "+AlligatorDaemonAnalyzer.ALLIGATOR_PORT);
	System.out.println("-precompute: means that the daemon will make a few precomputations in order to accelerate the next classification of samples sent with the Alligator client. The daemon takes more time to start when that option is activated");
	System.out.println("-automaticrestart: means that the daemon restarts when there is an exception occuring, e.g., a broken network connection.");
        //System.out.println("-learning specifies that all weight intervals will be studied in order to compute the most accurate weights to apply");
        System.out.println("<file> must be a reference to a file");
    }

    public static boolean checkArgs(String [] args) {
        return !(args.length < 4);
    }

    public static boolean hasArg(String [] args, String argument) {
        for(int i=0; i<args.length; i++) {
            if (args[i].equals(argument)) {
                return true;
            }
        }

        return false;
    }

    public static String analyzeFileArgument(String[] args, String argument) {
        for(int i=0; i<args.length-1; i++) {
            if (args[i].equals(argument)) {
                return args[i+1];
            }
        }
        return null;
    }

    public static boolean computeMaxPercentageOfInvalidProperties(String []args) {
        try {
            for(int i=0; i<args.length-1; i++) {
                if (args[i].toLowerCase().equals("-maxpercentageofinvalidproperties")) {
                    maxPercentageOfInvalidProperties = new Integer(args[i+1]).intValue();

                    if ((maxPercentageOfInvalidProperties <= 0) || (maxPercentageOfInvalidProperties > 100)) {
                        return false;
                    }
                    return true;
                }
            }
        } catch (Exception e) {
            return false;
        }

        return false;
    }

    public static boolean computeMaxClusterSize(String []args) {
        try {
            for(int i=0; i<args.length-1; i++) {
                if (args[i].toLowerCase().equals("-maxclustersize")) {
                    maxNbOfElementsPerCluster = new Integer(args[i+1]).intValue();
                    if (maxNbOfElementsPerCluster <= 0) {
                        return false;
                    }
                    return true;
                }
            }
        } catch (Exception e) {
            return false;
        }

        return false;
    }

    public static boolean analyseArgs(String [] args) {

        hasMaxPercentageOfInvalidProperties = hasArg(args, "-maxpercentageofinvalidproperties");
        debug = hasArg(args, "-debug");
        help = hasArg(args, "-help");
	nocsvname = hasArg(args, "-nocsvname");

        if (help) {
            return false;
        }

        if (hasMaxPercentageOfInvalidProperties) {
            if (!computeMaxPercentageOfInvalidProperties(args)) {
                TraceManager.addUser("Error: max percentage of invalid properties badly specified");
                return false;
            }
        }

        if (debug) {
            TraceManager.devPolicy = TraceManager.TO_CONSOLE;
        } else {
            TraceManager.devPolicy = TraceManager.TO_DEVNULL;
        }

        regularFile = analyzeFileArgument(args, "-regular");
        if ((regularFile == null) && (help == false)) {
            TraceManager.addUser("Error: Regular file not specified");
            return false;
        }

        malwareFile = analyzeFileArgument(args, "-malware");
        if ((malwareFile == null) && (help == false)){
            TraceManager.addUser("Error: malware file not specified");
            return false;
        }

        toExecuteFile = analyzeFileArgument(args, "-execute");
        if ((toExecuteFile == null) && (help == false)){
            TraceManager.addUser("Error: script file to be execute is not specified");
            return false;
        }


        //properties = hasArg(args, "-properties");

        maxClusterSize =  hasArg(args, "-maxclustersize");
        if (maxClusterSize) {
            if (!computeMaxClusterSize(args)) {
                TraceManager.addUser("Error: max cluster size  badly specified");
                return false;
            }
        }

	precompute = hasArg(args, "-precompute");
	automaticRestart = hasArg(args, "-automaticrestart");

        String s = analyzeFileArgument(args, "-port");
        if (s != null) {
            //System.out.println("Has port argument");
            try {
                port = new Integer(s).intValue();
            } catch (Exception e) {
                System.out.println("Exception on port attribute: " + e.getMessage());
                port = AlligatorDaemonAnalyzer.ALLIGATOR_PORT;
            }

            if (port < 1 || port>65535) {
                port = AlligatorDaemonAnalyzer.ALLIGATOR_PORT;
            }
        }

        return true;
    }


    public static void main(String[] args) {
        TraceManager.addUser("Alligator Daemon: AnaLyzing maLware wIth partitioninG and probAbiliTy-based algORithms Daemon");
        printCopyright();
        TraceManager.addUser("*** Your Alligator version is: " + DefaultText.getFullVersion() + " ***\n");

        if (!analyseArgs(args)) {
            printUsage();
            System.exit(-1);
        }

        Cluster regularCluster = new Cluster("Regular");
        if (hasMaxPercentageOfInvalidProperties) {
            regularCluster.setPercentageOfMaxInvalidProperties(maxPercentageOfInvalidProperties);
        }
	regularCluster.setNoCsvName(nocsvname);
        try {
            regularCluster.makeFromFile(regularFile, maxNbOfElementsPerCluster);
        } catch (ClusterException ce) {
            TraceManager.addUser("Error on regular cluster -> " + ce.getMessage());
            System.exit(-1);
        }

        Cluster malwareCluster = new Cluster("Malware");
        if (hasMaxPercentageOfInvalidProperties) {
            malwareCluster.setPercentageOfMaxInvalidProperties(maxPercentageOfInvalidProperties);
        }
	malwareCluster.setNoCsvName(nocsvname);
        try {
            malwareCluster.makeFromFile(malwareFile, maxNbOfElementsPerCluster);
        } catch (ClusterException ce) {
            TraceManager.addUser("Error on malware cluster -> " + ce.getMessage());
            System.exit(-1);
        }


        /* Loading execution file */
        String execute = null;
        try {
            execute = FileUtils.loadFile(toExecuteFile);
        } catch (Exception e) {
            TraceManager.addUser("Error on to execution file -> " + e.getMessage());
            System.exit(-1);
        }



        //AlligatorLanguageReader executor = new AlligatorLanguageReader(regularCluster, malwareCluster, toAnalyzeCluster);
        //if (!(executor.execute(execute))) {
        //       System.exit(-1);
        //}

        AlligatorDaemonAnalyzer ada = new AlligatorDaemonAnalyzer(regularCluster, malwareCluster, execute);
        ada.setPort(port);

	if (precompute) {
	    TraceManager.addDev("Precomputing");
	    ada.precompute();
	}
	
	// First start
	try {
		ada.startServer();
	    } catch (Exception e) {
		TraceManager.addDev("Server stopped because: " + e.getMessage()); 
	    }

	// Must restart the server?
	while(automaticRestart) {
	    if (automaticRestart) {
		try {
		    Thread.currentThread().sleep(1000);
		} catch (Exception e) {
		}
	    }
	    try {
		ada.startServer();
	    } catch (Exception e) {
		TraceManager.addDev("Server stopped because: " + e.getMessage()); 
	    }
	    
	}

        TraceManager.addUser("\nServer terminated");

        System.exit(0);
    }

} // Class Main
