/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

   ludovic.apvrille AT telecom-paristech.fr

   This software is a computer program whose purpose is to allow malware analysis

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class Learning
   * Starts the learning mode of alligator
   * Creation: 23/07/2012
   * @version 1.0 23/07/2012
   * @author Ludovic APVRILLE
   * @see
   */


import java.util.*;
//import java.io.*;

import analyzer.*;
import info.*;
import myutil.*;
import cluster.*;
//import java.awt.*;

public class Learning  {

    private static boolean hasMaxPercentageOfInvalidProperties = false;
    private static int maxPercentageOfInvalidProperties = 1;
    private static boolean debug = false;
    private static boolean help = true;
    private static boolean bruteForce2 = false;
    private static boolean oneOtherAverage = false;
    private static boolean twoOtherAverage = false;
    private static boolean random = false;
    private static boolean maxClusterSize = false;
    private static boolean randomAndOneMultipass = false;
    private static boolean randomAndTwoMultipass = false;
    private static boolean oneAfterOne = false;
    private static boolean dichotomic = false;
    private static boolean nocsvname = false;
    private static int randomTime;
    private static int maxNbOfPass;

    // Forget
    private static int maxForgetPercentage;
    private static boolean hasForget = false;
    private static boolean hasAgressiveForget = false;
    private static int maxNbOfForgetPass = 1;

    // Boost
    private static int maxBoostPercentage;
    private static boolean hasBoost = false;
    private static boolean hasAgressiveBoost = false;
    private static int maxNbOfBoostPass = 1;

    private static int maxNbOfElementsPerCluster = Integer.MAX_VALUE;

    private static int bestMode = 0;
    private static boolean bestSetOnPercentageOnly = false;
    private static boolean bestSetOnAverage = false;
    //    private static boolean bestSetOnSigmaOnly = false;

    //private static boolean report = false;
    private static String regularFile;
    private static String malwareFile;
    //private static String toAnalyzeFile;
    private static String toExecuteFile;
    private static String generateScriptFile;
    private static boolean generateScript;
    private static String generateScriptNoCommentFile;
    private static boolean generateScriptNoComment;

    private static boolean hasNormalizeValue = false;
    private static double normalizeValue = 0.0;

    private static String provideClean = null, provideMalware = null;

    private static boolean hasMaxNbOfSteps;
    private static int maxNbOfSteps = 1000000000;


    public static void printCopyright() {
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator: (C) Institut Mines Telecom / Telecom ParisTech, Ludovic APVRILLE, ludovic.apvrille@telecom-paristech.fr");
        System.out.println(AlligatorLanguageReader.icontext[0] + " http://perso.telecom-paristech.fr/~apvrille/alligator.html");
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator is released under a CECILL License. See http://www.cecill.info/index.en.html");
        System.out.println(AlligatorLanguageReader.icontext[0] + " Enjoy!!!\n");
    }

    public static void printUsage() {
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator learning: usage");
        System.out.println("Alligator learning <options> -regular <file> -malware <file> -execute <file>");
        System.out.println("<options> are optional.  Might be: -debug, -help, -maxclustersize <int>, -maxpercentageofinvalidproperties <int in [1..100]>,");
        System.out.println("-bruteforce2, -oneotheraverage,");
        System.out.println("-twootheraverage, -random <time of search in seconds>,");
        System.out.println("-oneafterone");
        System.out.println("-dichotomic");
        System.out.println("-randomandonemultipass <random time> <max nb of pass>, -randomandtwomultipass <random time> <max nb of pass>");
        System.out.println("-bestAll, -bestRegular, -bestMalware, -bestSetOnPercentageOnly, -bestSetOnAverage");
        System.out.println("-generatescript <file>, -generatescriptnocomment <file>");
        System.out.println("----");
        System.out.println("-maxclustersize <max nb of elements per cluster>: the maximum number of elements a cluster may contain. If the option is not provided, the cluster may contain up to MAX_INT (system-dependent) elements");
        System.out.println("-maxpercentageofinvalidproperties <int in [1..100]>: cluster elements having more than the provided percentage of invalid properties are not taken into account");
        //System.out.println("If -report is specified, then the report is saved in the corresponding file.\nOtherwise, it is displayed in the default terminal");
        System.out.println("-execute must refer to a file in alligator language");
        System.out.println("-nocsvname means that elements of csv files have no name provided in the first column");
        System.out.println("-normalizevalue <value>: maximum value in clusters used for normalization");
        System.out.println("-provideclean <file>: provides the clean cluster to generate the \"best\" script");
        System.out.println("-providemalware <file>: provides the malware cluster to generate the \"best\" script");
        System.out.println("-forget <maxpercentage> <maxnbofpass>: Forgets the elements that are the most difficult to\nclassifiy");
        System.out.println("-agressiveforget: -forget must be present to be taken into account. Continue to forget the worse elements\n even after having reached a 100% recognition rate");
        System.out.println("-boost <maxpercentage> <maxnbofpass>: Boosts the elements that are the most difficult to\nclassifiy");
        System.out.println("-agressiveboost: -boost must be present to be taken into account. Continue to boost the worse elements\n even after having reached a 100% recognition rate");
        System.out.println("-maxnbofsteps: maximum nb of steps per interval");
        //System.out.println("-learning specifies that all weight intervals will be studied in order to compute the most accurate weights to apply");
        System.out.println("<file> must be a reference to a file");

    }

    public static boolean checkArgs(String [] args) {
        return !(args.length < 6);
    }

    public static boolean hasArg(String [] args, String argument) {
        for(int i=0; i<args.length; i++) {
            if (args[i].toLowerCase().equals(argument)) {
                return true;
            }
        }

        return false;
    }

    public static String analyzeFileArgument(String[] args, String argument) {
        for(int i=0; i<args.length-1; i++) {
            if (args[i].toLowerCase().equals(argument)) {
                return args[i+1];
            }
        }
        return null;
    }

    public static boolean computeRandomTime(String []args) {
        try {
            for(int i=0; i<args.length-1; i++) {
                if (args[i].toLowerCase().equals("-random")) {
                    randomTime = new Integer(args[i+1]).intValue();
                    if (randomTime <= 0) {
                        return false;
                    }
                    return true;
                }
            }
        } catch (Exception e) {
            return false;
        }

        return false;
    }


    public static boolean computeMaxPercentageOfInvalidProperties(String []args) {
        try {
            for(int i=0; i<args.length-1; i++) {
                if (args[i].toLowerCase().equals("-maxpercentageofinvalidproperties")) {
                    maxPercentageOfInvalidProperties = new Integer(args[i+1]).intValue();

                    if ((maxPercentageOfInvalidProperties <= 0) || (maxPercentageOfInvalidProperties > 100)) {
                        return false;
                    }
                    return true;
                }
            }
        } catch (Exception e) {
            return false;
        }

        return false;
    }

    public static boolean computeMaxClusterSize(String []args) {
        try {
            for(int i=0; i<args.length-1; i++) {
                if (args[i].toLowerCase().equals("-maxclustersize")) {
                    maxNbOfElementsPerCluster = new Integer(args[i+1]).intValue();
                    if (maxNbOfElementsPerCluster <= 0) {
                        return false;
                    }
                    return true;
                }
            }
        } catch (Exception e) {
            return false;
        }

        return false;
    }

    public static boolean computeMaxNbOfSteps(String []args) {
        try {
            for(int i=0; i<args.length-1; i++) {
                if (args[i].toLowerCase().equals("-maxnbofsteps")) {
                    maxNbOfSteps = new Integer(args[i+1]).intValue();
                    if ((maxNbOfSteps <= 0) || (maxNbOfSteps >= 1000000000)) {
                        return false;
                    }
                    return true;
                }
            }
        } catch (Exception e) {
            return false;
        }

        return false;
    }

    public static boolean computeMaxNormalizeValue(String []args) {
        try {
            for(int i=0; i<args.length-1; i++) {
                if (args[i].toLowerCase().equals("-normalizevalue")) {
                    normalizeValue = new Double(args[i+1]).doubleValue();
                    if (normalizeValue <= 0) {
                        return false;
                    }
                    return true;
                }
            }
        } catch (Exception e) {
            return false;
        }

        return false;
    }

    public static boolean computeForgetArgs(String args[]) {
        try {
            for(int i=0; i<args.length-1; i++) {
                if (args[i].toLowerCase().equals("-forget")) {
                    maxForgetPercentage = new Integer(args[i+1]).intValue();
                    if ((maxForgetPercentage <= 0) || (maxForgetPercentage >= 100)) {
                        return false;
                    }
                    maxNbOfForgetPass = new Integer(args[i+2]).intValue();
                    if (maxNbOfForgetPass <= 0) {
                        return false;
                    }
                    return true;
                }
            }
        } catch (Exception e) {
            return false;
        }

        return false;
    }

    public static boolean computeBoostArgs(String args[]) {
        try {
            for(int i=0; i<args.length-1; i++) {
                if (args[i].toLowerCase().equals("-boost")) {
                    maxBoostPercentage = new Integer(args[i+1]).intValue();
                    if ((maxBoostPercentage <= 0) || (maxBoostPercentage >= 100)) {
                        return false;
                    }
                    maxNbOfBoostPass = new Integer(args[i+2]).intValue();
                    if (maxNbOfBoostPass <= 0) {
                        return false;
                    }
                    return true;
                }
            }
        } catch (Exception e) {
            return false;
        }

        return false;
    }

    public static boolean computeRandomTimeAndPass(String []args) {
        try {
            for(int i=0; i<args.length-1; i++) {
                if (args[i].toLowerCase().equals("-randomandonemultipass") || args[i].toLowerCase().equals("-randomandtwomultipass")) {
                    randomTime = new Integer(args[i+1]).intValue();
                    if (randomTime <= 0) {
                        return false;
                    }
                    maxNbOfPass = new Integer(args[i+2]).intValue();
                    if (maxNbOfPass <= 0) {
                        return false;
                    }
                    return true;
                }
            }
        } catch (Exception e) {
            return false;
        }

        return false;
    }

    public static boolean analyseArgs(String [] args) {
        hasMaxPercentageOfInvalidProperties = hasArg(args, "-maxpercentageofinvalidproperties");
        debug = hasArg(args, "-debug");
        help = hasArg(args, "-help");
        bruteForce2 = hasArg(args, "-bruteforce2");
        oneOtherAverage = hasArg(args, "-oneotheraverage");
        twoOtherAverage = hasArg(args, "-twootheraverage");
        random = hasArg(args, "-random");
        maxClusterSize =  hasArg(args, "-maxclustersize");
        randomAndOneMultipass = hasArg(args, "-randomandonemultipass");
        randomAndTwoMultipass = hasArg(args, "-randomandtwomultipass");
        oneAfterOne = hasArg(args, "-oneafterone");
        dichotomic = hasArg(args, "-dichotomic");
        bestSetOnPercentageOnly = hasArg(args, "-bestsetonpercentageonly");
        bestSetOnAverage = hasArg(args, "-bestsetonaverage");
        nocsvname = hasArg(args, "-nocsvname");
        hasForget = hasArg(args, "-forget");
        hasAgressiveForget = hasArg(args, "-agressiveforget");
        hasBoost = hasArg(args, "-boost");
        hasAgressiveBoost = hasArg(args, "-agressiveboost");
        hasMaxNbOfSteps = hasArg(args, "-maxnbofsteps");

        if (hasArg(args, "-normalizevalue")) {
            hasNormalizeValue = true;
            if (!computeMaxNormalizeValue(args)) {
                TraceManager.addUser("Error: max normalize value  badly specified");
                return false;
            }
        }

        if (hasMaxNbOfSteps) {
            if (!computeMaxNbOfSteps(args)) {
                TraceManager.addUser("Error: max nb of steps value  badly specified. Must be >0 and <1,000,000,000");
                return false;
            }
        }

        if (hasArg(args, "-bestall")) {
            bestMode = AlligatorEvaluationSet.BEST_TWO_PERCENTAGE;
            AlligatorEvaluationSet.bestMode = bestMode;
        }

        if (hasArg(args, "-bestregular")) {
            bestMode = AlligatorEvaluationSet.BEST_REGULAR_PERCENTAGE;
            AlligatorEvaluationSet.bestMode = bestMode;
        }

        if (hasArg(args, "-bestmalware")) {
            bestMode = AlligatorEvaluationSet.BEST_MALWARE_PERCENTAGE;
            AlligatorEvaluationSet.bestMode = bestMode;
        }

        if (hasMaxPercentageOfInvalidProperties) {
            if (!computeMaxPercentageOfInvalidProperties(args)) {
                TraceManager.addUser("Error: max percentage of invalid properties badly specified");
                return false;
            }
        }

        if (hasForget) {
            if (!computeForgetArgs(args)) {
                TraceManager.addUser("Error: Forget parameters badly specified");
                return false;
            }
        }

        if (hasBoost) {
            if (!computeBoostArgs(args)) {
                TraceManager.addUser("Error: Boost parameters badly specified");
                return false;
            }
        }

        if ((hasBoost) && (hasForget)) {
            TraceManager.addUser("Error: Boost and Forgetparameters badly specified");
            return false;
        }

        if (random) {
            if (!computeRandomTime(args)) {
                TraceManager.addUser("Error: random time badly specified");
                return false;
            }
        }

        if (maxClusterSize) {
            if (!computeMaxClusterSize(args)) {
                TraceManager.addUser("Error: max cluster size  badly specified");
                return false;
            }
        }

        if ((randomAndOneMultipass) || (randomAndTwoMultipass)) {
            if (!computeRandomTimeAndPass(args)) {
                TraceManager.addUser("Error: random or nb of pass badly specified");
                return false;
            }
        }

        if (help) {
            return false;
        }

        if (debug) {
            TraceManager.devPolicy = TraceManager.TO_CONSOLE;
        } else {
            TraceManager.devPolicy = TraceManager.TO_DEVNULL;
        }

        regularFile = analyzeFileArgument(args, "-regular");
        if ((regularFile == null) && (help == false)) {
            TraceManager.addUser("Error: Regular file not specified");
            return false;
        }

        malwareFile = analyzeFileArgument(args, "-malware");
        if ((malwareFile == null) && (help == false)){
            TraceManager.addUser("Error: malware file not specified");
            return false;
        }

        toExecuteFile = analyzeFileArgument(args, "-execute");
        if ((toExecuteFile == null) && (help == false)){
            TraceManager.addUser("Error: script file to be execute is not specified");
            return false;
        }

        generateScript = hasArg(args, "-generatescript");
        if (generateScript) {
            generateScriptFile = analyzeFileArgument(args, "-generatescript");
            if ((generateScriptFile == null) && (help == false)){
                TraceManager.addUser("Error: generatescript option used, but no file is specified");
                return false;
            }
        }

        generateScriptNoComment = hasArg(args, "-generatescriptnocomment");
        if (generateScriptNoComment) {
            generateScriptNoCommentFile = analyzeFileArgument(args, "-generatescriptnocomment");
            if ((generateScriptNoCommentFile == null) && (help == false)){
                TraceManager.addUser("Error: generatescriptnocomment option used, but no file is specified");
                return false;
            }
        }

        provideClean = analyzeFileArgument(args, "-provideclean");
        provideMalware = analyzeFileArgument(args, "-providemalware");

        return true;
    }


    public static void main(String[] args) {
        long time00 = System.currentTimeMillis();
        TraceManager.addUser("Alligator Learning: AnaLyzing maLware wIth partitioninG and probAbiliTy-based algORithms: Learning");
        printCopyright();
        TraceManager.addUser("*** Your Alligator version is: " + DefaultText.getFullVersion() + " ***\n");

        if (!analyseArgs(args)) {
            printUsage();
            System.exit(-1);
        }


        if (hasMaxNbOfSteps) {
            TraceManager.addDev("Setting max nb of steps to " + maxNbOfSteps);
            AlligatorScore.MAX_NB_OF_STEPS = maxNbOfSteps;
        }

        TraceManager.addDev("Making computations over Regular");
        Cluster regularCluster1 = new Cluster("Regular");

        if (hasMaxPercentageOfInvalidProperties) {
            regularCluster1.setPercentageOfMaxInvalidProperties(maxPercentageOfInvalidProperties);
        }

        if (hasNormalizeValue) {
            regularCluster1.setNormalizationValue(normalizeValue);
        }

        regularCluster1.setNoCsvName(nocsvname);

        try {
            regularCluster1.makeFromFile(regularFile, maxNbOfElementsPerCluster);
        } catch (ClusterException ce) {
            TraceManager.addUser("Error on regular cluster -> " + ce.getMessage());
            System.exit(-1);
        }

        Cluster malwareCluster1 = new Cluster("Malware");
        if (hasMaxPercentageOfInvalidProperties) {
            malwareCluster1.setPercentageOfMaxInvalidProperties(maxPercentageOfInvalidProperties);
        }
        if (hasNormalizeValue) {
            malwareCluster1.setNormalizationValue(normalizeValue);
        }
        malwareCluster1.setNoCsvName(nocsvname);
        try {
            malwareCluster1.makeFromFile(malwareFile, maxNbOfElementsPerCluster);
        } catch (ClusterException ce) {
            TraceManager.addUser("Error on malware cluster -> " + ce.getMessage());
            System.exit(-1);
        }

        Cluster regularCluster = new Cluster("Regular");
        if (hasMaxPercentageOfInvalidProperties) {
            regularCluster.setPercentageOfMaxInvalidProperties(maxPercentageOfInvalidProperties);
        }
        if (hasNormalizeValue) {
            regularCluster.setNormalizationValue(normalizeValue);
        }
        regularCluster.setNoCsvName(nocsvname);
        try {
            if (provideClean != null) {
                regularCluster.makeFromFile(provideClean, maxNbOfElementsPerCluster);
            } else {
                regularCluster.makeFromFile(regularFile, maxNbOfElementsPerCluster);
            }
        } catch (ClusterException ce) {
            TraceManager.addUser("Error on regular cluster -> " + ce.getMessage());
            System.exit(-1);
        }

        /* Loading execution file */
        String execute = null;
        try {
            execute = FileUtils.loadFile(toExecuteFile);
        } catch (Exception e) {
            TraceManager.addUser("Error on to execution file -> " + e.getMessage());
            System.exit(-1);
        }

        AlligatorLanguageReader executorRegular = new AlligatorLanguageReader(regularCluster1, malwareCluster1, regularCluster);
        executorRegular.setIgnoreReport(true);
        if (!(executorRegular.execute(execute))) {
            System.exit(-1);
        }

 
        TraceManager.addDev("Making computations over Malware");

        /*regularCluster1 = new Cluster("Regular");

        if (hasMaxPercentageOfInvalidProperties) {
            regularCluster1.setPercentageOfMaxInvalidProperties(maxPercentageOfInvalidProperties);
        }
        if (hasNormalizeValue) {
            regularCluster1.setNormalizationValue(normalizeValue);
        }
        regularCluster1.setNoCsvName(nocsvname);

        try {
            regularCluster1.makeFromFile(regularFile, maxNbOfElementsPerCluster);
        } catch (ClusterException ce) {
            TraceManager.addUser("Error on regular cluster -> " + ce.getMessage());
            System.exit(-1);
        }

        malwareCluster1 = new Cluster("Malware");
        if (hasMaxPercentageOfInvalidProperties) {
            malwareCluster1.setPercentageOfMaxInvalidProperties(maxPercentageOfInvalidProperties);
        }
        if (hasNormalizeValue) {
            malwareCluster1.setNormalizationValue(normalizeValue);
        }
        malwareCluster1.setNoCsvName(nocsvname);
        try {
            malwareCluster1.makeFromFile(malwareFile, maxNbOfElementsPerCluster);
        } catch (ClusterException ce) {
            TraceManager.addUser("Error on malware cluster -> " + ce.getMessage());
            System.exit(-1);
	    }*/

	regularCluster1.reinitPropertyRange();
	malwareCluster1.reinitPropertyRange();
	

        Cluster malwareCluster = new Cluster("Malware");
        malwareCluster.setNoCsvName(nocsvname);
        if (hasMaxPercentageOfInvalidProperties) {
            malwareCluster.setPercentageOfMaxInvalidProperties(maxPercentageOfInvalidProperties);
        }
        if (hasNormalizeValue) {
            malwareCluster.setNormalizationValue(normalizeValue);
        }
        try {
            if (provideMalware != null) {
                malwareCluster.makeFromFile(provideMalware, maxNbOfElementsPerCluster);
            } else {
                malwareCluster.makeFromFile(malwareFile, maxNbOfElementsPerCluster);
            }
            //malwareCluster.makeFromFile(malwareFile, maxNbOfElementsPerCluster);
        } catch (ClusterException ce) {
            TraceManager.addUser("Error on malware cluster -> " + ce.getMessage());
            System.exit(-1);
        }

        AlligatorLanguageReader executorMalware = new AlligatorLanguageReader(regularCluster1, malwareCluster1, malwareCluster);
        executorMalware.setIgnoreReport(true);
        if (!(executorMalware.execute(execute))) {
            System.exit(-1);
        }

        regularCluster1.purgeClusterElements();
        malwareCluster1.purgeClusterElements();
        malwareCluster.purgeClusterElements();
	regularCluster.purgeClusterElements();

        // Real learning phase
        long time0 = System.currentTimeMillis();
        AlligatorLearner al;

        al = learn(executorRegular, executorMalware);
        AlligatorEvaluationSet best = al.getBest();

        // Generate script
        String commandLine = "";
        for(int i=0; i<args.length; i++) {
            commandLine += args[i] + " ";
        }

        // Computing learning time
        long time1 = System.currentTimeMillis();
        String timeForReport = "# Overall learning time:" + (time1-time00)/((double)1000) + "s";
        String timeForUser = "Learning time:\n  Overall learning time:" + (time1-time00)/((double)1000) + "s";
	timeForReport += "\n#  Clustering time:" + (time0-time00)/((double)1000) + "s";
	timeForUser += "\n  Clustering time:" + (time0-time00)/((double)1000) + "s";
	timeForReport += "\n#  Multipliers computation time:" + (time1-time0)/((double)1000) + "s";
	timeForUser += "\n  Multipliers computation time:" + (time1-time0)/((double)1000) + "s";


        //TraceManager.addUser("Overall learning time:" + (time1-time00)/((double)1000) + "s");
        TraceManager.addUser("  Clustering time:" + (time0-time00)/((double)1000) + "s");
        TraceManager.addUser("  Multipliers computation time:" + (time1-time0)/((double)1000) + "s");

        // Normal script
        if ((best != null) && (al != null) && (generateScript)) {
            TraceManager.addUser("Generating script...");
            generateNormalScript(al, best, execute, commandLine, timeForReport);
        }

        // Script no comment
        if ((best != null) && (al != null) && (generateScriptNoComment)) {
            TraceManager.addUser("Generating no comment script...");
            generateNoCommentScript(al, best, execute, commandLine, timeForReport);
        }


        TraceManager.addUser("\nLearning successful, all done");
        TraceManager.addDev(timeForUser + "\n");

        System.exit(0);
    }

    public static AlligatorLearner learn(AlligatorLanguageReader  executorRegular, AlligatorLanguageReader executorMalware) {

        AlligatorEvaluationSet best;
        AlligatorLearner al = null;

        boolean go = true;

        int nbOfForgetPassDone = maxNbOfForgetPass;
        int nbOfBoostPassDone = maxNbOfBoostPass;

        while(go) {

            if (random) {
                TraceManager.addUser("\nRandom approach");
                al = new AlligatorLearner(executorRegular, executorMalware, bestSetOnPercentageOnly, bestSetOnAverage);
                best = al.random(randomTime);
                TraceManager.addUser("\nBest set:" + best);
                TraceManager.addUser("Multipliers to use: " + al.getMultipliers(best));

            } else if (dichotomic) {
                TraceManager.addUser("\nDichotomic brute force");
                al = new AlligatorLearner(executorRegular, executorMalware, bestSetOnPercentageOnly, bestSetOnAverage);
                best = al.dichotomicBruteForce();
                TraceManager.addUser("\nBest set:" + best);
                TraceManager.addUser("Multipliers to use: " + al.getMultipliers(best));

            } else if (oneAfterOne) {
                TraceManager.addUser("\nOne after one");
                al = new AlligatorLearner(executorRegular, executorMalware, bestSetOnPercentageOnly, bestSetOnAverage);
                best = al.oneAfterTheOther();
                TraceManager.addUser("\nBest set:" + best);
                TraceManager.addUser("Multipliers to use: " + al.getMultipliers(best));

            } else if (randomAndOneMultipass) {
                TraceManager.addUser("\nRandom and then multipass, one by one multiplier approach");
                al = new AlligatorLearner(executorRegular, executorMalware, bestSetOnPercentageOnly, bestSetOnAverage);
                best = al.random(randomTime);
                best = al.multipassFromBest(best, maxNbOfPass, false);
                TraceManager.addUser("\nBest set:" + best);
                TraceManager.addUser("Multipliers to use: " + al.getMultipliers(best));

            } else if (randomAndTwoMultipass) {
                TraceManager.addUser("\nRandom and then multipass, two multipliers at a time approach");
                al = new AlligatorLearner(executorRegular, executorMalware, bestSetOnPercentageOnly, bestSetOnAverage);
                best = al.random(randomTime);
                best = al.multipassFromBest(best, maxNbOfPass, true);
                TraceManager.addUser("\nBest set:" + best);
                TraceManager.addUser("Multipliers to use: " + al.getMultipliers(best));

            } else if (oneOtherAverage) {
                TraceManager.addUser("\nOne other average algorithm");
                al = new AlligatorLearner(executorRegular, executorMalware, bestSetOnPercentageOnly, bestSetOnAverage);
                //AlligatorEvaluationSet best = al.oneOtherAverage();
                best = al.oneOtherProgressiveMultipassAverage(false);
                /*TraceManager.addUser("\nBest set with standard average:" + best);
                  TraceManager.addUser("Percentage of recognized regular: " + best.getRegularPercentage());
                  TraceManager.addUser("Percentage of recognized malware: " + best.getMalwarePercentage());
                  TraceManager.addUser("Mutlipliers to use: " + al.getMultipliers(best));*/

                TraceManager.addUser("\n\nBest set with progressive multipass average:" + best);
                //TraceManager.addUser("Percentage of recognized regular: " + best1.getRegularPercentage());
                //TraceManager.addUser("Percentage of recognized malware: " + best1.getMalwarePercentage());
                TraceManager.addUser("Multipliers to use: " + al.getMultipliers(best));

            } else if (twoOtherAverage) {
                TraceManager.addUser("\nTwo other average algorithm");
                al = new AlligatorLearner(executorRegular, executorMalware, bestSetOnPercentageOnly, bestSetOnAverage);
                best = al.oneOtherProgressiveMultipassAverage(true);

                TraceManager.addUser("\n\nBest set with progressive multipass two-at-a-time average:" + best);
                TraceManager.addUser("Multipliers to use: " + al.getMultipliers(best));

            } else if (bruteForce2) {
                TraceManager.addUser("\nBrute forcing 2 over possible values");
                al = new AlligatorLearner(executorRegular, executorMalware, bestSetOnPercentageOnly, bestSetOnAverage);
                best = al.bruteForce2();
                TraceManager.addUser("\nBest set:" + best);
                //TraceManager.addUser("Percentage of recognized regular: " + best.getRegularPercentage());
                //TraceManager.addUser("Percentage of recognized malware: " + best.getMalwarePercentage());
                TraceManager.addUser("Multipliers to use: " + al.getMultipliers(best));

            } else {
                TraceManager.addUser("\nBrute forcing over possible values");
                al = new AlligatorLearner(executorRegular, executorMalware, bestSetOnPercentageOnly, bestSetOnAverage);
                al.bruteForce();

                best = al.getBestSet();
                TraceManager.addUser("\nBest set:" + best);
                //TraceManager.addUser("Percentage of recognized regular: " + best.getRegularPercentage());
                //TraceManager.addUser("Percentage of recognized malware: " + best.getMalwarePercentage());
                TraceManager.addUser("Multipliers to use: " + al.getMultipliers(best));
            }

            al.setBest(best);

            // Forget or boost - If none, just exit the learning phase
            if ((hasForget) && (nbOfForgetPassDone > 0)){
                nbOfForgetPassDone --;
                go = al.forget(best, maxForgetPercentage, hasAgressiveForget);
                TraceManager.addDev("\nForget remaining pass:" + nbOfForgetPassDone + " go=" + go);
            } if ((hasBoost) && (nbOfBoostPassDone > 0)) {
                nbOfBoostPassDone --;
                go = al.boost(best, maxForgetPercentage, hasAgressiveBoost);
                TraceManager.addDev("\nBoost remaining pass:" + nbOfBoostPassDone + " go=" + go);
            } else {
                go = false;
            }
        }

        return al;
    }

    public static void generateNormalScript(AlligatorLearner al, AlligatorEvaluationSet best, String execute, String commandLine, String timeReport) {
        Date date = new Date();

        // Display time and date using toString()
        StringBuffer sb = new StringBuffer("# Script generated by Alligator: " + date.toString() + "\n\n");

        sb.append("# Learning command line: " + commandLine + "\n\n");
	sb.append(timeReport +"\n\n");

        sb.append("# Results: " + best + "\n");
        sb.append("# Multipliers to use: " + al.getMultipliers(best) + "\n");
        int nbOfForgotten = al.getNbOfForgotten();
        if (nbOfForgotten > 0) {
            sb.append("\n# Forgotten elements in " + best.REGULAR + ":" +  al.getNbOfRegularForgotten());
            sb.append("\n# Forgotten elements in " + best.MALWARE + ":" +  al.getNbOfMalwareForgotten());
        }
        int nbOfBoost = al.getNbOfForgotten();
        if (nbOfBoost > 0) {
            sb.append("\n# Boost value in " + best.REGULAR + ":" +  al.getNbOfRegularBoost());
            sb.append("\n# Boost value in " + best.MALWARE + ":" +  al.getNbOfMalwareBoost());
        }
        sb.append("\n\n");




        // Reading the string, and interpreting it */
        Scanner scanner = new Scanner(execute);
        int index = 0;
        String compExpr;
        double weightr = 1, weightm = 1;
        //String newLine;

        while (scanner.hasNextLine()) {
            String line = scanner.nextLine();
            String trimmedLine = line.trim();

            TraceManager.addDev("Handling line: " + line +  " index=" + index);


            // We ignore weight setting
            if (trimmedLine.startsWith("#")) {
                // comment: recopy the line
                myAppend(sb, line+"\n");
            } else if ((trimmedLine.toLowerCase().startsWith(AlligatorLanguageReader.COMMANDS[1])) || (trimmedLine.toLowerCase().startsWith(AlligatorLanguageReader.COMMANDS[2]))) {
                myAppend(sb, "#Former weight specification: " + line + "\n");
            } else {
                if (trimmedLine.toLowerCase().startsWith(AlligatorLanguageReader.COMMANDS[3]+" ")) {
                    // compute
                    // We put the weight
                    compExpr = al.getAlligatorLanguageReaderRegular().getRegularComplexExpression(index);
                    weightr = al.getAlligatorLanguageReaderRegular().getRegularMultiplier(index, best.arrayOfValueIndexRegular[index]);

                    if (compExpr == null) {
                        myAppend(sb, "setmultiplierweight regular " + weightr + "\n");
                    } else {
                        myAppend(sb, "setcomplexweight regular " + weightr + " " + compExpr + "\n");
                    }

                    compExpr = al.getAlligatorLanguageReaderRegular().getMalwareComplexExpression(index);
                    weightm = al.getAlligatorLanguageReaderRegular().getMalwareMultiplier(index, best.arrayOfValueIndexMalware[index]);

                    if (compExpr == null) {
                        myAppend(sb, "setmultiplierweight malware " + weightm + "\n");
                    } else {
                        myAppend(sb, "setcomplexweight malware " + weightm + " " + compExpr + "\n");
                    }

                    index ++;
                }
                // We recopy the line
                myAppend(sb, line+"\n");
            }
        }

        scanner.close();

        myAppend(sb, "\n#Computing all score\ncomputeOverallScores\n");

        myAppend(sb, "\n#Printing summary of results");
        if (best.maxRegularDiff > 0) {
            myAppend(sb, "\nsetDiffValue regular " + best.maxRegularDiff);
        }
        if (best.maxMalwareDiff > 0) {
            myAppend(sb, "\nsetDiffValue malware " + best.maxMalwareDiff);
        }
        myAppend(sb, "\nprintResultSummary\n");

        try {
            FileUtils.saveFile(generateScriptFile, sb.toString());
            TraceManager.addUser("Script generated in " + generateScriptFile);
        } catch (Exception e) {
            TraceManager.addUser("Report could not be written to " + generateScriptFile + " because " + e.getMessage());
        }
    }

    public static void generateNoCommentScript(AlligatorLearner al, AlligatorEvaluationSet best, String execute, String commandLine, String timeReport) {

        Date date = new Date();

        // Display time and date using toString()
        StringBuffer sb = new StringBuffer("# Script generated by Alligator: " + date.toString() + "\n\n");
        sb.append("# Learning command line: " + commandLine + "\n\n");
	sb.append(timeReport +"\n\n");

        sb.append("# Results: " + best + "\n");
        sb.append("# Multipliers to use: " + al.getMultipliers(best) + "\n");
        int nbOfForgotten = al.getNbOfForgotten();
        if (nbOfForgotten > 0) {
            sb.append("\n# Forgotten elements in " + best.REGULAR + ":" +  al.getNbOfRegularForgotten());
            sb.append("\n# Forgotten elements in " + best.MALWARE + ":" +  al.getNbOfMalwareForgotten());
        }
        int nbOfBoost = al.getNbOfForgotten();
        if (nbOfBoost > 0) {
            sb.append("\n# Boost value in " + best.REGULAR + ":" +  al.getNbOfRegularBoost());
            sb.append("\n# Boost value in " + best.MALWARE + ":" +  al.getNbOfMalwareBoost());
        }
        sb.append("\n\n");


        // Reading the string, and interpreting it */
        Scanner scanner = new Scanner(execute);
        int index = 0;
        String compExpr;
        double weight;
        //String newLine;

        while (scanner.hasNextLine()) {
            String line = scanner.nextLine();
            String trimmedLine = line.trim();

            //TraceManager.addDev("Handling line: " + line +  " index=" + index);


            // We ignore weight setting
            if (trimmedLine.startsWith("#")) {
                // comment: recopy the line
                //myAppend(sb, line+"\n");
            } else if (trimmedLine.length() == 0) {
                // We ignore empty lines
            } else if ((trimmedLine.toLowerCase().startsWith(AlligatorLanguageReader.COMMANDS[1])) || (trimmedLine.toLowerCase().startsWith(AlligatorLanguageReader.COMMANDS[2]))) {
                //myAppend(sb, "#Former weight specification: " + line + "\n");
            } else {
                if (trimmedLine.toLowerCase().startsWith(AlligatorLanguageReader.COMMANDS[3])) {
                    // compute
                    // We put the weight
                    compExpr = al.getAlligatorLanguageReaderRegular().getRegularComplexExpression(index);
                    weight = al.getAlligatorLanguageReaderRegular().getRegularMultiplier(index, best.arrayOfValueIndexRegular[index]);

                    if (compExpr == null) {
                        myAppend(sb, "\nsetmultiplierweight regular " + weight + "\n");
                    } else {
                        myAppend(sb, "\nsetcomplexweight regular " + weight + " " + compExpr + "\n");
                    }

                    compExpr = al.getAlligatorLanguageReaderRegular().getMalwareComplexExpression(index);
                    weight = al.getAlligatorLanguageReaderRegular().getMalwareMultiplier(index, best.arrayOfValueIndexMalware[index]);

                    if (compExpr == null) {
                        myAppend(sb, "setmultiplierweight malware " + weight + "\n");
                    } else {
                        myAppend(sb, "setcomplexweight malware " + weight + " " + compExpr + "\n");
                    }

                    index ++;
                }
                // We recopy the line
                myAppend(sb, line+"\n");
            }
        }

        scanner.close();

        myAppend(sb, "\n#Computing all score\ncomputeOverallScores\n");

        myAppend(sb, "\n#Printing summary of results");
        if (best.maxRegularDiff > 0) {
            myAppend(sb, "\nsetDiffValue regular " + best.maxRegularDiff);
        }
        if (best.maxMalwareDiff > 0) {
            myAppend(sb, "\nsetDiffValue malware " + best.maxMalwareDiff);
        }
        myAppend(sb, "\nprintResultSummary\n");

        try {
            FileUtils.saveFile(generateScriptNoCommentFile, sb.toString());
            TraceManager.addUser("Script generated in " + generateScriptNoCommentFile);
        } catch (Exception e) {
            TraceManager.addUser("Report could not be written to " + generateScriptNoCommentFile + " because " + e.getMessage());
        }

    }


    public static void myAppend(StringBuffer sb, String s) {
        sb.append(s);
        //TraceManager.addDev("  -> " + s);
    }


} // Class Main
