/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

   ludovic.apvrille AT telecom-paristech.fr

   This software is a computer program whose purpose is to allow malware analysis

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class NormalizeAllPerProperty
   * Evaluate the maximum value of a set of cluster
   * Creation: 26/05/2014
   * @version 1.0 26/05/2014
   * @author Ludovic APVRILLE
   * @see
   */


//import java.util.*;
//import java.io.*;

import analyzer.*;
import info.*;
import myutil.*;
import cluster.*;
//import java.awt.*;
//import java.io.*;

public class NormalizeAllPerProperty {
    private static boolean hasMaxPercentageOfInvalidProperties = false;
    private static int maxPercentageOfInvalidProperties = 1;
    private static boolean debug = false;
    private static boolean help = true;
    //private static boolean report = false;
    //private static boolean properties = false;
    //private static String regularFile;
    //private static String malwareFile;
    private static String [] toAnalyzeFile;
    //private static String toExecuteFile;
    //private static String [] reportFile;
    private static boolean nocsvname = false;


    private static boolean maxClusterSize = false;
    private static int maxNbOfElementsPerCluster = Integer.MAX_VALUE;


    public static void printCopyright() {
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator: (C) Institut Mines Telecom / Telecom ParisTech, Ludovic APVRILLE, ludovic.apvrille@telecom-paristech.fr");
        System.out.println(AlligatorLanguageReader.icontext[0] + " http://perso.telecom-paristech.fr/~apvrille/alligator.html");
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator is released under a CECILL License. See http://www.cecill.info/index.en.html");
        System.out.println(AlligatorLanguageReader.icontext[0] + " Enjoy!!!\n");
    }

    public static void printUsage() {
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator MaxCkusterSize: usage");
        System.out.println("Alligator <options> -toanalyze <files>");
        System.out.println("<options> are optional.  Might be : -debug, -help, -report <files>, -maxclustersize <int>, -maxpercentageofinvalidproperties <int in [1..100]>");
        System.out.println("----");
        System.out.println("-maxclustersize means the maximum number of elements a cluster may contain. If the option is not provided, the cluster may contain up to MAX_INT (system-dependent) elements");
        System.out.println("-maxpercentageofinvalidproperties <int in [1..100]>: cluster elements having more than the provided percentage of invalid properties are not taken into account");
        System.out.println("-nocsvname means that elements of csv files have no name");
        System.out.println("<file> must be a reference to a file");
    }

    public static boolean checkArgs(String [] args) {
        return !(args.length < 3);
    }

    public static boolean hasArg(String [] args, String argument) {
        for(int i=0; i<args.length; i++) {
            if (args[i].equals(argument)) {
                return true;
            }
        }

        return false;
    }

    public static String analyzeFileArgument(String[] args, String argument) {
        for(int i=0; i<args.length-1; i++) {
            if (args[i].toLowerCase().equals(argument)) {
                return args[i+1];
            }
        }
        return null;
    }

    public static int getIndexOfNextArgAfter(String[] args, int startingIndex) {
        int cpt=startingIndex;
        if (startingIndex > args.length-1) {
            return startingIndex;
        }
        for(int i=startingIndex+1; i<args.length-1; i++) {
            if (args[i].startsWith("-")) {
                return i;
            }
        }
        return args.length;
    }


    public static String[] analyzeFilesArgument(String[] args, String argument) {
        String[] files;
        for(int i=0; i<args.length-1; i++) {
            if (args[i].equals(argument)) {
                int index = getIndexOfNextArgAfter(args, i);
                int nbOfFiles = index - i - 1;
                if (nbOfFiles < 1) {
                    return null;
                }
                TraceManager.addDev(nbOfFiles + " files to be analyzed");
                files = new String[nbOfFiles];
                for(int k=0; k<nbOfFiles; k++) {
                    files[k]=args[i+k+1];
                }
                return files;
            }
        }
        return null;
    }

    public static boolean computeMaxPercentageOfInvalidProperties(String []args) {
        try {
            for(int i=0; i<args.length-1; i++) {
                if (args[i].toLowerCase().equals("-maxpercentageofinvalidproperties")) {
                    maxPercentageOfInvalidProperties = new Integer(args[i+1]).intValue();

                    if ((maxPercentageOfInvalidProperties <= 0) || (maxPercentageOfInvalidProperties > 100)) {
                        return false;
                    }
                    return true;
                }
            }
        } catch (Exception e) {
            return false;
        }

        return false;
    }

    public static boolean computeMaxClusterSize(String []args) {
        try {
            for(int i=0; i<args.length-1; i++) {
                if (args[i].toLowerCase().equals("-maxclustersize")) {
                    maxNbOfElementsPerCluster = new Integer(args[i+1]).intValue();
                    if (maxNbOfElementsPerCluster <= 0) {
                        return false;
                    }
                    return true;
                }
            }
        } catch (Exception e) {
            return false;
        }

        return false;
    }

    public static boolean analyseArgs(String [] args) {

        hasMaxPercentageOfInvalidProperties = hasArg(args, "-maxpercentageofinvalidproperties");
        debug = hasArg(args, "-debug");
        help = hasArg(args, "-help");
        nocsvname = hasArg(args, "-nocsvname");

        if (help) {
            return false;
        }


        if (hasMaxPercentageOfInvalidProperties) {
            if (!computeMaxPercentageOfInvalidProperties(args)) {
                TraceManager.addUser("Error: max percentage of invalid properties badly specified");
                return false;
            }
        }

        if (debug) {
            TraceManager.devPolicy = TraceManager.TO_CONSOLE;
        } else {
            TraceManager.devPolicy = TraceManager.TO_DEVNULL;
        }


        toAnalyzeFile = analyzeFilesArgument(args, "-toanalyze");
        if ((toAnalyzeFile == null) && (help == false)){
            TraceManager.addUser("Error: files to be analyzed is not specified");
            return false;
        }


        maxClusterSize =  hasArg(args, "-maxclustersize");
        if (maxClusterSize) {
            if (!computeMaxClusterSize(args)) {
                TraceManager.addUser("Error: max cluster size  badly specified");
                return false;
            }
        }

        //TraceManager.addDev("-))))) Max cluster size=" + maxNbOfElementsPerCluster);

        /*properties = hasArg(args, "-properties");
          if (properties) {
          propertiesFile = analyzeFileArgument(args, "-properties");
          if ((propertiesFile == null) && (help == false)){
          TraceManager.addUser("Error: properties option used, but no property file is specified");
          return false;
          }
          }*/

        return true;
    }


    public static void main(String[] args) {
        TraceManager.addUser("Alligator: AnaLyzing maLware wIth partitioninG and probAbiliTy-based algORithms - MaxClusterSize");
        printCopyright();
        TraceManager.addUser("*** Your Alligator version is: " + DefaultText.getFullVersion() + " ***\n");

        if (!analyseArgs(args)) {
            printUsage();
            System.exit(-1);
        }

        TraceManager.addDev("Loading clusters");

        Cluster [] cluster = new Cluster[toAnalyzeFile.length];
        for(int i=0; i<toAnalyzeFile.length; i++) {
            TraceManager.addDev("Anayzing " + toAnalyzeFile[i]);
            cluster[i] = new Cluster(toAnalyzeFile[i]);
            if (hasMaxPercentageOfInvalidProperties) {
                cluster[i].setPercentageOfMaxInvalidProperties(maxPercentageOfInvalidProperties);
            }
            cluster[i].setNoCsvName(nocsvname);
            try {
                cluster[i].makeFromFile(toAnalyzeFile[i], maxNbOfElementsPerCluster);
            } catch (ClusterException ce) {
                TraceManager.addUser("Error on to be analyzed cluster " + toAnalyzeFile[i] + " -> " + ce.getMessage());
                System.exit(-1);
            }
        }



        // Normalization per property
        TraceManager.addDev("Normalizing");
        double max = -1;
        int k;
        double allMax = -1;
        for(int i=0; i<cluster[0].getTotalNbOfProperties(); i++) {
            max = -1;
            for(k=0; k<cluster.length; k++) {
                max = Math.max(cluster[k].getMaxForProperty(i), max);
            }
	    TraceManager.addDev("Max of prop #" + i + " = " + max);
            allMax = Math.max(allMax, max);
            for(k=0; k<cluster.length; k++) {
                cluster[k].normalizeProperties(i, max);
            }
        }

        if (allMax <= 1) {
            System.out.println("Files already normalized -> nothing to do");
            System.exit(1);
        }

        // Saving files
        String file = null;
        TraceManager.addDev("Savings csvs");
        for(k=0; k<cluster.length; k++) {
            StringBuffer sb = new StringBuffer();
            for(ClusterElement ce: cluster[k].getElements()) {
                if (!nocsvname) {
                    sb.append(ce.getName() + ",");
                }
                for(int j=0; j<ce.propertiesD.length; j++) {
                    if (ce.propertiesV[j]) {
                        sb.append(ce.propertiesD[j]);
                    } else {
                        sb.append("-");
                    }
                    if (j != (ce.propertiesD.length - 1)) {
                        sb.append(",");
                    }
                }
                sb.append("\n");
            }
            try {
                file = FileUtils.addBeforeFileExtension(toAnalyzeFile[k],"_N");
                FileUtils.saveFile(file, sb.toString());
                TraceManager.addUser("Normalized version of " + toAnalyzeFile[k] + " generated in " + file);
            } catch (Exception e) {
                TraceManager.addUser("Normalized version of " + toAnalyzeFile[k] + " could NOT be generated in " + file + " because " + e.getMessage());
            }
        }


        /*if (max != null) {
          System.out.println("\nAnalyze successful\nMax property value: " + max[0] + " in element #" + max[1] + " of property #" + max[2] + " in cluster " + toAnalyzeFile[index]);
          } else {
          System.out.println("Analyze failed (empty clusters?");
          }*/


        System.exit(0);
    }

} // Class Main
