/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

   ludovic.apvrille AT telecom-paristech.fr

   This software is a computer program whose purpose is to allow malware analysis

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class ScoreSorter
   * Sort list of results according to a given score provided as parameter
   * obtained from classification
   * Creation: 26/06/2014
   * @version 1.0 16/06/2014
   * @author Ludovic APVRILLE
   * @see
   */


//import java.util.*;
//import java.io.*;

import analyzer.*;
import classificationreport.*;
import info.*;
import myutil.*;
//import java.awt.*;
import java.io.*;

public class ScoreSorter

{
    private static boolean debug = false;
    private static boolean help = true;

    //private static boolean properties = false;
    private static String inputFile;
    private static String outputFile;

    private static boolean hasDiff = false;
    private static boolean hasAbsDiff = true;
    private static boolean hasMalware = false;
    private static boolean hasRegular = false;

    private static boolean hasReverse = false;


    public static void printCopyright() {
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator: (C) Institut Mines Telecom / Telecom ParisTech, Ludovic APVRILLE, ludovic.apvrille@telecom-paristech.fr");
        System.out.println(AlligatorLanguageReader.icontext[0] + " http://perso.telecom-paristech.fr/~apvrille/alligator.html");
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator is released under a CECILL License. See http://www.cecill.info/index.en.html");
        System.out.println(AlligatorLanguageReader.icontext[0] + " Enjoy!!!\n");
    }

    public static void printUsage() {
        System.out.println(AlligatorLanguageReader.icontext[0] + " ScoreSorter: usage");
        System.out.println("ScoreSorter <options> -input <file> -output <file>");
        System.out.println("<options> are optional.  Might be : -malware, -regular, -diff, -absdiff, -reverse, -cleantomalware, -malwaretoclean, -help, -debug");
        System.out.println("----");
        System.out.println("-malware means the malware score is used for classification, from the one with the highest malware score to the one with the lowest malware score.\n-regular classifies according to the highest regular score.\n-diff classifies according to the score difference between malware and regular (highest difference first): This is the default classification option. At last, -absdiff classifies with the highest absolute value first");
        System.out.println("-reverse means that the classification is done in the reverse order");
	 System.out.println("-cleantomalware equals -diff");
	 System.out.println("-malwareroclean equals -diff -reverse");
        System.out.println("<file> must be a reference to a file");
    }

    public static boolean checkArgs(String [] args) {
        return !(args.length < 4);
    }

    public static boolean hasArg(String [] args, String argument) {
        for(int i=0; i<args.length; i++) {
            if (args[i].equals(argument)) {
                return true;
            }
        }
        return false;
    }

    public static String analyzeFileArgument(String[] args, String argument) {
        for(int i=0; i<args.length-1; i++) {
            if (args[i].toLowerCase().equals(argument)) {
                return args[i+1];
            }
        }
        return null;
    }


    public static boolean analyseArgs(String [] args) {

        debug = hasArg(args, "-debug");
        help = hasArg(args, "-help");

        if (help) {
            return false;
        }

        if (debug) {
            TraceManager.devPolicy = TraceManager.TO_CONSOLE;
        } else {
            TraceManager.devPolicy = TraceManager.TO_DEVNULL;
        }

        inputFile = analyzeFileArgument(args, "-input");
        if ((inputFile == null) && (help == false)) {
            TraceManager.addUser("Error: input file not specified");
            return false;
        }

        outputFile = analyzeFileArgument(args, "-output");
        if ((outputFile == null) && (help == false)){
            TraceManager.addUser("Error: output file not specified");
            return false;
        }

	

	hasReverse = hasArg(args, "-reverse");

        hasDiff = hasArg(args, "-diff");
        hasAbsDiff =  hasArg(args, "-absdiff");
        hasMalware = hasArg(args, "-malware");
        hasRegular = hasArg(args, "-regular");

	if (hasArg(args, "-cleantomalware")) {
	    hasDiff = true;
	    hasReverse = false;
	}

	if (hasArg(args, "-malwaretoclean")) {
	    hasDiff = true;
	    hasReverse = true;
	}

        int cpt = 0;
        if (hasDiff) cpt ++;
        if (hasMalware) cpt ++;
        if (hasRegular) cpt ++;
        if (hasAbsDiff) cpt ++;

        if (cpt > 1) {
            TraceManager.addUser("Error: more than one classification option provided");
            return false;
        }

        // Defaut option -> hasDiff
        if (cpt == 0){
            hasDiff = true;
        }

        
	


        return true;
    }


    public static void main(String[] args) {
        long time00 = System.currentTimeMillis();
        TraceManager.addUser("Alligator: AnaLyzing maLware wIth partitioninG and probAbiliTy-based algORithms");
        printCopyright();
        TraceManager.addUser("*** Your Alligator version is: " + DefaultText.getFullVersion() + " ***\n");

        if (!analyseArgs(args)) {
            printUsage();
            System.exit(-1);
        }


        /* Loading report files */
        String report = null;
        try {
            report = FileUtils.loadFile(inputFile);
        } catch (Exception e) {
            TraceManager.addUser("Error when loading regular file -> " + e.getMessage());
            System.exit(-1);
        }
        /* Loading data into classification analyzer */
        ClassificationReportAnalyzer analyzer = new ClassificationReportAnalyzer("analyzer1");
        analyzer.loadReport(report, true);
        TraceManager.addDev("Summary:" + analyzer.getSummary());

        // Computing analyze time
        long time1 = System.currentTimeMillis();
        String timeForReport = "Classification time:" + (time1-time00)/((double)1000) + "s";

        TraceManager.addDev("Sorting");
        if (hasDiff) {
            analyzer.sortDiff(true);
        } else if (hasMalware){
            analyzer.sortMalware(true);
        } else if (hasRegular) {
            analyzer.sortRegular(true);
        } else {
            analyzer.sortAbsDiff(true);
        }

        if (!hasReverse) {
            analyzer.reverse(true);
        }



        String rep = analyzer.getStringCluster(true);

        try {
            TraceManager.addUser("\nSaving sorter report to " + outputFile);
            FileUtils.saveFile(outputFile, rep);
        } catch (FileException fe) {
            TraceManager.addUser("Report could not be saved: " + fe.getMessage());
            return;
        }


        /*if (!report) {
          TraceManager.addUser("\n" + rep);
          } else {
          // Saving report to file

          }*/

        TraceManager.addDev(rep);

        TraceManager.addUser("\nScore sorting successful - " + timeForReport);


        System.exit(0);
    }

} // Class Main
