/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

   ludovic.apvrille AT telecom-paristech.fr

   This software is a computer program whose purpose is to allow malware analysis

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class SliderFinder
   * Find the best compromise classification on reports
   * obtained from classification
   * Creation: 26/06/2014
   * @version 1.0 16/06/2014
   * @author Ludovic APVRILLE
   * @see
   */


//import java.util.*;
//import java.io.*;

import analyzer.*;
import classificationreport.*;
import info.*;
import myutil.*;
//import java.awt.*;
import java.io.*;

public class SliderFinder  {
    private static boolean debug = false;
    private static boolean help = true;
    private static boolean report = false;
    //private static boolean properties = false;
    private static String regularFile;
    private static String malwareFile;
    private static String reportFile;

    private static boolean maxClusterSize = false;
    private static int maxNbOfElementsPerCluster = Integer.MAX_VALUE;


    public static void printCopyright() {
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator: (C) Institut Mines Telecom / Telecom ParisTech, Ludovic APVRILLE, ludovic.apvrille@telecom-paristech.fr");
        System.out.println(AlligatorLanguageReader.icontext[0] + " http://perso.telecom-paristech.fr/~apvrille/alligator.html");
        System.out.println(AlligatorLanguageReader.icontext[0] + " Alligator is released under a CECILL License. See http://www.cecill.info/index.en.html");
        System.out.println(AlligatorLanguageReader.icontext[0] + " Enjoy!!!\n");
    }

    public static void printUsage() {
        System.out.println(AlligatorLanguageReader.icontext[0] + " SliderFinder: usage");
        System.out.println("Alligator <options> -regular <file> -malware <file>");
        System.out.println("<options> are optional.  Might be : -debug, -help, -report <file>, -maxclustersize <int>");
        System.out.println("----");
        System.out.println("If -report is specified, then the report is saved in the corresponding file.\nOtherwise, it is displayed in the default terminal");
        System.out.println("-maxclustersize means the maximum number of elements a cluster may contain. If the option is not provided, the cluster may contain up to MAX_INT (system-dependent) elements");
        System.out.println("<file> must be a reference to a file");
    }

    public static boolean checkArgs(String [] args) {
        return !(args.length < 4);
    }

    public static boolean hasArg(String [] args, String argument) {
        for(int i=0; i<args.length; i++) {
            if (args[i].equals(argument)) {
                return true;
            }
        }
        return false;
    }

    public static String analyzeFileArgument(String[] args, String argument) {
        for(int i=0; i<args.length-1; i++) {
            if (args[i].toLowerCase().equals(argument)) {
                return args[i+1];
            }
        }
        return null;
    }


    public static boolean computeMaxClusterSize(String []args) {
        try {
            for(int i=0; i<args.length-1; i++) {
                if (args[i].toLowerCase().equals("-maxclustersize")) {
                    maxNbOfElementsPerCluster = new Integer(args[i+1]).intValue();
                    if (maxNbOfElementsPerCluster <= 0) {
                        return false;
                    }
                    return true;
                }
            }
        } catch (Exception e) {
            return false;
        }

        return false;
    }


    public static boolean analyseArgs(String [] args) {

        debug = hasArg(args, "-debug");
        help = hasArg(args, "-help");

        if (help) {
            return false;
        }

        if (debug) {
            TraceManager.devPolicy = TraceManager.TO_CONSOLE;
        } else {
            TraceManager.devPolicy = TraceManager.TO_DEVNULL;
        }

        regularFile = analyzeFileArgument(args, "-regular");
        if ((regularFile == null) && (help == false)) {
            TraceManager.addUser("Error: Regular file not specified");
            return false;
        }

        malwareFile = analyzeFileArgument(args, "-malware");
        if ((malwareFile == null) && (help == false)){
            TraceManager.addUser("Error: malware file not specified");
            return false;
        }

        report = hasArg(args, "-report");
        if (report) {
            reportFile = analyzeFileArgument(args, "-report");
            if ((reportFile == null) && (help == false)){
                TraceManager.addUser("Error: report option used, but no reporting file is specified");
                return false;
            }
        }


        maxClusterSize =  hasArg(args, "-maxclustersize");
        if (maxClusterSize) {
            if (!computeMaxClusterSize(args)) {
                TraceManager.addUser("Error: max cluster size  badly specified");
                return false;
            }
        }



        return true;
    }


    public static void main(String[] args) {
        long time00 = System.currentTimeMillis();
        TraceManager.addUser("Alligator: AnaLyzing maLware wIth partitioninG and probAbiliTy-based algORithms");
        printCopyright();
        TraceManager.addUser("*** Your Alligator version is: " + DefaultText.getFullVersion() + " ***\n");

        if (!analyseArgs(args)) {
            printUsage();
            System.exit(-1);
        }

        


        /* Loading report files */
        String reportRegular = null;
        try {
            reportRegular = FileUtils.loadFile(regularFile);
        } catch (Exception e) {
            TraceManager.addUser("Error when loading regular file -> " + e.getMessage());
            System.exit(-1);
        }

	String reportMalware = null;
        try {
            reportMalware = FileUtils.loadFile(malwareFile);
        } catch (Exception e) {
            TraceManager.addUser("Error when loading malware file -> " + e.getMessage());
            System.exit(-1);
        }

	

	/* Loading data into classification analyzer */
	ClassificationReportAnalyzer analyzer = new ClassificationReportAnalyzer("analyzer1");
	analyzer.loadReport(reportRegular, true);
	analyzer.loadReport(reportMalware, false);
	TraceManager.addDev("Summary:" + analyzer.getSummary());

        // Computing analyze time
        long time1 = System.currentTimeMillis();
        String timeForReport = "Classification time:" + (time1-time00)/((double)1000) + "s";

	TraceManager.addDev("Computing slider points");
	analyzer.performSliderAnalysis();



        String rep = "";
        /*if (hasTest) {
          TraceManager.addDev("Making test report");
          rep = executor.getAllReports(size2, toAnalyzeFile[0], toAnalyzeFile[1], timeForReport);
          } else {
          rep = executor.getAllReports(timeForReport);
          }*/

        if (!report) {
            TraceManager.addUser("\n" + rep);
        } else {
            // Saving report to file
            try {
                TraceManager.addUser("\nSaving report in " + reportFile);
                FileUtils.saveFile(reportFile, rep);
            } catch (FileException fe) {
                TraceManager.addUser("Report could not be saved: " + fe.getMessage());
                return;
            }
        }

        TraceManager.addUser("\nSliderFinder successful - " + timeForReport);


        System.exit(0);
    }

} // Class Main
