/**Copyright or (C) or Copr. GET / ENST, Telecom-Paris, Ludovic Apvrille

   ludovic.apvrille AT enst.fr

   This software is a computer program whose purpose is to allow the 
   edition of TURTLE analysis, design and deployment diagrams, to 
   allow the generation of RT-LOTOS or Java code from this diagram, 
   and at last to allow the analysis of formal validation traces 
   obtained from external tools, e.g. RTL from LAAS-CNRS and CADP 
   from INRIA Rhone-Alpes.

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use, 
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info". 

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability. 

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or 
   data to be ensured and,  more generally, to use and operate it in the 
   same conditions as regards security. 

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class AlligatorDaemonAnalyzer
   * Creation: 14/01/2013
   * version 1.0 14/01/2013
   * @author Ludovic APVRILLE
   * @see 
   */

package analyzer;

import myutil.*;

import cluster.*;

import java.util.*;
import java.io.*;
import java.net.*;


public class AlligatorDaemonAnalyzer {
    public final static int ALLIGATOR_PORT = 7078;
	
    public final static String GUESS_REQUEST = "g";
    public final static String EXIT_SERVER = "s";
    public final static String INFO_FROM_SERVER = "i";
    public final static String UNKNOWN_REQUEST = "u";
    public final static String ERROR_FROM_SERVER = "e";
    public final static String DONE = "d";
    public final static String FULL_REPORT_REQUEST = "f";

    private AlligatorLanguageReader alr;
    private Cluster regularCluster;
    private Cluster malwareCluster;
	
    private String execute;
	
    private boolean connectionActive;
	
    private DataOutputStream outToClient;
	
    private int port;
	
    private boolean fullReport = false;
	
    // Saving elements of guess cluster
    //private Vector<MasterEpsilonClusters> masterEpsilonClusters;
	
	
	
    public AlligatorDaemonAnalyzer(Cluster _regularCluster, Cluster _malwareCluster, String _execute) {
	regularCluster = _regularCluster;
	malwareCluster = _malwareCluster;
	execute = _execute;
	//masterEpsilonClusters = null;
	port = ALLIGATOR_PORT;
    }
	
    public void setPort(int _port) {
	port = _port;
    }
	
    public void startServer() {
	//String response;
		
	connectionActive = true;
	try {
	    TraceManager.addDev("Getting localhost IP address");
	    byte[] addr = new byte[4];
	    addr[0] = 127;
	    addr[1] = 0;
	    addr[2] = 0;
	    addr[3] = 1;
	    //InetAddress ip = InetAddress.getByAdressLocalHost();
	    InetAddress ip = InetAddress.getByAddress(addr);
	    TraceManager.addDev("Creating socket");
	    ServerSocket alligatorSocket = new ServerSocket(port);
			
	    while(connectionActive) {
		TraceManager.addDev("Waiting for connection on " + ip.getHostAddress() + ":" + port);
		Socket connectionSocket = alligatorSocket.accept();
		BufferedReader inFromClient = new BufferedReader(new InputStreamReader(connectionSocket.getInputStream()));
		outToClient = new DataOutputStream(connectionSocket.getOutputStream());
		String clientSentence = inFromClient.readLine();
		TraceManager.addDev("Received: >" + clientSentence.substring(0, Math.min(50, clientSentence.length())) + "<");
		clientSentence = clientSentence.trim();
				
		//response = null;
				
		// Guess sample?
		if (clientSentence.startsWith(GUESS_REQUEST)) {
		    TraceManager.addDev("Handling sample");
		    handleSample(clientSentence.substring(1, clientSentence.length()));
					
		} else if (clientSentence.startsWith(FULL_REPORT_REQUEST)) {
		    fullReport = true;
		    TraceManager.addDev("Handling sample with full report");
		    handleSample(clientSentence.substring(1, clientSentence.length()));
					
		    // Stop server?
		} else if (clientSentence.startsWith(EXIT_SERVER)) {
		    TraceManager.addDev("Stopping server");
		    sendResponse(INFO_FROM_SERVER + "stopping server");
		    connectionActive = false;
					
		    // Unknown request
		} else {
		    TraceManager.addDev("Unknown request");
		    sendResponse(UNKNOWN_REQUEST + "request " + clientSentence + " is unkown");
		}
					
	    }
	    alligatorSocket.close();
        } catch (Exception e) {
	    TraceManager.addDev("Exception in server: " + e.getMessage());
        }	
		
    }

    public void precompute() {
	Cluster guessCluster = new Cluster("Guess");
	guessCluster.setNoCsvName(regularCluster.getNoCsvName());
	//guessCluster.canPurgeEpsilonClusterMemory(false);
		
	// Putting old state
	//guessCluster.setMasterEpsilonClusters(masterEpsilonClusters);
		
	// Setting property names
	guessCluster.initProperties(regularCluster.getTotalNbOfProperties());
	for(int i=0; i<guessCluster.getTotalNbOfProperties(); i++) {
	    guessCluster.setPropertyName(i, regularCluster.getPropertyName(i));
	}

	// Putting zero element
	ClusterElement ce = ClusterElement.makeZeroClusterElement("zero", regularCluster.getTotalNbOfProperties());
	
	try {
	    guessCluster.addSample(ce);
	} catch (Exception e) {
	    TraceManager.addDev("Precomputing failed: Guess sample badly formatted: " + e.getMessage());
	    return;
	}

	alr = new AlligatorLanguageReader(regularCluster, malwareCluster, guessCluster);
	try {
	    alr.advancedNormalized();
	    TraceManager.addDev("Executing the precomputing...");
	    alr.execute(execute);
	    // Save Master Epsilon clusters ....
	} catch (Exception e) {
	    TraceManager.addDev("Precomputing failed: " + e.getMessage());
	    e.printStackTrace();
	}
		
	// Saving state
	//masterEpsilonClusters = guessCluster.getMasterEpsilonClusters();
	
    }
	
	
    private void handleSample(String sample) throws Exception{
	sendInfoHandlingSample(sample);
	Cluster guessCluster = new Cluster("Guess");
	guessCluster.setNoCsvName(regularCluster.getNoCsvName());
	//guessCluster.canPurgeEpsilonClusterMemory(false);
		
	// Putting old state
	//guessCluster.setMasterEpsilonClusters(masterEpsilonClusters);
		
	// Setting property names
	guessCluster.initProperties(regularCluster.getTotalNbOfProperties());
	for(int i=0; i<guessCluster.getTotalNbOfProperties(); i++) {
	    guessCluster.setPropertyName(i, regularCluster.getPropertyName(i));
	}
		
	// Adding sample to guess cluster
	try {
	    guessCluster.addSampleFromString(sample);
	    guessCluster.checkForPropertiesComputed();
	} catch (Exception e) {
	    sendError("Guess sample badly formatted: " + e.getMessage());
	    return;
	}
		
	if (guessCluster.getElements().size() == 0) {
	    sendError("Guess sample has been removed because too many of its properties are invalid");
	    return;
	}
		
		
	alr = new AlligatorLanguageReader(regularCluster, malwareCluster, guessCluster);
	try {
	    alr.advancedNormalized();
	    alr.execute(execute);
	    // Save Master Epsilon clusters ....
	} catch (Exception e) {
	    TraceManager.addDev("Exception during sample analysis: " + e.getMessage());
	    e.printStackTrace();
	    sendError("Error during sample analysis:" + e.getMessage());
	    return;
	}
		
	// Saving state
	//masterEpsilonClusters = guessCluster.getMasterEpsilonClusters();
	//TraceManager.addDev("\nPrinting saved epsilon clusters:");
	//for(MasterEpsilonClusters mec: masterEpsilonClusters) {
	//	TraceManager.addDev("" + mec.toString());
	//}
		
	if (fullReport) {
	    String s = alr.getFullReportForSample(0);
	    TraceManager.addDev("Getting full report for sample:" + s + " and sending it back to the client");
	    sendAllDone(s);
	    TraceManager.addDev("Full report sent back to the client");
	} else {
	    TraceManager.addDev("Sending all done back to client");
	    sendAllDone(alr.getReportForSample(0));
	    TraceManager.addDev("All done sent back toclient");
	}
		
    }
	
    private void sendAllDone(String s) throws Exception {
	sendResponse(DONE + s);
    }
	
    private void sendError(String error) throws Exception{
	if (error.length() > 100) {
	    String tmp1 = error.substring(0, 50);
	    String tmp2 = error.substring(error.length()-50, error.length());
	    error = tmp1 + "..." + tmp2;
	}
	sendResponse(ERROR_FROM_SERVER+ error);
    }
	
    private void sendInfoHandlingSample(String sample) throws Exception{
	sendResponse(INFO_FROM_SERVER+ "handling sample: >" + sample.substring(0, Math.min(50, sample.length())) + "<");
    }
	
    private void sendResponse(String response) throws Exception {
	response = Conversion.replaceAllString(response, "\n", "@");
	outToClient.writeBytes(response + "\n");
    }
	
    
} // Class 	
