/**Copyright or (C) or Copr. GET / ENST, Telecom-Paris, Ludovic Apvrille
   Copyright or (C) or Copr. FORTINET, Axelle Apvrille

   ludovic.apvrille AT enst.fr
   aapvrille AT fortinet.com

   This software is a computer program whose purpose is to allow the
   edition of TURTLE analysis, design and deployment diagrams, to
   allow the generation of RT-LOTOS or Java code from this diagram,
   and at last to allow the analysis of formal validation traces
   obtained from external tools, e.g. RTL from LAAS-CNRS and CADP
   from INRIA Rhone-Alpes.

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class AlligatorEvaluationSet
   * Creation: 24/07/2012
   * version 1.0 24/07/2012
   * @author Ludovic APVRILLE, Axelle APVRILLE
   * @see
   */

package analyzer;

import myutil.*;


public class AlligatorEvaluationSet {

    public static String REGULAR = "Regular";
    public static String MALWARE = "Malware";


    public static final int BEST_TWO_PERCENTAGE = 0;
    public static final int BEST_REGULAR_PERCENTAGE = 1;
    public static final int BEST_MALWARE_PERCENTAGE = 2;
    public static final int DEFAULT_MIN_PERCENTAGE = 0;
    public static final int DEFAULT_WEIGHT_PERCENTAGE = 1;

    public static int bestMode = BEST_TWO_PERCENTAGE;

    public int [] arrayOfValueIndexRegular; // Values of multipliers to use in algorithms
    public int [] arrayOfValueIndexMalware; // Values of multipliers to use in algorithms
    public int nbOfRegularRecognized;
    public int nbOfRegularFailed;
    public int nbOfMalwareRecognized;
    public int nbOfMalwareFailed;

    public double maxRegularDiff;
    public double maxMalwareDiff;

    public double average;
    //public double sigma;
    public double minDiff; // min difference of score for rightly recognized elements
    public double maxDiff; // max difference of score for badly recognized elements

    public static int regularMin = DEFAULT_MIN_PERCENTAGE;
    public static int malwareMin = DEFAULT_MIN_PERCENTAGE;

    public static int regularPercentageWeight = DEFAULT_WEIGHT_PERCENTAGE;
    public static int malwarePercentageWeight = DEFAULT_WEIGHT_PERCENTAGE;

    public AlligatorEvaluationSet(int nbOfValues) {
        arrayOfValueIndexRegular = new int[nbOfValues];
        arrayOfValueIndexMalware = new int[nbOfValues];
    }

    public AlligatorEvaluationSet(AlligatorTreeNode node, int depth) {
        arrayOfValueIndexRegular = new int[depth/2];
        arrayOfValueIndexMalware = new int[depth/2];


        for(int i=(depth/2)-1; i>=0; i--) {
            arrayOfValueIndexMalware[i] = node.getMyIndex();
            node = node.getFather();
        }

        for(int i=(depth/2)-1; i>=0; i--) {
            arrayOfValueIndexRegular[i] = node.getMyIndex();
            node = node.getFather();
        }
    }

    public String toString() {
        String s = "Multipliers (" + REGULAR + " then " + MALWARE + "):";
        for(int i=0; i<arrayOfValueIndexRegular.length; i++) {
            s += arrayOfValueIndexRegular[i] + " ";
        }
        for(int i=0; i<arrayOfValueIndexMalware.length; i++) {
            s += arrayOfValueIndexMalware[i] + " ";
        }

        s += " " + REGULAR + " recognized:" + nbOfRegularRecognized;
        s += " " + REGULAR + " failed:" + nbOfRegularFailed;
        s += " " + MALWARE + " recognized:" + nbOfMalwareRecognized;
        s += " " + MALWARE + " failed:" + nbOfMalwareFailed;

        s+= "  Percentage per cluster: " + REGULAR + ":" + getRegularPercentage() + "% " + MALWARE + ": " + getMalwarePercentage() + "%";
        s+= "  Average identification percentage: " + ((int)(10*(getRegularPercentage() + getMalwarePercentage())/2)/10.0) + "%";

	//        s+= "  Standard deviation between " + MALWARE + " and " + REGULAR + " scores:" + Math.sqrt(sigma);
        s+= "  Average between " + MALWARE + " and " + REGULAR + " scores:" + average;
        s+= "  Min difference between " +  MALWARE + " and " + REGULAR + " scores (rightly classified):" + minDiff;
	s+= "  Max difference between " +  MALWARE + " and " + REGULAR + " scores (badly classified):" + maxDiff;
	s+= "  Difference between Min and Max:" + (minDiff - maxDiff);


        if (!respectMinRegularPercentage())  {
            s+= " [WARNING: this configuration does NOT respect specified minimum percentage of " + REGULAR + " : " + getRegularPercentage()  + "%<" + regularMin + "%]";
        }

        if (!respectMinMalwarePercentage())  {
            s+= " [WARNING: this configuration does NOT respect specified minimum percentage of " + MALWARE + " : " + getMalwarePercentage()  + "%<" + malwareMin + "%]";
        }

        return s;

    }

    public int getIndex(int index) {
        if (index < arrayOfValueIndexRegular.length) {
            return arrayOfValueIndexRegular[index];
        }

        return arrayOfValueIndexMalware[index-arrayOfValueIndexRegular.length];
    }

    public double getRegularPercentage() {
        return 100.0 * nbOfRegularRecognized / (nbOfRegularRecognized + nbOfRegularFailed);
    }

    public double getMalwarePercentage() {
        return 100.0 * nbOfMalwareRecognized / (nbOfMalwareRecognized + nbOfMalwareFailed);
    }

    public double getOverallPercentage() {
        return (regularPercentageWeight*getRegularPercentage() + malwarePercentageWeight*getMalwarePercentage())/(regularPercentageWeight + malwarePercentageWeight);
    }

    public boolean respectMinPercentage() {
        return (getRegularPercentage() >= regularMin) && (getMalwarePercentage() > malwareMin);
    }

    public boolean respectMinRegularPercentage() {
        return (getRegularPercentage() >= regularMin);
    }

    public boolean respectMinMalwarePercentage() {
        return (getMalwarePercentage() >= malwareMin);
    }

    public double getBestPercentage() {
        switch(bestMode) {
        case BEST_TWO_PERCENTAGE:
            return getOverallPercentage();
        case BEST_REGULAR_PERCENTAGE:
            return getRegularPercentage();
        case BEST_MALWARE_PERCENTAGE:
            return getMalwarePercentage();
        }

        return getOverallPercentage();
    }

    // Returns the best if best
    // Otherwise a new one ...
    public AlligatorEvaluationSet getBest(AlligatorEvaluationSet currentBest, boolean _bestSetOnPercentageOnly, boolean _bestSetOnAverage) {
	//TraceManager.addDev("Get Best 0");
        if (currentBest == null) {
            return this.fullClone();
        }

        
        double totalPercentage = getBestPercentage();
        double bestPercentage = currentBest.getBestPercentage();

	//TraceManager.addDev("Get Best 1");
        if (totalPercentage > bestPercentage) {
	    //TraceManager.addDev("Get best 2");
            if ( (getRegularPercentage() >= regularMin) && (getMalwarePercentage() >= malwareMin) ) {
                return this.fullClone();
            } else {
                if ((currentBest.getRegularPercentage() <= regularMin) || (currentBest.getMalwarePercentage() <= malwareMin)) {
                    return this.fullClone();
                }
            }
        } else if ((totalPercentage == bestPercentage) && (!_bestSetOnPercentageOnly)) {
            boolean respectBothMin = (getRegularPercentage() >= regularMin) && (getMalwarePercentage() >= malwareMin);
            boolean otherRespectBothMin = (currentBest.getRegularPercentage() >= regularMin) && (currentBest.getMalwarePercentage() >= malwareMin);


            if (respectBothMin && !otherRespectBothMin) {
                return this.fullClone();
            }

            if (otherRespectBothMin && !respectBothMin) {
                return currentBest;
            }


            // So, the two are in the same situation with respecting min percentage
            // -> Use standard deviation between regular and malware score in each sample
            // so as to select the best of the two
            if (_bestSetOnAverage) {
                if (average > currentBest.average) {
                    return this.fullClone();
                }
            } else {
                /*if (sigma > currentBest.sigma) {
                  return this.fullClone();
                  }*/
                if (minDiff > currentBest.minDiff) {
                    return this.fullClone();
                }
            }



        }

        return currentBest;
    }

    public AlligatorEvaluationSet getBestNoClone(AlligatorEvaluationSet currentBest, boolean _bestSetOnPercentageOnly, boolean _bestSetOnAverage) {
        if (currentBest == null) {
            return this;
        }

	//TraceManager.addDev("Get Best 0-n");

        /*if (_bestSetOnSigmaOnly) {
            if (sigma > currentBest.sigma) {
                return this;
            } else {
                return currentBest;
            }
	    }*/

        double totalPercentage = getBestPercentage();
        double bestPercentage = currentBest.getBestPercentage();

        if (totalPercentage > bestPercentage &&
            getRegularPercentage() >= regularMin &&
            getMalwarePercentage() >= malwareMin) {
            return this;
        }

	//TraceManager.addDev("Get Best 1-n");

        if (totalPercentage > bestPercentage) {
	    //TraceManager.addDev("Get Best 2-n");
            if ( (getRegularPercentage() >= regularMin) && (getMalwarePercentage() >= malwareMin) ) {
                return this;
            } else {
		TraceManager.addDev("Does not respect min percentage");
                if ((currentBest.getRegularPercentage() >= regularMin) || (currentBest.getMalwarePercentage() >= malwareMin)) {
                    return currentBest;
                }
		return this;
            }
        } else if ((totalPercentage == bestPercentage) && (!_bestSetOnPercentageOnly)) {
	    //TraceManager.addDev("Get Best 3-n");
            boolean respectBothMin = (getRegularPercentage() >= regularMin) && (getMalwarePercentage() >= malwareMin);
            boolean otherRespectBothMin = (currentBest.getRegularPercentage() >= regularMin) && (currentBest.getMalwarePercentage() >= malwareMin);


            if (respectBothMin && !otherRespectBothMin) {
                return this;
            }

            if (otherRespectBothMin && !respectBothMin) {
                return currentBest;
            }


            // So, the two are in the same situation with respecting min percentage
            // -> Use standard deviation between regular and malware score in each sample
            // so as to select the best of the two
            if (_bestSetOnAverage) {
                if (average > currentBest.average) {
                    return this;
                }
            } else {
		//TraceManager.addDev("Testing minDiff");
                /*  if (sigma > currentBest.sigma) {
                    return this;
                    }*/
                if (getDiff() > currentBest.getDiff()) {
                    return this;
                }
            }

        }

	TraceManager.addDev("Get Best 4-n");
        return currentBest;
    }


    public AlligatorEvaluationSet fullClone() {
        AlligatorEvaluationSet set = new AlligatorEvaluationSet(arrayOfValueIndexRegular.length);

        for(int i=0; i<arrayOfValueIndexRegular.length; i++) {
            set.arrayOfValueIndexRegular[i] = arrayOfValueIndexRegular[i];
            set.arrayOfValueIndexMalware[i] = arrayOfValueIndexMalware[i];
        }
        set.nbOfRegularRecognized = nbOfRegularRecognized;
        set.nbOfRegularFailed = nbOfRegularFailed;
        set.nbOfMalwareRecognized = nbOfMalwareRecognized;
        set.nbOfMalwareFailed = nbOfMalwareFailed;
	
	set.average = average;
	//set.sigma = sigma;
	set.minDiff = minDiff;
	set.maxDiff = maxDiff;
	set.maxRegularDiff = maxRegularDiff;
	set.maxMalwareDiff = maxMalwareDiff;

        return set;
    }

    public void reinitAnalysisValues() {
        nbOfRegularRecognized = 0;
        nbOfRegularFailed = 0;
        nbOfMalwareRecognized = 0;
        nbOfMalwareFailed = 0;
    }

    public boolean isOptimalSet(boolean _bestSetOnPercentageOnly) {
        if (_bestSetOnPercentageOnly) {
            return ((nbOfRegularFailed == 0) && (nbOfMalwareFailed==0));
        }

        return false;
    }

    public double getDiff() {
	double diff = (minDiff - maxDiff);

	return diff;
    }



} // Class
