/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille
   Copyright or (C) or Copr. FORTINET, Axelle Apvrille

   ludovic.apvrille AT enst.fr
   aapvrille AT fortinet.com

   This software is a computer program whose purpose is to allow malware analysis

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class AlligatorLanguageReader
   * Element of a cluster
   * Creation: 20/07/2012
   * @version 1.0 20/07/2012
   * @author Ludovic APVRILLE, Axelle APVRILLE
   * @see
   */

package analyzer;

import cluster.*;
import myutil.*;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.text.*;



public class AlligatorLanguageReader implements StepProvider {
    public static final String[] COMMANDS = {"printclustersummary",  "setcomplexweight", "setmultiplierweight",
                                             /*3 */ "compute", "jumplineinreport", "printresultsummary", "set", "printclusteraveragevalues", "generateplot",
                                             /* 9 */    "printpropertycorrelations", "jumplineintopreport", "printexclusivepropertycorrelations", "percentagemin",
                                             /* 13 */ "setpropertyweight", "setpropertyweights", "setpropertyweightsfromcolumn", "percentageweight",
                                             /* 17 */ "computeoverallscores", "setpropertyweightbyname", "setprintintermediatescore", "setclustername",
                                             /* 21 */ "printstatistics", "selectpropertiesrange", "removepropertiesafter", "removepropertiesbefore",
                                             /* 25 */ "setmaxnbofvaluesperproperty", "setdiffvalue", "setbasicscore", "printprobastatistics",
                                             /*29 */  "printcorrelationstatistics", "printproximityelements"};

    public static final String[] CLUSTERS = {"regular",  "malware", "guess"};

    public static final String[] METRICS = {"deviation", "proximity", "proximitywithlimitedproperties", "correlations",
                                            /* 4 */ "probabilitydifference", "probabilityfactor", "epsiloncluster", "deviationpercentage", "deviationweightpercentage", "weightdeviation",
                                            /* 10 */ "correlation", "addvaluetoscore", "closestaverageoneachproperty",
                                            /* 13 */ "degressiveproximity", "proximitywithlimitedpropertiesf", "proximityratio", "svm",
					    /* 17 */ "inversedeviation", "inverseweightdeviation"};

    public static final String[] OPTIONS = {"printintermediatescore"};

    public static final String[] PLOT_METRICS = {"deviation", "alldeviation", "proximity", "allproximity", "probabilityfactor", "allprobabilityfactor", "correlation",
                                                 /*7*/ "allcorrelation", "probabilitydifference", "allprobabilitydifference",
                                                 /* 10 */ "epsilon", "allepsilon", "deviationpercentage", "alldeviationpercentage", "deviationweightpercentage","alldeviationweightpercentage",
                                                 /* 16 */ "weightdeviation", "allweightdeviation"};


    public static final String[] icontext = {" -,===,oo<", " -,===,oo="};

    private Cluster regularCluster;
    private Cluster malwareCluster;

    private Cluster guessCluster;

    private double[] regularScore, malwareScore;

    private Vector<AlligatorScore> regularScores[], malwareScores[]; // for each ClusterElement
    public int[] classificationWeight; // for each cluster element


    private int nbOfComputationsDone;

    private double minRegularMultiplier, maxRegularMultiplier, regularStep;
    private double minMalwareMultiplier, maxMalwareMultiplier, malwareStep;
    private boolean regularHasComplexExpresion, malwareHasComplexExpresion;
    private String regularComplexExpression, malwareComplexExpression;

    //private DoubleExpressionEvaluator dee;

    private int proximity = 0;
    private int nbOfLimitedProperties = 0;
    private Cluster[] clusters = new Cluster[2];
    //private boolean mostCommonValueAndProbabilitiesComputed = false;
    //private boolean computeSortedPropertyValues = false;
    private boolean printStatistics = false;
    private boolean printProbaStatistics = false;
    private boolean printCorrelationStatistics = false;

    // options
    private boolean printIntermediateScore = false;
    private boolean ignoreReport = false;
    private double diffValueMalware = -1;
    private double diffValueRegular = -1;
    private boolean basicScore = false;
    private boolean printProximityElements = false;

    // for handling computation step
    private int iteration;

    // For displaying progression
    private ProgressionDisplayer pd;

    private int nbOfRelevantLines;
    private int currentRelevantLine;

    private int nbOfRelevantComputationLines;
    private int currentRelevantComputationLine;

    private boolean optimizeMemory = false;


    public AlligatorLanguageReader(Cluster _regularCluster, Cluster _malwareCluster, Cluster _guessCluster) {
        regularCluster = _regularCluster;
        malwareCluster = _malwareCluster;
        guessCluster = _guessCluster;
        clusters[0] = malwareCluster;
        clusters[1] = regularCluster;

        pd = new ProgressionDisplayer(guessCluster);
        pd.setProgressionIcons(icontext);
        pd.setStepProvider(this);
        pd.go();
    }

    public Cluster getGuessCluster() {
        return guessCluster;
    }

    public static String getIconText(int index) {
        return icontext[index%icontext.length];
    }

    public void setOptimizeMemory(boolean _optimizeMemory) {
	optimizeMemory = _optimizeMemory;
    }

    public String getAllReports(int nbOfElementsInCluster2, String name1, String name2, String timeReport) {
        /* Make extra report */
        TraceManager.addDev("Getting all reports");
        String extra = "";
        extra += "\nEvaluating the two clusters " + name1 + " and " + name2 + "\n";
        extra += "\n";

        int malware1 = 0, clean1 = 0, malware2 = 0, clean2 = 0;
        int lightMalware1 = 0, lightClean1 = 0;
        int mediumMalware1 = 0, mediumClean1 = 0;
        int strongMalware1 = 0, strongClean1 = 0;
        int lightMalware2 = 0, lightClean2 = 0;
        int mediumMalware2 = 0, mediumClean2 = 0;
        int strongMalware2 = 0, strongClean2 = 0;

        int cpt = 0;
        int limit = guessCluster.getElements().size() - nbOfElementsInCluster2;
        double diff;

        for(ClusterElement ce: guessCluster.getElements()) {
            //computeScores(cpt);

            if (regularScore[cpt] < malwareScore[cpt]) {
                //                score = "*" + malwareCluster.getName() + "*";
                if (cpt <  limit) {
                    malware1 ++;
                } else {
                    malware2 ++;
                }

                if (diffValueMalware > 0) {
                    diff = malwareScore[cpt] - regularScore[cpt];
                    if (diff < (diffValueMalware/3)) {
                        if (cpt <  limit) {
                            lightMalware1 ++;
                        } else {
                            lightMalware2 ++;
                        }

                    } else if ((2*diff)<(2*diffValueMalware/3)) {
                        if (cpt <  limit) {
                            mediumMalware1 ++;
                        } else {
                            mediumMalware2 ++;
                        }

                    } else {
                        if (cpt <  limit) {
                            strongMalware1 ++;
                        } else {
                            strongMalware2 ++;
                        }
                    }
                }

            }  else {
                //               score = " " + regularCluster.getName() + " ";
                if (cpt <  limit) {
                    clean1 ++;
                } else {
                    clean2 ++;
                }

                if (diffValueRegular > 0) {
                    diff = regularScore[cpt] - malwareScore[cpt];
                    if (diff < (diffValueRegular/3)) {
                        if (cpt <  limit) {
                            lightClean1 ++;
                        } else {
                            lightClean2 ++;
                        }

                    } else if ((diff)<(2*diffValueRegular/3)) {
                        if (cpt <  limit) {
                            mediumClean1 ++;
                        } else {
                            mediumClean2 ++;
                        }

                    } else {
                        if (cpt <  limit) {
                            strongClean1 ++;
                        } else {
                            strongClean2 ++;
                        }

                    }
                }
            }
            cpt ++;
        }
        int total1 = clean1 + malware1;
        int total2 = clean2 + malware2;

        double rateFalse = ((int)(((double)malware1+clean2)/(total1+total2)*10000))/100.0;
        double ratePositive = ((int)((100-rateFalse)*100))/100;
        double percentage1 = ((int)(((double)clean1)/(total1)*10000))/100.0;
        double percentage2 = ((int)(((double)malware1)/(total1)*10000))/100.0;
        double percentage3 = ((int)(((double)clean2)/(total2)*10000))/100.0;
        double percentage4 = ((int)(((double)malware2)/(total2)*10000))/100.0;

        // False positive and negative
        extra += "Good classification: " + (clean1 + malware2) + " (" + ratePositive + "%)\n";
        extra += "Wrong classification: " + (malware1 + clean2) + " (" + rateFalse + "%)\n";
        extra += "  FP i.e. falsy classified in " + malwareCluster.getName() + ": " + malware1 + " (" + percentage2+"%)\n";
        extra += "  FN i.e. falsy classified in " + regularCluster.getName() + ": " + clean2 + " (" + percentage3 + "%)\n\n";

        extra += name1 + ": " + clean1 + " " + regularCluster.getName() + "(s) found, " + malware1 + " " + malwareCluster.getName() + "(s) found\n";
        extra += "  Percentage of " + regularCluster.getName() + ": " + percentage1 + "\n";
        extra += "  Percentage of " + malwareCluster.getName() + ": " + percentage2 + "\n";
        extra += "\n";

        extra += name2 + ": " + clean2 + " " + regularCluster.getName() + "(s) found, " + malware2 + " " + malwareCluster.getName() + "(s) found\n";

        extra += "  Percentage of " + regularCluster.getName() + ": " + percentage3 + "\n";
        extra += "  Percentage of " + malwareCluster.getName() + ": " + percentage4 + "\n";


        return  guessCluster.getAllReports(printIntermediateScore, nbOfElementsInCluster2, name1, name2, timeReport + "\n\n" + extra);
    }

    public String getAllReports(String timeReport) {
        return  guessCluster.getAllReports(printIntermediateScore, 0, null, null, timeReport);
    }

    public void setIgnoreReport(boolean b) {
        ignoreReport = b;
    }

    public int getCurrentStep() {
        return currentRelevantComputationLine;
    }

    public int getMaxStep() {
        return nbOfRelevantComputationLines;
    }

    public double getDiffValueMalware() {
        return diffValueMalware;
    }

    public double getDiffValueRegular() {
        return diffValueRegular;
    }


    @SuppressWarnings("unchecked")
    public boolean execute(String execute) {
        minRegularMultiplier = 1.0;
        maxRegularMultiplier = 1.0;
        regularStep = 0;

        minMalwareMultiplier = 1.0;
        maxMalwareMultiplier = 1.0;
        malwareStep = 0;

        regularHasComplexExpresion = false;
        malwareHasComplexExpresion = false;


        /* Report init */
        if (!ignoreReport) {
            guessCluster.addToReport("------------------------------------------------------------------------------------------------", Cluster.ONLY_TEXT, "\n");
            guessCluster.addToReport("", Cluster.NAME, ":\n");
        }

	/* Cluster init */
	regularCluster.init();
	malwareCluster.init();
	guessCluster.init();

        /* Computing properties, and normalize them */

        // Check whether properties are computer, or not
        TraceManager.addDev("Computing properties of regular");
        regularCluster.checkForPropertiesComputed();
        TraceManager.addDev("Computing properties of malware");
        malwareCluster.checkForPropertiesComputed();
        TraceManager.addDev("Computing properties of guess");
        guessCluster.checkForPropertiesComputed();

        if (!regularCluster.isNormalized() || !malwareCluster.isNormalized() || !guessCluster.isNormalized()) {
            normalize();
        }

        TraceManager.addDev(regularCluster.getShortDescription());
        TraceManager.addDev(malwareCluster.getShortDescription());
        TraceManager.addDev(guessCluster.getShortDescription());

        //TraceManager.addDev(guessCluster.getFullDescription());


        // Initializing scores
        regularScore = new double[guessCluster.getElements().size()];
        malwareScore = new double[guessCluster.getElements().size()];

        Vector<AlligatorScore> tmpVector [] = new Vector[guessCluster.getElements().size()];
        regularScores = (Vector<AlligatorScore>[])tmpVector;

        tmpVector = new Vector[guessCluster.getElements().size()];
        malwareScores = (Vector<AlligatorScore>[])tmpVector;


        // Checking syntax and counting the number of relevant lines */
        nbOfRelevantLines = 0;
        currentRelevantComputationLine = 0;
        int lineNbSyntax = 0;
        regularCluster.setName(CLUSTERS[0]);
        malwareCluster.setName(CLUSTERS[1]);
        guessCluster.setName(CLUSTERS[2]);
        Scanner scannerSyntax = new Scanner(execute);
        while (scannerSyntax.hasNextLine()) {
            String lineSyntax = scannerSyntax.nextLine();
            lineNbSyntax++;

            try {
                if (executeLine(lineSyntax.trim(), true)) {
                    nbOfRelevantLines++;
                }

            } catch (AlligatorSyntaxException ase) {
                TraceManager.addUser("Error in line #" + lineNbSyntax + ": " + lineSyntax + "\n" + ase.getMessage());
                pd.setMonitor(false);
                scannerSyntax.close();
                return false;
            }
        }
        scannerSyntax.close();
        regularCluster.setName(CLUSTERS[0]);
        malwareCluster.setName(CLUSTERS[1]);
        guessCluster.setName(CLUSTERS[2]);
        nbOfRelevantComputationLines = currentRelevantComputationLine;
        currentRelevantComputationLine = 0;

        //regularCluster.setNbOfPropertiesToConsider(regularCluster.getTotalNbOfProperties());
        //malwareCluster.setNbOfPropertiesToConsider(regularCluster.getTotalNbOfProperties());
        //guessCluster.setNbOfPropertiesToConsider(regularCluster.getTotalNbOfProperties());

        // Computing memory to be purged
        int linePurgeProximity = -1;
        int linePurgeDeviations = -1;
	int linePurgeProbabilities = -1;
	int linePurgeWeightDeviations = -1;
        int linePurgeCorrelations = -1;
        int linePurgeEpsilonClusters = -1;
	int linePurgeSVMModel = -1;
        int lineNb = 0;


        Scanner scannerLines = new Scanner(execute);
        while (scannerLines.hasNextLine()) {
            String linePurge = scannerLines.nextLine().toLowerCase();
            lineNb++;

            if (linePurge.indexOf("correlation") > -1) {
                linePurgeCorrelations = lineNb;
            }

            if ((linePurge.indexOf("deviation") > -1) && (linePurge.indexOf("weight") == -1))  {
                linePurgeDeviations = lineNb;
            }

	    if (linePurge.indexOf("weightdeviation") > -1) {
                linePurgeWeightDeviations = lineNb;
            }

	    if (linePurge.indexOf("probability") > -1) {
                linePurgeProbabilities = lineNb;
            }

            if (linePurge.indexOf("proximity") > -1) {
                linePurgeProximity = lineNb;
            }

            if (linePurge.indexOf("epsiloncluster") > -1) {
                linePurgeEpsilonClusters = lineNb;
            }
	    if (linePurge.indexOf("SVM") > -1) {
                linePurgeSVMModel = lineNb;
            }

        }
        scannerLines.close();

        // Reading the string, and interpreting it */
        lineNb = 0;
        Scanner scanner = new Scanner(execute);
        while (scanner.hasNextLine()) {
            String line = scanner.nextLine();
            lineNb++;

            try {
                if (executeLine(line.trim())) {
                    currentRelevantLine ++;
                }

                // Purge memory
		//if (optimizeMemory) {
		    if (lineNb == linePurgeProximity) {
			TraceManager.addDev("\n--- Purging memory of proximity\n");
			guessCluster.purgeProximity();
		    }

		    if (lineNb == linePurgeDeviations) {
			TraceManager.addDev("\n--- Purging memory of deviation\n");
			guessCluster.purgeDeviationsOfElements();
		    }

		    if (lineNb == linePurgeWeightDeviations) {
			TraceManager.addDev("\n--- Purging memory of weight deviation\n");
			guessCluster.purgeWeightDeviationsOfElements();
			if (optimizeMemory) {
			    regularCluster.purgeAllWeightDeviations();
			    malwareCluster.purgeAllWeightDeviations();
			}
		    }

		    if ((lineNb == linePurgeProbabilities) && ((linePurgeCorrelations == -1) || (linePurgeCorrelations < linePurgeProbabilities))) {
			TraceManager.addDev("\n--- Purging memory of probabilities\n");
			if (optimizeMemory) {
			    regularCluster.purgeAllProbabilities();
			    malwareCluster.purgeAllProbabilities();
			}
		    }

		    if (lineNb == linePurgeCorrelations) {
			TraceManager.addDev("\n--- Purging memory of correlation\n");
			guessCluster.purgeAllCorrelations();
			if (optimizeMemory) {
			    regularCluster.purgeAllCorrelations();
			    malwareCluster.purgeAllCorrelations();
			}

			if ((linePurgeProbabilities == -1) || (linePurgeProbabilities < linePurgeCorrelations)) {
			    if (optimizeMemory) {
				regularCluster.purgeAllProbabilities();
				malwareCluster.purgeAllProbabilities();
			    }
			}

		    }

		    if (lineNb == linePurgeEpsilonClusters) {
			TraceManager.addDev("\n--- Purging memory of epsilon clusters\n");
			guessCluster.purgeAllEpsilonClusters();
			if (optimizeMemory) {
			    regularCluster.purgeAllEpsilonClusters();
			    malwareCluster.purgeAllEpsilonClusters();
			}
		    }

		    if (lineNb == linePurgeSVMModel) {
			TraceManager.addDev("\n--- Purging memory of SVM Model\n");
			guessCluster.purgeAllSVMModels();
			if (optimizeMemory) {
			    regularCluster.purgeAllSVMModels();
			    malwareCluster.purgeAllSVMModels();
			}
		    }
		    //}

            } catch (AlligatorSyntaxException ase) {
                TraceManager.addUser("Error in line #" + lineNb + ": " + line + "\n" + ase.getMessage());
                pd.setMonitor(false);
                scanner.close();
                return false;
            }
        }
        scanner.close();

        pd.setGo(false);

        // Put scores
        if (!ignoreReport) {
            int cpt = 0;
            for(ClusterElement ce: guessCluster.getElements()) {
                ce.addToReport("\n  " + regularCluster.getName() + " score: " + regularScore[cpt] + "\n");
                ce.addToReport("  " + malwareCluster.getName() + " score: " + malwareScore[cpt] + "\n");
                if (malwareScore[cpt] <= regularScore[cpt]) {
                    ce.addToReport("*  -> " + regularCluster.getName() + "\n");
                } else {
                    ce.addToReport("* ->>>>> " + malwareCluster.getName() + "\n");
                }
                cpt ++;
            }

            // Put statistics
            if (printStatistics) {
                computeStatistics();
            }

            if (printProbaStatistics) {
                computeProbaStatistics();
            }

            if (printCorrelationStatistics) {
                computeCorrelationStatistics();
            }
        }

        return true;

    }

    public boolean isRegular(int cpt) {
        return (malwareScore[cpt] <= regularScore[cpt]);
    }

    public String getClassificationString(int index) {
        if (isRegular(index)) {
            return regularCluster.getName();
        }
        return malwareCluster.getName();
    }

    public void computeStatistics() {
        int index;
        TraceManager.addDev("Computing statistics");
        guessCluster.addToTopReport("*** STATS ***");

        guessCluster.addToTopReport("     highest scores:");
        index = getHighestScore(regularScore);
        guessCluster.addToTopReport("Element of " + guessCluster.getName() + " with the highest " + regularCluster.getName() + " score: " + guessCluster.getElements().get(index).getName()
                                    + " (classified as " + getClassificationString(index) + ")");
        index = getHighestScore(malwareScore);
        guessCluster.addToTopReport("Element of " + guessCluster.getName() + " with the highest " + malwareCluster.getName() + " score: " + guessCluster.getElements().get(index).getName()
                                    + " (classified as " + getClassificationString(index) + ")");


        guessCluster.addToTopReport("     lowest scores:");
        index = getLowestScore(regularScore);
        guessCluster.addToTopReport("Element of " + guessCluster.getName() + " with the lowest " + regularCluster.getName() + " score: " + guessCluster.getElements().get(index).getName()
                                    + " (classified as " + getClassificationString(index) + ")");
        index = getLowestScore(malwareScore);
        guessCluster.addToTopReport("Element of " + guessCluster.getName() + " with the lowest " + malwareCluster.getName() + " score: " + guessCluster.getElements().get(index).getName()
                                    + " (classified as " + getClassificationString(index) + ")");

        guessCluster.addToTopReport("     closest to the average score:");
        index = getClosestAverageScore(regularScore);
        guessCluster.addToTopReport("Element of " + guessCluster.getName() + " being the closest to the average of " + regularCluster.getName() + " scores: " + guessCluster.getElements().get(index).getName()
                                    + " (classified as " + getClassificationString(index) + ")");
        index = getClosestAverageScore(malwareScore);
        guessCluster.addToTopReport("Element of " + guessCluster.getName() + " being the closest to the average of " + malwareCluster.getName() + " scores: " + guessCluster.getElements().get(index).getName()
                                    + " (classified as " + getClassificationString(index) + ")");

        guessCluster.addToTopReport("     Highest score difference:");
        index = getHighestScoreDifference(regularScore, malwareScore);
        guessCluster.addToTopReport("Element of " + guessCluster.getName() + " having the highest absolute score difference between " + regularCluster.getName() + " and "  + malwareCluster.getName() + ": " + guessCluster.getElements().get(index).getName()
                                    + " (classified as " + getClassificationString(index) + ")");

        guessCluster.addToTopReport("     Lowest score difference:");
        index = getLowestScoreDifference(regularScore, malwareScore);
        guessCluster.addToTopReport("Element of " + guessCluster.getName() + " having the lowest absolute score difference between " + regularCluster.getName() + " and "  + malwareCluster.getName() + ": " + guessCluster.getElements().get(index).getName()
                                    + " (classified as " + getClassificationString(index) + ")");




        TraceManager.addDev("Computing statistics: done");
    }

    public void computeProbaStatistics() {
        // Should compute higest proba difference
        int index;
	

	regularCluster.computeSortedPropertyValues();
	malwareCluster.computeSortedPropertyValues();
	
        TraceManager.addDev("Computing proba statistics");
        guessCluster.addToTopReport("\n\n*** STATS on PROBA ***");
        guessCluster.addToTopReport(regularCluster.getStringProbabilities(malwareCluster) + "\n\n");

        TraceManager.addDev("Computing proba statistics: done");
    }

    public void computeCorrelationStatistics() {
        // Should compute higest proba difference
        int index;
	pd.setPercentageProvider(regularCluster);
        regularCluster.computeCorrelations();
        TraceManager.addDev("Regular: computing list of interesting correlations");
        regularCluster.computeListOfInterestingCorrelations();

        pd.setPercentageProvider(malwareCluster);
        malwareCluster.computeCorrelations();
        TraceManager.addDev("Malware: computing list of interesting correlations");
        malwareCluster.computeListOfInterestingCorrelations();
	pd.setPercentageProvider(guessCluster);

        TraceManager.addDev("Computing correlation statistics");
        guessCluster.addToTopReport("*** STATS on CORRELATION ***");

	guessCluster.addToTopReport("\tInteresting correlations of " + regularCluster.getName() + ":\n" + regularCluster.getInterestingCorrelations() + "\n");
        guessCluster.addToTopReport("\tInteresting correlations of " + malwareCluster.getName() + ":\n" + malwareCluster.getInterestingCorrelations() + "\n\n");

        TraceManager.addDev("Computing correlation statistics: done");
    }

    public void advancedNormalized() {
        if (regularCluster.isNormalized() && malwareCluster.isNormalized() && !guessCluster.isNormalized()) {
            TraceManager.addDev("Advanced normalized ....");

            boolean mustRenormalize = false;
            int indexOfNormalizePb = -1;

            // Must see whether one value in guess is higher than 1 after division by the normlization value of the corresponding property
            double max, maxGuess;
            for(int i=0; i<regularCluster.getTotalNbOfProperties(); i++) {
                max = regularCluster.getNormalizedValue(i);
                if (max > 0) {
                    maxGuess = guessCluster.getMaxForProperty(i);
                    if (maxGuess > max) {
                        mustRenormalize = true;
                        indexOfNormalizePb = i;
                        break;
                    }
                }
            }

            if (mustRenormalize) {
                TraceManager.addDev("Must renormalize because of property #" + indexOfNormalizePb);
                regularCluster.purgeClusterElements();
                malwareCluster.purgeClusterElements();
            } else {
                TraceManager.addDev("No need to renormalize -> normalizing guess cluster");
                for(int j=0; j<regularCluster.getTotalNbOfProperties(); j++) {
                    guessCluster.normalizeProperties(j, regularCluster.getNormalizedValue(j));
                }


            }
        }
    }

    public void normalize() {


        if (guessCluster.hasNormalizationValue()) {
            TraceManager.addDev("Normalize process with max value = " + guessCluster.getNormalizationValue());
            regularCluster.normalizeAllProperties(guessCluster.getNormalizationValue());
            malwareCluster.normalizeAllProperties(guessCluster.getNormalizationValue());
            guessCluster.normalizeAllProperties(guessCluster.getNormalizationValue());
        } else {
            TraceManager.addDev("Normal normalize process");
            double max;
            for(int i=0; i<regularCluster.getTotalNbOfProperties(); i++) {
                max = regularCluster.getMaxForProperty(i);
                max = Math.max(malwareCluster.getMaxForProperty(i), max);
                max = Math.max(guessCluster.getMaxForProperty(i), max);
                if (max > 1) {
                    //TraceManager.addDev("NORMALIZING PROP " + i);
                    regularCluster.normalizeProperties(i, max);
                    malwareCluster.normalizeProperties(i, max);
                    guessCluster.normalizeProperties(i, max);
                }
            }
        }
    }

    public void selectPropertiesRange(String split[], boolean _checkSyntaxOnly) throws AlligatorSyntaxException {

        TraceManager.addDev("Selecting range of properties");
        int index;
        int nb = 0;

        try {
            index =  Integer.valueOf(split[1]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Index is invalid. Current value: " + split[1]);
        }

        try {
            nb = Integer.valueOf(split[2]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Nb of properties is invalid. Current value: " + split[2]);
        }

        if (nb <= 0) {
            throw new AlligatorSyntaxException("Nb of properties is invalid. Current value: " + nb);
        }

        if (_checkSyntaxOnly) {
            return;
        }


        regularCluster.setPropertiesRange(index, nb);
        malwareCluster.setPropertiesRange(index, nb);
        guessCluster.setPropertiesRange(index, nb);
    }

    public void setMaxNbOfValuesPerProperty(String split[], boolean _checkSyntaxOnly) throws AlligatorSyntaxException {

        TraceManager.addDev("Remove properties after");
        int nb = 0;

        try {
            nb = Integer.valueOf(split[1]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Value is invalid. Must be >1. Current value: " + split[1]);
        }

        if (nb < 2) {
            throw new AlligatorSyntaxException("Value is invalid. Must be >1. Current value: " + split[1]);
        }

        if (_checkSyntaxOnly) {
            return;
        }

        regularCluster.setMaxNbOfValues(nb);
        malwareCluster.setMaxNbOfValues(nb);
        guessCluster.setMaxNbOfValues(nb);

    }

    public void removePropertiesAfter(String split[], boolean _checkSyntaxOnly) throws AlligatorSyntaxException {

        TraceManager.addDev("Remove properties after");
        int index = 0;

        try {
            index = Integer.valueOf(split[1]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Index is invalid. Current value: " + split[1]);
        }

        if (index <= 0) {
            throw new AlligatorSyntaxException("Index is invalid. Current value: " + index);
        }

        if (_checkSyntaxOnly) {
            return;
        }

        if (!regularCluster.setPropertiesRange(0, index)) {
            throw new AlligatorSyntaxException("Index of is invalid. Current value: " + index);
        }

        if (!malwareCluster.setPropertiesRange(0, index)) {
            throw new AlligatorSyntaxException("Index of is invalid. Current value: " + index);
        }

        if (!guessCluster.setPropertiesRange(0, index)) {
            throw new AlligatorSyntaxException("Index of is invalid. Current value: " + index);
        }
    }

    public void removePropertiesBefore(String split[], boolean _checkSyntaxOnly) throws AlligatorSyntaxException {

        TraceManager.addDev("Removing properties before");
        int index = 0;

        try {
            index = Integer.valueOf(split[1]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Index of is invalid. Current value: " + split[1]);
        }

        if (index <= 0) {
            throw new AlligatorSyntaxException("Index of is invalid. Current value: " + index);
        }

        if (_checkSyntaxOnly) {
            TraceManager.addDev("Check syntax only");
            return;
        }

        if (!regularCluster.setPropertiesRange(index, regularCluster.getTotalNbOfProperties())) {
            throw new AlligatorSyntaxException("Index of is invalid. Current value: " + index);
        }

        if (!malwareCluster.setPropertiesRange(index, malwareCluster.getTotalNbOfProperties())) {
            throw new AlligatorSyntaxException("Index of is invalid. Current value: " + index);
        }

	if (!guessCluster.setPropertiesRange(index, guessCluster.getTotalNbOfProperties())) {
            throw new AlligatorSyntaxException("Index of is invalid. Current value: " + index);
        }

        TraceManager.addDev("Nb of properties:" + regularCluster.getTotalNbOfProperties() + " FirstIndex=" + regularCluster.cd.firstProp + " nbOfPropertiesToConsider=" + regularCluster.cd.nbOfProps);

    }

    public String getReportForSample(int indexOfSample) {
        String ret = "";
        if (regularScore[indexOfSample] >= malwareScore[indexOfSample]) {
            ret += "1";
        } else {
            ret += "2";
        }

        ret += regularCluster.getName() + " " +  regularScore[indexOfSample] + " " + malwareCluster.getName() + " " + malwareScore[indexOfSample];
        return ret;
    }

    public String getFullReportForSample(int indexOfSample) {
        return getReportForSample(indexOfSample) + "\n" + guessCluster.getFullReportForSample(indexOfSample);
    }


    public boolean executeLine(String line) throws AlligatorSyntaxException {
        return executeLine(line, false);
    }

    // Returns true if the line is "usefull" i.e. it contains a execution of something.
    // REturns false otherwise
    public boolean executeLine(String line, boolean checkSyntaxOnly) throws AlligatorSyntaxException {
        //TraceManager.addDev("Executing line: " + line);

        if (line.length() == 0) {
            return false;
        }

        if (line.startsWith("#")) {
            // comment line
            return false;
        }

        line = Conversion.replaceAllString(line.trim(), "  ", " ");
        TraceManager.addDev("Executing line: " + line);

        String split[] = line.trim().split(" ");


        // Command
        String command = split[0];
        checkForCommandName(command);

        if (isCommand(0, command)) {
            executePrintSummary(split, checkSyntaxOnly);
            return true;
        }

        if (isCommand(1, command)) {
            setComplexWeight(split, checkSyntaxOnly);
            return true;
        }

        if (isCommand(2, command)) {
            setMultiplierWeight(split, checkSyntaxOnly);
            return true;
        }



        if (isCommand(3, command)) {
            currentRelevantComputationLine ++;
            executeComputation(split, checkSyntaxOnly);
            return true;
        }

        if (isCommand(4, command)) {
            if (checkSyntaxOnly) {
                return true;
            }
            if (!ignoreReport) {
                guessCluster.jumpLineInReport();
            }
            return true;
        }

        if (isCommand(5, command)) {
            if (checkSyntaxOnly) {
                return true;
            }
            if (!ignoreReport) {
                printResultSummary();
            }
            return true;
        }

        if (isCommand(6, command)) {
            setOptions(split, checkSyntaxOnly);
            return true;
        }

        if (isCommand(7, command)) {
            if ((!ignoreReport) || (checkSyntaxOnly)) {
                executePrintAverageValues(split, checkSyntaxOnly);
            }
            return true;
        }

        if (isCommand(8, command)) {
            executePlot(split, checkSyntaxOnly);
            if (checkSyntaxOnly) {
                return true;
            }
            pd.setMonitor(false);
            return true;
        }

        if (isCommand(9, command)) {
            if ((!ignoreReport) || (checkSyntaxOnly)) {
                executePrintPropertiesCorrelations(split, checkSyntaxOnly);
            }
            return true;
        }

        if (isCommand(10, command)) {
            if (checkSyntaxOnly) {
                return true;
            }
            if (!ignoreReport) {
                guessCluster.jumpLineInTopReport();
            }
            return true;
        }

        if (isCommand(11, command)) {
            if (!(ignoreReport) || (checkSyntaxOnly)) {
                executePrintExclusivePropertiesCorrelations(split, checkSyntaxOnly);
            }
            return true;
        }

        if (isCommand(12, command)) {
            percentageMin(split, checkSyntaxOnly);
            return true;
        }

        if (isCommand(13, command)) {
            setPropertyWeight(split, checkSyntaxOnly);
            return true;
        }

        if (isCommand(14, command)) {
            setPropertyWeights(split, checkSyntaxOnly);
            return true;
        }

        if (isCommand(15, command)) {
            setPropertyWeightsFromColumn(split, checkSyntaxOnly);
            return true;
        }

        if (isCommand(16, command)) {
            percentageWeight(split, checkSyntaxOnly);
            return true;
        }

        if (isCommand(17, command)) {
            if (checkSyntaxOnly) {
                return true;
            }
            computeOverallResults();
            return true;
        }

        if (isCommand(18, command)) {
            setPropertyWeightByName(split, checkSyntaxOnly);
            return true;
        }

        if (isCommand(19, command)) {
            if (checkSyntaxOnly) {
                return true;
            }
            TraceManager.addDev("--------------------------------------------- Setting intermediate score");
            printIntermediateScore = true;
            return true;
        }

        if (isCommand(20, command)) {
            setClusterName(split);
            return true;
        }

        if (isCommand(21, command)) {
            printStatistics();
            return true;
        }

        if (isCommand(22, command)) {
            selectPropertiesRange(split, checkSyntaxOnly);
            return true;
        }

        if (isCommand(23, command)) {
            removePropertiesAfter(split, checkSyntaxOnly);
            return true;
        }

        if (isCommand(24, command)) {
            removePropertiesBefore(split, checkSyntaxOnly);
            return true;
        }

        if (isCommand(25, command)) {
            setMaxNbOfValuesPerProperty(split, checkSyntaxOnly);
            return true;
        }

        if (isCommand(26, command)) {
            setDiffValue(split, checkSyntaxOnly);
            return true;
        }

        // Basic score
        if (isCommand(27, command)) {
            if (!checkSyntaxOnly) {
                basicScore = true;
            }
            return true;
        }

        if (isCommand(28, command)) {
            if (!checkSyntaxOnly) {
                printProbaStatistics = true;
            }
            return true;
        }

        if (isCommand(29, command)) {
            if (!checkSyntaxOnly) {
                printCorrelationStatistics = true;
            }
            return true;
        }

	if (isCommand(30, command)) {
            if (!checkSyntaxOnly) {
                printProximityElements = true;
            }
            return true;
        }

        throw new AlligatorSyntaxException("Unknown command");
    }

    public void executePrintSummary(String []split, boolean checkSyntaxOnly) throws AlligatorSyntaxException {

        String cluster = null;
        try {
            cluster = split[1];
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Missing cluster name");
        }

        if (!((isRegular(cluster) || isMalware(cluster) || isGuess(cluster)))) {
            throw new AlligatorSyntaxException("Invalid cluster name:" + cluster);
        }

        if (checkSyntaxOnly) {
            return;
        }


        String report = null;
        if (isRegular(cluster)) {
            report = regularCluster.getShortDescription();
        } else if (isMalware(cluster)) {
            report = malwareCluster.getShortDescription();
        } else if (isGuess(cluster)) {
            report = guessCluster.getShortDescription();
        }
        guessCluster.addToTopReport(report);
    }




    public void executePrintAverageValues(String []split, boolean checkSyntaxOnly) throws AlligatorSyntaxException {
        String cluster = null;
        try {
            cluster = split[1];
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Missing cluster name");


        }


        if (!((isRegular(cluster) || isMalware(cluster) || isGuess(cluster)))) {
            throw new AlligatorSyntaxException("Invalid cluster name:" + cluster);
        }

        if (checkSyntaxOnly) {
            return;
        }

        Cluster cl;

        String s = ": Average values of properties: ";

        String report = null;
        if (isRegular(cluster)) {
            cl = regularCluster;
        } else if (isMalware(cluster)) {
            cl = malwareCluster;
        } else {
            cl = guessCluster;
        }

        cl.computeMeanPerProperties();
        report = cl.getAverageValues();
        guessCluster.addToTopReport(cl.getName() + s + report);
    }

    public void setDiffValue(String []split, boolean checkSyntaxOnly) throws AlligatorSyntaxException {
        if (split.length < 2) {
            throw new AlligatorSyntaxException("Missing cluster name");
        }

        String cluster = split[1];

        if (!((isRegular(cluster) || isMalware(cluster)))) {
            throw new AlligatorSyntaxException("Invalid cluster name:" + cluster);
        }

        if (split.length < 3) {
            throw new AlligatorSyntaxException("Missing difference value");
        }

        double value;
        try {
            value = Double.valueOf(split[2]).doubleValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Invalid value: " + split[2]);
        }

        if (checkSyntaxOnly) {
            return;
        }

        if (isRegular(cluster)) {
            diffValueRegular = value;
        } else {
            diffValueMalware = value;
        }
    }

    public void executePrintPropertiesCorrelations(String []split, boolean checkSyntaxOnly) throws AlligatorSyntaxException {

        if (split.length < 2) {
            throw new AlligatorSyntaxException("Missing cluster name");
        }

        if (split[1].toLowerCase().equals("regular")) {
            if (checkSyntaxOnly) {
                return;
            }
            guessCluster.addToTopReport("Interesting correlations of " + regularCluster.getName() + ":\n" + regularCluster.getShortInterestingCorrelations() + "\n");
            return;
        }

        if (split[1].toLowerCase().equals("malware")) {
            if (checkSyntaxOnly) {
                return;
            }
            guessCluster.addToTopReport("Interesting correlations of " + malwareCluster.getName() + ":\n" + malwareCluster.getShortInterestingCorrelations() + "\n");
            return;
        }

        throw new AlligatorSyntaxException("Unknown cluster name: " + split[1]);

    }

    public void executePrintExclusivePropertiesCorrelations(String []split, boolean checkSyntaxOnly) throws AlligatorSyntaxException {
        String cluster = null;
        try {
            cluster = split[1];
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Missing cluster name");
        }
        Cluster cl = null;



        if (isRegular(cluster)) {
            cl = regularCluster;
        }

        if (isMalware(cluster)) {
            cl = malwareCluster;
        }

        if (cl == null) {
            throw new AlligatorSyntaxException("Invalid cluster name");
        }

        if (checkSyntaxOnly) {
            return;
        }

        guessCluster.addToTopReport("Exclusive interesting correlations of " + cl.getName() + "\n" + guessCluster.getCorrelationsOfCluster(cl));

    }

    public void executeComputation(String []split, boolean checkSyntaxOnly) throws AlligatorSyntaxException {
        String metric = null;
        try {
            metric = split[1];
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Missing metric");
        }

        if (!(isMetric(0, metric) || isMetric(1, metric)|| isMetric(2, metric) || isMetric(3, metric) || isMetric(4, metric) || isMetric(5, metric) || isMetric(6, metric)
              || isMetric(7, metric) || isMetric(8, metric) || isMetric(9, metric) || isMetric(10, metric) || isMetric(11, metric) || isMetric(12, metric) ||isMetric(13, metric) || isMetric(14, metric) || isMetric(15, metric) || isMetric(16, metric) || isMetric(17, metric)|| isMetric(18, metric))) {
            throw new AlligatorSyntaxException("Invalid computation name: " + metric);
        }

        TraceManager.addDev("metric=" + metric);

        if (checkSyntaxOnly) {
            return;
        }

        // If 0 weights, don't do the computation
        if ((maxMalwareMultiplier ==0) && (maxRegularMultiplier ==0)) {
            return;
        }

        TraceManager.addDev("Performing the computations=" + metric);

        AlligatorInfoScore ais = getNewAlligatorInfoScore();

        pd.setMonitor(true);



        // Execution is according to metric
        if (isMetric(0, metric)) {
            computeDeviation(ais);
        }

        if (isMetric(1, metric)) {
            computeProximity(ais, split);
        }

        if (isMetric(2, metric)) {
            computeProximityWithLimitedProperties(ais, split);
        }

        if (isMetric(3, metric)) {
            computeCorrelations(ais, split);
        }

        TraceManager.addDev("Testing 4=" + metric);
        if (isMetric(4, metric)) {
            computeProbabilityDifference(ais, split);
        }

        if (isMetric(5, metric)) {
            computeProbabilityFactor(ais, split);
        }

        if (isMetric(6, metric)) {
            computeEpsilonClusters(ais, split);
        }

        if (isMetric(7, metric)) {
            computeDeviationPercentage(ais);
        }

        if (isMetric(8, metric)) {
            computeDeviationWeightPercentage(ais);
        }

        if (isMetric(9, metric)) {
            computeWeightDeviation(ais);
        }

        if (isMetric(10, metric)) {
            computeCorrelations(ais, split);
        }

        if (isMetric(11, metric)) {
            addValueToScore(ais, split);
        }

        if (isMetric(12, metric)) {
            closestAverageOnEachProperty(ais, split);
        }

        if (isMetric(13, metric)) {
            computeDegressiveProximity(ais, split);
        }

        if (isMetric(14, metric)) {
            computeProximityWithLimitedProperties_FirstVersion(ais, split);
        }

        if (isMetric(15, metric)) {
            computeProximityRatio(ais, split);
        }

	if (isMetric(16, metric)) {
            computeSVM(ais);
        }

	if (isMetric(17, metric)) {
            computeInverseDeviation(ais);
        }

	if (isMetric(18, metric)) {
            computeInverseWeightDeviation(ais);
        }

        pd.setMonitor(false);

    }

    public void percentageMin(String []split, boolean checkSyntaxOnly) throws AlligatorSyntaxException {
        String cluster = null;
        int min = 0;
        // syntax: percentageMin <cluster> <min>
        try {
            cluster = split[1];
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Missing cluster name");
        }

        try {
            min = Integer.valueOf(split[2]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Percentage argument is invalid");
        }
        if (min < 0) {
            throw new AlligatorSyntaxException("Percentage must be >= 0");
        }

        if (!((isRegular(cluster) || isMalware(cluster)))) {
            throw new AlligatorSyntaxException("Invalid cluster name: " + cluster);
        }

        if (checkSyntaxOnly) {
            return;
        }

        if (isRegular(cluster)) {
            TraceManager.addUser("Regular min percentage: "+min);
            AlligatorEvaluationSet.regularMin = min;
        } else  {
            TraceManager.addUser("Malware min percentage: "+min);
            AlligatorEvaluationSet.malwareMin = min;
        }
    }

    public void percentageWeight(String []split, boolean checkSyntaxOnly) throws AlligatorSyntaxException {
        String cluster = null;
        int weight = 0;
        // syntax: percentageMin <cluster> <min>
        try {
            cluster = split[1];
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Missing cluster name");
        }

        try {
            weight = Integer.valueOf(split[2]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("weight argument is invalid");
        }
        if (weight < 0) {
            throw new AlligatorSyntaxException("Weight must be >= 0");
        }

        if (!((isRegular(cluster) || isMalware(cluster)))) {
            throw new AlligatorSyntaxException("Invalid cluster name: " + cluster);
        }

        if (checkSyntaxOnly) {
            return;
        }


        if (isRegular(cluster)) {
            TraceManager.addUser("Regular weight : "+weight);
            AlligatorEvaluationSet.regularPercentageWeight = weight;
        } else {
            TraceManager.addUser("Malware weight : "+weight);
            AlligatorEvaluationSet.malwarePercentageWeight = weight;
        }
    }

    public void setPropertyWeight(String []split, boolean checkSyntaxOnly) throws AlligatorSyntaxException {
        int propertyIndex = 0;
        int propertyWeight = 1;
        try {
            propertyIndex = Integer.valueOf(split[1]).intValue();
            propertyIndex = propertyIndex - 2;
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Property index is invalid");
        }
        if ((propertyIndex < 0) || (propertyIndex >= guessCluster.getTotalNbOfProperties())){
            throw new AlligatorSyntaxException("Invalid index: must be between 2 and " + (guessCluster.getTotalNbOfProperties()+1));
        }

        try {
            propertyWeight = Integer.valueOf(split[2]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Property weight is invalid");
        }

        if (propertyWeight < 0){
            throw new AlligatorSyntaxException("Invalid property weight value: must be greater or equal to 0");
        }

        if (checkSyntaxOnly) {
            return;
        }

        regularCluster.setPropertyWeight(propertyIndex, propertyWeight);
        malwareCluster.setPropertyWeight(propertyIndex, propertyWeight);
        guessCluster.setPropertyWeight(propertyIndex, propertyWeight);


    }

    public void setPropertyWeights(String []split, boolean checkSyntaxOnly) throws AlligatorSyntaxException {

        int propertyWeight;

        if (split.length < 2) {
            throw new AlligatorSyntaxException("Should specify at least one weight");
        }

        for(int i=1; i<split.length; i++) {
            try {
                propertyWeight = Integer.valueOf(split[i]).intValue();
            } catch (Exception e) {
                throw new AlligatorSyntaxException("Property weight is invalid at index " + i);
            }

            if (propertyWeight < 0){
                throw new AlligatorSyntaxException("Invalid property weight value for index " + i + ": must be greater or equal to 0");
            }

            if (!checkSyntaxOnly) {
                regularCluster.setPropertyWeight(i-1, propertyWeight);
                malwareCluster.setPropertyWeight(i-1, propertyWeight);
                guessCluster.setPropertyWeight(i-1, propertyWeight);
            }

        }
    }

    public void setPropertyWeightsFromColumn(String []split, boolean checkSyntaxOnly) throws AlligatorSyntaxException {

        int startColumn;
        try {
            startColumn = Integer.valueOf(split[1]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Property index is invalid");
        }
        if ((startColumn < 0) || (startColumn >= guessCluster.getTotalNbOfProperties())){
            throw new AlligatorSyntaxException("Invalid index: must be between 2 and " + (guessCluster.getTotalNbOfProperties()+1));
        }


        int propertyWeight;

        if (split.length < 2) {
            throw new AlligatorSyntaxException("Should specify at least one weight");
        }

        for(int i=2; i<split.length; i++) {
            try {
                propertyWeight = Integer.valueOf(split[i]).intValue();
            } catch (Exception e) {
                throw new AlligatorSyntaxException("Property weight is invalid at index " + (i + startColumn) + ": value=" + split[i]);
            }

            if (propertyWeight < 0){
                throw new AlligatorSyntaxException("Invalid property weight value for index " + (i+startColumn) + ": must be greater or equal to 0");
            }

            if (!checkSyntaxOnly) {
                regularCluster.setPropertyWeight(i+startColumn-4, propertyWeight);
                malwareCluster.setPropertyWeight(i+startColumn-4, propertyWeight);
                guessCluster.setPropertyWeight(i+startColumn-4, propertyWeight);
            }

        }
    }

    public void setPropertyWeightByName(String []split, boolean checkSyntaxOnly) throws AlligatorSyntaxException {

        if (split.length < 3) {
            throw new AlligatorSyntaxException("Should specify a name and a weight");
        }

        String name = split[1];
        int propertyWeight = 0;
        try {
            propertyWeight = Integer.valueOf(split[2]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Property weight is invalid value=" + split[1]);
        }


        if (propertyWeight < 0){
            throw new AlligatorSyntaxException("Property weight is invalid value=" + split[1] + ": value must be >0");
        }

        int colr, colm, colg;

        colr = regularCluster.getPropertyIndexByName(name);
        colm = malwareCluster.getPropertyIndexByName(name);
        colg = guessCluster.getPropertyIndexByName(name);

        if (!((colr != -1) && (colr == colm) && (colr == colg))) {
            throw new AlligatorSyntaxException("Property " + name + " does not have the same index in all clusters OR invalid property name");
        }

        if (checkSyntaxOnly) {
            return;
        }


        TraceManager.addDev("Setting weight " + propertyWeight + " to property " + name + " of index " + colr);
        regularCluster.setPropertyWeight(colr, propertyWeight);
        malwareCluster.setPropertyWeight(colr, propertyWeight);
        guessCluster.setPropertyWeight(colr, propertyWeight);


    }

    public void setClusterName(String []split) throws AlligatorSyntaxException {

        if (split.length < 3) {
            throw new AlligatorSyntaxException("Should specify an old cluster name (<regular>, <malware>, etc.) and a new name");
        }

        String clusterType = split[1];
        String name = split[2];
        String cluster0, cluster1, cluster2;


        // Checking that the new name is valid
        if (name.toLowerCase().compareTo(CLUSTERS[0]) == 0) {
            throw new AlligatorSyntaxException("The new name (" + name + ") is not a valid name");
        }

        if (name.toLowerCase().compareTo(CLUSTERS[1]) == 0) {
            throw new AlligatorSyntaxException("The new name (" + name + ") is not a valid name");
        }

        if (name.toLowerCase().compareTo(CLUSTERS[2]) == 0) {
            throw new AlligatorSyntaxException("The new name (" + name + ") is not a valid name");
        }

        cluster0 = regularCluster.getName().toLowerCase();
        cluster1 = regularCluster.getName().toLowerCase();
        cluster2 = regularCluster.getName().toLowerCase();


        if ((clusterType.toLowerCase().compareTo(CLUSTERS[0]) == 0) || (clusterType.toLowerCase().compareTo(cluster0) == 0)) {
            if ((name.toLowerCase().compareTo(cluster1) == 0) || (name.toLowerCase().compareTo(cluster2) == 0)) {
                throw new AlligatorSyntaxException("The new name (" + name + ") is already used for another cluster");
            }

            regularCluster.setName(name);
            AlligatorEvaluationSet.REGULAR = name;
            return;
        }

        if ((clusterType.toLowerCase().compareTo(CLUSTERS[1]) == 0) || (clusterType.toLowerCase().compareTo(cluster1) == 0)) {
            if ((name.toLowerCase().compareTo(cluster0) == 0) || (name.toLowerCase().compareTo(cluster2) == 0)) {
                throw new AlligatorSyntaxException("The new name (" + name + ") is already used for another cluster");
            }
            malwareCluster.setName(name);
            AlligatorEvaluationSet.MALWARE = name;
            return;
        }

        if ((clusterType.toLowerCase().compareTo(CLUSTERS[2]) == 0) || (clusterType.toLowerCase().compareTo(cluster2) == 0)) {
            if ((name.toLowerCase().compareTo(cluster0) == 0) || (name.toLowerCase().compareTo(cluster1) == 0)) {
                throw new AlligatorSyntaxException("The new name (" + name + ") is already used for another cluster");
            }
            guessCluster.setName(name);
            return;
        }


        throw new AlligatorSyntaxException("Invalid cluster type: "   + clusterType);

    }

    public void printStatistics() throws AlligatorSyntaxException {
        printStatistics = true;
    }


    public AlligatorInfoScore getNewAlligatorInfoScore() {
        AlligatorInfoScore ais = new AlligatorInfoScore(nbOfComputationsDone);
        nbOfComputationsDone ++;

        // Regular
        if (regularHasComplexExpresion) {
            ais.regularComplexExpression = regularComplexExpression;
        } else {
            ais.regularComplexExpression = null;
        }
        ais.regularMinMultiplier = minRegularMultiplier;
        ais.regularMaxMultiplier = maxRegularMultiplier;
        ais.regularStep = regularStep;


        // Malware
        if (malwareHasComplexExpresion) {
            ais.malwareComplexExpression = malwareComplexExpression;
        } else {
            ais.malwareComplexExpression = null;
        }
        ais.malwareMinMultiplier = minMalwareMultiplier;
        ais.malwareMaxMultiplier = maxMalwareMultiplier;
        ais.malwareStep = malwareStep;

        return ais;

    }

    public void setMultiplierWeight(String []split, boolean checkSyntaxOnly) throws  AlligatorSyntaxException{
        // Cluster
        String cluster = null;
        try {
            cluster = split[1];
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Missing cluster name");
        }
        // Expression

        double d;
        try {
            d = Double.valueOf(split[2]).doubleValue();
        } catch (Exception e) {
            if (!extractMultipierInterval(cluster, split[2], null)) {
                throw new AlligatorSyntaxException("Invalid multiplier: " + split[2]);
            }
            return;
        }


        if (!(isRegular(cluster) || isMalware(cluster))) {
            throw new AlligatorSyntaxException("Invalid cluster name:" + cluster);
        }

        if (checkSyntaxOnly) {
            return;
        }

        if (isRegular(cluster)) {
            regularHasComplexExpresion = false;
            minRegularMultiplier = d;
            maxRegularMultiplier = d;
            regularStep = 0;
        } else if (isMalware(cluster)) {
            malwareHasComplexExpresion = false;
            minMalwareMultiplier = d;
            maxMalwareMultiplier = d;
            malwareStep = 0;
        }



    }

    public void setComplexWeight(String []split, boolean checkSyntaxOnly) throws  AlligatorSyntaxException{
        String cluster = null;
        String expr = null;
        try {
            cluster = split[1];
            expr = split[3];
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Invalid arguments");
        }


        // Multiplier
        double d;
        try {
            d = Double.valueOf(split[2]).doubleValue();
        } catch (Exception e) {
            if (!extractMultipierInterval(cluster, split[2], expr)) {
                throw new AlligatorSyntaxException("Invalid multiplier: " + split[2]);
            }
            return;
        }

        if (!(isRegular(cluster) || isMalware(cluster))) {
            throw new AlligatorSyntaxException("Invalid cluster name:" + cluster);
        }

        if (checkSyntaxOnly) {
            return;
        }


        if (isRegular(cluster)) {
            regularHasComplexExpresion = true;
            minRegularMultiplier = d;
            maxRegularMultiplier = d;
            regularComplexExpression = expr;
            regularStep = 0;
        } else if (isMalware(cluster)) {
            malwareHasComplexExpresion = true;
            minMalwareMultiplier = d;
            maxMalwareMultiplier = d;
            malwareComplexExpression = expr;
            malwareStep = 0;
        }


        TraceManager.addDev("Complex expression set to " + expr);

    }

    public boolean extractMultipierInterval(String cluster, String multi, String expr) throws AlligatorSyntaxException {
        int index0 = multi.indexOf('-');
        int index1 = multi.indexOf(',');

        if ((index0 == -1) || (index1 == -1)) {
            return false;
        }


        String nb0 = multi.substring(0, index0);
        String nb1 = multi.substring(index0+1, index1);
        String nb2 = multi.substring(index1+1, multi.length());

        double d0, d1, d2;
        try {
            d0 = Double.valueOf(nb0).doubleValue();
            d1 = Double.valueOf(nb1).doubleValue();
            d2 = Double.valueOf(nb2).doubleValue();
        } catch (Exception e) {
            return false;
        }

        if (d0 > d1) {
            return false;
        }


        if (d2<0) {
            return false;
        }

        if ((d2 == 0) && (d0 != d1)) {
            return false;
        }

        TraceManager.addDev("Setting complex multiplier min: " + d0 + " max: " + d1 + " step: " + d2);
        if (isRegular(cluster)) {
            regularHasComplexExpresion = (!(expr == null));
            minRegularMultiplier = d0;
            maxRegularMultiplier = d1;
            regularComplexExpression = expr;
            regularStep = d2;
        } else if (isMalware(cluster)) {
            malwareHasComplexExpresion = (!(expr == null));
            minMalwareMultiplier = d0;
            maxMalwareMultiplier = d1;
            malwareComplexExpression = expr;
            malwareStep = d2;
        } else {
            throw new AlligatorSyntaxException("Invalid cluster name:" + cluster);
        }


        return true;

    }




    public boolean isMetric(int index, String s) {
        return METRICS[index].equals(s.toLowerCase());

    }


    public boolean isRegular(String _s) {
        boolean b = isClusterOfCLUSTER(_s, 0);
        if (b) {
            return true;
        }

        return (regularCluster.getName().compareTo(_s) == 0);
    }

    public boolean isMalware(String _s) {
        boolean b = isClusterOfCLUSTER(_s, 1);
        if (b) {
            return true;
        }

        return (malwareCluster.getName().compareTo(_s) == 0);
    }

    public boolean isGuess(String _s) {
        boolean b = isClusterOfCLUSTER(_s, 2);
        if (b) {
            return true;
        }

        return (guessCluster.getName().compareTo(_s) == 0);
    }

    public boolean isClusterOfCLUSTER(String name, int index) {
        return name.toLowerCase().equals(CLUSTERS[index]);
    }

    public boolean checkForCommandName(String _name) {
        return isInStringArray(COMMANDS, _name);
    }

    public boolean checkForClusterName(String _name) {
        return isInStringArray(COMMANDS, _name);
    }

    public boolean isInStringArray(String[] array, String inArray) {
        inArray = inArray.toLowerCase();
        for(int i=0; i<array.length; i++) {
            if (array[i].compareTo(inArray) == 0) {
                return true;
            }
        }
        return false;
    }

    public boolean isCommand(int index, String reference) {
        if (COMMANDS[index].equals(reference.toLowerCase())) {
            return true;
        }
        return false;
    }

    // SVM
    public void computeSVM(AlligatorInfoScore ais) throws AlligatorSyntaxException {
        ais.stageInfo = 0;
        DoublePoint dp;

        guessCluster.computeSVMModel(regularCluster, malwareCluster);

        if (!ignoreReport) {
            if (printIntermediateScore) {
                guessCluster.addToReport("  SVM: ", Cluster.SVM, "\n");
            }
        }

        int cpt = 0;
	int categorie;
        double value0, value1;
        //double lastScore0, lastScore1;
	TraceManager.addDev("Classify with SVM");
        for(ClusterElement ce: guessCluster.getElements()) {
	    categorie = guessCluster.getSVMCategorie(ce);
	    //TraceManager.addDev("Categorie=" + categorie);
	    value0 = (categorie == Cluster.REGULAR) ? 1.0 : 0.0;
	    value1 = (categorie == Cluster.MALWARE) ? 1.0 : 0.0;
            dp = computeScores(ais, value0, value1, cpt);
            /*lastScore0 = computeScoreRegular(ais, value, cpt);
              value = ce.getTotalDeviation(malwareCluster);
              lastScore1 = computeScoreMalware(ais, value, cpt);*/
            addScoreToReport(ce, dp);
            cpt ++;
        }
      }

    // Deviation
    public void computeDeviation(AlligatorInfoScore ais) throws AlligatorSyntaxException {
        ais.stageInfo = 0;
        DoublePoint dp;

        guessCluster.computeStandardDeviationWith(regularCluster);
        guessCluster.computeStandardDeviationWith(malwareCluster);

        if (!ignoreReport) {
            if (printIntermediateScore) {
                guessCluster.addToReport("  Deviation: ", Cluster.DEVIATION, "\n");
            }
        }

        int cpt = 0;
        double value0, value1;
        //double lastScore0, lastScore1;
        for(ClusterElement ce: guessCluster.getElements()) {
            value0 = ce.getTotalDeviation(regularCluster);
            value1 = ce.getTotalDeviation(malwareCluster);
            dp = computeScores(ais, value0, value1, cpt);
            addScoreToReport(ce, dp);
            cpt ++;
        }
    }

    // Inverse deviation
    public void computeInverseDeviation(AlligatorInfoScore ais) throws AlligatorSyntaxException {
        ais.stageInfo = 0;
        DoublePoint dp;

        guessCluster.computeStandardDeviationWith(regularCluster);
        guessCluster.computeStandardDeviationWith(malwareCluster);

        if (!ignoreReport) {
            if (printIntermediateScore) {
                guessCluster.addToReport("  Inverse deviation: ", Cluster.INVERSE_DEVIATION, "\n");
            }
        }

        int cpt = 0;
        double value0, value1;
        //double lastScore0, lastScore1;
        for(ClusterElement ce: guessCluster.getElements()) {
            value0 = ce.getTotalDeviation(regularCluster);
            value1 = ce.getTotalDeviation(malwareCluster);
	    value0 = 1/(value0+0.000000001);
	    value1 = 1/(value1+0.000000001);
            dp = computeScores(ais, value0, value1, cpt);
            addScoreToReport(ce, dp);
            cpt ++;
        }
    }

    // Weight Deviation
    public void computeWeightDeviation(AlligatorInfoScore ais) throws AlligatorSyntaxException {
        ais.stageInfo = 16;

        guessCluster.computeWeightDeviationWith(regularCluster);
        guessCluster.computeWeightDeviationWith(malwareCluster);

        if (!ignoreReport) {
            if (printIntermediateScore) {
                guessCluster.addToReport("  Weight deviation: ", Cluster.WEIGHT_DEVIATION, "\n");
            }
        }

        int cpt = 0;
        double value0, value1;
        DoublePoint dp;
        //double lastScore0, lastScore1;
        for(ClusterElement ce: guessCluster.getElements()) {
            value0 = ce.getTotalWeightDeviation(regularCluster);
            value1 = ce.getTotalWeightDeviation(malwareCluster);
            dp = computeScores(ais, value0, value1, cpt);
            addScoreToReport(ce, dp);
            cpt ++;
        }
    }

    // Inverse weight Deviation
    public void computeInverseWeightDeviation(AlligatorInfoScore ais) throws AlligatorSyntaxException {
        ais.stageInfo = 18;

        guessCluster.computeWeightDeviationWith(regularCluster);
        guessCluster.computeWeightDeviationWith(malwareCluster);

        if (!ignoreReport) {
            if (printIntermediateScore) {
                guessCluster.addToReport("  Inverse weight deviation: ", Cluster.INVERSE_WEIGHT_DEVIATION, "\n");
            }
        }

        int cpt = 0;
        double value0, value1;
        DoublePoint dp;
        //double lastScore0, lastScore1;
        for(ClusterElement ce: guessCluster.getElements()) {
            value0 = ce.getTotalWeightDeviation(regularCluster);
            value1 = ce.getTotalWeightDeviation(malwareCluster);
	    value0 = 1/(value0+0.000000001);
	    value1 = 1/(value1+0.000000001);
            dp = computeScores(ais, value0, value1, cpt);
            addScoreToReport(ce, dp);
            cpt ++;
        }
    }

    // Deviation percentage
    public void computeDeviationPercentage(AlligatorInfoScore ais) throws AlligatorSyntaxException {
        ais.stageInfo = 13;

        guessCluster.computeDeviationPercentageBetween(regularCluster, malwareCluster);

        if (!ignoreReport) {
            if (printIntermediateScore) {
                guessCluster.addToReport("  Deviation Percentage: ", Cluster.DEVIATION_PERCENTAGE, "\n", regularCluster, malwareCluster);
            }
        }

        int cpt = 0;
        //double value;
        //double lastScore0, lastScore1;
        DoublePoint dp;

        for(ClusterElement ce: guessCluster.getElements()) {
            dp = computeScores(ais, ce.lastPercentageDeviationWithCluster1, ce.lastPercentageDeviationWithCluster2, cpt);
            addScoreToReport(ce, dp);
            cpt ++;
        }
    }

    // Deviation weight percentage
    public void computeDeviationWeightPercentage(AlligatorInfoScore ais) throws AlligatorSyntaxException {
        ais.stageInfo = 14;

        guessCluster.computeDeviationWeightPercentageBetween(regularCluster, malwareCluster);

        if (!ignoreReport) {
            if (printIntermediateScore) {
                guessCluster.addToReport("  Deviation Weight Percentage: ", Cluster.DEVIATION_WEIGHT_PERCENTAGE, "\n", regularCluster, malwareCluster);
            }
        }

        int cpt = 0;
        //double value;
        DoublePoint dp;

        for(ClusterElement ce: guessCluster.getElements()) {
            dp = computeScores(ais, ce.lastPercentageDeviationWithCluster1, ce.lastPercentageDeviationWithCluster2, cpt);
            addScoreToReport(ce, dp);
            cpt ++;
        }
        //guessCluster.purgeDeviations();
        //regularCluster.purgeDeviations();
        //malwareCluster.purgeDeviations();

    }

    public void computeProximity(AlligatorInfoScore ais, String[] split) throws AlligatorSyntaxException {
        ais.stageInfo = 1;
        int nb;
        try {
            nb = Integer.valueOf(split[2]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Proximity argument is invalid");
        }

        if(nb<1) {
            throw new AlligatorSyntaxException("Proximity argument must be greater than 0 (currently:: " + split[2] + ")");
        }

        if (nb > proximity) {
            guessCluster.computeClosestElements(nb, clusters);
            proximity = nb;
        }

        guessCluster.setNbOfClosestElements(nb);

        if (!ignoreReport) {
            if (printIntermediateScore) {
		if (printProximityElements) {
		    guessCluster.addToReport("  Proximity " + nb + " elements: ", Cluster.FULL_CLOSEST, "\n");
		} else {
		    guessCluster.addToReport("  Proximity " + nb + " elements: ", Cluster.CLOSEST, "\n");
		}
            }
        }

        int cpt = 0;
        double value0, value1;
        DoublePoint dp;
        for(ClusterElement ce: guessCluster.getElements()) {
            value0 = (100 * ce.getNbOfClosestInCluster(regularCluster)) / nb;
            value1 = (100 * ce.getNbOfClosestInCluster(malwareCluster)) / nb;
            dp = computeScores(ais, value0, value1, cpt);
            addScoreToReport(ce, dp);
            cpt ++;
        }
    }

    public void computeProximityRatio(AlligatorInfoScore ais, String[] split) throws AlligatorSyntaxException {
        ais.stageInfo = 15;
        int nb;
        try {
            nb = Integer.valueOf(split[2]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Proximity argument is invalid");
        }

        if(nb<1) {
            throw new AlligatorSyntaxException("Proximity argument must be greater than 0 (currently:: " + split[2] + ")");
        }

        if (nb > proximity) {
            guessCluster.computeClosestElements(nb, clusters);
            proximity = nb;
        }

        guessCluster.setNbOfClosestElements(nb);

        if (!ignoreReport) {
            if (printIntermediateScore) {
		if (printProximityElements) {
		    guessCluster.addToReport("  Proximity ratio " + nb + " elements: ", Cluster.FULL_CLOSEST, "\n");
		} else {
		    guessCluster.addToReport("  Proximity ratio " + nb + " elements: ", Cluster.CLOSEST, "\n");
		}
            }
        }

        int cpt = 0;
        double value0, value1;
        DoublePoint dp;
	double totalSize = regularCluster.size() + malwareCluster.size();
        for(ClusterElement ce: guessCluster.getElements()) {
            value0 = (100 * ce.getNbOfClosestInCluster(regularCluster)) / nb;
            value1 = (100 * ce.getNbOfClosestInCluster(malwareCluster)) / nb;
	    TraceManager.addDev("Ratio1 value0 =" + value0 + " value1 = " + value1);

	    value0 = value0 * malwareCluster.size() / totalSize;
            value1 = value1 * regularCluster.size() / totalSize;
            dp = computeScores(ais, value0, value1, cpt);
            TraceManager.addDev("Ratio2 value0 =" + value0 + " value1 = " + value1);
            addScoreToReport(ce, dp);
            cpt ++;
        }
    }

    public void computeDegressiveProximity(AlligatorInfoScore ais, String[] split) throws AlligatorSyntaxException {
        ais.stageInfo = 13;
        int nb;
        try {
            nb = Integer.valueOf(split[2]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Degressive proximity argument is invalid");
        }

        if(nb<2) {
            throw new AlligatorSyntaxException("Degressive proximity argument must be greater than 1 (currently:: " + split[2] + ")");
        }

        if (nb > proximity) {
            guessCluster.computeClosestElements(nb, clusters);
            proximity = nb;
        }

        guessCluster.setNbOfClosestElements(nb);

        if (!ignoreReport) {
            if (printIntermediateScore) {
		if (printProximityElements) {
		    guessCluster.addToReport("  Degressive proximity " + nb + " elements: ", Cluster.FULL_CLOSEST, "\n");
		} else {
		    guessCluster.addToReport("  Degressive proximity " + nb + " elements: ", Cluster.CLOSEST, "\n");
		}
            }
        }

        int cpt = 0;
        double value0, value1;
        DoublePoint dp;
        // y = a*x*x + c
        double a = (1-nb) / (nb*nb - 1);
        double c  = nb - a;
        for(ClusterElement ce: guessCluster.getElements()) {
            value0 = ce.getClosestParabolicInCluster(regularCluster, a, c) / nb;
            value1 = ce.getClosestParabolicInCluster(malwareCluster, a, c) / nb;
            dp = computeScores(ais, value0, value1, cpt);
            //TraceManager.addDev("nb=" + nb + " found = " + ce.getNbOfClosestInCluster(malwareCluster) + " lastScore 1= " + lastScore1 + " original value=" + value);
            addScoreToReport(ce, dp);
            cpt ++;
        }

    }


    public void computeProximityWithLimitedProperties(AlligatorInfoScore ais, String[] split) throws AlligatorSyntaxException {
        ais.stageInfo = 2;

        int nb;
        int nbOfProperties;
        try {
            nb = Integer.valueOf(split[2]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Proximity argument is invalid");
        }

        if(nb<1) {
            throw new AlligatorSyntaxException("Nb of elements must be greater than 0 (currently:: " + split[2] + ")");
        }

        try {
            nbOfProperties = Integer.valueOf(split[3]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Property argument is invalid");
        }

        if(nbOfProperties<0) {
            throw new AlligatorSyntaxException("Property argument must be greater than 0 (currently: " + split[3] + ")");
        }

        if (nbOfProperties > regularCluster.getTotalNbOfProperties()) {
            throw new AlligatorSyntaxException("Property argument must be less than " + regularCluster.getTotalNbOfProperties() + "(currently:" + split[3] + ")");
        }

        if ((nb > proximity) || (nbOfLimitedProperties != nbOfProperties)) {
            //TraceManager.addDev("Nb Of properties = " + nbOfProperties);
            guessCluster.computeClosestElementsWithProperties(nb, nbOfProperties, clusters);
            proximity = nb;
            nbOfLimitedProperties = nbOfProperties;
        }

        guessCluster.setNbOfClosestElements(nb);
        if (!ignoreReport) {
            if (printIntermediateScore) {
		if (printProximityElements) {
		    guessCluster.addToReport("  Proximity " + nb + " elements & " + nbOfProperties + " properties: ", Cluster.FULL_CLOSEST, "\n");
		} else {
		    guessCluster.addToReport("  Proximity " + nb + " elements & " + nbOfProperties + " properties: ", Cluster.CLOSEST, "\n");
		}
            }
        }

        int cpt = 0;
        double value0, value1;
        DoublePoint dp;
        for(ClusterElement ce: guessCluster.getElements()) {
            value0 = 100 * ce.getNbOfClosestInCluster(regularCluster) / nb;
            value1 = 100 * ce.getNbOfClosestInCluster(malwareCluster) / nb;
            dp = computeScores(ais, value0, value1, cpt);
            addScoreToReport(ce, dp);
            cpt ++;
        }
    }

    public void computeProximityWithLimitedProperties_FirstVersion(AlligatorInfoScore ais, String[] split) throws AlligatorSyntaxException {
        ais.stageInfo = 14;

        int nb;
        int nbOfProperties;
        try {
            nb = Integer.valueOf(split[2]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Proximity argument is invalid");
        }

        if(nb<1) {
            throw new AlligatorSyntaxException("Nb of elements must be greater than 0 (currently:: " + split[2] + ")");
        }

        try {
            nbOfProperties = Integer.valueOf(split[3]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Property argument is invalid");
        }

        if(nbOfProperties<0) {
            throw new AlligatorSyntaxException("Property argument must be greater than 0 (currently: " + split[3] + ")");
        }

        if (nbOfProperties > regularCluster.getTotalNbOfProperties()) {
            throw new AlligatorSyntaxException("Property argument must be less than " + regularCluster.getTotalNbOfProperties() + "(currently:" + split[3] + ")");
        }

        if ((nb > proximity) || (nbOfLimitedProperties != nbOfProperties)) {
            //TraceManager.addDev("Nb Of properties = " + nbOfProperties);
            guessCluster.computeClosestElementsWithProperties_FirstVersion(nb, nbOfProperties, clusters);
            proximity = nb;
            nbOfLimitedProperties = nbOfProperties;
        }

        guessCluster.setNbOfClosestElements(nb);
        if (!ignoreReport) {
            if (printIntermediateScore) {
                guessCluster.addToReport("  Proximity " + nb + " elements & " + nbOfProperties + " properties: ", Cluster.CLOSEST, "\n");
            }
        }

        int cpt = 0;
        double value0, value1;
        DoublePoint dp;
        for(ClusterElement ce: guessCluster.getElements()) {
            value0 = 100 * ce.getNbOfClosestInCluster(regularCluster) / nb;
            value1 = 100 * ce.getNbOfClosestInCluster(malwareCluster) / nb;
            dp = computeScores(ais, value0, value1, cpt);
            addScoreToReport(ce, dp);
            cpt ++;
        }
    }

    // Correlations
    public void computeCorrelations(AlligatorInfoScore ais, String []split)  throws AlligatorSyntaxException  {
        ais.stageInfo = 3;

        double d;
        try {
            d = Double.valueOf(split[2]).doubleValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Invalid correlation threshold: " + split[2]);
        }

        if ((d<=0) || (d>=1)) {
            throw new AlligatorSyntaxException("Invalid correlation threshold: must be between 0 and 1 " + split[2]);
        }

        //guessCluster.setInfoOnCurrentComputation("Computing correlations in regular");
        pd.setPercentageProvider(regularCluster);
        regularCluster.computeCorrelations();
        regularCluster.setCorrelationsThreshold(d);
        TraceManager.addDev("Regular: computing list of interesting correlations");
        regularCluster.computeListOfInterestingCorrelations();

        //guessCluster.setInfoOnCurrentComputation("Computing correlations in malware");
        pd.setPercentageProvider(malwareCluster);
        malwareCluster.computeCorrelations();
        malwareCluster.setCorrelationsThreshold(d);
        TraceManager.addDev("Malware: computing list of interesting correlations");
        malwareCluster.computeListOfInterestingCorrelations();

        pd.setPercentageProvider(guessCluster);

        TraceManager.addDev("Printing list of interesting correlations");
        TraceManager.addDev(regularCluster.getName() + " Interesting correlations:\n" + regularCluster.getInterestingCorrelations() + "\n");
        TraceManager.addDev(malwareCluster.getName() + " Interesting correlations:\n" + malwareCluster.getInterestingCorrelations() + "\n");

        guessCluster.addInterestingCorrelation(regularCluster.getListOfInterestingCorrelations());
        guessCluster.addInterestingCorrelation(malwareCluster.getListOfInterestingCorrelations());
        guessCluster.removeNonExclusiveCorrelations();

        if (!ignoreReport) {
            if (printIntermediateScore) {
                guessCluster.addToReport("  Correlations: ", Cluster.CORRELATION, "\n");
            }
        }

        int cpt = 0;
        double value0, value1;
        DoublePoint dp;
        for(ClusterElement ce: guessCluster.getElements()) {
            value0 = ce.getPercentageOfCorrelations(regularCluster);
            value1 = ce.getPercentageOfCorrelations(malwareCluster);
            dp = computeScores(ais, value0, value1, cpt);
            addScoreToReport(ce, dp);
            cpt ++;
        }

    }


    // Closest average per property
    public void closestAverageOnEachProperty(AlligatorInfoScore ais, String []split)  throws AlligatorSyntaxException  {
        ais.stageInfo = 12;

        if (split.length > 2) {
            throw new AlligatorSyntaxException("No attribute must be provided with " + split[1]);
        }

        regularCluster.computeMeanPerProperties();
        malwareCluster.computeMeanPerProperties();

        int cpt = 0;
        double value1, value2;
        DoublePoint dp;
        Point p;
        for(ClusterElement ce: guessCluster.getElements()) {
            // Nb of properties closest to the average in each cluster
            p = Cluster.nbOfClosestToAveragePerProperty(ce, regularCluster, malwareCluster);
            ce.distanceToAverage = p;
        }
        if (!ignoreReport) {
            if (printIntermediateScore) {
                guessCluster.addToReport("  Distance to average: ", Cluster.DISTANCE_TO_AVERAGE, "\n", regularCluster, malwareCluster);
            }
        }
        for(ClusterElement ce: guessCluster.getElements()) {
            p = ce.distanceToAverage;
            //TraceManager.addDev("p.x=" + p.x + " p.y=" +p.y);
            value1 = ((double)p.x) / regularCluster.getCurrentNbOfProperties() * 100 ;
            value2 = ((double)p.y) / regularCluster.getCurrentNbOfProperties() * 100 ;
            dp = computeScores(ais, value1, value2, cpt);
            addScoreToReport(ce, dp);
            cpt ++;
        }


    }

    public void addValueToScore(AlligatorInfoScore ais, String []split)  throws AlligatorSyntaxException  {
        ais.stageInfo = 11;

        Integer param;
        try {
            param = Integer.valueOf(split[2]).intValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Invalid value to add: must provide an int value");
        }

        if (param<=0) {
            throw new AlligatorSyntaxException("Invalid value to add :"+ split[2] + ". Must be >0");
        }



        if (!ignoreReport) {
            if (printIntermediateScore) {
                guessCluster.addToReport("  Value to add: " + param, Cluster.VALUE_TO_ADD, "\n");
            }
        }

        int cpt = 0;
        double value0, value1;
        DoublePoint dp;
        for(ClusterElement ce: guessCluster.getElements()) {
            value0 = computeScoreRegular(ais, param, cpt);
            value1 = computeScoreMalware(ais, param, cpt);
            addScoreToReport(ce, new DoublePoint(value0, value1));
            cpt ++;
        }

    }

    public void computeProbabilityDifference(AlligatorInfoScore ais, String[] split) throws AlligatorSyntaxException {
        ais.stageInfo = 4;
        double nb;
        try {
            nb = Double.valueOf(split[2]).doubleValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Probability argument is invalid");
        }

        if(nb>=1||nb<=0) {
            throw new AlligatorSyntaxException("Probability argument must be between 0 and 1 (currently: " + split[2] + ")");
        }

        
	//regularCluster.computeMostCommonValueAndProbabilities();
	//malwareCluster.computeMostCommonValueAndProbabilities();

	regularCluster.computeSortedPropertyValues();
	malwareCluster.computeSortedPropertyValues();
        	
        TraceManager.addDev("PROBA DIFFERENCE");

	if (regularCluster.cd.sortedPropertyValues == null) {
	    TraceManager.addDev("NULL sorted in regular");
	} else {
	    TraceManager.addDev("regular: size of sortedPropertyValues: " + regularCluster.cd.sortedPropertyValues.length);
	}

	if (malwareCluster.cd.sortedPropertyValues == null) {
	    TraceManager.addDev("NULL sorted in malware");
	} else {
	    TraceManager.addDev("malware: size of sortedPropertyValues: " + malwareCluster.cd.sortedPropertyValues.length);
	}

        String p = regularCluster.getStringProbabilities(malwareCluster);
        TraceManager.addDev("--- proba\n" + p);
        //p = malwareCluster.getStringProbabilities();
        //TraceManager.addDev("--- malware proba:\n" + p);

        regularCluster.setProbabilityDifference(nb);
        malwareCluster.setProbabilityDifference(nb);
        guessCluster.setProbabilityDifference(nb);

        String regProp = regularCluster.computeStringAndTotalProbaDifferenceWithCluster(malwareCluster);
        String malProp = malwareCluster.computeStringAndTotalProbaDifferenceWithCluster(regularCluster);

        //guessCluster.clearProbabilityDifferences();
        //guessCluster.addAllProbabilityDifferences(regularCluster.getProbabilityDifferences());
        //guessCluster.addAllProbabilityDifferences(malwareCluster.getProbabilityDifferences());
        if (!ignoreReport) {
            if (printIntermediateScore) {
                guessCluster.addToReport("  Properties with >= " + nb + " probability difference in a given value between " + regularCluster.getName() +  " and " + malwareCluster.getName() + "; " + regularCluster.getTotalProbabilityDifference() +
                                         " properties in " + regularCluster.getName() +  ": (" + regProp + "); " + malwareCluster.getTotalProbabilityDifference() + " properties in " + malwareCluster.getName() + ": (" + malProp + ")", Cluster.PROBA_DIFFERENCE,"\n", regularCluster, malwareCluster);
            }
        }

        int cpt = 0;
        //double value;
        DoublePoint dp;
        DoublePoint dps;
        for(ClusterElement ce: guessCluster.getElements()) {
            dp = ce.getPercentageProbaDifference(regularCluster, malwareCluster);
            dps = computeScores(ais, dp.getX(), dp.getY(), cpt);
            addScoreToReport(ce, dps);
            cpt ++;
        }

    }

    public void computeProbabilityFactor(AlligatorInfoScore ais, String[] split) throws AlligatorSyntaxException {
        ais.stageInfo = 5;
        double nb;
        try {
            nb = Double.valueOf(split[2]).doubleValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Probability argument is invalid");
        }

        if(nb<=0) {
            throw new AlligatorSyntaxException("Probability argument must be greater than 0 (currently: " + split[2] + ")");
        }

	//regularCluster.computeMostCommonValueAndProbabilities();
	//malwareCluster.computeMostCommonValueAndProbabilities();
	
	regularCluster.computeSortedPropertyValues();
	malwareCluster.computeSortedPropertyValues();
	
        //TraceManager.addDev("Regular property values:\n" + regularCluster.printSortedPropertyValues());
        //TraceManager.addDev("Malware property values:\n" + malwareCluster.printSortedPropertyValues());

        guessCluster.setProbabilityFactor(nb);
        regularCluster.setProbabilityFactor(nb);
        malwareCluster.setProbabilityFactor(nb);

        regularCluster.computeTotalProbaFactorWithCluster(malwareCluster);
        malwareCluster.computeTotalProbaFactorWithCluster(regularCluster);

        String regProp = regularCluster.computeStringAndTotalProbaFactorWithCluster(malwareCluster);
        String malProp = malwareCluster.computeStringAndTotalProbaFactorWithCluster(regularCluster);

        //guessCluster.addToReport("  Properties with at least a factor of " + nb + " between property values probabilities ;"+ regularCluster.getTotalProbabilityFactor() + " valid factor differences in regular, " + malwareCluster.getTotalProbabilityFactor() + " valid factor differences in malware ", Cluster.PROBA_FACTOR,"\n", regularCluster, malwareCluster);
        if (!ignoreReport) {
            if (printIntermediateScore) {
                guessCluster.addToReport("  Properties with >= " + nb + " probability factor in a given value between " + regularCluster.getName() + " and " +
                                         malwareCluster.getName() + "; " + regularCluster.getTotalProbabilityFactor() +
                                         " properties in " + regularCluster.getName() + ": (" + regProp + "); " + malwareCluster.getTotalProbabilityFactor() +
                                         " properties in " + malwareCluster.getName() + ": (" + malProp + ")", Cluster.PROBA_DIFFERENCE,"\n", regularCluster, malwareCluster);

            }
        }

        int cpt = 0;
        //double value;
        double lastScore0, lastScore1;
        DoublePoint dp;
        DoublePoint dps;
        for(ClusterElement ce: guessCluster.getElements()) {
            dp = ce.getPercentageProbaFactorSummary(regularCluster, malwareCluster);
            dps = computeScores(ais, dp.getX(), dp.getY(), cpt);
            addScoreToReport(ce, dps);
            cpt ++;
        }

    }

    public void computeEpsilonClusters(AlligatorInfoScore ais, String[] split) throws AlligatorSyntaxException {
        ais.stageInfo = 6;
        double nb;

        pd.setMonitor(true);
        try {
            nb = Double.valueOf(split[2]).doubleValue();
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Distance threshold argument is invalid");
        }

        if(nb<=0) {
            throw new AlligatorSyntaxException("Distance threshold argument must be greater than 0 (currently: " + split[2] + ")");
        }

	if (regularCluster.cd.masterEpsilonClusters != null) {
	    guessCluster.cd.masterEpsilonClusters = regularCluster.cd.masterEpsilonClusters;
	}

        guessCluster.clearEpsilonClusters();
        guessCluster.loadEpsilonClusters(nb);
        if (guessCluster.isNullEpsilonCluster()) {
            guessCluster.addAllToEpsilonClusters(nb, regularCluster, true);
            guessCluster.addAllToEpsilonClusters(nb, malwareCluster, false);
            guessCluster.saveEpsilonClusters(nb);
	    regularCluster.cd.masterEpsilonClusters = guessCluster.cd.masterEpsilonClusters;
	    malwareCluster.cd.masterEpsilonClusters = guessCluster.cd.masterEpsilonClusters;
        }


        //TraceManager.addDev("Epsilon clusters: " + guessCluster.getSummaryOnEpsilonClusters());

        if (!ignoreReport) {
            if (printIntermediateScore) {
                guessCluster.addToReport("  Repartition of elements of " + regularCluster.getName() + " and " + malwareCluster.getName() + " being in the sample epsilon cluster,  internal distance < " + guessCluster.getEpsilonClusterThreshold() + "; " , Cluster.DISTANCE_EPSILON_CLUSTER,"\n", regularCluster, malwareCluster);
            }
        }

        int cpt = 0;
        DoublePoint dp;
        DoublePoint dps;

        int total = guessCluster.getElements().size();

        for(ClusterElement ce: guessCluster.getElements()) {
            if ((cpt % 10) == 0) {
                guessCluster.setPercentage((cpt * 100) / total);
            }
            dp = ce.getPercentageNeighboursInEpsilonCluster(regularCluster, malwareCluster);
            dps = computeScores(ais, dp.getX(), dp.getY(), cpt);
            addScoreToReport(ce, dps);
            cpt ++;
        }

        guessCluster.resetInfoOnComputation();

        pd.setMonitor(false);
    }

    public DoublePoint normalizeScore(double score1, double score2) {
	if (score1 == 0) {
	    if (score2 == 0) {
		return new DoublePoint(score1, score2);
	    }
	    return new DoublePoint(score1, 1);
	}

	if (score2 == 0) {
	    return new DoublePoint(1, score2);
	}

	// score1 and score2 are not equal to zero -> can compute the inverse
	if (score1 > score2) {
	    score2 = score2 / score1;
	    score1 = 1;
	    return new DoublePoint(score1, score2);
	}
	
	return new DoublePoint(score1/score2, 1);
    }

    public DoublePoint computeScores(AlligatorInfoScore ais, double valueRegular, double valueMalware, int index) throws AlligatorSyntaxException {

	// Normalize scores
	// The highest score should be equal to 1. 
	DoublePoint dp = normalizeScore(valueRegular, valueMalware);

        // Value Regular
        AlligatorScore asR = new AlligatorRegularScore(ais, dp.x);
        if (regularScores[index] == null) {
            regularScores[index] = new Vector<AlligatorScore>();
        }
        regularScores[index].add(asR);

        // Value Malware
        AlligatorScore asM = new AlligatorMalwareScore(ais, dp.y);
        if (malwareScores[index] == null) {
            malwareScores[index] = new Vector<AlligatorScore>();
        }
        malwareScores[index].add(asM);

        // Computing complex values
        asR.computeComplexExpression();
        asM.computeComplexExpression();

        if (basicScore) {
            if (asR.intermediateScore > asM.intermediateScore) {
                asR.intermediateScore = 1;
                asM.intermediateScore = 0;
            } else if (asR.intermediateScore < asM.intermediateScore) {
                asM.intermediateScore = 1;
                asR.intermediateScore = 0;
            } else {
                asM.intermediateScore = 0;
                asR.intermediateScore = 0;
            }
        }

        asR.computeAverageScore();
        asM.computeAverageScore();

        DoublePoint dpp = new DoublePoint(asR.intermediateScore, asM.intermediateScore);
        dpp.addExtra(asR.lastScore, asM.lastScore);

        return dpp;

    }

    public double computeScoreRegular(AlligatorInfoScore ais, double value, int index) throws AlligatorSyntaxException {
        AlligatorScore as = new AlligatorRegularScore(ais, value);
        if (regularScores[index] == null) {
            regularScores[index] = new Vector<AlligatorScore>();
        }
        regularScores[index].add(as);

        as.computeComplexExpression();
        as.computeAverageScore();

        return as.lastScore;
    }

    public double computeScoreMalware(AlligatorInfoScore ais, double value, int index) throws AlligatorSyntaxException {
        AlligatorScore as = new AlligatorMalwareScore(ais, value);
        if (malwareScores[index] == null) {
            malwareScores[index] = new Vector<AlligatorScore>();
        }
        malwareScores[index].add(as);

        as.computeComplexExpression();
        as.computeAverageScore();

        return as.lastScore;
    }


    public void addScoreToReport(ClusterElement ce, DoublePoint dp) {
        if ((printIntermediateScore) && (!ignoreReport)){
            String r = "    Raw score (" + regularCluster.getName() + ", " + malwareCluster.getName() + "): " + dp.getX()+", " + dp.getY();
            DoublePoint extra = dp.getExtra();
            if (extra != null) {
                r += "\tMultiplied score: " + extra.getX() + ", " + extra.getY();
            }

            ce.addToReport(r + "\n");
        }
    }

    public void computeScores(int indexOfClusterElement) {

        // Regular
        regularScore[indexOfClusterElement] = 0;
        for(AlligatorScore as: regularScores[indexOfClusterElement]) {
            regularScore[indexOfClusterElement] += as.lastScore;
        }

        // Malware
        malwareScore[indexOfClusterElement] = 0;
        for(AlligatorScore as: malwareScores[indexOfClusterElement]) {
            malwareScore[indexOfClusterElement] += as.lastScore;
        }
    }

    public void printResultSummary() {
        TraceManager.addDev("Printing result summary");
        int cpt = 0;
        int malware = 0, regular = 0;
        int lightMalware = 0, mediumMalware = 0, strongMalware = 0;
        int lightRegular = 0, mediumRegular = 0, strongRegular = 0;
        String score;
        double diff;
        double perc1, perc2, perc3;
        String sperc1, sperc2, sperc3;
        StringBuffer overallRes = new StringBuffer();
        guessCluster.addToTopReport("\nResults summary:");

        // Computing and displaying overall results

        for(ClusterElement ce: guessCluster.getElements()) {
            //computeScores(cpt);

            if (regularScore[cpt] < malwareScore[cpt]) {
                score = "*" + malwareCluster.getName() + "*";
                malware ++;

                if (diffValueMalware > 0) {
                    diff = malwareScore[cpt] - regularScore[cpt];
                    if (diff < (diffValueMalware/3)) {
                        lightMalware ++;
                    } else if ((2*diff)<(2*diffValueMalware/3)) {
                        mediumMalware ++;
                    } else {
                        strongMalware++;
                    }
                }

            }  else {
                score = " " + regularCluster.getName() + " ";
                regular++;

                if (diffValueRegular > 0) {
                    diff = regularScore[cpt] - malwareScore[cpt];
                    if (diff < (diffValueRegular/3)) {
                        lightRegular ++;
                    } else if ((diff)<(2*diffValueRegular/3)) {
                        mediumRegular ++;
                    } else {
                        strongRegular++;
                    }
                }
            }
            overallRes.append(ce.getName() + ": " + score + "\t(" + regularCluster.getName() + ":" + regularScore[cpt] + ", " + malwareCluster.getName()  + ":" + malwareScore[cpt] + ")\n");

            cpt ++;
        }
        int total = malware + regular;
        guessCluster.addToTopReport(regular + " " + regularCluster.getName() + "(s) found, " + malware + " " + malwareCluster.getName() + "(s) found in guess");
        guessCluster.addToTopReport("Percentage of " + regularCluster.getName() + ": " + (((double)(regular))/(total)*100));
        guessCluster.addToTopReport("Percentage of " + malwareCluster.getName() + ": " + (((double)(malware))/(total)*100));

        //guessCluster.addToTopReport("diff regular:" + diffValueRegular + " malware: " + diffValueMalware);
        DecimalFormat formatter = new DecimalFormat("#0.00");

        if (diffValueRegular > 0) {
            perc1 = (((double)(lightRegular))/total*100);
            perc2 = (((double)(mediumRegular))/total*100);
            perc3 = (((double)(strongRegular))/total*100);
            sperc1 = formatter.format(perc1);
            sperc2 = formatter.format(perc2);
            sperc3 = formatter.format(perc3);

            guessCluster.addToTopReport(regularCluster.getName() + ": Light:" + lightRegular + " (" + sperc1 + "%)  Medium:" + mediumRegular + " (" + sperc2 + "%)  Strong:" + strongRegular + " (" + sperc3 + "%)");
        }

        if (diffValueMalware > 0) {
            perc1 = (((double)(lightMalware))/total*100);
            perc2 = (((double)(mediumMalware))/total*100);
            perc3 = (((double)(strongMalware))/total*100);
            sperc1 = formatter.format(perc1);
            sperc2 = formatter.format(perc2);
            sperc3 = formatter.format(perc3);
            guessCluster.addToTopReport(malwareCluster.getName() + ": Light:" + lightMalware + " (" + sperc1 + "%)  Medium:" + mediumMalware + " (" + sperc2 + "%)  Strong:" + strongMalware + " (" + sperc3 + "%)");
        }

        guessCluster.addToTopReport(overallRes.toString());
    }

    public void computeOverallResults() {
        TraceManager.addDev("Computing overall results");
        int cpt = 0;

        // Computing  overall results

        for(cpt=0; cpt<guessCluster.getElements().size(); cpt++) {
            computeScores(cpt);
        }

    }



    public void setOptions(String []split, boolean checkSyntaxOnly) throws AlligatorSyntaxException {
        String option = split[1];

        if (!(isOption(0, option))) {
            throw new AlligatorSyntaxException("Invalid option: " + option);
        }

        if (checkSyntaxOnly) {
            return;
        }

        printIntermediateScore = true;
        TraceManager.addDev("----------------------------------- Setting intermediate score");

    }

    public boolean isOption(int index, String option) {
        return (OPTIONS[index].equals(option.toLowerCase()));
    }


    /* Methods added for learning */
    public void computeAllValues() {
        for(int i=0; i<regularScores.length; i++) {
            for(AlligatorScore as: regularScores[i]) {
                as.computeAllPossibleScores();
            }
            for(AlligatorScore as: malwareScores[i]) {
                as.computeAllPossibleScores();
            }
        }
    }

    public int getIteration() {
        return iteration;
    }

    public int getNbOfSteps() {
        return regularScores[0].size();
    }

    public int getNbOfValuesForStep(int index, boolean isRegular) {
        if (isRegular) {
            return regularScores[0].get(index).nbOfScores;
        } else {
            return malwareScores[0].get(index).nbOfScores;
        }
    }

    public double getMinMultiplierForStep(int index, boolean isRegular) {
        if (isRegular) {
            return regularScores[0].get(index).ais.regularMinMultiplier;
        } else {
            return malwareScores[0].get(index).ais.malwareMinMultiplier;
        }
    }

    public double getMaxMultiplierForStep(int index, boolean isRegular) {
        if (isRegular) {
            return regularScores[0].get(index).ais.regularMaxMultiplier;
        } else {
            return malwareScores[0].get(index).ais.malwareMaxMultiplier;
        }
    }

    // Score (regular, malware) for a given sample and for a given evaluation set
    public double[] getScoreOf(AlligatorEvaluationSet aes, int index) {
        double[] scores = new double[2];
        scores[0] = 0;
        scores[1] = 0;

        for(AlligatorScore as: regularScores[index]) {
            scores[0] += as.getScore(aes.getIndex(index));
        }

        for(AlligatorScore as: malwareScores[index]) {
            scores[1] += as.getScore(aes.getIndex(index));
        }

        return scores;
    }


    public void computeOne(AlligatorEvaluationSet aes, boolean isRegular) {
        double scoreMalware = 0;
        double scoreRegular = 0;
        int index;
        aes.average = 0;
        //aes.sigma = 0;
        aes.minDiff = Double.MAX_VALUE;
        aes.maxDiff = 0;
        double tmp;
        double scoreDiff;

        if (classificationWeight == null) {
            prepareAdvancedLearning();
        }

        for(int i=0; i<regularScores.length; i++) {
            scoreRegular = 0;
            scoreMalware = 0;
            index = 0;
            if (classificationWeight[i] > 0) {
                for(AlligatorScore as: regularScores[i]) {
                    tmp = scoreRegular;
                    scoreRegular += as.getScore(aes.getIndex(index));
                    if (scoreRegular == Double.POSITIVE_INFINITY) {
                        //TraceManager.addDev("*** ScoreRegular=" + scoreRegular + " intermediate=" + as.getScore(aes.getIndex(index)) + " tmp=" + tmp);
                    }
                    index ++;
                }
                for(AlligatorScore as: malwareScores[i]) {
                    scoreMalware += as.getScore(aes.getIndex(index));
                    index ++;
                }
                //TraceManager.addDev("ScoreMalware=" + scoreMalware + " scoreRegular=" + scoreRegular);
                scoreDiff = Math.abs(scoreRegular-scoreMalware);

                //scoreInRegular[i] = scoreRegular;
                //scoreInMalware[i] = scoreMalware;

                if (isRegular) {
                    if (scoreRegular > scoreMalware) {
                        aes.nbOfRegularRecognized  += classificationWeight[i];
                        aes.minDiff = Math.min(scoreDiff, aes.minDiff);

                        aes.average += scoreDiff;
                        //aes.sigma += (((scoreDiff)*(scoreDiff))/regularScores.length);
                    } else {
                        aes.nbOfRegularFailed  += classificationWeight[i];
                        aes.maxDiff = Math.max(scoreDiff, aes.maxDiff);
                    }

                    if (aes.maxRegularDiff < (scoreRegular - scoreMalware)) {
                        aes.maxRegularDiff = scoreRegular - scoreMalware;
                    }
                } else {
                    if (scoreRegular >= scoreMalware) {
                        aes.nbOfMalwareFailed += classificationWeight[i];
                        aes.maxDiff = Math.max(scoreDiff, aes.maxDiff);
                    } else {
                        aes.nbOfMalwareRecognized += classificationWeight[i];
                        aes.minDiff = Math.min(scoreDiff, aes.minDiff);
                        aes.average += scoreDiff;
                        //aes.sigma += (((scoreDiff)*(scoreDiff))/regularScores.length);
                    }

                    if (aes.maxMalwareDiff < (scoreMalware - scoreRegular)) {
                        aes.maxMalwareDiff = scoreMalware - scoreRegular;
                    }
                }
            }
        }
        //TraceManager.addDev("sigma=" + aes.sigma + " average=" + aes.average + " length=" + regularScores.length);
        int totalNb = aes.nbOfMalwareRecognized + aes.nbOfRegularRecognized;
        aes.average = aes.average / totalNb;
        //aes.sigma = (aes.sigma / regularScores.length) - (aes.average*aes.average);
        //TraceManager.addDev("sigma=" + aes.sigma + " average=" + aes.average);
        //aes.sigma = aes.sigma - (aes.average*aes.average);
        //TraceManager.addDev("sigma=" + aes.sigma + " average=" + aes.average);
    }

    public void computeSet(Vector<AlligatorEvaluationSet> set, boolean isRegular) {

        double scoreMalware = 0;
        double scoreRegular = 0;
        iteration = 0;
        int index;
        double scoreDiff;

        if (classificationWeight == null) {
            prepareAdvancedLearning();
        }

        for(AlligatorEvaluationSet aes: set) {
            aes.average = 0;
            //aes.sigma = 0;
            aes.minDiff = Double.MAX_VALUE;
            for(int i=0; i<regularScores.length; i++) {
                if (classificationWeight[i] > 0) {
                    scoreRegular = 0;
                    scoreMalware = 0;
                    index = 0;
                    for(AlligatorScore as: regularScores[i]) {
                        scoreRegular += as.getScore(aes.getIndex(index));
                        index ++;
                    }
                    for(AlligatorScore as: malwareScores[i]) {
                        scoreMalware += as.getScore(aes.getIndex(index));
                        index ++;
                    }
                    scoreDiff = Math.abs(scoreRegular-scoreMalware);

                    if (isRegular) {
                        if (scoreRegular > scoreMalware) {
                            aes.nbOfRegularRecognized += classificationWeight[i];
                            aes.minDiff = Math.min(scoreDiff, aes.minDiff);
                            aes.average += scoreDiff;
                            //aes.sigma += (((scoreDiff)*(scoreDiff))/regularScores.length);
                            //TraceManager.addDev("min diff=" + aes.minDiff);
                        } else {
                            aes.nbOfRegularFailed += classificationWeight[i];
                            aes.maxDiff = Math.max(scoreDiff, aes.maxDiff);
                        }

                        if (aes.maxRegularDiff < (scoreRegular - scoreMalware)) {
                            aes.maxRegularDiff = scoreRegular - scoreMalware;
                        }

                    } else {
                        if (scoreRegular >= scoreMalware) {
                            aes.nbOfMalwareFailed += classificationWeight[i];
                            aes.maxDiff = Math.max(scoreDiff, aes.maxDiff);
                        } else {
                            aes.nbOfMalwareRecognized += classificationWeight[i];
                            aes.minDiff = Math.min(scoreDiff, aes.minDiff);
                            aes.average += scoreDiff;
                            //aes.sigma += (((scoreDiff)*(scoreDiff))/regularScores.length);
                        }


                        if (aes.maxMalwareDiff < (scoreMalware - scoreRegular)) {
                            aes.maxMalwareDiff = scoreMalware - scoreRegular;
                        }
                    }
                }
            }
            iteration ++;
            aes.average = aes.average / regularScores.length;
            //aes.sigma = (aes.sigma / regularScores.length) - (aes.average*aes.average);
        }

    }

    public String getRegularComplexExpression(int indexOfScore) {
        if (indexOfScore >= regularScores[0].size()) {
            return null;
        }
        return regularScores[0].get(indexOfScore).ais.regularComplexExpression;
    }

    public String getMalwareComplexExpression(int indexOfScore) {
        if (indexOfScore >= malwareScores[0].size()) {
            return null;
        }
        return malwareScores[0].get(indexOfScore).ais.malwareComplexExpression;
    }

    public double getRegularMultiplier(int indexOfScore, int indexOfMultiplier) {
        if (indexOfScore >= regularScores[0].size()) {
            return -1.0;
        }
        return regularScores[0].get(indexOfScore).getRegularMultiplier(indexOfMultiplier);
    }

    public double getMalwareMultiplier(int indexOfScore, int indexOfMultiplier) {
        if (indexOfScore >= malwareScores[0].size()) {
            return -1.0;
        }
        return malwareScores[0].get(indexOfScore).getMalwareMultiplier(indexOfMultiplier);
    }

    public int getRegularAverageIndex(int indexOfScore) {
        return regularScores[0].get(indexOfScore).getAverageIndex();
    }

    public int getMalwareAverageIndex(int indexOfScore) {
        return malwareScores[0].get(indexOfScore).getAverageIndex();
    }

    public void reinitOverallScore() {
        regularScore = new double[guessCluster.getElements().size()];
        malwareScore = new double[guessCluster.getElements().size()];
    }

    public void addRegularToScore(int step, int index) {
        for(int i=0; i<regularScores.length; i++) {
            //for(AlligatorScore as: regularScores[i]) {
            regularScore[i] += regularScores[i].get(step).allPossibleScores[index];
            //}
        }
    }

    public void removeRegularToScore(int step, int index) {
        for(int i=0; i<regularScores.length; i++) {
            //for(AlligatorScore as: regularScores[i]) {
            regularScore[i] -= regularScores[i].get(step).allPossibleScores[index];
            //}
        }
    }

    public void addMalwareToScore(int step, int index) {
        for(int i=0; i<malwareScores.length; i++) {
            //for(AlligatorScore as: malwareScores[i]) {
            malwareScore[i] += malwareScores[i].get(step).allPossibleScores[index];
            //}
        }
    }

    public void removeMalwareToScore(int step, int index) {
        for(int i=0; i<malwareScores.length; i++) {
            //for(AlligatorScore as: malwareScores[i]) {
            malwareScore[i] -= malwareScores[i].get(step).allPossibleScores[index];
            //}
        }
    }

    public void fillAlligatorEvaluationSet(AlligatorEvaluationSet set, boolean isRegular) {
        for(int i=0; i<regularScore.length; i++) {
            if (isRegular) {
                if (regularScore[i] > malwareScore[i]) {
                    set.nbOfRegularRecognized += classificationWeight[i];
                } else {
                    set.nbOfRegularFailed += classificationWeight[i];
                }
            } else {
                if (regularScore[i] >= malwareScore[i]) {
                    set.nbOfMalwareFailed += classificationWeight[i];
                } else {
                    set.nbOfMalwareRecognized += classificationWeight[i];
                }
            }
        }
    }

    public void executePlot(String []split, boolean checkSyntaxOnly) throws AlligatorSyntaxException {
        // Extract the metrics to plot and the file name
        String metric = split[1];

        TraceManager.addDev("Working with plot metric: >" + metric + "< plot metrics 1=>" + PLOT_METRICS[1] + "<");

        String s = "";

        if (split.length < 3) {
            throw new AlligatorSyntaxException("Invalid number of arguments in plot command");
        }

        pd.setMonitor(true);


        if (metric.toLowerCase().equals(PLOT_METRICS[0])) {
            String fileName = split[2];

            testFile(fileName);
            // Metric deviation of guess cluster
            if (checkSyntaxOnly) {
                return ;
            }
            s = getStringPlotDeviation(guessCluster);
            saveInFile(fileName, s);


        } else if (metric.toLowerCase().equals(PLOT_METRICS[1])) {
            // Metric deviation of all clusters

            String fileNameRegular, fileNameMalware, fileNameGuess;

            try {
                fileNameRegular = split[2];
                fileNameMalware = split[3];
                fileNameGuess = split[4];
            } catch (Exception e) {
                throw new AlligatorSyntaxException("Invalid number of arguments in allprobabilitydifference");
            }

            // Testing the files for save
            testFile(fileNameRegular);
            testFile(fileNameMalware);
            testFile(fileNameGuess);

            if (checkSyntaxOnly) {
                return;
            }


            // Metric deviation of regular cluster
            regularCluster.computeStandardDeviationWith(regularCluster);
            regularCluster.computeStandardDeviationWith(malwareCluster);
            s = getStringPlotDeviation(regularCluster);
            saveInFile(fileNameRegular, s);

            // Metric deviation of malware cluster
            malwareCluster.computeStandardDeviationWith(regularCluster);
            malwareCluster.computeStandardDeviationWith(malwareCluster);
            s = getStringPlotDeviation(malwareCluster);
            saveInFile(fileNameMalware, s);

            // Metric deviation of guess cluster
            s = getStringPlotDeviation(guessCluster);
            saveInFile(fileNameGuess, s);


            // Proximity of guess elements
        } else if (metric.toLowerCase().equals(PLOT_METRICS[2])) {
            String fileName = split[2];
            testFile(fileName);
            s = getStringPlotProximity(guessCluster);
            saveInFile(fileName, s);

            // All proximities
        } else if (metric.toLowerCase().equals(PLOT_METRICS[3])) {
            // Metric deviation of all clusters

            String fileNameRegular, fileNameMalware, fileNameGuess;

            try {
                fileNameRegular = split[2];
                fileNameMalware = split[3];
                fileNameGuess = split[4];
            } catch (Exception e) {
                pd.setMonitor(false);
                throw new AlligatorSyntaxException("Invalid number of arguments in allprobabilitydifference");
            }

            // Testing the files for save
            testFile(fileNameRegular);
            testFile(fileNameMalware);
            testFile(fileNameGuess);

            if (checkSyntaxOnly) {
                return;
            }


            // Metric deviation of regular cluster
            regularCluster.computeClosestElements(proximity, clusters);
            s = getStringPlotProximity(regularCluster);
            saveInFile(fileNameRegular, s);

            // Metric deviation of malware cluster
            malwareCluster.computeClosestElements(proximity, clusters);
            s = getStringPlotProximity(malwareCluster);
            saveInFile(fileNameMalware, s);

            // Metric deviation of guess cluster
            s = getStringPlotProximity(guessCluster);
            saveInFile(fileNameGuess, s);

            // Probability factor of guess elements
        } else if (metric.toLowerCase().equals(PLOT_METRICS[4])) {
            String fileName = split[2];
            testFile(fileName);
            if (checkSyntaxOnly) {
                return;
            }
            s = getStringPlotProbabilityFactor(guessCluster);
            saveInFile(fileName, s);

            // All probability factor
        } else if (metric.toLowerCase().equals(PLOT_METRICS[5])) {
            // Metric deviation of all clusters

            String fileNameRegular, fileNameMalware, fileNameGuess;

            try {
                fileNameRegular = split[2];
                fileNameMalware = split[3];
                fileNameGuess = split[4];
            } catch (Exception e) {
                pd.setMonitor(false);
                throw new AlligatorSyntaxException("Invalid number of arguments in allprobabilityfactor");
            }

            // Testing the files for save
            testFile(fileNameRegular);
            testFile(fileNameMalware);
            testFile(fileNameGuess);

            if (checkSyntaxOnly) {
                return;
            }


            // Metric  of regular cluster
            s = getStringPlotProbabilityFactor(regularCluster);
            saveInFile(fileNameRegular, s);

            // Metric  of malware cluster
            s = getStringPlotProbabilityFactor(malwareCluster);
            saveInFile(fileNameMalware, s);

            // Metric deviation of guess cluster
            s = getStringPlotProbabilityFactor(guessCluster);
            saveInFile(fileNameGuess, s);

            // Correlation of guess elements
        } else if (metric.toLowerCase().equals(PLOT_METRICS[6])) {
            String fileName = split[2];
            testFile(fileName);

            if (checkSyntaxOnly) {
                return;
            }

            s = getStringPlotCorrelation(guessCluster);
            saveInFile(fileName, s);

            // All correlation
        } else if (metric.toLowerCase().equals(PLOT_METRICS[7])) {
            // Metric deviation of all clusters
            String fileNameRegular, fileNameMalware, fileNameGuess;

            try {
                fileNameRegular = split[2];
                fileNameMalware = split[3];
                fileNameGuess = split[4];
            } catch (Exception e) {
                pd.setMonitor(false);
                throw new AlligatorSyntaxException("Invalid number of arguments in allcorrelation");
            }

            // Testing the files for save
            testFile(fileNameRegular);
            testFile(fileNameMalware);
            testFile(fileNameGuess);

            if (checkSyntaxOnly) {
                return;
            }


            Vector<PropertiesCorrelation> savedList = regularCluster.getCloneListOfCorrelations();
            regularCluster.addInterestingCorrelation(malwareCluster.getListOfInterestingCorrelations());
            s = getStringPlotCorrelation(regularCluster);
            saveInFile(fileNameRegular, s);
            regularCluster.setListOfCorrelations(savedList);

            savedList = malwareCluster.getCloneListOfCorrelations();
            malwareCluster.addInterestingCorrelation(regularCluster.getListOfInterestingCorrelations());
            s = getStringPlotCorrelation(malwareCluster);
            saveInFile(fileNameMalware, s);
            malwareCluster.setListOfCorrelations(savedList);

            s = getStringPlotCorrelation(guessCluster);
            saveInFile(fileNameGuess, s);

            // Probability difference of guess elements
        } else if (metric.toLowerCase().equals(PLOT_METRICS[8])) {
            String fileName = split[2];
            testFile(fileName);

            if (checkSyntaxOnly) {
                return;
            }

            s = getStringPlotProbabilityDifference(guessCluster);
            saveInFile(fileName, s);

            // All probability difference
        } else if (metric.toLowerCase().equals(PLOT_METRICS[9])) {
            // Metric deviation of all clusters
            String fileNameRegular, fileNameMalware, fileNameGuess;

            try {
                fileNameRegular = split[2];
                fileNameMalware = split[3];
                fileNameGuess = split[4];
            } catch (Exception e) {
                pd.setMonitor(false);
                throw new AlligatorSyntaxException("Invalid number of arguments in allprobabilitydifference");
            }

            // Testing the files for save
            testFile(fileNameRegular);
            testFile(fileNameMalware);
            testFile(fileNameGuess);

            if (checkSyntaxOnly) {
                return;
            }


            // Metric difference of regular cluster
            //Vector<ProbabilityDifference> vect = regularCluster.getCloneProbabilityDifferences();
            //regularCluster.addAllProbabilityDifferences(malwareCluster.getProbabilityDifferences());
            s = getStringPlotProbabilityDifference(regularCluster);
            saveInFile(fileNameRegular, s);
            //regularCluster.setProbabilityDifferences(vect);

            // Metric deviation of malware cluster
            //vect = malwareCluster.getCloneProbabilityDifferences();
            //malwareCluster.addAllProbabilityDifferences(regularCluster.getProbabilityDifferences());
            s = getStringPlotProbabilityDifference(malwareCluster);
            saveInFile(fileNameMalware, s);
            //malwareCluster.setProbabilityDifferences(vect);

            // Metric deviation of guess cluster
            s = getStringPlotProbabilityDifference(guessCluster);
            saveInFile(fileNameGuess, s);

            // epsilon results of guess elements
        } else if (metric.toLowerCase().equals(PLOT_METRICS[10])) {
            String fileName = split[2];
            testFile(fileName);

            if (checkSyntaxOnly) {
                return;
            }

            s = getStringPlotEpsilon(guessCluster);
            saveInFile(fileName, s);


            // All epsilon clusters
        } else if (metric.toLowerCase().equals(PLOT_METRICS[11])) {
            // Metric deviation of all clusters
            String fileNameRegular, fileNameMalware, fileNameGuess;

            //TraceManager.addDev("All eps 0");


            try {
                fileNameRegular = split[2];
                fileNameMalware = split[3];
                fileNameGuess = split[4];
            } catch (Exception e) {
                pd.setMonitor(false);
                TraceManager.addDev("All eps 1");
                throw new AlligatorSyntaxException("Invalid number of arguments in all epsilon");
            }

            // Testing the files for save
            testFile(fileNameRegular);
            testFile(fileNameMalware);
            testFile(fileNameGuess);

            if (checkSyntaxOnly) {
                return;
            }

            //TraceManager.addDev("All eps 1");

            // For regular
            guessCluster.setInfoOnCurrentComputation(" Computing plots for epsilon clusters of regular");
            Vector<EpsilonCluster> clusterTmp = regularCluster.getEpsilonClusters();
            regularCluster.setEpsilonClusters(guessCluster.getEpsilonClusters());
            regularCluster.setEpsilonClusterDistanceThreshold(guessCluster.getEpsilonClusterDistanceThreshold());
            s = getStringPlotEpsilon(regularCluster);
            saveInFile(fileNameRegular, s);
            regularCluster.setEpsilonClusters(clusterTmp);

            // Malware
            guessCluster.setInfoOnCurrentComputation(" Computing plots for epsilon clusters of malware");
            clusterTmp = malwareCluster.getEpsilonClusters();
            malwareCluster.setEpsilonClusters(guessCluster.getEpsilonClusters());
            malwareCluster.setEpsilonClusterDistanceThreshold(guessCluster.getEpsilonClusterDistanceThreshold());
            s = getStringPlotEpsilon(malwareCluster);
            saveInFile(fileNameMalware, s);
            malwareCluster.setEpsilonClusters(clusterTmp);

            // Guess cluster
            guessCluster.setInfoOnCurrentComputation(" Computing plots for epsilon clusters of guess");
            s = getStringPlotEpsilon(guessCluster);
            saveInFile(fileNameGuess, s);


            // Deviation percentage
        } else if (metric.toLowerCase().equals(PLOT_METRICS[12])) {
            String fileName = split[2];

            testFile(fileName);

            if (checkSyntaxOnly) {
                return;
            }

            // Metric deviation percentage of guess cluster
            s = getStringPlotDeviationPercentage(guessCluster);
            saveInFile(fileName, s);


            // All deviation percentage
        } else if (metric.toLowerCase().equals(PLOT_METRICS[13])) {
            // Metric deviation pecentage of all clusters

            String fileNameRegular, fileNameMalware, fileNameGuess;

            try {
                fileNameRegular = split[2];
                fileNameMalware = split[3];
                fileNameGuess = split[4];
            } catch (Exception e) {
                throw new AlligatorSyntaxException("Invalid number of arguments in alldeviationpercentage");
            }

            // Testing the files for save
            testFile(fileNameRegular);
            testFile(fileNameMalware);
            testFile(fileNameGuess);

            if (checkSyntaxOnly) {
                return;
            }


            // Metric deviation of regular cluster
            regularCluster.computeDeviationPercentageBetween(regularCluster, malwareCluster);
            s = getStringPlotDeviationPercentage(regularCluster);
            saveInFile(fileNameRegular, s);

            // Metric deviation of malware cluster
            malwareCluster.computeDeviationPercentageBetween(regularCluster, malwareCluster);
            s = getStringPlotDeviationPercentage(malwareCluster);
            saveInFile(fileNameMalware, s);

            // Metric deviation of guess cluster
            s = getStringPlotDeviationPercentage(guessCluster);
            saveInFile(fileNameGuess, s);

            // Deviation weight percentage
        } else if (metric.toLowerCase().equals(PLOT_METRICS[14])) {
            String fileName = split[2];

            testFile(fileName);

            if (checkSyntaxOnly) {
                return;
            }

            // Metric deviation weight percentage of guess cluster
            s = getStringPlotDeviationWeightPercentage(guessCluster);
            saveInFile(fileName, s);


            // All deviation percentage
        } else if (metric.toLowerCase().equals(PLOT_METRICS[15])) {
            // Metric deviation pecentage of all clusters

            String fileNameRegular, fileNameMalware, fileNameGuess;

            try {
                fileNameRegular = split[2];
                fileNameMalware = split[3];
                fileNameGuess = split[4];
            } catch (Exception e) {
                throw new AlligatorSyntaxException("Invalid number of arguments in alldeviationweightpercentage");
            }

            // Testing the files for save
            testFile(fileNameRegular);
            testFile(fileNameMalware);
            testFile(fileNameGuess);

            if (checkSyntaxOnly) {
                return;
            }


            // Metric deviation of regular cluster
            regularCluster.computeDeviationWeightPercentageBetween(regularCluster, malwareCluster);
            s = getStringPlotDeviationWeightPercentage(regularCluster);
            saveInFile(fileNameRegular, s);

            // Metric deviation of malware cluster
            malwareCluster.computeDeviationWeightPercentageBetween(regularCluster, malwareCluster);
            s = getStringPlotDeviationWeightPercentage(malwareCluster);
            saveInFile(fileNameMalware, s);

            // Metric deviation of guess cluster
            s = getStringPlotDeviationWeightPercentage(guessCluster);
            saveInFile(fileNameGuess, s);


            // Weight deviation
        } else if (metric.toLowerCase().equals(PLOT_METRICS[16])) {
            String fileName = split[2];

            testFile(fileName);

            if (checkSyntaxOnly) {
                return;
            }

            // Metric deviation of guess cluster
            s = getStringPlotWeightDeviation(guessCluster);
            saveInFile(fileName, s);

            //Weight deviation
        } else if (metric.toLowerCase().equals(PLOT_METRICS[17])) {
            // Metric deviation of all clusters

            String fileNameRegular, fileNameMalware, fileNameGuess;

            try {
                fileNameRegular = split[2];
                fileNameMalware = split[3];
                fileNameGuess = split[4];
            } catch (Exception e) {
                throw new AlligatorSyntaxException("Invalid number of arguments in allprobabilitydifference");
            }

            // Testing the files for save
            testFile(fileNameRegular);
            testFile(fileNameMalware);
            testFile(fileNameGuess);

            if (checkSyntaxOnly) {
                return;
            }


            // Metric deviation of regular cluster
            regularCluster.computeStandardDeviationWith(regularCluster);
            regularCluster.computeStandardDeviationWith(malwareCluster);
            s = getStringPlotWeightDeviation(regularCluster);
            saveInFile(fileNameRegular, s);

            // Metric deviation of malware cluster
            malwareCluster.computeStandardDeviationWith(regularCluster);
            malwareCluster.computeStandardDeviationWith(malwareCluster);
            s = getStringPlotWeightDeviation(malwareCluster);
            saveInFile(fileNameMalware, s);

            // Metric deviation of guess cluster
            s = getStringPlotWeightDeviation(guessCluster);
            saveInFile(fileNameGuess, s);



        } else {
            pd.setMonitor(false);
            throw new AlligatorSyntaxException("Unknown plot metrics");
        }

        pd.setMonitor(false);


    }

    public String getStringPlotProximity(Cluster cl) {
        StringBuffer s = new StringBuffer("#Regular and malware proximity\n");
        for(ClusterElement ce: cl.getElements()) {
            s.append(ce.getName());
            s.append(" ");
            s.append(ce.getNbOfClosestInCluster(regularCluster));
            s.append(" ");
            s.append(ce.getNbOfClosestInCluster(malwareCluster));
            s.append("\n");
        }
        return s.toString();
    }


    public String getStringPlotDeviation(Cluster cl) {
        StringBuffer s = new StringBuffer("#Regular and malware deviation\n");
        for(ClusterElement ce: cl.getElements()) {
            s.append(ce.getName());
            s.append(" ");
            s.append(ce.getTotalDeviation(regularCluster));
            s.append(" ");
            s.append(ce.getTotalDeviation(malwareCluster));
            s.append("\n");
        }
        return s.toString();
    }

    public String getStringPlotWeightDeviation(Cluster cl) {
        StringBuffer s = new StringBuffer("#Regular and malware deviation\n");
        for(ClusterElement ce: cl.getElements()) {
            s.append(ce.getName());
            s.append(" ");
            s.append(ce.getTotalWeightDeviation(regularCluster));
            s.append(" ");
            s.append(ce.getTotalWeightDeviation(malwareCluster));
            s.append("\n");
        }
        return s.toString();
    }

    public String getStringPlotDeviationPercentage(Cluster cl) {
        StringBuffer s = new StringBuffer("#Regular and malware deviation percentage\n");
        for(ClusterElement ce: cl.getElements()) {
            s.append(ce.getName());
            s.append(" ");
            s.append(ce.lastPercentageDeviationWithCluster1);
            s.append(" ");
            s.append(ce.lastPercentageDeviationWithCluster2);
            s.append("\n");
        }
        return s.toString();
    }

    public String getStringPlotDeviationWeightPercentage(Cluster cl) {
        StringBuffer s = new StringBuffer("#Regular and malware deviation weight percentage\n");
        for(ClusterElement ce: cl.getElements()) {
            s.append(ce.getName());
            s.append(" ");
            s.append(ce.lastPercentageDeviationWithCluster1);
            s.append(" ");
            s.append(ce.lastPercentageDeviationWithCluster2);
            s.append("\n");
        }
        return s.toString();
    }


    public String getStringPlotCorrelation(Cluster cl) {
        StringBuffer s = new StringBuffer("#Regular and malware correlation\n");
        for(ClusterElement ce: cl.getElements()) {
            s.append(ce.getName());
            s.append(" ");
            s.append(ce.getPercentageOfCorrelations(regularCluster));
            s.append(" ");
            s.append(ce.getPercentageOfCorrelations(malwareCluster));
            s.append("\n");
        }
        return s.toString();
    }

    public String getStringPlotProbabilityFactor(Cluster cl) {
        StringBuffer s = new StringBuffer("#Regular and malware probability factor / difference\n");
        DoublePoint dp;

        for(ClusterElement ce: cl.getElements()) {
            dp = ce.getPercentageProbaFactorSummary(regularCluster, malwareCluster);
            s.append(ce.getName());
            s.append(" ");
            s.append(dp.getX());
            s.append(" ");
            s.append(dp.getY());
            s.append("\n");
        }
        return s.toString();
    }

    public String getStringPlotProbabilityDifference(Cluster cl) {
        StringBuffer s = new StringBuffer("#Regular and malware probability factor / difference\n");
        DoublePoint dp;
        for(ClusterElement ce: cl.getElements()) {
            dp = ce.getPercentageProbaDifference(regularCluster, malwareCluster);
            s.append(ce.getName());
            s.append(" ");
            s.append(dp.getX());
            s.append(" ");
            s.append(dp.getY());
            s.append("\n");
        }
        return s.toString();
    }

    public String getStringPlotEpsilon(Cluster cl) {
        StringBuffer s = new StringBuffer("#Regular and malware probability factor / difference\n");
        DoublePoint dp;
        int total = cl.getElements().size();
        int cpt = 0;

        for(ClusterElement ce: cl.getElements()) {
            guessCluster.setPercentage(cpt * 100 / total);
            cpt ++;
            dp = ce.getPercentageNeighboursInEpsilonCluster(regularCluster, malwareCluster);
            s.append(ce.getName());
            s.append(" ");
            s.append(dp.getX());
            s.append(" ");
            s.append(dp.getY());
            s.append("\n");
        }


        return s.toString();
    }

    public void testFile(String fileName) throws AlligatorSyntaxException {
        try {
            if (!FileUtils.checkFileForSave(new File(fileName))) {
                throw new AlligatorSyntaxException("Plot file cannot be created: " + fileName);
            }
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Plot file cannot be created: " + fileName +  " because: " + e.getMessage());
        }

    }


    public void saveInFile(String fileName, String data) throws AlligatorSyntaxException {
        try {
            FileUtils.saveFile(fileName, data);
        } catch (Exception e) {
            throw new AlligatorSyntaxException("Plot data could not be saved: " + fileName +  " because: " + e.getMessage());
        }
    }


    public int getHighestScore(double scores[]) {
        double score = -1;
        int index = -1;
        for(int i=0; i<scores.length; i++) {
            if (scores[i] > score) {
                score = scores[i];
                index = i;
            }
        }
        return index;
    }

    public int getLowestScore(double scores[]) {
        double score = Double.MAX_VALUE;
        int index = -1;
        for(int i=0; i<scores.length; i++) {
            if (scores[i] < score) {
                score = scores[i];
                index = i;
            }
        }
        return index;
    }


    public int getClosestAverageScore(double scores[]) {
        double average = 0;
        int i;
        double diff = Double.MAX_VALUE;
        double current;
        int index = -1;

        for(i=0; i<scores.length; i++) {
            average +=scores[i];
        }
        average /= scores.length;

        for(i=0; i<scores.length; i++) {
            current = Math.abs(scores[i]-average);
            if (current < diff) {
                diff = current;
                index = i;
            }
        }

        return index;

    }

    public int getHighestScoreDifference(double scores1[], double scores2[]) {
        double diff = -1;
        int index = -1;
        double current;
        for(int i=0; i<scores1.length; i++) {
            current = Math.abs(scores1[i]-scores2[i]);
            if (current > diff) {
                diff = current;
                index = i;
            }
        }
        return index;

    }

    public int getLowestScoreDifference(double scores1[], double scores2[]) {
        double diff = Double.MAX_VALUE;
        int index = -1;
        double current;

        for(int i=0; i<scores1.length; i++) {
            current = Math.abs(scores1[i]-scores2[i]);
            if (current < diff) {
                diff = current;
                index = i;
            }
        }
        return index;

    }

    public void setAlligatorStepsTo(int nbOfSteps) {
        AlligatorScore as;
        for(int i=0; i<regularScores.length; i++) {
            regularScores[i].get(0).ais.regularStep = nbOfSteps;
            regularScores[i].get(0).ais.malwareStep = nbOfSteps;
            malwareScores[i].get(0).ais.regularStep = nbOfSteps;
            malwareScores[i].get(0).ais.malwareStep = nbOfSteps;
        }
    }

    public void prepareAdvancedLearning() {
        int i;

        classificationWeight = new int[guessCluster.size()];
        for (i=0; i<guessCluster.size(); i++) {
            classificationWeight[i] = 1;
        }

        //scoreInRegular = new double[guessCluster.size()];
        //scoreInMalware = new double[guessCluster.size()];
    }

    public int getNbOfForgotten() {
        int nb = 0;
        for (int i=0; i<classificationWeight.length; i++) {
            if (classificationWeight[i] == 0) {
                nb ++;
            }
        }
        return nb;
    }

    public int getNbOfBoost() {
        int nb = 0;
        for (int i=0; i<classificationWeight.length; i++) {
            if (classificationWeight[i] > 1) {
                nb += classificationWeight[1] - 1;
            }
        }
        return nb;
    }

    public int getNbOfGuessSamples() {
        return guessCluster.size();
    }

    public void fillWithClassificationValue(AlligatorEvaluationSet aes, boolean isRegular, Vector<AlligatorElementReference> _elts) {
        double scoreMalware = 0;
        double scoreRegular = 0;
        int index;

        double tmp;
        double scoreDiff;
        //AlligatorElementReference aer;

        if (classificationWeight == null) {
            prepareAdvancedLearning();
        }

        for(int i=0; i<regularScores.length; i++) {
            scoreRegular = 0;
            scoreMalware = 0;
            index = 0;
            if (classificationWeight[i] > 0) {
                for(AlligatorScore as: regularScores[i]) {
                    tmp = scoreRegular;
                    scoreRegular += as.getScore(aes.getIndex(index));
                    if (scoreRegular == Double.POSITIVE_INFINITY) {
                        //TraceManager.addDev("*** ScoreRegular=" + scoreRegular + " intermediate=" + as.getScore(aes.getIndex(index)) + " tmp=" + tmp);
                    }
                    index ++;
                }
                for(AlligatorScore as: malwareScores[i]) {
                    scoreMalware += as.getScore(aes.getIndex(index));
                    index ++;
                }
                //TraceManager.addDev("ScoreMalware=" + scoreMalware + " scoreRegular=" + scoreRegular);
                scoreDiff = Math.abs(scoreRegular-scoreMalware);

                //scoreInRegular[i] = scoreRegular;
                //scoreInMalware[i] = scoreMalware;

                if (isRegular) {
                    if (scoreRegular <= scoreMalware) {
                        _elts.add(new AlligatorElementReference(i, isRegular, scoreDiff));
                    }


                } else {
                    if (scoreRegular >= scoreMalware) {
                        _elts.add(new AlligatorElementReference(i, isRegular, scoreDiff));
                    }
                }
            }
        }
    }

    public void fillWithClassificationMinValue(AlligatorEvaluationSet aes, boolean isRegular, Vector<AlligatorElementReference> _elts) {
        double scoreMalware = 0;
        double scoreRegular = 0;
        int index;

        double tmp;
        double scoreDiff;
        //AlligatorElementReference aer;

        if (classificationWeight == null) {
            prepareAdvancedLearning();
        }

        for(int i=0; i<regularScores.length; i++) {
            scoreRegular = 0;
            scoreMalware = 0;
            index = 0;
            if (classificationWeight[i] > 0) {
                for(AlligatorScore as: regularScores[i]) {
                    tmp = scoreRegular;
                    scoreRegular += as.getScore(aes.getIndex(index));
                    if (scoreRegular == Double.POSITIVE_INFINITY) {
                        //TraceManager.addDev("*** ScoreRegular=" + scoreRegular + " intermediate=" + as.getScore(aes.getIndex(index)) + " tmp=" + tmp);
                    }
                    index ++;
                }
                for(AlligatorScore as: malwareScores[i]) {
                    scoreMalware += as.getScore(aes.getIndex(index));
                    index ++;
                }
                //TraceManager.addDev("ScoreMalware=" + scoreMalware + " scoreRegular=" + scoreRegular);
                scoreDiff = Math.abs(scoreRegular-scoreMalware);

                //scoreInRegular[i] = scoreRegular;
                //scoreInMalware[i] = scoreMalware;

                if (isRegular) {
                    if (!(scoreRegular <= scoreMalware)) {
                        _elts.add(new AlligatorElementReference(i, isRegular, scoreDiff));
                    }


                } else {
                    if (!(scoreRegular >= scoreMalware)) {
                        _elts.add(new AlligatorElementReference(i, isRegular, scoreDiff));
                    }
                }
            }
        }
    }

} // Class
