/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

   ludovic.apvrille AT telecom-paristech.fr

   This software is a computer program whose purpose is to allow malware analysis

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class AlligatorLearner
   * Class that tests different values of weight, and that proposes the "best" values
   * Creation: 23/07/2012
   * @version 1.0 23/07/2012
   * @author Ludovic APVRILLE
   * @see
   */

package analyzer;

//import cluster.*;
import myutil.*;

import java.util.*;
import java.math.*;



public class AlligatorLearner  {

    public static int MAX_VALUES_BRUTE_FORCE = 100000;

    private boolean bestSetOnPercentageOnly = false;
    private boolean bestSetOnAverage = false;
    //private boolean bestSetOnSigmaOnly = false;


    private  AlligatorLanguageReader readerRegular;
    private  AlligatorLanguageReader readerMalware;

    private Vector<AlligatorEvaluationSet> set;

    private  AlligatorLearnerExecutor executorRegular, executorMalware;

    private int step;

    private AlligatorEvaluationSet best;


    public AlligatorLearner(AlligatorLanguageReader _readerRegular, AlligatorLanguageReader _readerMalware, boolean _bestSetOnPercentageOnly, boolean _bestSetOnAverage) {
        readerRegular = _readerRegular;
        readerMalware = _readerMalware;
        bestSetOnPercentageOnly = _bestSetOnPercentageOnly;
        bestSetOnAverage = _bestSetOnAverage;

        //TraceManager.addDev("Best set on average : " + bestSetOnAverage);

        evaluateMaxNbOfScoresInAlligatorScore();
    }

    public void setBest(AlligatorEvaluationSet _best) {
        best = _best;
    }

    public AlligatorEvaluationSet getBest() {
        return best;
    }

    public AlligatorLanguageReader getAlligatorLanguageReaderRegular() {
        return readerRegular;
    }

    public void bruteForce() {
        // Set up one thread per cluster
        AlligatorLearnerExecutor executorRegular = new AlligatorLearnerExecutor(readerRegular, "Regular");
        AlligatorLearnerExecutor executorMalware = new AlligatorLearnerExecutor(readerMalware, "Malware");

        // For each cluster element of each cluster, we compute all possible values

        executorRegular.go();
        executorMalware.go();

        executorRegular.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);
        executorMalware.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);

        executorRegular.waitForDone();
        executorMalware.waitForDone();

        // Computing sets of attemps
        TraceManager.addDev("Computing sets of attempts");
        int nbOfSteps = readerRegular.getNbOfSteps();
        int [] nbOfValuesPerStepRegular = new int[nbOfSteps];
        int [] nbOfValuesPerStepMalware = new int[nbOfSteps];

        BigInteger bi = BigInteger.ONE;
        for(int i=0; i<nbOfSteps; i++) {
            nbOfValuesPerStepRegular[i] = readerRegular.getNbOfValuesForStep(i, true);
            nbOfValuesPerStepMalware[i] = readerRegular.getNbOfValuesForStep(i, false);
            TraceManager.addDev("Nb Of value for step " + i + " regular:" + nbOfValuesPerStepRegular[i] + " malware:" + nbOfValuesPerStepMalware[i]);
            bi.multiply(new BigInteger(""+nbOfValuesPerStepMalware[i])).multiply(new BigInteger(""+nbOfValuesPerStepRegular[i]));
        }

        TraceManager.addDev("Nb of sets to compute: " + bi.toString());

        // Make a tree of all possible values
        AlligatorTreeNode root = new AlligatorTreeNode(null);
        Vector<AlligatorTreeNode> leafs = new Vector<AlligatorTreeNode>();
        Vector<AlligatorTreeNode> tmp;
        AlligatorTreeNode node;

        leafs.add(root);
        //int index = 0;
        for(int i=0; i<nbOfSteps; i++) {
            tmp = new Vector<AlligatorTreeNode>();
            for(int j=0; j<nbOfValuesPerStepRegular[i]; j++) {
                for(AlligatorTreeNode oldNode: leafs) {
                    node = new AlligatorTreeNode(oldNode);
                    tmp.add(node);
                    oldNode.addChildren(node);
                }
            }
            leafs = tmp;
        }
        for(int i=0; i<nbOfSteps; i++) {
            tmp = new Vector<AlligatorTreeNode>();
            for(int j=0; j<nbOfValuesPerStepMalware[i]; j++) {
                for(AlligatorTreeNode oldNode: leafs) {
                    node = new AlligatorTreeNode(oldNode);
                    tmp.add(node);
                    oldNode.addChildren(node);
                }
            }
            leafs = tmp;
        }

        // We now have all leafs -> print the leafs
        int depth = leafs.get(0).getDepth();
        TraceManager.addDev("Depth=" + depth + " nb of leafs=" + leafs.size());
        /*for(AlligatorTreeNode n: leafs) {
          TraceManager.addDev(n.getOrigin(""));
          }*/

        // Building the set of evaluation sets
        set = new Vector<AlligatorEvaluationSet>();
        for(AlligatorTreeNode n: leafs) {
            AlligatorEvaluationSet aset = new AlligatorEvaluationSet(n, depth);
            set.add(aset);
        }

        // Evaluation set
        TraceManager.addDev("For information: Set(0):" + set.get(0).toString() + "\n");
        TraceManager.addDev("Last Set element:" + set.get(set.size()-1).toString() + "\n");

        /*for(AlligatorEvaluationSet aset: set) {
          TraceManager.addDev(aset.toString());
          }*/



        executorRegular.setSet(set);
        executorMalware.setSet(set);

        executorRegular.setStep(AlligatorLearnerExecutor.COMPUTE_VALUES_FOR_REGULAR);
        executorMalware.setStep(AlligatorLearnerExecutor.COMPUTE_VALUES_FOR_MALWARE);

        int ite1, ite2;

        Runtime runtime = Runtime.getRuntime();
        long maxMemory, allocatedMemory;
        //long freeMemory;


        while((executorRegular.getStep() != AlligatorLearnerExecutor.DONE) || (executorMalware.getStep() != AlligatorLearnerExecutor.DONE)) {
            ite1 = executorRegular.getIteration();
            ite2 = executorMalware.getIteration();
            maxMemory = runtime.maxMemory()/1024;
            allocatedMemory = runtime.totalMemory()/1024;
            //freeMemory = runtime.freeMemory()/1024;

            System.out.print("Completion on regular: " + (int)((100*ite1/(double)(set.size()))) + "% Completion on malware: " +  (int)((100*ite2/(double)(set.size()))) + "% allocated memory:" + allocatedMemory + "/" + maxMemory + "\r");
            try {
                ((Thread)(Thread.currentThread())).sleep(250);
            } catch (Exception e) {
            }
        }
        System.out.println("\n");






        // For each step, must determine how many values can be obtained
        // then compute the set of all possible tries: set of array with each array containing for the index the value to use
        // Compute all scores for each element of the set, and then evaluate the result: if better than another one, keep it.
        // Agregate is then given

        executorRegular.setStep(AlligatorLearnerExecutor.EXIT);
        executorMalware.setStep(AlligatorLearnerExecutor.EXIT);


        try {
            executorRegular.join();
            executorMalware.join();
        } catch (InterruptedException ie) {
        }

        /*for(AlligatorEvaluationSet aset: set) {
          TraceManager.addDev(aset.toString());
          }*/
    }

    public AlligatorEvaluationSet bruteForce2() {
        // Set up one thread per cluster
        AlligatorLearnerExecutor executorRegular = new AlligatorLearnerExecutor(readerRegular, "Regular");
        AlligatorLearnerExecutor executorMalware = new AlligatorLearnerExecutor(readerMalware, "Malware");

        // For each cluster element of each cluster, we compute all possible values

        executorRegular.go();
        executorMalware.go();

        executorRegular.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);
        executorMalware.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);

        executorRegular.waitForDone();
        executorMalware.waitForDone();

        // Computing sets of attemps
        TraceManager.addDev("Computing sets of attempts");
        int nbOfSteps = readerRegular.getNbOfSteps();
        int [] nbOfValuesPerStepRegular = new int[nbOfSteps];
        int [] nbOfValuesPerStepMalware = new int[nbOfSteps];

        BigInteger bi = BigInteger.ONE;
        for(int i=0; i<nbOfSteps; i++) {
            nbOfValuesPerStepRegular[i] = readerRegular.getNbOfValuesForStep(i, true);
            nbOfValuesPerStepMalware[i] = readerRegular.getNbOfValuesForStep(i, false);
            //TraceManager.addDev("Nb Of value for step " + i + " regular:" + nbOfValuesPerStepRegular[i] + " malware:" + nbOfValuesPerStepMalware[i]);
            //nb = nb * nbOfValuesPerStepMalware[i] * nbOfValuesPerStepRegular[i];
            bi.multiply(new BigInteger(""+nbOfValuesPerStepMalware[i])).multiply(new BigInteger(""+nbOfValuesPerStepRegular[i]));
        }

        TraceManager.addDev("Nb of sets to compute: " + bi.toString());

        AlligatorEvaluationSet current = new AlligatorEvaluationSet(nbOfSteps);
        //AlligatorEvaluationSet best = new AlligatorEvaluationSet(nbOfSteps);


        readerRegular.reinitOverallScore();
        readerMalware.reinitOverallScore();



        AlligatorEvaluationSet best = bruteForce2Tree(0, nbOfSteps * 2, nbOfValuesPerStepRegular, nbOfValuesPerStepMalware, current, null);

        return best;

    }


    // Returns the best set, and the worst one?
    public AlligatorEvaluationSet bruteForce2Tree(int currentStep, int nbOfSteps, int [] nbOfValuesPerStepRegular, int [] nbOfValuesPerStepMalware, AlligatorEvaluationSet current, AlligatorEvaluationSet best) {

        // Leaf?
        if (currentStep == nbOfSteps) {

            // Compute the content of current: nb of recognized
            current.reinitAnalysisValues();
            readerRegular.fillAlligatorEvaluationSet(current, true);
            readerMalware.fillAlligatorEvaluationSet(current, false);
            //TraceManager.addDev("in Leaf, current=" + current);
            AlligatorEvaluationSet set = current.getBest(best, bestSetOnPercentageOnly, bestSetOnAverage);
            //TraceManager.addDev("in Leaf, current=" + current + "\nbest=" + set);
            return set;
        }


        //Regular computations
        if (currentStep < (nbOfSteps/2)) {
            // Computations over all possible values
            for(int i=0; i<nbOfValuesPerStepRegular[currentStep]; i++) {
                //Add to score the current one to compute, in the two clusters
                readerRegular.addRegularToScore(currentStep, i);
                readerMalware.addRegularToScore(currentStep, i);

                // Recursive call
                current.arrayOfValueIndexRegular[currentStep] = i;
                best = bruteForce2Tree(currentStep+1, nbOfSteps, nbOfValuesPerStepRegular, nbOfValuesPerStepMalware, current, best);

                if (best.isOptimalSet(bestSetOnPercentageOnly)) {
                    return best;
                }

                // Previous version of the score
                readerRegular.removeRegularToScore(currentStep, i);
                readerMalware.removeRegularToScore(currentStep, i);
            }

            // Malware computations
        } else {
            // Computations over all possible values
            int cs = currentStep - (nbOfSteps / 2);
            for(int i=0; i<nbOfValuesPerStepMalware[cs]; i++) {
                //Add to score the current one to compute, in the two clusters
                readerRegular.addMalwareToScore(cs, i);
                readerMalware.addMalwareToScore(cs, i);

                // Recursive call
                current.arrayOfValueIndexMalware[cs] = i;
                best = bruteForce2Tree(currentStep+1, nbOfSteps, nbOfValuesPerStepRegular, nbOfValuesPerStepMalware, current, best);

                if (best.isOptimalSet(bestSetOnPercentageOnly)) {
                    return best;
                }

                // Previous version of the score
                readerRegular.removeMalwareToScore(cs, i);
                readerMalware.removeMalwareToScore(cs, i);
            }
        }

        return best;
    }



    public AlligatorEvaluationSet getBestSet() {
        double sumPercentage = 0;
        double tmp;
        AlligatorEvaluationSet bestOne = null;

        int cpt = 0;


        for(AlligatorEvaluationSet aset: set) {
            /*if ((cpt % 10) == 0) {
              printPercentage(100*cpt/(double)(set.size()), " in analyzing best");
              }*/
            //      cpt ++;
            
                tmp = aset.getBestPercentage();
                if ((tmp > sumPercentage) && (aset.respectMinPercentage()))  {
                    bestOne = aset;
                    sumPercentage = tmp;
                } else if ((tmp == sumPercentage) && (aset.respectMinPercentage()) && (!bestSetOnPercentageOnly) && (bestOne != null)) {
                    // Compare with standard deviation as well
                    if (bestSetOnAverage) {
                        if (aset.average >  bestOne.average) {
                            //TraceManager.addDev("Better average found\n");
                            bestOne = aset;
                            sumPercentage = tmp;
                        }
                    } else {
                        /*if (aset.sigma >  bestOne.sigma) {
                        //TraceManager.addDev("Better sigma found\n");
                        bestOne = aset;
                        sumPercentage = tmp;
                        }*/
                        if (aset.getDiff() > bestOne.getDiff()) {
                            bestOne = aset;
                            sumPercentage = tmp;
                        }
                    }
                }
            
        }

        if (bestOne == null) {
            for(AlligatorEvaluationSet aset: set) {
                tmp = aset.getBestPercentage();
                if (tmp > sumPercentage)  {
                    bestOne = aset;
                    sumPercentage = tmp;
                } else if ((tmp == sumPercentage) && (!bestSetOnPercentageOnly) && (bestOne != null)) {
                    // Compare with standard deviation as well
                    // Compare with standard deviation as well
                    if (bestSetOnAverage) {
                        if (aset.average >  bestOne.average) {
                            //TraceManager.addDev("Better average found\n");
                            bestOne = aset;
                            sumPercentage = tmp;
                        }
                    } else {
                        //if (aset.sigma >  bestOne.sigma) {
                        //TraceManager.addDev("Better sigma found\n");
                        //    bestOne = aset;
                        //    sumPercentage = tmp;
                        //}
                        if (aset.getDiff() > bestOne.getDiff()) {
                            bestOne = aset;
                            sumPercentage = tmp;
                        }
                    }
                }
            }
        }

        return bestOne;
    }


    public String getMultipliers(AlligatorEvaluationSet set) {
        String s = "";
        for(int i=0; i<set.arrayOfValueIndexRegular.length; i++) {
            s+="#" + (i+1) + ":(" + readerRegular.getRegularMultiplier(i, set.arrayOfValueIndexRegular[i]) + ","+ readerRegular.getMalwareMultiplier(i, set.arrayOfValueIndexMalware[i]) + ") ";
        }
        return s;

    }

    public AlligatorEvaluationSet oneOtherAverage() {
        // Set up one thread per cluster
        AlligatorLearnerExecutor executorRegular = new AlligatorLearnerExecutor(readerRegular, "Regular");
        AlligatorLearnerExecutor executorMalware = new AlligatorLearnerExecutor(readerMalware, "Malware");

        // For each cluster element of each cluster, we compute all possible values

        executorRegular.go();
        executorMalware.go();

        executorRegular.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);
        executorMalware.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);

        executorRegular.waitForDone();
        executorMalware.waitForDone();

        // Computing sets of attemps
        TraceManager.addDev("Computing sets of attempts");
        int nbOfSteps = readerRegular.getNbOfSteps();
        int [] nbOfValuesPerStepRegular = new int[nbOfSteps];
        int [] nbOfValuesPerStepMalware = new int[nbOfSteps];

        long nb = 1;
        for(int i=0; i<nbOfSteps; i++) {
            nbOfValuesPerStepRegular[i] = readerRegular.getNbOfValuesForStep(i, true);
            nbOfValuesPerStepMalware[i] = readerRegular.getNbOfValuesForStep(i, false);
            //TraceManager.addDev("Nb Of value for step " + i + " regular:" + nbOfValuesPerStepRegular[i] + " malware:" + nbOfValuesPerStepMalware[i]);
            nb = nb + nbOfValuesPerStepMalware[i] + nbOfValuesPerStepRegular[i];
        }

        TraceManager.addDev("Nb of sets to compute: " + nb);

        set = new Vector<AlligatorEvaluationSet>();
        AlligatorEvaluationSet average = new AlligatorEvaluationSet(nbOfSteps);
        AlligatorEvaluationSet aes;

        // Filling the average
        for(int i=0; i<nbOfSteps; i++) {
            average.arrayOfValueIndexRegular[i] = readerRegular.getRegularAverageIndex(i);
            average.arrayOfValueIndexMalware[i] = readerRegular.getMalwareAverageIndex(i);
        }

        readerRegular.computeOne(average, true);
        readerMalware.computeOne(average, false);
        TraceManager.addDev("Average index: " + average);



        int []bestIndex = new int[nbOfSteps*2];
        for(int i=0; i<nbOfSteps; i++) {
            // For each steps we generate a set of AlligatorEvaluationSet -> we compute the best one
            // We then keep track of the best index to be used
            // Regular, then malware

            // Regular computation
            if (nbOfValuesPerStepRegular[i] == 1) {
                bestIndex[i] = 0;
            } else {

                set.clear();
                for(int j=0; j<nbOfValuesPerStepRegular[i]; j++) {
                    aes = average.fullClone();
                    aes.arrayOfValueIndexRegular[i] = j;
                    set.add(aes);
                }
                readerRegular.computeSet(set, true);
                readerMalware.computeSet(set, false);
                aes = getBestSet();

                //TraceManager.addDev("List of all :" + set);
                bestIndex[i] = aes.arrayOfValueIndexRegular[i];
                //TraceManager.addDev("Best set step " + i + " = " + aes);
                TraceManager.addDev("Best regular index found for step " + i + " = " + bestIndex[i]);

            }

            // Malware computation
            if (nbOfValuesPerStepMalware[i] == 1) {
                bestIndex[i+nbOfSteps] = 0;
            } else {

                set.clear();
                for(int j=0; j<nbOfValuesPerStepMalware[i]; j++) {
                    aes = average.fullClone();
                    aes.arrayOfValueIndexMalware[i] = j;
                    set.add(aes);
                }
                readerRegular.computeSet(set, true);
                readerMalware.computeSet(set, false);
                aes = getBestSet();

                //TraceManager.addDev("List of all :" + set);
                bestIndex[i+nbOfSteps] = aes.arrayOfValueIndexMalware[i];
                //TraceManager.addDev("Best set step " + i + " = " + aes);
                TraceManager.addDev("Best malware index found for step " + i + " = " + bestIndex[i+nbOfSteps]);

            }


        }

        // From all bestIndex, building the best AlligatorEvaluationSet, and computing results
        set.clear();
        aes = average.fullClone();

        for(int i=0; i<nbOfSteps; i++) {
            aes.arrayOfValueIndexRegular[i] = bestIndex[i];
            aes.arrayOfValueIndexMalware[i] = bestIndex[i+nbOfSteps];
        }
        set.add(aes);
        readerRegular.computeSet(set, true);
        readerMalware.computeSet(set, false);

        return aes;


    }

    public AlligatorEvaluationSet oneAfterTheOther() {
        TraceManager.addDev("One after the other");

        double currentPercentage = 0;
        // Set up one thread per cluster
        AlligatorLearnerExecutor executorRegular = new AlligatorLearnerExecutor(readerRegular, "Regular");
        AlligatorLearnerExecutor executorMalware = new AlligatorLearnerExecutor(readerMalware, "Malware");

        // For each cluster element of each cluster, we compute all possible values

        executorRegular.go();
        executorMalware.go();

        executorRegular.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);
        executorMalware.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);

        executorRegular.waitForDone();
        executorMalware.waitForDone();

        // Computing sets of attemps
        TraceManager.addDev("Computing sets of attempts");
        int nbOfSteps = readerRegular.getNbOfSteps();
        int [] nbOfValuesPerStepRegular = new int[nbOfSteps];
        int [] nbOfValuesPerStepMalware = new int[nbOfSteps];

        long nb = 1;
        for(int i=0; i<nbOfSteps; i++) {
            nbOfValuesPerStepRegular[i] = readerRegular.getNbOfValuesForStep(i, true);
            nbOfValuesPerStepMalware[i] = readerRegular.getNbOfValuesForStep(i, false);
            //TraceManager.addDev("Nb Of value for step " + i + " regular:" + nbOfValuesPerStepRegular[i] + " malware:" + nbOfValuesPerStepMalware[i]);
            nb = nb + nbOfValuesPerStepMalware[i] + nbOfValuesPerStepRegular[i];
        }

        //TraceManager.addDev("Nb of sets to compute: " + nb);

        set = new Vector<AlligatorEvaluationSet>();
        AlligatorEvaluationSet study = new AlligatorEvaluationSet(nbOfSteps);
        AlligatorEvaluationSet aes, aes1;

        TraceManager.addDev("Filling 0 values");
        // Filling the average
        for(int i=0; i<nbOfSteps; i++) {
            study.arrayOfValueIndexRegular[i] = 0;
            study.arrayOfValueIndexMalware[i] = 0;
        }


        //readerRegular.computeOne(average, true);
        //readerMalware.computeOne(average, false);
        TraceManager.addDev("filling done");

        int []bestIndex = new int[nbOfSteps*2];
        //aes = average;
        for(int i=0; i<nbOfSteps; i++) {
            //if ((i % 10) == 0) {
            //elapsedTime = (System.currentTimeMillis() - startTime)/1000;
            printReducedAnalysedPercentage(100*i/(double)(nbOfSteps), currentPercentage);
            TraceManager.addDev("\nComputing step #" + i);
            //}
            // For each steps we generate a set of AlligatorEvaluationSet -> we compute the best one
            // We then keep track of the best index to be used
            // Regular, then malware

            // All computations

            // Preparing the set to evalute
            set.clear();
            for(int j=0; j<nbOfValuesPerStepRegular[i]; j++) {
                aes = study.fullClone();
                aes.arrayOfValueIndexRegular[i] = j;
                for (int k=0; k<nbOfValuesPerStepMalware[i]; k++) {
                    aes1=aes.fullClone();
                    aes1.arrayOfValueIndexMalware[i] = k;
                    set.add(aes1);
                }
            }

            TraceManager.addDev("First set:" + set.get(0));
            TraceManager.addDev("Second set:" + set.get(1));
            TraceManager.addDev("Last set:" + set.get(set.size()-1) + "\n");

            readerRegular.computeSet(set, true);
            readerMalware.computeSet(set, false);
            study = getBestSet();
            currentPercentage = study.getBestPercentage();

            //TraceManager.addDev("List of all :" + set);

            bestIndex[i] = study.arrayOfValueIndexRegular[i];
            bestIndex[i+nbOfSteps] = study.arrayOfValueIndexMalware[i];
            //study.arrayOfValueIndexRegular[i] = bestIndex[i];
            //study.arrayOfValueIndexMalware[i] = bestIndex[i+nbOfSteps];
            TraceManager.addDev("Best set step " + i + " = " + study);
            TraceManager.addDev("Best regular/malware index found for step " + i + " = " + bestIndex[i] + "/" + bestIndex[i+nbOfSteps]);

        }

        // From all bestIndex, building the best AlligatorEvaluationSet, and computing results
        set.clear();
        aes = study.fullClone();

        for(int i=0; i<nbOfSteps; i++) {
            aes.arrayOfValueIndexRegular[i] = bestIndex[i];
            aes.arrayOfValueIndexMalware[i] = bestIndex[i+nbOfSteps];
        }
        set.add(aes);
        readerRegular.computeSet(set, true);
        readerMalware.computeSet(set, false);

        return aes;


    }


    public AlligatorEvaluationSet dichotomicBruteForce() {
        int i;

        AlligatorLearnerExecutor executorRegular = new AlligatorLearnerExecutor(readerRegular, "Regular");
        AlligatorLearnerExecutor executorMalware = new AlligatorLearnerExecutor(readerMalware, "Malware");

        // For each cluster element of each cluster, we compute all possible values

        executorRegular.go();
        executorMalware.go();

        executorRegular.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);
        executorMalware.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);

        executorRegular.waitForDone();
        executorMalware.waitForDone();

        // Prepare data sets
        // Does not take into account steps between multipliers
        int nbOfSteps = readerRegular.getNbOfSteps();

        int nbOfValuesPerTry = (int)(Math.pow(MAX_VALUES_BRUTE_FORCE, 1.0/(nbOfSteps*2)));
        nbOfValuesPerTry = Math.min(10, Math.max(nbOfValuesPerTry, 3));

        //    Get the min and max multiplier for each step, and
        //For each step, we store the current min and max multipliers
        int[] currentMinRegularIndex = new int[nbOfSteps];
        int[] currentMinMalwareIndex = new int[nbOfSteps];
        int[] currentMaxRegularIndex = new int[nbOfSteps];
        int[] currentMaxMalwareIndex = new int[nbOfSteps];

        //readerRegular.setAlligatorStepsTo(1);
        //readerMalware.setAlligatorStepsTo(1);

        for(i=0; i<nbOfSteps; i++) {
            //Regular
            currentMinRegularIndex[i] = 0;
            currentMaxRegularIndex[i] = readerRegular.getNbOfValuesForStep(i, true)-1;
            //Malware
            currentMinMalwareIndex[i] = 0;
            currentMaxMalwareIndex[i] = readerRegular.getNbOfValuesForStep(i, false)-1;
        }



        set = new Vector<AlligatorEvaluationSet>();
        AlligatorEvaluationSet best = new AlligatorEvaluationSet(nbOfSteps);
        AlligatorEvaluationSet previousBest = null, test;
        for(i=0; i<nbOfSteps; i++) {
            best.arrayOfValueIndexRegular[i] = readerRegular.getRegularAverageIndex(i);
            best.arrayOfValueIndexMalware[i] = readerRegular.getMalwareAverageIndex(i);
            TraceManager.addDev("NbOfSteps= " + nbOfSteps + "  Regular : minIndex=" + currentMinRegularIndex[i] + " maxIndex=" + currentMaxRegularIndex[i] + "  Malware : minIndex=" + currentMinMalwareIndex[i] + " maxIndex=" + currentMaxMalwareIndex[i] );
        }

        boolean canGo;

        canGo = !noMoreValuesIn(currentMinRegularIndex, currentMaxRegularIndex);
        TraceManager.addDev("canGo=" + canGo);
        canGo = canGo && !noMoreValuesIn(currentMinMalwareIndex, currentMaxMalwareIndex);

        TraceManager.addDev("canGo=" + canGo);

        while(canGo) {
            // Elaborate the sets
            TraceManager.addDev("Elaborating " + nbOfValuesPerTry * nbOfSteps + "sets");
            set.clear();
            elaborateSet(null, nbOfSteps, 0, nbOfValuesPerTry, currentMinRegularIndex, currentMinMalwareIndex, currentMaxRegularIndex, currentMaxMalwareIndex);
            TraceManager.addDev("Size of set=" + set.size());

            // Find the best
            readerRegular.computeSet(set, true);
            readerMalware.computeSet(set, false);
            best = getBestSet();

            TraceManager.addDev("Best:" + best);

            if (previousBest != null) {
                TraceManager.addDev("Testing with previous\nprevious=" + previousBest + "\nbest=" + best);
                test = previousBest.getBestNoClone(best, false, false);
                if (test == previousBest) {
                    TraceManager.addDev("---> Selecting previous");
                    canGo = false;
                    best = test;
                } else {
                    TraceManager.addDev("---> Selecting best bestPercentage:" + best.getBestPercentage() + " previous=" + previousBest.getBestPercentage());

                }
            }

            if (canGo) {
                previousBest = best;
                // Recompute search intervals
                //Cut the search interval in two.
                int index;
                int diff;
                for(i=0; i<nbOfSteps; i++) {
                    //regular
                    index = best.arrayOfValueIndexRegular[i];
                    diff = (currentMaxRegularIndex[i] - currentMinRegularIndex[i])/2;
                    if (diff < 2) {
                        currentMaxRegularIndex[i] = index;
                        currentMinRegularIndex[i] = index;
                    } else {
                        currentMinRegularIndex[i] = Math.max(index - (diff/2), 0);
                        currentMaxRegularIndex[i] = Math.min(currentMinRegularIndex[i]+diff, readerRegular.getNbOfValuesForStep(i, true)-1);
                    }

                    TraceManager.addDev("Index=" + index + " New regular range at index " + i + "=(" + currentMinRegularIndex[i] + "," + currentMaxRegularIndex[i] + ")");

                    //malware
                    index = best.arrayOfValueIndexMalware[i];
                    diff = (currentMaxMalwareIndex[i] - currentMinMalwareIndex[i])/2;
                    if (diff < 2) {
                        currentMaxMalwareIndex[i] = index;
                        currentMinMalwareIndex[i] = index;
                    } else {
                        currentMinMalwareIndex[i] = Math.max(index - diff, 0);
                        currentMaxMalwareIndex[i] = Math.min(currentMinMalwareIndex[i]+diff, readerRegular.getNbOfValuesForStep(i, false)-1);
                    }

                    TraceManager.addDev("Index=" + index + " New malware range at index " + i + "=(" + currentMinMalwareIndex[i] + "," + currentMaxMalwareIndex[i] + ")");
                }



                canGo = !noMoreValuesIn(currentMinRegularIndex, currentMaxRegularIndex);
                TraceManager.addDev("canGo=" + canGo);
                canGo = canGo && !noMoreValuesIn(currentMinMalwareIndex, currentMaxMalwareIndex);
            }

            //canGo = false;
        }


        set.clear();
        AlligatorEvaluationSet aes = best.fullClone();

        /*for(int i=0; i<nbOfSteps; i++) {
          aes.arrayOfValueIndexRegular[i] = bestIndex[i];
          aes.arrayOfValueIndexMalware[i] = bestIndex[i+nbOfSteps];
          }*/
        set.add(aes);
        //readerRegular.computeSet(set, true);
        //readerMalware.computeSet(set, false);

        return aes;
    }

    private void elaborateSet(AlligatorEvaluationSet aes, int nbOfSteps, int currentStep, int nbOfValuesPerTry, int[] currentMinRegularIndex, int[] currentMinMalwareIndex, int[] currentMaxRegularIndex, int[] currentMaxMalwareIndex) {
        if (currentStep == nbOfSteps) {
            set.add(aes);
            return;
        }

        if (aes == null) {
            aes = new AlligatorEvaluationSet(nbOfSteps);
        }

        AlligatorEvaluationSet tmp, tmp1;

        double diffr = (currentMaxRegularIndex[currentStep] - currentMinRegularIndex[currentStep])/(nbOfValuesPerTry-1);
        double diffm = (currentMaxMalwareIndex[currentStep] - currentMinMalwareIndex[currentStep])/(nbOfValuesPerTry-1);

        if (diffr == 0) {
        }
        for(int i=0; i<nbOfValuesPerTry; i++) {
            tmp = aes.fullClone();
            tmp.arrayOfValueIndexRegular[currentStep] = Math.min((int)(diffr * i + currentMinRegularIndex[currentStep]), currentMaxRegularIndex[currentStep]);
            for(int j=0; j<nbOfValuesPerTry; j++) {
                tmp1 = tmp.fullClone();
                tmp1.arrayOfValueIndexMalware[currentStep] = Math.min((int)(diffr * j + currentMinMalwareIndex[currentStep]), currentMaxMalwareIndex[currentStep]);
                elaborateSet(tmp1, nbOfSteps, currentStep + 1, nbOfValuesPerTry, currentMinRegularIndex, currentMinMalwareIndex, currentMaxRegularIndex, currentMaxMalwareIndex);
            }
        }
    }


    private boolean noMoreValuesIn(int[] mins, int[] maxs) {
        for (int i=0; i<mins.length; i++) {
            if (maxs[i] - mins[i] > 0) {
                return false;
            }
        }
        return true;
    }

    public AlligatorEvaluationSet oneOtherProgressiveAverage() {
        // Set up one thread per cluster
        AlligatorLearnerExecutor executorRegular = new AlligatorLearnerExecutor(readerRegular, "Regular");
        AlligatorLearnerExecutor executorMalware = new AlligatorLearnerExecutor(readerMalware, "Malware");

        // For each cluster element of each cluster, we compute all possible values

        executorRegular.go();
        executorMalware.go();

        executorRegular.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);
        executorMalware.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);

        executorRegular.waitForDone();
        executorMalware.waitForDone();

        // Computing sets of attemps
        TraceManager.addDev("Computing sets of attempts");
        int nbOfSteps = readerRegular.getNbOfSteps();
        int [] nbOfValuesPerStepRegular = new int[nbOfSteps];
        int [] nbOfValuesPerStepMalware = new int[nbOfSteps];

        long nb = 1;
        for(int i=0; i<nbOfSteps; i++) {
            nbOfValuesPerStepRegular[i] = readerRegular.getNbOfValuesForStep(i, true);
            nbOfValuesPerStepMalware[i] = readerRegular.getNbOfValuesForStep(i, false);
            //TraceManager.addDev("Nb Of value for step " + i + " regular:" + nbOfValuesPerStepRegular[i] + " malware:" + nbOfValuesPerStepMalware[i]);
            nb = nb + nbOfValuesPerStepMalware[i] + nbOfValuesPerStepRegular[i];
        }

        TraceManager.addDev("Nb of sets to compute: " + nb);

        set = new Vector<AlligatorEvaluationSet>();
        AlligatorEvaluationSet average = new AlligatorEvaluationSet(nbOfSteps);
        AlligatorEvaluationSet aes;

        // Filling the average
        for(int i=0; i<nbOfSteps; i++) {
            average.arrayOfValueIndexRegular[i] = readerRegular.getRegularAverageIndex(i);
            average.arrayOfValueIndexMalware[i] = readerRegular.getMalwareAverageIndex(i);
        }

        readerRegular.computeOne(average, true);
        readerMalware.computeOne(average, false);
        TraceManager.addDev("Average index: " + average);



        int []bestIndex = new int[nbOfSteps*2];
        int percentage = 0;
        for(int i=0; i<nbOfSteps; i++) {
            // For each steps we generate a set of AlligatorEvaluationSet -> we compute the best one
            // We then keep track of the best index to be used
            // Regular, then malware
            // Regular computation
            percentage ++;
            if (nbOfValuesPerStepRegular[i] == 1) {
                bestIndex[i] = 0;
            } else {

                set.clear();
                for(int j=0; j<nbOfValuesPerStepRegular[i]; j++) {
                    aes = average.fullClone();
                    aes.arrayOfValueIndexRegular[i] = j;
                    set.add(aes);
                }
                readerRegular.computeSet(set, true);
                readerMalware.computeSet(set, false);
                aes = getBestSet();

                //TraceManager.addDev("List of all :" + set);
                bestIndex[i] = aes.arrayOfValueIndexRegular[i];
                //TraceManager.addDev("Best set step " + i + " = " + aes);
                //TraceManager.addDev("Best regular index found for step " + i + " = " + bestIndex[i]);

                printPercentage(100*percentage/(double)(2*nbOfSteps));
                average.arrayOfValueIndexRegular[i] = bestIndex[i];

            }

            // Malware computation
            percentage ++;
            if (nbOfValuesPerStepMalware[i] == 1) {
                bestIndex[i+nbOfSteps] = 0;
            } else {

                set.clear();
                for(int j=0; j<nbOfValuesPerStepMalware[i]; j++) {
                    aes = average.fullClone();
                    aes.arrayOfValueIndexMalware[i] = j;
                    set.add(aes);
                }
                readerRegular.computeSet(set, true);
                readerMalware.computeSet(set, false);
                aes = getBestSet();

                //TraceManager.addDev("List of all :" + set);
                bestIndex[i+nbOfSteps] = aes.arrayOfValueIndexMalware[i];
                //TraceManager.addDev("Best set step " + i + " = " + aes);
                //TraceManager.addDev("Best malware index found for step " + i + " = " + bestIndex[i+nbOfSteps]);
                average.arrayOfValueIndexMalware[i] = bestIndex[i+nbOfSteps];
                printPercentage(100*percentage/(double)(2*nbOfSteps));
            }


        }

        // From all bestIndex, building the best AlligatorEvaluationSet, and computing results
        set.clear();
        aes = average.fullClone();

        for(int i=0; i<nbOfSteps; i++) {
            aes.arrayOfValueIndexRegular[i] = bestIndex[i];
            aes.arrayOfValueIndexMalware[i] = bestIndex[i+nbOfSteps];
        }
        set.add(aes);
        readerRegular.computeSet(set, true);
        readerMalware.computeSet(set, false);

        return aes;


    }


    public AlligatorEvaluationSet oneOtherProgressiveMultipassAverage(boolean twoTogether) {
        // Set up one thread per cluster
        AlligatorLearnerExecutor executorRegular = new AlligatorLearnerExecutor(readerRegular, "Regular");
        AlligatorLearnerExecutor executorMalware = new AlligatorLearnerExecutor(readerMalware, "Malware");

        // For each cluster element of each cluster, we compute all possible values

        executorRegular.go();
        executorMalware.go();

        executorRegular.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);
        executorMalware.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);

        executorRegular.waitForDone();
        executorMalware.waitForDone();

        // Computing sets of attemps
        TraceManager.addDev("Computing sets of attempts");
        int nbOfSteps = readerRegular.getNbOfSteps();
        int [] nbOfValuesPerStepRegular = new int[nbOfSteps];
        int [] nbOfValuesPerStepMalware = new int[nbOfSteps];

        long nb = 1;
        for(int i=0; i<nbOfSteps; i++) {
            nbOfValuesPerStepRegular[i] = readerRegular.getNbOfValuesForStep(i, true);
            nbOfValuesPerStepMalware[i] = readerRegular.getNbOfValuesForStep(i, false);
            //TraceManager.addDev("Nb Of value for step " + i + " regular:" + nbOfValuesPerStepRegular[i] + " malware:" + nbOfValuesPerStepMalware[i]);
            nb = nb + nbOfValuesPerStepMalware[i] + nbOfValuesPerStepRegular[i];
        }

        TraceManager.addDev("Nb of sets to compute: " + nb);

        set = new Vector<AlligatorEvaluationSet>();
        AlligatorEvaluationSet average = new AlligatorEvaluationSet(nbOfSteps);
        //AlligatorEvaluationSet aes;

        // Filling the average
        for(int i=0; i<nbOfSteps; i++) {
            average.arrayOfValueIndexRegular[i] = readerRegular.getRegularAverageIndex(i);
            average.arrayOfValueIndexMalware[i] = readerRegular.getMalwareAverageIndex(i);
        }

        readerRegular.computeOne(average, true);
        readerMalware.computeOne(average, false);
        TraceManager.addDev("Average index: " + average);

        return getBestMultiPass(twoTogether, 100, nbOfSteps, nbOfValuesPerStepRegular, nbOfValuesPerStepMalware, average);
    }


    public AlligatorEvaluationSet multipassFromBest(AlligatorEvaluationSet currentBest, int nbOfPass, boolean twoTogether) {

        if (currentBest.isOptimalSet(bestSetOnPercentageOnly)) {
            return currentBest;
        }

        // Set up one thread per cluster
        if ((executorRegular == null) || (executorMalware == null)) {
            executorRegular = new AlligatorLearnerExecutor(readerRegular, "Regular");
            executorMalware = new AlligatorLearnerExecutor(readerMalware, "Malware");

            // For each cluster element of each cluster, we compute all possible values

            executorRegular.go();
            executorMalware.go();

            executorRegular.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);
            executorMalware.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);

            executorRegular.waitForDone();
            executorMalware.waitForDone();
        }

        // Computing sets of attempts
        TraceManager.addDev("Computing nb of values per computation step");
        int nbOfSteps = readerRegular.getNbOfSteps();
        int [] nbOfValuesPerStepRegular = new int[nbOfSteps];
        int [] nbOfValuesPerStepMalware = new int[nbOfSteps];

        //long nb = 1;
        for(int i=0; i<nbOfSteps; i++) {
            nbOfValuesPerStepRegular[i] = readerRegular.getNbOfValuesForStep(i, true);
            nbOfValuesPerStepMalware[i] = readerRegular.getNbOfValuesForStep(i, false);
            //TraceManager.addDev("Nb Of value for step " + i + " regular:" + nbOfValuesPerStepRegular[i] + " malware:" + nbOfValuesPerStepMalware[i]);
            //nb = nb + nbOfValuesPerStepMalware[i] + nbOfValuesPerStepRegular[i];
        }

 
        //TraceManager.addDev("Nb of sets to compute: " + nb);

        set = new Vector<AlligatorEvaluationSet>();

        TraceManager.addDev("Nb of compute:" + currentBest.arrayOfValueIndexRegular.length);
        TraceManager.addDev("Best before multipass: " + currentBest);

        if (currentBest.isOptimalSet(bestSetOnPercentageOnly)) {
            return currentBest;
        }

        return getBestMultiPass(twoTogether, nbOfPass, nbOfSteps, nbOfValuesPerStepRegular, nbOfValuesPerStepMalware, currentBest);
    }

    public AlligatorEvaluationSet getBestMultiPass(boolean twoTogether, int nbOfRemainingPass, int nbOfSteps, int [] nbOfValuesPerStepRegular, int [] nbOfValuesPerStepMalware, AlligatorEvaluationSet currentBest) {
        //System.out.println("Remaining pass #" + nbOfRemainingPass + " current best percentage: " +  ((int)(currentBest.getBestPercentage()*10))/10.0 + "%\r");

        if (currentBest.isOptimalSet(bestSetOnPercentageOnly)) {
            return currentBest;
        }

        AlligatorEvaluationSet current = currentBest.fullClone();
        AlligatorEvaluationSet aes, aes1;

        int []bestIndex = new int[nbOfSteps*2];
        int percentage = 0;
        for(int i=0; i<nbOfSteps; i++) {
            // For each steps we generate a set of AlligatorEvaluationSet -> we compute the best one
            // We then keep track of the best index to be used
            // Regular, then malware
            // Regular computation
	    set.clear();

            if (twoTogether) {
                percentage ++;

                for(int j=0; j<nbOfValuesPerStepRegular[i]; j++) {
                    if (j % 10 == 0) {
                        printPercentage(Math.min(100, 100*(((double)(percentage)+1.0/nbOfValuesPerStepRegular[i]))/(double)(nbOfSteps)), "in building  sets. Step: " + i + " Nb Of sets: " + (nbOfValuesPerStepRegular[i]*nbOfValuesPerStepMalware[i]) + " Remaining pass #" + nbOfRemainingPass + " current best percentage: " +  ((int)(currentBest.getBestPercentage()*10))/10.0 + "%      ");
                    }
                    aes = current.fullClone();
                    aes.arrayOfValueIndexRegular[i] = j;
                    for (int k=0; k<nbOfValuesPerStepMalware[i]; k++) {
                        aes1=aes.fullClone();
                        aes1.arrayOfValueIndexMalware[i] = k;
                        set.add(aes1);
                    }
                }

                readerRegular.computeSet(set, true);
                readerMalware.computeSet(set, false);
                aes = getBestSet();
                bestIndex[i] = aes.arrayOfValueIndexRegular[i];
                bestIndex[i+nbOfSteps] = aes.arrayOfValueIndexMalware[i];
                current.arrayOfValueIndexRegular[i] = bestIndex[i];
                current.arrayOfValueIndexMalware[i] = bestIndex[i+nbOfSteps];

            } else {
                percentage ++;
                if (nbOfValuesPerStepRegular[i] == 1) {
                    bestIndex[i] = 0;
                } else {

                    set.clear();
                    for(int j=0; j<nbOfValuesPerStepRegular[i]; j++) {
                        aes = current.fullClone();
                        aes.arrayOfValueIndexRegular[i] = j;
                        set.add(aes);
                    }
                    readerRegular.computeSet(set, true);
                    readerMalware.computeSet(set, false);
                    aes = getBestSet();

                    //TraceManager.addDev("List of all :" + set);
                    bestIndex[i] = aes.arrayOfValueIndexRegular[i];
                    //TraceManager.addDev("Best set step " + i + " = " + aes);
                    //TraceManager.addDev("Best regular index found for step " + i + " = " + bestIndex[i]);

                    printPercentage(100*percentage/(double)(2*nbOfSteps), "Remaining pass #" + nbOfRemainingPass + " current best percentage: " +  ((int)(currentBest.getBestPercentage()*10))/10.0 + "%      ");
                    current.arrayOfValueIndexRegular[i] = bestIndex[i];

                }

                // Malware computation
                percentage ++;
                if (nbOfValuesPerStepMalware[i] == 1) {
                    bestIndex[i+nbOfSteps] = 0;
                } else {

                    set.clear();
                    for(int j=0; j<nbOfValuesPerStepMalware[i]; j++) {
                        aes = current.fullClone();
                        aes.arrayOfValueIndexMalware[i] = j;
                        set.add(aes);
                    }
                    readerRegular.computeSet(set, true);
                    readerMalware.computeSet(set, false);
                    aes = getBestSet();

                    //TraceManager.addDev("List of all :" + set);
                    bestIndex[i+nbOfSteps] = aes.arrayOfValueIndexMalware[i];
                    //TraceManager.addDev("Best set step " + i + " = " + aes);
                    //TraceManager.addDev("Best malware index found for step " + i + " = " + bestIndex[i+nbOfSteps]);
                    current.arrayOfValueIndexMalware[i] = bestIndex[i+nbOfSteps];
                    printPercentage(100*percentage/(double)(2*nbOfSteps), "Remaining pass #" + nbOfRemainingPass + " current best percentage: " +  ((int)(currentBest.getBestPercentage()*10))/10.0 + "%      ");
                }
            }


        }

        // From all bestIndex, building the best AlligatorEvaluationSet, and computing results
        aes = current.fullClone();
        aes.reinitAnalysisValues();

        for(int i=0; i<nbOfSteps; i++) {
            aes.arrayOfValueIndexRegular[i] = bestIndex[i];
            aes.arrayOfValueIndexMalware[i] = bestIndex[i+nbOfSteps];
        }

        readerRegular.computeOne(aes, true);
        readerMalware.computeOne(aes, false);

        AlligatorEvaluationSet tmp = aes.getBestNoClone(currentBest, bestSetOnPercentageOnly, bestSetOnAverage);

        if (tmp.isOptimalSet(bestSetOnPercentageOnly)) {
            return tmp;
        } else if (tmp == currentBest) {
            return currentBest;
        } else {

            TraceManager.addUser("Last best:" + tmp + "\n\n");

            nbOfRemainingPass --;
            if (nbOfRemainingPass > 0) {
                return getBestMultiPass(twoTogether, nbOfRemainingPass, nbOfSteps, nbOfValuesPerStepRegular, nbOfValuesPerStepMalware, aes);
            } else {
                return currentBest;
            }
        }

    }


    public AlligatorEvaluationSet random(int maxTime) {
        TraceManager.addUser("Random algorithm for " + maxTime + " seconds");
        long startTime = System.currentTimeMillis();
        //long finishTime = startTime + maxTime * 1000;

        // Set up one thread per cluster
        if ((executorRegular == null) || (executorMalware == null)) {
            executorRegular = new AlligatorLearnerExecutor(readerRegular, "Regular");
            executorMalware = new AlligatorLearnerExecutor(readerMalware, "Malware");

            // For each cluster element of each cluster, we compute all possible values

            executorRegular.go();
            executorMalware.go();

            executorRegular.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);
            executorMalware.setStep(AlligatorLearnerExecutor.COMPUTE_ALL_VALUES);

            executorRegular.waitForDone();
            executorMalware.waitForDone();
        }

        // Computing sets of attemps
        //TraceManager.addDev("Computing sets of attempts");
        int nbOfSteps = readerRegular.getNbOfSteps();
        int [] nbOfValuesPerStepRegular = new int[nbOfSteps];
        int [] nbOfValuesPerStepMalware = new int[nbOfSteps];

        //long nb = 1;
        for(int i=0; i<nbOfSteps; i++) {
            nbOfValuesPerStepRegular[i] = readerRegular.getNbOfValuesForStep(i, true);
            nbOfValuesPerStepMalware[i] = readerRegular.getNbOfValuesForStep(i, false);
            //TraceManager.addDev("Nb Of value for step " + i + " regular:" + nbOfValuesPerStepRegular[i] + " malware:" + nbOfValuesPerStepMalware[i]);
            //nb = nb + nbOfValuesPerStepMalware[i] + nbOfValuesPerStepRegular[i];
        }

        //TraceManager.addDev("Nb of sets to compute: " + nb);

        //set = new Vector<AlligatorEvaluationSet>();
        AlligatorEvaluationSet best = new AlligatorEvaluationSet(nbOfSteps);
        AlligatorEvaluationSet aes;

        // Filling the average = best at the beginning
        for(int i=0; i<nbOfSteps; i++) {
            best.arrayOfValueIndexRegular[i] = readerRegular.getRegularAverageIndex(i);
            best.arrayOfValueIndexMalware[i] = readerRegular.getMalwareAverageIndex(i);
        }

        readerRegular.computeOne(best, true);
        readerMalware.computeOne(best, false);
        TraceManager.addDev("Average: " + best);

        if (best.isOptimalSet(bestSetOnPercentageOnly)) {
            return best;
        }

        long elapsedTime = (System.currentTimeMillis() - startTime)/1000;
        set = new Vector<AlligatorEvaluationSet>();
        //set.add(best);

        long nbAnalyzed = 0;
        int nbOfBestFound = 1;
        Random generator = new Random();
        AlligatorEvaluationSet tmp;
        double currentPercentage = best.getBestPercentage();

        while(elapsedTime < maxTime) {
            if ((nbAnalyzed % 10) == 0) {
                elapsedTime = (System.currentTimeMillis() - startTime)/1000;
                printAnalysedPercentage(100*elapsedTime/(double)(maxTime), nbAnalyzed, nbOfBestFound, currentPercentage);
            }
            aes = new AlligatorEvaluationSet(nbOfSteps);
            for(int i=0; i<nbOfSteps; i++) {
		//	if (nbOfValuesPerStepRegular[i] < 1) {
		//TraceManager.addDev("valueR=" + nbOfValuesPerStepRegular[i] + " ValueM=" + nbOfValuesPerStepMalware[i]);
		    //}
	
                aes.arrayOfValueIndexRegular[i] = generator.nextInt(nbOfValuesPerStepRegular[i]);
                aes.arrayOfValueIndexMalware[i] = generator.nextInt(nbOfValuesPerStepMalware[i]);
            }
            readerRegular.computeOne(aes, true);
            readerMalware.computeOne(aes, false);
            set.add(best);
            set.add(aes);
            tmp = getBestSet();

            if (tmp != best) {
                nbOfBestFound++;
                currentPercentage = tmp.getBestPercentage();
                TraceManager.addUser("Last best:" + tmp + "\n\n");
            }

            best = tmp;

            if (best.isOptimalSet(bestSetOnPercentageOnly)) {
                return best;
            }

            set.clear();
            nbAnalyzed++;


        }

        return best;

    }

    public void printAnalysedPercentage(double percentage, long nbAnalyzed, int nbOfBestFound, double currentPercentage) {
        System.out.print(" " + AlligatorLanguageReader.getIconText(step) + " " + (int)(percentage*10)/10.0 + "% done, analyzed samples: " + nbAnalyzed + ", # best found:" + nbOfBestFound + ", best %:" + (int)(currentPercentage*10)/10.0 + "%          \r");
        step++;
    }

    public void printReducedAnalysedPercentage(double percentage, double currentPercentage) {
        System.out.print(" " + AlligatorLanguageReader.getIconText(step) + " " + (int)(percentage*10)/10.0 + "% done, best %:" + (int)(currentPercentage*10)/10.0 + "%          \r");
        step++;
    }

    public void printPercentage(double percentage) {
        System.out.print(" " + AlligatorLanguageReader.getIconText(step) + " " + (int)(percentage*10)/10.0 + "% done          \r");
        step ++;
    }

    public void printPercentage(double percentage, String s) {
        System.out.print(" " + AlligatorLanguageReader.getIconText(step) + " " + (int)(percentage*10)/10.0 + "% done " + s + "          \r");
        step ++;
    }

    public void evaluateMaxNbOfScoresInAlligatorScore() {
        // We limit the memory to 1GB
        long max = 1073741824L;

        // We manipulate double scores i.e. 8B per score
        // We take an average for each algorithm
        // Nb of elements:
        long nb = readerRegular.getGuessCluster().getElements().size() + readerMalware.getGuessCluster().getElements().size();

        // We multiply that number per 8 (size of double)
        nb = nb * 8;

        // We need to multiply this by the number of algorithm, multiplied by 2 (regular, malware)
        nb = nb * readerRegular.getNbOfSteps() * 2;

        TraceManager.addDev("Maximum nb of values per algo:" + max/nb);
        AlligatorScore.MAX_NB_OF_SCORES = (int)(max/nb);


    }

    // Returns whether it could forget elements or not
    public boolean forget(AlligatorEvaluationSet aes, int maxPercentage, boolean force) {
        // If recognition percentage is 100% -> must forget the one with the lowest  min
        // If recognition is not 100% -> forget the badly classified ones one with the highest maxDiff
        // Forget for at mostthe provided percentage

        int currentForgotten = getNbOfForgotten();

        int total = readerRegular.getNbOfGuessSamples();
        total += readerMalware.getNbOfGuessSamples();

        int maxNbOfElementsThatCanBeForgotten = total * maxPercentage / 100;

        int maxToForget = maxNbOfElementsThatCanBeForgotten - currentForgotten;

        // Max nb of forgotten samples reached?
        if (maxToForget <= 0) {
            TraceManager.addDev("Maximum nb of forgotten elements reached");
            return false;
        }

        Vector<AlligatorElementReference> elements;

        // Recognition rate of 100% -> only in agressiveForget
        // Those which are too close to the
        if ((aes.nbOfRegularFailed == 0) && (aes.nbOfMalwareFailed == 0)) {
            if (!force) {
                return false;
            }
            elements = new Vector<AlligatorElementReference>(aes.nbOfRegularFailed + aes.nbOfMalwareFailed + aes.nbOfRegularRecognized + aes.nbOfMalwareRecognized);
            readerRegular.fillWithClassificationMinValue(aes, true, elements);
            readerMalware.fillWithClassificationMinValue(aes, false, elements);

            TraceManager.addDev("Forget/nb of elements: " + elements.size());

            Collections.sort(elements);

            if (elements.size() > 0) {
                TraceManager.addDev(" First:" + elements.get(0));
                if (elements.size() > 1) {
                    TraceManager.addDev(" Second:" + elements.get(1));
                }
            }

            AlligatorElementReference aer;
            for(int i=0; i<Math.min(elements.size(), maxToForget); i++) {
                aer = elements.get(i);
                if (aer.isRegular) {
                    readerRegular.classificationWeight[aer.index] = 0;
                } else {
                    readerMalware.classificationWeight[aer.index] = 0;
                }
            }

            return true;
        }

        // Needs to forget the baldy recognized ones, with a maximum of maxToForgot
        // Classify the badly recognized ones according to their score difference
        elements = new Vector<AlligatorElementReference>(aes.nbOfRegularFailed + aes.nbOfMalwareFailed);
        readerRegular.fillWithClassificationValue(aes, true, elements);
        readerMalware.fillWithClassificationValue(aes, false, elements);

        TraceManager.addDev("Forget/nb of elements: " + elements.size());

        if (elements.size() == 0) {
            return false;
        }

        Collections.sort(elements);
        Collections.reverse(elements);

        if (elements.size() > 0) {
            TraceManager.addDev(" First:" + elements.get(0));
            if (elements.size() > 1) {
                TraceManager.addDev(" Second:" + elements.get(1));
            }
        }

        AlligatorElementReference aer;
        for(int i=0; i<Math.min(elements.size(), maxToForget); i++) {
            aer = elements.get(i);
            if (aer.isRegular) {
                readerRegular.classificationWeight[aer.index] = 0;
            } else {
                readerMalware.classificationWeight[aer.index] = 0;
            }
        }


        return true;
    }



    public boolean boost(AlligatorEvaluationSet aes, int maxPercentage, boolean force) {
        // If recognition percentage is 100% -> must boost the ones with the lowest  min
        // If recognition is not 100% -> boost the badly classified ones one with the highest maxDiff
        // Boost for at most the provided percentage


        int total = readerRegular.getNbOfGuessSamples();
        total += readerMalware.getNbOfGuessSamples();

        int totalNb = total * maxPercentage / 100;

        int maxToBoost = total * maxPercentage / 100;

        Vector<AlligatorElementReference> elements;

        // Recognition rate of 100% -> only in agressiveBoost
        // Those which are too close to the max
        if ((aes.nbOfRegularFailed == 0) && (aes.nbOfMalwareFailed == 0)) {
            if (!force) {
                return false;
            }
            elements = new Vector<AlligatorElementReference>(aes.nbOfRegularFailed + aes.nbOfMalwareFailed + aes.nbOfRegularRecognized + aes.nbOfMalwareRecognized);
            readerRegular.fillWithClassificationMinValue(aes, true, elements);
            readerMalware.fillWithClassificationMinValue(aes, false, elements);

            TraceManager.addDev("Boost/nb of elements: " + elements.size());

            Collections.sort(elements);

            if (elements.size() > 0) {
                TraceManager.addDev(" First:" + elements.get(0));
                if (elements.size() > 1) {
                    TraceManager.addDev(" Second:" + elements.get(1));
                }
            }

            AlligatorElementReference aer;
            for(int i=0; i<Math.min(elements.size(), maxToBoost); i++) {
                aer = elements.get(i);
                if (aer.isRegular) {
                    readerRegular.classificationWeight[aer.index] ++;
                } else {
                    readerMalware.classificationWeight[aer.index] ++;
                }
            }

            return true;
        }

        // Needs to boost the baldy recognized ones, with a maximum of maxToBoost
        // Classify the badly recognized ones according to their score difference
        elements = new Vector<AlligatorElementReference>(aes.nbOfRegularFailed + aes.nbOfMalwareFailed);
        readerRegular.fillWithClassificationValue(aes, true, elements);
        readerMalware.fillWithClassificationValue(aes, false, elements);

        TraceManager.addDev("Boost/nb of elements: " + elements.size());

        if (elements.size() == 0) {
            return false;
        }

        Collections.sort(elements);
        Collections.reverse(elements);

        if (elements.size() > 0) {
            TraceManager.addDev(" First:" + elements.get(0));
            if (elements.size() > 1) {
                TraceManager.addDev(" Second:" + elements.get(1));
            }
        }

        AlligatorElementReference aer;
        for(int i=0; i<Math.min(elements.size(), maxToBoost); i++) {
            aer = elements.get(i);
            if (aer.isRegular) {
                readerRegular.classificationWeight[aer.index] ++;
            } else {
                readerMalware.classificationWeight[aer.index] ++;
            }
        }


        return true;



    }

    public int getNbOfForgotten() {
        int currentForgotten = readerRegular.getNbOfForgotten();
        currentForgotten += readerMalware.getNbOfForgotten();
        return currentForgotten;
    }

    public int getNbOfBoost() {
        int currentBoost = getNbOfRegularBoost();
	currentBoost += getNbOfMalwareBoost();
        return currentBoost;
    }

    public int getNbOfRegularForgotten() {
        return readerRegular.getNbOfForgotten();
    }

    public int getNbOfMalwareForgotten() {
        return readerMalware.getNbOfForgotten();
    }

    public int getNbOfRegularBoost() {
        return readerRegular.getNbOfBoost();
    }

    public int getNbOfMalwareBoost() {
        return readerMalware.getNbOfBoost();
    }

}
