/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

   ludovic.apvrille AT telecom-paristech.fr

   This software is a computer program whose purpose is to allow malware analysis

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class AlligatorLearnerExecutor
   * Executes algorithm in a different thread
   * Creation: 23/07/2012
   * @version 1.0 23/07/2012
   * @author Ludovic APVRILLE
   * @see
   */

package analyzer;


import myutil.*;

import java.util.*;



public class AlligatorLearnerExecutor extends Thread  {

    public static final int DONE = 1;

    public static final int COMPUTE_ALL_VALUES = 2;
    public static final int COMPUTE_VALUES_FOR_REGULAR = 3;
    public static final int COMPUTE_VALUES_FOR_MALWARE = 4;
    public static final int EXIT = 5;

    private AlligatorLanguageReader reader;

    private int step;

    private String name;

    private Vector<AlligatorEvaluationSet> set;


    public AlligatorLearnerExecutor(AlligatorLanguageReader _reader, String _name) {
        reader = _reader;
        name = _name;
        step = DONE;
    }

    public void setSet(Vector<AlligatorEvaluationSet> _set) {
        set = _set;
    }

    public void go() {
        this.start();
    }

    public synchronized void waitForDone() {
        while(step != DONE) {
            try {
                wait();
            } catch (InterruptedException ie) {
            }
        }
    }

    public synchronized void waitForCommand() {
        while(step == DONE) {
            try {
                wait();
            } catch (InterruptedException ie) {
            }
        }
    }

    public synchronized void setStep(int stepID) {
        step = stepID;
        notifyAll();
    }

    public int getStep() {
        return step;
    }

    public void run() {
        TraceManager.addDev("Starting computing values");
        boolean go = true;

        while(go == true) {
            waitForCommand();

            switch(step) {
            case COMPUTE_ALL_VALUES:
                reader.computeAllValues();
                break;
            case COMPUTE_VALUES_FOR_REGULAR:
                reader.computeSet(set, true);
                break;
            case COMPUTE_VALUES_FOR_MALWARE:
                reader.computeSet(set, false);
                break;

            case EXIT:
                go = false;
                break;
            }

            setStep(DONE);
        }

        TraceManager.addDev("All done in thread " + name);

        // Then extract the number of steps
    }

    public int getIteration() {
        return reader.getIteration();
    }


}
