/**Copyright or (C) or Copr. GET / ENST, Telecom-Paris, Ludovic Apvrille

   ludovic.apvrille AT enst.fr

   This software is a computer program whose purpose is to allow the
   edition of TURTLE analysis, design and deployment diagrams, to
   allow the generation of RT-LOTOS or Java code from this diagram,
   and at last to allow the analysis of formal validation traces
   obtained from external tools, e.g. RTL from LAAS-CNRS and CADP
   from INRIA Rhone-Alpes.

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class AlligatorScore
   * Creation: 23/07/2012
   * version 1.0 23/07/2012
   * @author Ludovic APVRILLE
   * @see
   */

package analyzer;

import myutil.*;


public  abstract class AlligatorScore {

    public static int MAX_NB_OF_SCORES = 100;
    public static int MAX_NB_OF_STEPS = 1000;

    public AlligatorInfoScore ais;

    public double intermediateScore; // rawscore on which complex expression has been applied
    public boolean scoreComputed;
    public double lastScore;

    private static DoubleExpressionEvaluator dee;

    public int nbOfScores;

    public double[] allPossibleScores;

    public AlligatorScore(AlligatorInfoScore _ais, double _intermediateScore) {
        ais = _ais;
        scoreComputed = false;
        intermediateScore = _intermediateScore;
    }

    public void evaluateExpression(String expr) throws AlligatorSyntaxException {
        // Replace "x" with value
        String expr1 = Conversion.replaceAllString(expr, "x", String.format("%f", intermediateScore));

        if (dee == null) {
            dee = new DoubleExpressionEvaluator();
        }

        intermediateScore = dee.getResultOf(expr1);

        //TraceManager.addDev("Computing expr:" + expr1 + " result=" + intermediateScore);  

        if (dee.hasError()) {
            throw new AlligatorSyntaxException("Invalid expression:" + expr + " because: " + dee.getError());
        }

    }

    public double getScore(int index) {
        return allPossibleScores[index];
    }

    public double getRegularMultiplier(int index) {
        return ais.regularMinMultiplier + ais.regularStep * index;
    }

    public double getMalwareMultiplier(int index) {
        return ais.malwareMinMultiplier + ais.malwareStep * index;
    }

    public int getAverageIndex() {
        return nbOfScores/2;
    }

    public double getMultipliedScore(int multiplier) {
	return intermediateScore * multiplier;
    }

    public abstract void computeMaxScore();
    public abstract void computeAverageScore();
    public abstract void computeComplexExpression() throws AlligatorSyntaxException;
    public abstract void computeAllPossibleScores();


} // Class
