/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

   ludovic.apvrille AT telecom-paristech.fr

   This software is a computer program whose purpose is to allow malware analysis

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.
   /**
   * Class ClassificationReportAnalyzer
   * Element of a classification report
   * Creation: 16/06/2014
   * @version 1.0 16/06/2014
   * @author Ludovic APVRILLE
   * @see
   */

package classificationreport;

import myutil.*;


import java.util.*;
import java.awt.*;


public class ClassificationReportAnalyzer  {
    private String name;
    private String type1, type2;

    public int MAX_NB_OF_POINTS = 1000;

    public ClassificationReportElement[] regulars, malwares;

    private ClassificationReportPoint bestPoint;
    private ClassificationReportPoint[] points;

    public ClassificationReportAnalyzer(String _name) {
        name = _name;
    }


    public boolean loadReport(String contentOfReport, boolean isRegular) {
        ArrayList<ClassificationReportElement> elements = new ArrayList<ClassificationReportElement>();

        // Analyze report line by line
        // Searches for lines of the form "name: type (type1:double, type2:double)" with type is type1 or *type2*

        Scanner scanner = new Scanner(contentOfReport);
        String split[];

        double score1, score2;
        String type;
        boolean typeRegular;
        String name;
        ClassificationReportElement cre;
        int cpt = 0;

        while (scanner.hasNextLine()) {
            String line = scanner.nextLine();
            line = line.trim();
            line = Conversion.replaceAllString(line, "\t", " ");
            line = Conversion.replaceAllString(line, "  ", " ");
            //TraceManager.addDev("line=" + line);
            split = line.split(" ");
            if (split.length == 4) {
                //TraceManager.addDev("0=" + split[0] + "/1=" + split[1] + "/2=" +split[2] + "/3=" + split[3]);
                if (split[0].charAt(split[0].length()-1) == ':') {
                    // name
                    name = split[0].substring(0, split[0].length()-1);
                    //type
                    type = split[1];
                    if (type.startsWith("*")) {
                        typeRegular = false;
                        if (type2 == null) {
                            type2 = type.substring(1, type.length()-1);
                        }
                    } else {
                        typeRegular = true;
                        type1 = type;
                    }
                    // Regular score
                    score1 = getScore(split[2]);
                    score2 = getScore(split[3]);
                    if ((score1 >= 0) && (score2 >= 0)) {
                        cre = new ClassificationReportElement(name);
                        cre.indexInReport = cpt;
                        cre.isClassifiedAsRegular = typeRegular;
                        cre.scoreRegular = score1;
                        cre.scoreMalware = score2;
                        elements.add(cre);
                        cpt++;
                        //TraceManager.addDev("Good line=" + line);
                        //TraceManager.addDev("  Element=" + cre);
                    }
                }
            }
        }

        if (elements.size() == 0) {
            TraceManager.addDev("No element found");
            return false;
        }

        if (isRegular) {
            regulars = elements.toArray(new ClassificationReportElement[elements.size()]);
        } else {
            malwares = elements.toArray(new ClassificationReportElement[elements.size()]);
        }

        return true;
    }

    private double getScore(String s) {
        int index = s.indexOf(':');
        if (index == -1) {
            return -1;
        }

        s = s.substring(index+1, s.length()-1);

        double d;
        try {
            d = Double.valueOf(s).doubleValue();
        } catch (Exception e) {
            return -1;
        }

        return d;
    }

    public double getMaxDifference() {
        double best = 0;
        double tmp;

        for (int i=0; i<regulars.length; i++) {
            tmp = Math.abs(regulars[i].scoreRegular - regulars[i].scoreMalware);
            best = Math.max(tmp, best);
        }

        for (int j=0; j<malwares.length; j++) {
            tmp = Math.abs(malwares[j].scoreRegular - malwares[j].scoreMalware);
            best = Math.max(tmp, best);
        }

        return best;
    }


    public boolean performSliderAnalysis() {
        double diff = getMaxDifference() + 1;
        TraceManager.addDev("Max difference =" + diff);

        diff = diff * 2;
        int nbOfPoints = MAX_NB_OF_POINTS;
        if (diff < MAX_NB_OF_POINTS) {
            nbOfPoints = Math.max(2, (int)(diff));
        }

        TraceManager.addDev("Computing for " + nbOfPoints);

        points = new ClassificationReportPoint[nbOfPoints];
        ClassificationReportPoint point;

        int currentDec;
        for(int i=0; i<nbOfPoints; i++) {
            currentDec = (int)((i * diff /(nbOfPoints-1))-(diff/2));
            // Computing the points for currentDec applied to regular score
            TraceManager.addDev("current dec=" + currentDec);
            point = getPoint(currentDec);
            points[i] = point;

            if (bestPoint == null) {
                bestPoint = point;
            } else {
                if (point.isBetter(bestPoint)) {
                    bestPoint = point;
                }
            }
        }

        TraceManager.addDev("Best = " + bestPoint);
        return true;
    }

    private ClassificationReportPoint getPoint(int dec) {
        int x = 0;
        int y = 0;

        for(int i=0; i<regulars.length; i++) {
            if (regulars[i].isRegular(dec)) {
                x ++;
            }
        }

        for(int j=0; j<malwares.length; j++) {
            if (malwares[j].isMalware(dec)) {
                y ++;
            }
        }

        ClassificationReportPoint crp = new ClassificationReportPoint(dec, x, y);
        return crp;
    }




    public boolean generateReport() {
        return true;
    }

    public String getSummary() {
        String ret = "";
        if (regulars == null) {
            ret += "No element in first report";
        } else {
            ret += regulars.length + " elements in first report";
        }
        ret += "  ";
        if (malwares == null) {
            ret += "No element in second report";
        } else {
            ret += malwares.length + " elements in second report";
        }
        return ret;
    }

    // Sorting according to Malware score
    public boolean sortMalware(boolean regular) {
        ClassificationReportElement.CLASSIFICATION_MODE = ClassificationReportElement.MALWARE_CLASSIFICATION;
        return sort(regular);
    }

    public boolean sortRegular(boolean regular) {
        ClassificationReportElement.CLASSIFICATION_MODE = ClassificationReportElement.REGULAR_CLASSIFICATION;
        return sort(regular);
    }

    public boolean sortDiff(boolean regular) {
        ClassificationReportElement.CLASSIFICATION_MODE = ClassificationReportElement.DIFF_CLASSIFICATION;
        return sort(regular);
    }
    
    public boolean sortAbsDiff(boolean regular) {
        ClassificationReportElement.CLASSIFICATION_MODE = ClassificationReportElement.ABS_DIFF_CLASSIFICATION;
        return sort(regular);
    }

    public boolean sort(boolean isRegular) {
        if (isRegular) {
            Arrays.sort(regulars);
        } else {
            Arrays.sort(malwares);
        }
        return true;
    }

    public void reverse(boolean isRegular) {
        if (isRegular) {
            reverse(regulars);
        } else {
            reverse(malwares);
        }
    }


    public void reverse(ClassificationReportElement []data) {
        int left = 0;
        int right = data.length - 1;

        while( left < right ) {
            // swap the values at the left and right indices
            ClassificationReportElement temp = data[left];
            data[left] = data[right];
            data[right] = temp;

            // move the left and right index pointers in toward the center
            left++;
            right--;
        }
    }

    public String  getStringCluster(boolean isRegular) {
        if (isRegular) {
            return getStringCluster(regulars);
        } else {
            return getStringCluster(malwares);
        }
    }

    public String getStringCluster(ClassificationReportElement []cres) {
        String rep="";
        for (int i=0; i<cres.length; i++) {
            rep += cres[i].toString() + "\n";
        }
        return rep;
    }

}
