/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

   ludovic.apvrille AT telecom-paristech.fr

   This software is a computer program whose purpose is to allow malware analysis

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class Cluster
   * Notion of cluster
   * Creation: 22/06/2012
   * @version 2.0 25/09/2014
   * @author Ludovic APVRILLE
   * @see
   */

package cluster;

import myutil.*;
import java.io.*;
import java.util.*;
import java.awt.*;

//import edu.berkeley.compbio.jlibsvm.ContinuousModel;
import edu.berkeley.compbio.jlibsvm.DiscreteModel;
import edu.berkeley.compbio.jlibsvm.ImmutableSvmParameterPoint;
import edu.berkeley.compbio.jlibsvm.MutableSvmProblem;
import edu.berkeley.compbio.jlibsvm.SVM;
import edu.berkeley.compbio.jlibsvm.SolutionModel;
//import edu.berkeley.compbio.jlibsvm.SvmException;
//import edu.berkeley.compbio.jlibsvm.binary.BinaryClassificationSVM;
import edu.berkeley.compbio.jlibsvm.binary.C_SVC;
import edu.berkeley.compbio.jlibsvm.binary.MutableBinaryClassificationProblemImpl;
//import edu.berkeley.compbio.jlibsvm.binary.Nu_SVC;
//import edu.berkeley.compbio.jlibsvm.kernel.GammaKernel;
//import edu.berkeley.compbio.jlibsvm.kernel.GaussianRBFKernel;
import edu.berkeley.compbio.jlibsvm.kernel.LinearKernel;
//import edu.berkeley.compbio.jlibsvm.kernel.PolynomialKernel;
//import edu.berkeley.compbio.jlibsvm.kernel.PrecomputedKernel;
//import edu.berkeley.compbio.jlibsvm.kernel.SigmoidKernel;
//import edu.berkeley.compbio.jlibsvm.labelinverter.ByteLabelInverter;
//import edu.berkeley.compbio.jlibsvm.multi.MultiClassificationSVM;
//import edu.berkeley.compbio.jlibsvm.multi.MutableMultiClassProblemImpl;
//import edu.berkeley.compbio.jlibsvm.oneclass.OneClassSVC;
//import edu.berkeley.compbio.jlibsvm.regression.EpsilonSVR;
//import edu.berkeley.compbio.jlibsvm.regression.MutableRegressionProblemImpl;
//import edu.berkeley.compbio.jlibsvm.regression.Nu_SVR;
//import edu.berkeley.compbio.jlibsvm.regression.RegressionSVM;
//import edu.berkeley.compbio.jlibsvm.scaler.NoopScalingModel;
import edu.berkeley.compbio.jlibsvm.util.SparseVector;

//import java.applet.Applet;
//import java.awt.*;
//import java.awt.event.ActionEvent;
//import java.awt.event.ActionListener;
//import java.awt.event.MouseEvent;
//import java.io.BufferedReader;
//import java.io.DataOutputStream;
//import java.io.File;
//import java.io.FileOutputStream;
//import java.io.FileReader;
//import java.io.IOException;
//import java.util.HashSet;
//import java.util.Set;
//import java.util.StringTokenizer;
//import java.util.Vector;



public class Cluster implements PercentageProvider {

    public static int MAX_DISPLAYED_PROPS = 500;

    public final static int ERROR = -1;
    public final static int REGULAR = 0;
    public final static int MALWARE = 1;


    public static final int ONLY_TEXT = 0;
    public static final int NAME = 1;
    public static final int DEVIATION = 2;
    public static final int CLOSEST = 3;
    public static final int PROBA_DIFFERENCE = 4;
    public static final int CORRELATION = 5;
    public static final int PROBA_FACTOR = 6;
    public static final int DISTANCE_EPSILON_CLUSTER = 7;
    public static final int DEVIATION_PERCENTAGE = 8;
    public static final int DEVIATION_WEIGHT_PERCENTAGE = 9;
    public static final int WEIGHT_DEVIATION = 10;
    public static final int VALUE_TO_ADD = 11;
    public static final int DEGRESSIVE_CLOSEST = 12;
    public static final int DISTANCE_TO_AVERAGE = 13;
    public static final int FULL_CLOSEST = 14;
    public static final int FULL_DEGRESSIVE_CLOSEST = 15;
    public static final int SVM = 16;
    public static final int INVERSE_DEVIATION = 17;
    public static final int INVERSE_WEIGHT_DEVIATION = 18;


    public static final String COMMENT_SYMBOL = "#";
    public static final String PROPERTY_SYMBOL = "!";
    public static final String SEPARATOR_SYMBOL = ",";

    private String name;
    private Vector<ClusterElement> elements;
    //private ClusterElement[] elementsA;
    private int nbOfProperties = 0;
    //public int firstProperty = 0;
    //public int nbOfPropertiesToConsider = 0;
    public boolean propertiesDComputed = false;

    // Classified values DoubleInteger: value, nb of occurences
    /*public ArrayList<DoubleInteger>[] sortdPropertyValues;
    public int[] sortedPropertyTotal;  // total nb of valid properties per property index
    public int totalProbabilityDifference; // total nb of proba diff with another cluster
    public int totalProbabilityFactor; // total nb of proba factor with another cluster

    public double probaDifference;
    public double probaFactor;*/

    public String[] namesOfProperties;

    //public int[] weightOfProperties;
    //public int[] originalWeightOfProperties;

    /*public double[] meanPerProperties;
    public double[] weightedMeanPerProperties;
    public boolean[] meanPerPropertiesV;
    public boolean meanPerPropertiesComputed;
    public double[] standardDeviationPerProperties;
    //public double[] weightedStandardDeviationPerProperties;
    public boolean standardDeviationPerPropertiesV [];
    public boolean standardDeviationPerPropertiesComputed;

    public double[][] correlations; //Use of -2 when invalid

    public Vector<PropertiesCorrelation> listOfCorrelations;


    public double[] probabilityOfMostCommonValue; // -1 not relevant
    public double[] mostCommonValue;
    //public Vector<ProbabilityDifference> probaDifferences;

    //public double probaThreshold = 0.5;

    public double correlationThreadhold = 0.75;

    //public double probaRelation = 10;
    //public double lowProbaThreshold = 0.1;
    //public double highProbaThreshold = 1;

    public int nbOfClosestComputed = 0;*/

    private String topReport;

    // Epsilon clusters
    /*private Vector<MasterEpsilonClusters> masterEpsilonClusters;
    private Vector<EpsilonCluster> epsilonClusters;
    private double epsilonClusterDistanceThreshold;
    private boolean canPurgeEpsilonClusterMemory = true;*/

    private int percentageOfCurrentComputation;
    private String infoOnCurrentComputation = "unknown";


    // Options
    private boolean nocsvname = false;

    // Normalized
    private boolean normalized = false;
    private double[] normalizedValues;
    private double normalizationValue = -1; // User provided: in that case, this value is used for all properties

    private ClusterAlgoMultithreadManager manager;

    private static final int PERCENTAGE_OF_MAX_INVALID_PROPERTIES = 50;
    private int percentageOfMaxInvalidProperties = PERCENTAGE_OF_MAX_INVALID_PROPERTIES;

    // SVM
    //SolutionModel svmSolutionModel;

    public ClusterData cd; // current data
    private Vector<ClusterData> datas;



    public Cluster(String _name) {
        name = _name;
	datas = new Vector<ClusterData>();
	cd = new ClusterData();
	datas.add(cd);
    }


    /* ---------- HANDLING CLUSTER DATA ------------*/

    // The cluster change of clusterData each time a setOfProperties (min, max) is defined.
    // Also, if a weight is changed, a the deviaiton in the current profile is not null, 
    // then a new profile is created as well
    // Same when appending a cluster
    // Creation occurs when the Data is not null i.e. firstProperty is not equal to -1

    // Must be called before any computation
    public boolean appendCluster(Cluster cluster) {
        if (nbOfProperties != cluster.nbOfProperties) {
            return false;
        }
        for(ClusterElement ce: cluster.getElements()) {
            ce.mainCluster = this;
        }
        elements.addAll(cluster.getElements());
        name = name + "_then_" + cluster.getName();

        return true;
    }

    public void init() {
	setPropertiesRange(0, nbOfProperties);
	TraceManager.addDev("Nb Of datas in cluster " + getName() + " = " + datas.size());
    }

    public void reinitPropertyRange() {
	setPropertiesRange(0, nbOfProperties);
    }

    public boolean setPropertiesRange(int lowIndex, int nb) {
        if (lowIndex < 0) {
            return false;
        }

        if (lowIndex >= nbOfProperties) {
            return false;
        }

        if (lowIndex+nb-1 > nbOfProperties) {
            return false;
        }

        for(ClusterElement ce: elements) {
            //ce.setRangeOfProperties(lowIndex, nb);
	    ce.propertyRangeChanged();
        }

	// Find a similar profile i.e. same range & same property weigths
	ClusterData cd_toUse = getSimilarPropertyProfile(lowIndex, nb, cd.maxNbOfValues);
	if (cd_toUse == null) {
	    cd_toUse = new ClusterData(cd, lowIndex, nb);
	    datas.add(cd_toUse);
	    purgeProximity();
	} 
	cd = cd_toUse;

	TraceManager.addDev("Current cd=" + cd.toString() + " datas size:" + datas.size());
	
        return true;
    }

    public ClusterData getSimilarPropertyProfile(int lowIndex, int nb, int maxVal) {
	for(ClusterData c: datas) {
	    if ((c.firstProp == lowIndex) && (c.nbOfProps == nb) && (c.maxNbOfValues == maxVal)) {
		return c;
	    }
	}
	return null;
    }

    public Vector<ClusterData> getDatas() {
	return datas;
    }

    public void setDatas(Vector<ClusterData> _datas) {
	datas = _datas;
	cd = datas.get(0);
    }

    /* ---------- END OF HANDLING CLUSTER DATA ------------*/

 
    public void setName(String _name) {
        name = _name;
    }

    public String getName() {
        return name;
    }

    public void setNoCsvName(boolean _nocsvname) {
        nocsvname = _nocsvname;
    }

    public boolean getNoCsvName() {
        return nocsvname;
    }

    public void setNormalizationValue(double _val) {
        normalizationValue = _val;
    }

    public double getNormalizationValue() {
        return normalizationValue;
    }

    public boolean hasNormalizationValue() {
        return (normalizationValue != -1);
    }

    /*public void setNbOfPropertiesToConsider(int value) {
	if ((value > 0) && (value <= nbOfProperties)) {
	    nbOfPropertiesToConsider = nbOfProperties;
	}
	}*/

    
    // Percentage must be between 1 and 100, included
    public void setPercentageOfMaxInvalidProperties(int _percentage) {
        TraceManager.addDev("Setting max percentage of invalid properties to " + _percentage + " in cluster " + getName());
        percentageOfMaxInvalidProperties = _percentage;
    }

    // Should be called before any computation on the cluster
    public void initProperties(int _nbOfProperties) {
	cd.nbOfProps = _nbOfProperties;
	cd.firstProp = 0;
	cd.lastProp = _nbOfProperties;
        nbOfProperties = _nbOfProperties;
        namesOfProperties = new String[nbOfProperties];
        cd.weightOfProperties = new int[nbOfProperties];
        for(int j=0; j<nbOfProperties; j++) {
            cd.weightOfProperties[j] = 1;
        }
        cd.weightedMeanPerProperties = null;
        //standardDeviationPerPropertiesComputed = false;
	if (elements != null) {
	    for (ClusterElement ce: elements) {
		ce.propertyWeightChanged();
	    }
	}
	
        //originalWeightOfProperties = weightOfProperties;
    }

    

    public int getTotalNbOfProperties() {
        return nbOfProperties;
    }

    public int getCurrentNbOfProperties() {
        return cd.nbOfProps;
    }

    public int size() {
        return elements.size();
    }

    public void setPropertyName(int cpt, String propName) {
        if (namesOfProperties == null) {
            return;
        }

        if (cpt >= nbOfProperties) {
            return;
        }

        namesOfProperties[cpt] = propName;
    }

    public int getPropertyIndexByName(String propname) {
        if (namesOfProperties == null) {
            return -1;
        }

        for(int j=0; j<nbOfProperties; j++) {
            if (namesOfProperties[j].compareTo(propname) == 0) {
                return j;
            }
        }

        return -1;

    }

    public String getPropertyName(int index) {
        if (namesOfProperties == null) {
            return "";
        }

        if ((index < 0) || (index >= namesOfProperties.length)) {
            return "";
        }

        String prop = namesOfProperties[index];

        if (prop == null) {
            return "";
        }

        return prop;
    }

    /*public Vector<MasterEpsilonClusters> getMasterEpsilonClusters() {
        return cd.masterEpsilonClusters;
	}*/

    /*public void setMasterEpsilonClusters(Vector<MasterEpsilonClusters> vect) {
        cd.masterEpsilonClusters = vect;
	}*/

    public void setEpsilonClusters(Vector<EpsilonCluster> vect) {
        cd.epsilonClusters = vect;
    }

    public void setEpsilonClusterDistanceThreshold(double value) {
	if (cd.epsilonClusterDistanceThreshold == -1) {
	    cd.epsilonClusterDistanceThreshold = value;
	    return;
	}

	// Otherwise -> must create a new profile
	/*ClusterData cdToUse = cd.clone();
	cdToUse.epsilonClusterDistanceThreshold = value;
	cd = cdToUse;
	datas.add(cd);*/
    }

    public Vector<EpsilonCluster> getEpsilonClusters() {
        return cd.epsilonClusters;
    }

    public double getEpsilonClusterDistanceThreshold() {
        return cd.epsilonClusterDistanceThreshold;
    }

    public Vector<ClusterElement> getElements() {
        return elements;
    }

    // returns :max value, index of element, index of property
    public int[] getMaxPropertyValue() {
        if (elements.size() ==0) {
            return null;
        }

        int ret[] = new int[3];
        ret[0] = -1;
        ret[1] = -1;
        ret[2] = -1;

        int index = 0;

        for(ClusterElement ce: elements) {
            for (int i=0; i<ce.originalPropertiesD.length; i++) {
                if (ce.originalPropertiesV[i]) {
                    if (ce.originalPropertiesD[i] > ret[0]) {
                        ret[0] = (int)(Math.ceil(ce.originalPropertiesD[i]));
                        ret[1] = index;
                        ret[2] = i;
                    }
                }
            }
            index ++;
        }

        if (ret[0] == -1) {
            return null;
        }

        return ret;

    }



    public double getProbabilityFactor() {
        return cd.probaFactor;
    }

    public double getProbabilityDifference() {
        return cd.probaDifference;
    }

    public void setProbabilityFactor(double _probaFactor) {
        cd.probaFactor = _probaFactor;
    }

    public void setProbabilityDifference(double proba) {
        if ((proba >= 0) && (proba <= 1)) {
            cd.probaDifference = proba;
        }
    }

    public int getTotalProbabilityDifference() {
        return cd.totalProbabilityDifference;
    }

    public int getTotalProbabilityFactor() {
        return cd.totalProbabilityFactor;
    }

    public void setCorrelationsThreshold(double _d) {
	cd.correlationThreshold = _d;
    }

    public void addInterestingCorrelation(Vector<PropertiesCorrelation> _list) {
        if (cd.listOfCorrelations == null) {
            cd.listOfCorrelations = new Vector<PropertiesCorrelation>();
        }


        TraceManager.addDev("Adding " + _list.size() + " correlations");
        cd.listOfCorrelations.addAll(_list);
    }

    public Vector<PropertiesCorrelation> getListOfCorrelations() {
        return cd.listOfCorrelations;
    }

    public Vector<PropertiesCorrelation> getCloneListOfCorrelations() {
        if (cd.listOfCorrelations == null) {
            return null;
        }

        Vector<PropertiesCorrelation> list = new Vector<PropertiesCorrelation>();
        for (PropertiesCorrelation pc: cd.listOfCorrelations) {
            list.add(pc);
        }

        return list;
    }

    public void setListOfCorrelations(Vector<PropertiesCorrelation> list) {
        cd.listOfCorrelations = list;
    }

    public void addSample(ClusterElement ce) throws ClusterException {
	if (elements == null) {
	    elements = new Vector<ClusterElement>();
	}
	if (ce.properties.length != nbOfProperties) {
	    TraceManager.addDev("Bad nb of properties");
	    throw new ClusterException("Bad number of properties in added sample (" + ce.properties.length + "). should be:" + nbOfProperties);
	}
	ce.mainCluster = this;
	elements.add(ce);
        cd.elementsA = null;

    }

    public void addSampleFromString(String sample) throws ClusterException {
        int index;
        String namece, strLine = sample;

        elements = new Vector<ClusterElement>();
        ClusterElement ce;


        index = strLine.indexOf(SEPARATOR_SYMBOL);
        if (index != -1) {
            if (nocsvname) {
                if (elements == null) {
                    namece = "Element_0";
                } else {
                    namece = "Element_" + elements.size();
                }
                strLine = strLine.trim();
            } else {
                namece = strLine.substring(0, index);
                strLine = strLine.substring(index+1, strLine.length()).trim();
            }
            if (strLine.length() > 0) {
                ce = new ClusterElement(namece);
                ce.mainCluster = this;
                ce.properties = strLine.split(SEPARATOR_SYMBOL);
                if (nbOfProperties == -1) {
                    nbOfProperties = ce.properties.length;
                    cd.nbOfProps = nbOfProperties;
                    cd.firstProp = 0;
		    cd.lastProp = cd.nbOfProps;
                    namesOfProperties = new String[nbOfProperties];
                } else {
                    if (nbOfProperties != ce.properties.length) {
                        throw new ClusterException("Sample " + sample + " is badly formatted (invalid number of properties, should be " + nbOfProperties + ")");
                    }
                }
                elements.add(ce);
                cd.elementsA = null;

            } else {
                throw new ClusterException("Sample " + sample + " is badly formatted (invalid number of properties, should be " + nbOfProperties + ")");
            }
        } else {
            throw new ClusterException("Sample " + sample + " is badly formatted (invalid number of properties, should be " + nbOfProperties + ")");
        }


    }


    public void makeFromFile(String refToFile) throws ClusterException {
        makeFromFile(refToFile, Integer.MAX_VALUE);
    }

    public void makeFromFile(String refToFile, int maxNbOfElementsPerCluster) throws ClusterException {
        nbOfProperties = -1;
        //TraceManager.addDev("Predicting size of cluster ; nbOfProperties=" + nbOfProperties);
        int size = predictFromFile(refToFile, maxNbOfElementsPerCluster);
        //TraceManager.addDev("Loading cluster");
        makeFromFile(refToFile, maxNbOfElementsPerCluster, size);
    }

    public int predictFromFile(String refToFile, int maxNbOfElementsPerCluster) throws ClusterException {
        //ClusterElement ce;
        //String []ss;
        //String namece;
        int cpt = 0;
        int index;
        String propertyLine = null;
        int nbOfElementsPerCluster = 0;

        File file = new File(refToFile);
        if (!FileUtils.checkFileForOpen(file)) {
            throw new ClusterException("File " + refToFile + " could not be opened");
        }

        try {
            FileInputStream fstream = new FileInputStream(refToFile);
            DataInputStream in = new DataInputStream(fstream);
            BufferedReader br = new BufferedReader(new InputStreamReader(in));
            String strLine;
            while (((strLine = br.readLine()) != null) && (nbOfElementsPerCluster < maxNbOfElementsPerCluster))  {
                strLine = strLine.trim();
                //TraceManager.addDev(cpt  + " / line: " + strLine);
                if (strLine.length() > 0) {
                    // start with a comment ?
                    if (!strLine.startsWith(COMMENT_SYMBOL)) {
                        // Is line a property name line ?
                        if (strLine.startsWith(PROPERTY_SYMBOL)) {
                            propertyLine = strLine.substring(1, strLine.length());
                            index = strLine.indexOf(SEPARATOR_SYMBOL);
                            if (index != -1) {
                                propertyLine = strLine.substring(index+1, strLine.length()).trim();
                            } else {
                                propertyLine = null;
                            }
                        } else {
                            // Extract name
                            index = strLine.indexOf(SEPARATOR_SYMBOL);
                            if (index != -1) {
                                //namece = strLine.substring(0, index);
                                if (!nocsvname) {
                                    strLine = strLine.substring(index+1, strLine.length()).trim();
                                }
                                if (strLine.length() > 0) {
                                    if (nbOfProperties == -1) {

                                        nbOfProperties = strLine.split(SEPARATOR_SYMBOL).length;
                                        TraceManager.addDev("Setting nb of properties to " + nbOfProperties);
                                        cd.nbOfProps = nbOfProperties;
                                        cd.firstProp = 0;
					cd.lastProp = nbOfProperties;
                                        namesOfProperties = new String[nbOfProperties];
                                    } else {
                                        if (nbOfProperties != strLine.split(SEPARATOR_SYMBOL).length) {
                                            in.close();
                                            br.close();
                                            throw new ClusterException("File " + refToFile + " badly formatted (invalid number of properties in line " + (cpt + 1) +  ": should be " + nbOfProperties + ")");
                                        }
                                    }
                                    if (cpt % 1000 == 0) {
                                        //TraceManager.addDev("Nb Of lines:" + cpt);
                                    }
                                    nbOfElementsPerCluster ++;

                                } else {
                                    in.close();
                                    br.close();
                                    throw new ClusterException("File " + refToFile + " badly formatted (no property in line " + (cpt + 1) + ")");
                                }
                            } else {
                                in.close();
                                br.close();
                                throw new ClusterException("File " + refToFile + " badly formatted (no property in line " + (cpt +1) + ")");
                            }
                        }
                    }
                }
                cpt ++;
            }
            //Close the input stream
            br.close();
            in.close();


        }catch (Exception e){//Catch exception if any
            TraceManager.addDev("Exception: " + e.getMessage());
            throw new ClusterException("File " + refToFile + " is badly formatted in line: " + (cpt + 1));

        }

        return nbOfElementsPerCluster;
    }


    // Return nb Of elements in cluster
    public void makeFromFile(String refToFile, int maxNbOfElementsPerCluster, int sizeOfVector) throws ClusterException {

        if (sizeOfVector > 0) {
            TraceManager.addDev("Creating a cluster with " + sizeOfVector + " elements");
            elements = new Vector<ClusterElement>(sizeOfVector);
        }
        ClusterElement ce;
        //String []ss;
        String namece;
        int cpt = 0;
        int index;
        String propertyLine = null;
        int nbOfElementsPerCluster = 0;

        File file = new File(refToFile);
        if (!FileUtils.checkFileForOpen(file)) {
            throw new ClusterException("File " + refToFile + " could not be opened");
        }

        try {
            FileInputStream fstream = new FileInputStream(refToFile);
            DataInputStream in = new DataInputStream(fstream);
            BufferedReader br = new BufferedReader(new InputStreamReader(in));
            String strLine;
            while (((strLine = br.readLine()) != null) && (nbOfElementsPerCluster < maxNbOfElementsPerCluster))  {
                strLine = strLine.trim();
                //TraceManager.addDev(cpt  + " / line: " + strLine);
                if (strLine.length() > 0) {
                    // start with a comment ?
                    if (!strLine.startsWith(COMMENT_SYMBOL)) {
                        // Is line a property name line ?
                        if (strLine.startsWith(PROPERTY_SYMBOL)) {
                            propertyLine = strLine.substring(1, strLine.length());
                            index = strLine.indexOf(SEPARATOR_SYMBOL);
                            if (index != -1) {
                                propertyLine = strLine.substring(index+1, strLine.length()).trim();
                            } else {
                                propertyLine = null;
                            }
                        } else {
                            // Extract name
                            index = strLine.indexOf(SEPARATOR_SYMBOL);
                            if (index != -1) {
                                if (nocsvname) {
                                    namece = "Elt_" + (cpt+1);
                                    //strLine = strLine.substring(index+1, strLine.length()).trim();
                                } else {
                                    namece = strLine.substring(0, index);
                                    strLine = strLine.substring(index+1, strLine.length()).trim();
                                }
                                if (strLine.length() > 0) {
                                    ce = new ClusterElement(namece);
                                    ce.mainCluster = this;
                                    ce.properties = strLine.split(SEPARATOR_SYMBOL);
                                    if (nbOfProperties == -1) {
                                        nbOfProperties = ce.properties.length;
                                        namesOfProperties = new String[nbOfProperties];
                                        cd.nbOfProps = nbOfProperties;
                                        cd.firstProp = 0;
                                    } else {
                                        if (nbOfProperties != ce.properties.length) {
                                            in.close();
                                            br.close();
                                            throw new ClusterException("File " + refToFile + " badly formatted (invalid number of properties in line " + (cpt + 1) +  ": should be " + nbOfProperties + ")");
                                        }
                                    }
                                    elements.add(ce);
                                    if (elements.size() % 1000 == 0) {
                                        //TraceManager.addDev("Size of cluster:" + elements.size());
                                    }
                                    cd.elementsA = null;
                                    nbOfElementsPerCluster ++;
                                    makePropertiesOf(ce);

                                } else {
                                    in.close();
                                    br.close();
                                    throw new ClusterException("File " + refToFile + " badly formatted (no property in line " + (cpt + 1) + ")");
                                }
                            } else {
                                in.close();
                                br.close();
                                throw new ClusterException("File " + refToFile + " badly formatted (no property in line " + (cpt +1) + ")");
                            }
                        }
                    }
                }
                cpt ++;
            }
            //Close the input stream
            br.close();
            in.close();

            TraceManager.addDev("Working on properties");
            // Working with properties
            if (propertyLine != null) {
                // property line starts at first property, i.e. name of sample have been remove before
                String props[] = propertyLine.split(SEPARATOR_SYMBOL);
                for(int i=0; i<props.length; i++) {
                    setPropertyName(i, props[i].trim());
                }
            }

	    initProperties(nbOfProperties);
            //originalWeightOfProperties = cd.weightOfProperties;

        } catch (Exception e){//Catch exception if any
            throw new ClusterException("File " + refToFile + " is badly formatted in line: " + (cpt + 1));
        }
    }

    public void setPropertyWeight(int index, int weight) {
        TraceManager.addDev("Setting weight " + weight + " to property " + index + " (" + namesOfProperties[index] + ")");
        cd.weightOfProperties[index] = weight;
        cd.weightedMeanPerProperties = null;
        //standardDeviationPerPropertiesComputed = false;
	if (elements != null) {
	    for (ClusterElement ce: elements) {
		ce.propertyWeightChanged();
	    }
	    /*ClusterElement ce = elements.get(0);
	    if (ce != null) {
		ce.deviations = null;
		}*/
	}
    }



    public String getShortDescription() {
        return name + " - " + elements.size() + " elements in cluster, nb of properties: " + nbOfProperties;
    }

    public String getFullDescription() {
        StringBuffer sb = new StringBuffer(getShortDescription() + "\n");
        for(ClusterElement ce: elements) {
            sb.append(ce.getDescription() + "\n");
        }
        return sb.toString();

    }

    public String getMeanPerPropertiesDescription() {
        String s = "";
        for(int i=cd.firstProp; i<cd.lastProp; i++) {
            if (cd.meanPerPropertiesV[i]) {
                s += " " + cd.meanPerProperties[i];
            } else {
                s += " invalid";
            }
        }
        return s;
    }

    public String getWeightedMeanPerPropertiesDescription() {
        String s = "";
        for(int i=cd.firstProp; i<cd.lastProp; i++) {
            if (cd.meanPerPropertiesV[i]) {
                s += " " + cd.meanPerProperties[i];
            } else {
                s += " invalid";
            }
        }
        return s;
    }

    public String getAverageValues() {
        String s = "";
        for(int i=cd.firstProp; i<cd.lastProp; i++) {
            if (cd.meanPerPropertiesV[i]) {
                if (namesOfProperties[i] != null) {
                    s += " (" + i + "," +  namesOfProperties[i] + "," + cd.meanPerProperties[i] + ") ";
                } else {
                    s += " (" + i + "," +  cd.meanPerProperties[i] + ") ";
                }
            } else {
                if (namesOfProperties[i] != null) {
                    s += " (" + i + "," + namesOfProperties[i] + ",invalid)";
                } else {
                    s += " (" + i + ",invalid)";
                }
            }
        }
        return s;
    }

    public void computeMeanPerProperties() {

        // compute int properties, if necessary
        checkForPropertiesComputed();

        if (cd.meanPerProperties != null) {
            //TraceManager.addDev("DO NOT Compute mean properties of:" + name + ": already done");
            return;
        }

        TraceManager.addDev("Computing mean properties of: " + name);

        cd.meanPerProperties = new double[cd.nbOfProps];
        //cd.weightedMeanPerProperties = new double[cd.nbOfProps];
        cd.meanPerPropertiesV = new boolean[cd.nbOfProps];

        // Compute mean
        int computed = 0;
        int cpt = 0;

        for(int j=0; j<cd.nbOfProps; j++) {
            cd.meanPerProperties[j] = 0.0;
            computed = 0;
            cpt = 0;
            for (ClusterElement ce: elements) {
                if (ce.propertiesV[j+cd.firstProp]) {

                    cd.meanPerProperties[j] += ce.propertiesD[j+cd.firstProp];
                    cpt ++;
                    //if (j == 28) {
                    //TraceManager.addDev("Average of value of prop " + j + " / adding: " + ce.propertiesD[j] + " total=" + meanPerProperties[j] + " cpt=" + cpt);
                    //}
                    //cd.weightedMeanPerProperties[j] += (ce.propertiesD[j+cd.firstProp] * cd.weightOfProperties[j+cd.firstProp]);
                    computed +=  cd.weightOfProperties[j+cd.firstProp];
                }
            }
            if (computed > 0) {
                cd.meanPerProperties[j] = cd.meanPerProperties[j] / cpt;
                //cd.weightedMeanPerProperties[j] = cd.weightedMeanPerProperties[j] / computed;
                cd.meanPerPropertiesV[j] = true;
                //TraceManager.addDev("Average of value " + j + " for cluster " + name + " = " + meanPerProperties[j]);
            } else {
                cd.meanPerPropertiesV[j] = false;
            }
        }

        TraceManager.addDev("\nMean properties computed\n");

        //printMeanValues();

    }

    public void computeWeightedMeanPerProperties() {

        // compute int properties, if necessary
        checkForPropertiesComputed();

        if (cd.weightedMeanPerProperties != null) {
            //TraceManager.addDev("DO NOT Compute mean properties of:" + name + ": already done");
            return;
        }

        TraceManager.addDev("Computing weighted mean properties of: " + name);

        cd.weightedMeanPerProperties = new double[cd.nbOfProps];
        cd.weightedMeanPerPropertiesV = new boolean[cd.nbOfProps];

        // Compute mean
        int computed = 0;
        int cpt = 0;

        for(int j=0; j<cd.nbOfProps; j++) {
            cd.weightedMeanPerProperties[j] = 0.0;
            computed = 0;
            cpt = 0;
            for (ClusterElement ce: elements) {
                if (ce.propertiesV[j+cd.firstProp]) {

                    //cd.meanPerProperties[j] += ce.propertiesD[j+cd.firstProp];
                    cpt ++;
                    //if (j == 28) {
                    //TraceManager.addDev("Average of value of prop " + j + " / adding: " + ce.propertiesD[j] + " total=" + meanPerProperties[j] + " cpt=" + cpt);
                    //}
                    cd.weightedMeanPerProperties[j] += (ce.propertiesD[j+cd.firstProp] * cd.weightOfProperties[j+cd.firstProp]);
                    computed +=  cd.weightOfProperties[j+cd.firstProp];
                }
            }
            if (computed > 0) {
                //cd.meanPerProperties[j] = cd.meanPerProperties[j] / cpt;
                cd.weightedMeanPerProperties[j] = cd.weightedMeanPerProperties[j] / computed;
                cd.weightedMeanPerPropertiesV[j] = true;
                //TraceManager.addDev("Average of value " + j + " for cluster " + name + " = " + meanPerProperties[j]);
            } else {
                cd.weightedMeanPerPropertiesV[j] = false;
            }
        }

        TraceManager.addDev("\nWeighted mean properties computed\n");
    }


    public boolean isDeviationComputedWith(Cluster cluster) {
        ClusterElement ce = elements.get(0);
        return ce.hasDeviationWith(cluster);

    }

    public boolean isWeightDeviationComputedWith(Cluster cluster) {
        ClusterElement ce = elements.get(0);
        return ce.hasWeightDeviationWith(cluster);

    }

    public void computeStandardDeviationWith(Cluster compCluster) {
        checkForPropertiesComputed();
        // Compute double properties, if necessary

        if (compCluster.cd.meanPerProperties == null) {
            compCluster.computeMeanPerProperties();
        }

        if (isDeviationComputedWith(compCluster)) {
	    TraceManager.addDev(" *** Deviation already computed ***");
            return;
        }

	TraceManager.addDev("Computing deviation ..." + " nbOfPropToConsider=" + cd.nbOfProps);

        setInfoOnCurrentComputation("Computing standard deviation with " + compCluster.getName() );
        int cpt = 0;

        int computed, computedW;
        for(ClusterElement ce: elements) {
            percentageOfCurrentComputation = (100 * cpt) / elements.size();
            cpt ++;
            ClusterDeviationElement cde = new ClusterDeviationElement(compCluster, cd.nbOfProps);
            ce.addDeviationElement(cde);
            computed = 0;
            computedW = 0;
            for(int j=0; j<cd.nbOfProps; j++) {
                cde.standardDeviation[j] = 0;
                if ((compCluster.cd.meanPerPropertiesV[j]) && (ce.propertiesV[j+cd.firstProp])){
                    cde.standardDeviationV[j] = true;
                    cde.standardDeviation[j] += (Math.pow(ce.propertiesD[j+cd.firstProp] - compCluster.cd.meanPerProperties[j], 2));
                    computed ++;
                    //TraceManager.addDev("Deviation with " + compCluster.getName() + "=" + cde.standardDeviation[j] + " at step " + j + " value in ce=" + ce.propertiesD[j] + " value of mean=" + compCluster.meanPerProperties[j]);
                    //cde.weightedDeviation[j] +=  compCluster.cd.weightOfProperties[j] * cde.standardDeviation[j];
                    //computedW += compCluster.cd.weightOfProperties[j];
                } else {
                    cde.standardDeviationV[j] = false;
                }
            }
            if (computed == 0) {
                cde.totalDeviationV = false;

            } else {
                cde.totalDeviationV = true;
                cde.totalDeviation = 0;
                for(int k=0; k<cde.standardDeviation.length; k++) {
                    cde.totalDeviation += cde.standardDeviation[k];
                    //TraceManager.addDev("Total Deviation with " + compCluster.getName() + "=" + cde.totalDeviation + " at step " + k);
                    //cde.totalWeightDeviation += cde.weightedDeviation[k];
                }
                cde.totalDeviation = Math.pow(cde.totalDeviation / computed, 0.5);
                //cde.totalWeightDeviation = Math.pow(cde.totalWeightDeviation / computedW, 0.5);

                //TraceManager.addDev("Total deviation with " + compCluster.getName() + "=" + cde.totalDeviation);
            }
        }

        resetInfoOnComputation();
    }


    public void computeWeightDeviationWith(Cluster compCluster) {
        checkForPropertiesComputed();
        // Compute double properties, if necessary

        if (compCluster.cd.weightedMeanPerProperties == null) {
            compCluster.computeWeightedMeanPerProperties();
        }

        if (isWeightDeviationComputedWith(compCluster)) {
	    TraceManager.addDev(" *** Deviation already computed ***");
            return;
        }

	TraceManager.addDev("Computing weight deviation ..." + " nbOfPropToConsider=" + cd.nbOfProps);

        setInfoOnCurrentComputation("Computing weight deviation with " + compCluster.getName() );
        int cpt = 0;

        int computed, computedW;
        for(ClusterElement ce: elements) {
            percentageOfCurrentComputation = (100 * cpt) / elements.size();
            cpt ++;
            ClusterWeightDeviationElement cwde = new ClusterWeightDeviationElement(compCluster, cd.nbOfProps);
            ce.addWeightDeviationElement(cwde);
            computed = 0;
            computedW = 0;
	    double tmpsd;
            for(int j=0; j<cd.nbOfProps; j++) {
                cwde.weightDeviation[j] = 0;
                if ((compCluster.cd.weightedMeanPerPropertiesV[j]) && (ce.propertiesV[j+cd.firstProp])){
                    cwde.weightDeviationV[j] = true;
                    tmpsd = (Math.pow(ce.propertiesD[j+cd.firstProp] - compCluster.cd.weightedMeanPerProperties[j], 2));
                    computed ++;
                    //TraceManager.addDev("Deviation with " + compCluster.getName() + "=" + cde.standardDeviation[j] + " at step " + j + " value in ce=" + ce.propertiesD[j] + " value of mean=" + compCluster.meanPerProperties[j]);
                    cwde.weightDeviation[j] +=  compCluster.cd.weightOfProperties[j] * tmpsd;
                    computedW += compCluster.cd.weightOfProperties[j];
                } else {
                    cwde.weightDeviationV[j] = false;
                }
            }
            if (computed == 0) {
                cwde.totalWeightDeviationV = false;

            } else {
                cwde.totalWeightDeviationV = true;
                cwde.totalWeightDeviation = 0;
                for(int k=0; k<cwde.weightDeviation.length; k++) {
                    cwde.totalWeightDeviation += cwde.weightDeviation[k];
                    //TraceManager.addDev("Total Deviation with " + compCluster.getName() + "=" + cde.totalDeviation + " at step " + k);
                    cwde.totalWeightDeviation += cwde.weightDeviation[k];
                }
                cwde.totalWeightDeviation = Math.pow(cwde.totalWeightDeviation / computedW, 0.5);

                //TraceManager.addDev("Total deviation with " + compCluster.getName() + "=" + cde.totalDeviation);
            }
        }

        resetInfoOnComputation();
    }



    public String getStringDeviations() {
        String s = "";
        for(ClusterElement ce: elements) {
            s += ce.getDeviations();
        }
        return s;
    }


    public String getDeviationStarComparisonString(int index0, int index1) {
        String s = "";
        for(ClusterElement ce: elements) {
            s += ce.getDeviationStarComparisonString(index0, index1);
        }
        return s;
    }

    public void computeDeviationPercentageBetween(Cluster cluster1, Cluster cluster2) {
        computeStandardDeviationWith(cluster1);
        computeStandardDeviationWith(cluster2);

        DoublePoint dp;
        for(ClusterElement ce:elements) {
            dp = ce.getPercentageDeviationsWith(cluster1, cluster2);
            ce.lastPercentageDeviationWithCluster1 = dp.getX();
            ce.lastPercentageDeviationWithCluster2 = dp.getY();
        }

    }

    public void computeDeviationWeightPercentageBetween(Cluster cluster1, Cluster cluster2) {
        computeStandardDeviationWith(cluster1);
        computeStandardDeviationWith(cluster2);

        DoublePoint dp;
        for(ClusterElement ce:elements) {
            dp = ce.getWeightPercentageDeviationsWith(cluster1, cluster2);
            ce.lastPercentageDeviationWithCluster1 = dp.getX();
            ce.lastPercentageDeviationWithCluster2 = dp.getY();
        }

    }

    public String getMinStringDeviations() {
        String s = "";
        for(ClusterElement ce: elements) {
            s += ce.getMinDeviations();
        }
        return s;
    }

    public void checkForPropertiesComputed() {
        if (!propertiesDComputed) {
            TraceManager.addDev("Computing values of properties");
            for (ClusterElement ce: elements) {
                makePropertiesOf(ce);
            }
            propertiesDComputed = true;
            removeElementsWithNoValidProperties((nbOfProperties*percentageOfMaxInvalidProperties)/100);
        }

        resetInfoOnComputation();
    }

    public void makePropertiesOf(ClusterElement ce) {
        if (ce.propertiesD == null) {
            if (ce.originalPropertiesD != null) {
		//TraceManager.addDev("Original properties are still here!");
		ce.propertiesD = ce.originalPropertiesD;
                ce.propertiesV = ce.originalPropertiesV;
            } else {
                ce.propertiesD = new double[nbOfProperties];
                ce.propertiesV = new boolean[nbOfProperties];
                for(int k=0; k<nbOfProperties; k++) {
                    try {
                        ce.propertiesD[k] = Double.parseDouble(ce.properties[k].trim());
                        ce.propertiesV[k] = true;
                    } catch (Exception e) {
                        //TraceManager.addDev("Exception on property: " + e.getMessage() + " - parsed double:" + ce.properties[k].trim());
                        ce.propertiesV[k] = false;
                    }
                }
                ce.originalPropertiesD = ce.propertiesD;
                ce.originalPropertiesV = ce.propertiesV;
                ce.properties = null;
            }
        }
    }


    public void removeElementsWithNoValidProperties(int maxInvalid) {
        Vector<ClusterElement> toRemove = new Vector<ClusterElement>();

        for(ClusterElement ce: elements) {
            if (ce.hasMoreThanInvalidProperties(maxInvalid)) {
                toRemove.add(ce);
            }
        }

        if (toRemove.size() == 0) {
            return;
        }

        cd.elementsA = null;

        TraceManager.addUser("\n");
        for(ClusterElement rm: toRemove) {
            elements.remove(rm);
            TraceManager.addUser(rm.getName() + " has been removed because too many of its properties are invalid");
        }
        TraceManager.addUser("\n");


    }



    public double getMaxForProperty(int index) {

        checkForPropertiesComputed();

        double max = 0;
        for(ClusterElement ce: elements) {
            if (ce.propertiesV[index]) {
                if (ce.propertiesD[index] > max) {
                    max = ce.propertiesD[index];
                }
            }
        }
        return max;
    }

    public void normalizeAllProperties(double _value) {
        normalized = true;
        if (normalizedValues == null) {
            normalizedValues = new double[nbOfProperties];
        }
        for(ClusterElement ce: elements) {
            for (int i=0; i<ce.propertiesV.length; i++) {
                if (ce.propertiesV[i]) {
                    ce.propertiesD[i] = ce.propertiesD[i] / _value;
                }
                normalizedValues[i] = _value;
            }
        }




    }

    //Must be called only when no range has been selected
    public void normalizeProperties(int index, double max) {
        // For each property: find the max value, and divide each property by the max value of its set

        if (max != 0) {
            for(ClusterElement ce: elements) {
                if (ce.propertiesV[index]) {
                    ce.propertiesD[index] = ce.propertiesD[index] / max;
                }
            }
        }

        normalized = true;
        if (normalizedValues == null) {
            normalizedValues = new double[nbOfProperties];
        }
        normalizedValues[index] = max;
    }

    public void setMaxNbOfValues(int nb) {
	if (nb < 1) {
	    return;
	}

	double dec = 1.0/(2*(nb-1.0));
	TraceManager.addDev("dec=" + dec);
	double v;
	int vi;

        for(ClusterElement ce: elements) {
            for(int index=0; index<nbOfProperties; index++) {
                if (ce.originalPropertiesV[index]) {
		    v = ce.originalPropertiesD[index] + dec;
		    v = v * (nb-1);
		    vi = (int)(Math.floor(v));
                    ce.originalPropertiesD[index] = vi / ((double)nb - 1.0);
                }
            }
            if (ce.originalPropertiesD != ce.propertiesD) {
                for(int index=0; index<nbOfProperties; index++) {
                    if (ce.propertiesV[index]) {
			v = ce.propertiesD[index] + dec;
			v = v * (nb-1);
			vi = (int)(Math.floor(v));
			ce.propertiesD[index] = vi / ((double)nb - 1.0);
                    }
                }
            }
        }

	// Find a similar profile i.e. same range & same property weigths
	ClusterData cd_toUse = getSimilarPropertyProfile(cd.firstProp, cd.nbOfProps, nb);
	if (cd_toUse == null) {
	    cd_toUse = new ClusterData(cd, cd.firstProp, cd.nbOfProps);
	    cd_toUse.maxNbOfValues = nb;
	    datas.add(cd_toUse);
	    purgeProximity();
	} 
	cd = cd_toUse;

	TraceManager.addDev("Current cd=" + cd.toString() + " datas size:" + datas.size());
    }

    public boolean isNormalized() {
        return normalized;
    }

    public double getNormalizedValue(int index) {
        if (normalizedValues == null) {
            return 0;
        }

        return normalizedValues[index];
    }

    public void setNbOfClosestElements(int nb) {
        cd.nbOfClosestComputed = nb;
    }


    public void computeClosestElements(int _nbOfElements, Cluster[] withClusters) {
        checkForPropertiesComputed();
        int i;

        setInfoOnCurrentComputation("Computing " + _nbOfElements + " closest elements");

        for(i=0; i<withClusters.length; i++) {
            withClusters[i].checkForPropertiesComputed();
        }

        if (manager == null) {
            manager = new ClusterAlgoMultithreadManager(this);
        }

        if (cd.elementsA == null) {
            cd.elementsA = new ClusterElement[elements.size()];
            cd.elementsA = elements.toArray(cd.elementsA);
        }

        manager.computeClosestElements(_nbOfElements, withClusters);

        cd.nbOfClosestComputed = _nbOfElements;

        resetInfoOnComputation();
    }

    // index indicates the index of ClusterElement in elements
    public void computeClosestElementsRange(int nbOfElements, Cluster[] withClusters, int startIndex, int endIndex, boolean updatePercentage) {
        double distance;
        double d;
        ClusterElement ce;
        //ClusterElementDistance found;
        //double maxDist;
        int range = endIndex - startIndex;

        setInfoOnCurrentComputation("Computing closest elements");

        //TraceManager.addDev("Computing closest from " + startIndex + " to " + endIndex + "(not included)\n");

        for(int cpt=startIndex; cpt<endIndex; cpt++) {
            ce = cd.elementsA[cpt];
            ce.closest = new Vector<ClusterElementDistance>(nbOfElements);
            if (updatePercentage) {
                percentageOfCurrentComputation = (100 * cpt) / (range);

            }

            distance = 0;

            for(int j=0; j<withClusters.length; j++) {
                for(ClusterElement otherElement: withClusters[j].getElements()) {
                    if (ce.closest.size() < nbOfElements) {
                        d = ce.computeSquareDistanceWith(otherElement);
                        ce.addSortedInClosest(new ClusterElementDistance(otherElement, d));
                        distance = ce.getMaxDistanceInClosest();
                    } else {
                        d = ce.computeSquareDistanceWithLimit(otherElement, distance);
                        if (d < distance) {
                            ce.closest.removeElementAt(ce.closest.size()-1);
                            ce.addSortedInClosest(new ClusterElementDistance(otherElement, d));
                            distance = ce.getMaxDistanceInClosest();
                        }
                    }
                }

            }
        }
    }


    /*public void oldComputeClosestElementsWithProperties(int nbOfElements, int nbOfConsideredProperties, Cluster[] withClusters) {
        checkForPropertiesComputed();
        int i;
        double distance;
        double d;
        //ClusterElementDistance found;
        //double maxDist;

        setInfoOnCurrentComputation("Computing " + nbOfElements + " closest elts for " + nbOfConsideredProperties + " most relevant props");

        for(i=0; i<withClusters.length; i++) {
            withClusters[i].checkForPropertiesComputed();
        }

        int cpt = 0;
        for(ClusterElement ce: elements) {
            percentageOfCurrentComputation = (100 * cpt) / elements.size();
            cpt ++;
            if ((ce.closest.size() < nbOfElements) || (ce.nbOfConsideredProperties != nbOfConsideredProperties)) {
                ce.closest = new Vector<ClusterElementDistance>(nbOfElements);
                ce.nbOfConsideredProperties = nbOfConsideredProperties;
                distance = 0;

                for(int j=0; j<withClusters.length; j++) {
                    for(ClusterElement otherElement: withClusters[j].getElements()) {
                        // Compute distance with element
                        d = ce.computeSquareDistanceConsideredPropertiesWith(nbOfConsideredProperties, otherElement);
                        //TraceManager.addDev("Distance=" + d);
                        //otherElement.distance = d;

                        if (ce.closest.size() < nbOfElements) {
                            ce.addSortedInClosest(new ClusterElementDistance(otherElement, d));
                            distance = ce.getMaxDistanceInClosest();
                        } else {
                            if (d < distance) {
                                ce.closest.removeElementAt(ce.closest.size()-1);
                                ce.addSortedInClosest(new ClusterElementDistance(otherElement, d));
                                distance = ce.getMaxDistanceInClosest();
                            }
                        }
                    }
                }

            }
        }
        nbOfClosestComputed = nbOfElements;

        resetInfoOnComputation();
	}*/

    public void computeClosestElementsWithProperties(int nbOfElements, int nbOfConsideredProperties, Cluster[] withClusters) {
        checkForPropertiesComputed();
        int i;

        setInfoOnCurrentComputation("Computing " + nbOfElements + " closest elts for " + nbOfConsideredProperties + " most relevant props");

        for(i=0; i<withClusters.length; i++) {
            withClusters[i].checkForPropertiesComputed();
        }

        if (manager == null) {
            manager = new ClusterAlgoMultithreadManager(this);
        }

        if (cd.elementsA == null) {
            cd.elementsA = new ClusterElement[elements.size()];
            cd.elementsA = elements.toArray(cd.elementsA);
        }

        manager.computeClosestElementsWithProperties(nbOfElements, nbOfConsideredProperties, withClusters);

        cd.nbOfClosestComputed = nbOfElements;

        resetInfoOnComputation();

    }

    public void computeClosestElementsWithProperties_FirstVersion(int nbOfElements, int nbOfConsideredProperties, Cluster[] withClusters) {
        checkForPropertiesComputed();
        int i;

        setInfoOnCurrentComputation("Computing " + nbOfElements + " closest elts for " + nbOfConsideredProperties + " most relevant props");

        for(i=0; i<withClusters.length; i++) {
            withClusters[i].checkForPropertiesComputed();
        }

        if (manager == null) {
            manager = new ClusterAlgoMultithreadManager(this);
        }

        if (cd.elementsA == null) {
            cd.elementsA = new ClusterElement[elements.size()];
            cd.elementsA = elements.toArray(cd.elementsA);
        }

        manager.computeClosestElementsWithProperties_FirstVersion(nbOfElements, nbOfConsideredProperties, withClusters);

        cd.nbOfClosestComputed = nbOfElements;

        resetInfoOnComputation();

    }

    public void computeClosestElementsWithPropertiesRange(int nbOfElements, int nbOfConsideredProperties, Cluster[] withClusters, int startIndex, int endIndex, boolean updatePercentage) {
        checkForPropertiesComputed();
        double distance;
        double d;
        ClusterElement ce;

        int range = endIndex - startIndex;

        TraceManager.addDev("nbOfConsideredProperties  = " + nbOfConsideredProperties);
        setInfoOnCurrentComputation("Computing closest elts w.r.t. " + nbOfConsideredProperties + " props");


        for(int cpt=startIndex; cpt<endIndex; cpt++) {
            ce=cd.elementsA[cpt];
            if (updatePercentage) {
                percentageOfCurrentComputation = (100 * cpt) / (range);
            }

            if ((ce.closest == null) || (ce.closest.size() < nbOfElements) || (nbOfConsideredProperties != cd.nbOfProps)) {
                ce.closest = new Vector<ClusterElementDistance>(nbOfElements);
                //ce.nbOfConsideredProperties = cd.nbOfProps;
                distance = 0;

                for(int j=0; j<withClusters.length; j++) {
                    for(ClusterElement otherElement: withClusters[j].getElements()) {
                        if (ce.closest.size() < nbOfElements) {
                            d = ce.computeSquareDistanceConsideredPropertiesWith(nbOfConsideredProperties, otherElement);
                            ce.addSortedInClosest(new ClusterElementDistance(otherElement, d));
                            distance = ce.getMaxDistanceInClosest();
                        } else {
                            d = ce.computeSquareDistanceConsideredPropertiesWith(nbOfConsideredProperties, otherElement);
                            if (d < distance) {
                                ce.closest.removeElementAt(ce.closest.size()-1);
                                ce.addSortedInClosest(new ClusterElementDistance(otherElement, d));
                                distance = ce.getMaxDistanceInClosest();
                            }
                        }
                    }
                }

            }
        }
    }

    public String getClosestString() {
        String ret = "";
        for(ClusterElement ce: elements) {
            ret += ce.getClosestString() + "\n\n";
        }
        return ret;

    }

    public String getClosestStringByClusterName() {
        String ret = "";
        for(ClusterElement ce: elements) {
            ret += ce.getClosestStringByClusterName() + "\n\n";
        }
        return ret;

    }

    // Used by Correlation
    public void computeStandardDeviationPerPropertiesForCorrelation() {
        cd.standardDeviationPerProperties = new double[cd.nbOfProps];
        //weightedStandardDeviationPerProperties = new double[nbOfProperties];
        cd.standardDeviationPerPropertiesV = new boolean[cd.nbOfProps];
        int cpt = 0;
        //int wcpt = 0;
        double sum;
        double tmp;
	int ind;

        for(int i=0; i<cd.nbOfProps; i++) {
            if (cd.meanPerPropertiesV[i]) {
                cpt = 0;
                //wcpt = 0;
                sum = 0.0;
                //wsum = 0.0;
		
                for(ClusterElement ce: elements) {
		    ind = i+cd.firstProp;
                    if (ce.propertiesV[ind]) {
                        tmp = Math.pow(ce.propertiesD[ind] - cd.meanPerProperties[i], 2);
                        sum = sum + tmp;
                        cpt ++;
                        //wsum = wsum +  weightOfProperties[i] * tmp;
                        //wcpt += weightOfProperties[i];
                    }
                }
                if (cpt == 0) {
                    cd.standardDeviationPerPropertiesV[i] = false;
                } else {
                    cd.standardDeviationPerPropertiesV[i] = true;
                    cd.standardDeviationPerProperties[i] = Math.pow(sum, 0.5);
                    //weightedStandardDeviationPerProperties[i] = Math.pow(wsum/wcpt, 0.5);
                }
            } else {
                cd.standardDeviationPerPropertiesV[i] = false;
            }
        }
    }

    public String getStandardDeviationString() {
        String ret = "";
        for(int i=0; i<cd.nbOfProps; i++) {
            if (cd.standardDeviationPerPropertiesV[i]) {
                ret += cd.standardDeviationPerProperties[i] + " ";
            } else {
                ret += "- ";
            }
        }
        return ret;
    }

    public void computeCorrelations() {
        setInfoOnCurrentComputation("Computing correlations of " + getName());

        checkForPropertiesComputed();

        TraceManager.addDev("Mean properties for cluster " + getName());
        if (cd.meanPerProperties == null) {
            computeMeanPerProperties();
        }

        //TraceManager.addDev("Standard dev by prop\n");
        if (cd.standardDeviationPerProperties == null) {
            computeStandardDeviationPerPropertiesForCorrelation();
        }

        TraceManager.addDev("Real correlation\n");
        if (cd.correlations == null) {
            TraceManager.addDev("NbOfPropertiesToConsider:" + cd.nbOfProps);
            cd.correlations = new double[cd.nbOfProps][cd.nbOfProps];

            if (manager == null) {
                manager = new ClusterAlgoMultithreadManager(this);
            }

            manager.computeCorrelations(cd.nbOfProps);

            resetInfoOnComputation();
        }
    }

    public void computeCorrelations(int startIndex, int endIndex, boolean updatePercentage) {
        int range = endIndex - startIndex;
        double value;
        int iindex, jindex;
        //boolean ok;

        //int cpt = 0;

        TraceManager.addDev("Computing list of interesting correlations");
        setInfoOnCurrentComputation("Computing list of interesting correlations of " + getName());

        for(int i=startIndex; i<endIndex; i++) {
            if (updatePercentage) {
                percentageOfCurrentComputation = (100 * i) / range;
            }
            //cpt ++;
            for(int j=0; j<cd.nbOfProps; j++) {

                if (j<=i) {
                    cd.correlations[i][j] = -4;
                } else {
                    // Compute correlation
                    if (cd.standardDeviationPerPropertiesV[i] &&  cd.standardDeviationPerPropertiesV[j]) {
                        value = 0.0;
                        //ok = true;
                        cd.correlations[i][j] = 0;
                        for(ClusterElement ce: elements) {
                            if (ce.propertiesV[i+cd.firstProp] && ce.propertiesV[j+cd.firstProp]) {
                                value = (ce.propertiesD[i+cd.firstProp] - cd.meanPerProperties[i]) * (ce.propertiesD[j+cd.firstProp] - cd.meanPerProperties[j]);
                                cd.correlations[i][j] += value;
                            } /*else {
                                if ((ce.propertiesV[i] && ! ce.propertiesV[j]) || (ce.propertiesV[i] && ! ce.propertiesV[j])) {
                                ok = false;
                                break;
                                }
                                }*/
                        }
                        //if (ok) {
                        cd.correlations[i][j] = cd.correlations[i][j] / (cd.standardDeviationPerProperties[i] * cd.standardDeviationPerProperties[j]);
                        /*} else {
                          correlations[i][j] = -2;
                          }*/


                    } else {
                        cd.correlations[i][j] = -3;
                    }

                }
            }
        }



    }


    // Correlations with the similar properties are removed
    public void removeNonExclusiveCorrelations() {
        Vector<PropertiesCorrelation> toRemove = new Vector<PropertiesCorrelation>();
        int index = 0;
        PropertiesCorrelation pc2;

        for(PropertiesCorrelation pc1: cd.listOfCorrelations) {
            if (!toRemove.contains(pc1)) {
                for(int i=index+1; i<cd.listOfCorrelations.size(); i++) {
                    pc2 = cd.listOfCorrelations.get(i);
                    if (pc2.samePropertiesAs(pc1)) {
                        if (!toRemove.contains(pc1)) {
                            toRemove.add(pc1);
                        }
                        toRemove.add(pc2);
                    }
                }
            }
            index ++;
        }


        for(PropertiesCorrelation pc: toRemove) {
            cd.listOfCorrelations.remove(pc);
        }
        TraceManager.addDev("Removed " + toRemove.size() + " similar correlations ; remaining correlations:" + cd.listOfCorrelations.size());
    }

    public Vector<PropertiesCorrelation> getListOfInterestingCorrelations() {
	return cd.listOfCorrelations;
    }

    public void  computeListOfInterestingCorrelations() {
	if (cd.listOfCorrelations != null) {
	    return;
	}

        PropertiesCorrelation cr;
        cd.listOfCorrelations = new Vector<PropertiesCorrelation>();

        setInfoOnCurrentComputation("Computing list of interesting correlations of " + getName());

        //int cpt = 0;

        for(int i=0; i<cd.nbOfProps; i++) {
            percentageOfCurrentComputation = (100 * i) / cd.nbOfProps;
            //cpt ++;
            for(int j=0; j<cd.nbOfProps; j++) {

                if (cd.correlations[i][j] > cd.correlationThreshold) {
                    cr = new PropertiesCorrelation();
                    cr.property1Index = i;
                    cr.property2Index = j;
                    cr.cluster = this;
                    cr.correlationValue =  cd.correlations[i][j];

                    // Compute line -> the two most probable values
                    double[] x = getTheTwoMostFrequentValues(i+cd.firstProp);
                    double[] y = getTheTwoMostFrequentValues(j+cd.firstProp);

                    //Computed values
                    //TraceManager.addDev("x0=" + x[0] + " y0=" + y[0] + " x1=" + x[1] + " y1=" + y[1]);


                    if ((x != null) && (y != null)) {
                        // Computing a and b
                        if (x[0] != x[1]) {
                            cr.a = (y[0] - y[1])/(x[0] - x[1]);
                            cr.b = y[0] - x[0] * cr.a;

                            // Adding to the list!
                            cd.listOfCorrelations.add(cr);
                        }
                    }
                }
            }
        }

        resetInfoOnComputation();

    }

    public String printSortedPropertyValues() {
        String ret = "";
        for(int i=0; i<cd.nbOfProps; i++) {
            ret += printSortedPropertyValue(i) + "\n";
        }
        return ret;
    }

    public String printSortedPropertyValue(int index) {
        String ret = "" + index + ": ";
        DoubleInteger di;
        ArrayList<DoubleInteger> al = cd.sortedPropertyValues[index];
        for(Object o: al) {
            di = (DoubleInteger)(o);
            ret += "(" + di.d + "," + di.i + ") ";
        }
        return ret;
    }


    public void computeSortedPropertyValues() {
	/*if (cd.sortedPropertyValues != null) {
	    TraceManager.addDev("Non null sorted values");
	    if (cd.sortedPropertyValues[0] != null) {
		TraceManager.addDev("Non null sorted values and values already computed");
		return;
	    }
	    }*/
	TraceManager.addDev("Computing sorted values");
        for(int i=cd.firstProp; i<cd.lastProp; i++) {
            ComputeAndSortPropertyValue(i);
        }
    }

    public void computeTotalProbaDifferenceWithCluster(Cluster cl) {
        cd.totalProbabilityDifference = 0;

        //ArrayList<DoubleInteger> al;
        //DoubleInteger di;
        double val1, val2;

        for(int i=0; i<cd.sortedPropertyValues.length; i++) {
            for(DoubleInteger di: cd.sortedPropertyValues[i]) {
		//                di = (DoubleInteger)o;

                val1 = di.i/(double)(cd.sortedPropertyTotal[i]);
                val2 = cl.getProbability(i, di.d);

                if(val1 > (val2 +  cd.probaDifference)) {
                    //TraceManager.addDev("Increasing proba diff");
                    cd.totalProbabilityDifference ++;
                    break;
                }
            }
        }
    }

    public String computeStringAndTotalProbaDifferenceWithCluster(Cluster cl) {
        cd.totalProbabilityDifference = 0;
        String props = "";

        //ArrayList al;
        DoubleInteger di;
        double val1, val2;
        boolean comma = false;

        for(int i=0; i<cd.sortedPropertyValues.length; i++) {
            for(Object o: cd.sortedPropertyValues[i]) {
                di = (DoubleInteger)o;

                val1 = di.i/(double)(cd.sortedPropertyTotal[i]);
                val2 = cl.getProbability(i+cd.firstProp, di.d);

                if(val1 > (val2 +  cd.probaDifference)) {
                    TraceManager.addDev("index " + (i+cd.firstProp) + ": proba difference for value " + di.d + " proba1=" + val1 + " proba2=" + val2);
                    cd.totalProbabilityDifference ++;

                    if (i<=MAX_DISPLAYED_PROPS) {
                        if (comma) {
                            props += ",";
                        } else {
                            comma = true;
                        }
                        if (i == MAX_DISPLAYED_PROPS) {
                            props += "...";
                        } else {
                            props += (i+cd.firstProp);
                        }
                    }
                    break;
                }
            }
        }
        return props;
    }

    public String computeStringAndTotalProbaFactorWithCluster(Cluster cl) {
        String props = "";
        cd.totalProbabilityFactor = 0;

        //ArrayList al;
        DoubleInteger di;
        double val1, val2;
        boolean comma = false;

        for(int i=0; i<cd.sortedPropertyValues.length; i++) {
            for(Object o: cd.sortedPropertyValues[i]) {
                di = (DoubleInteger)o;

                val1 = di.i/(double)(cd.sortedPropertyTotal[i]);
                val2 = cl.getProbability(i+cd.firstProp, di.d);

                if(val1 > (val2 *  cd.probaFactor)) {
                    //TraceManager.addDev("Increasing proba factor");
                    cd.totalProbabilityFactor ++;

                    if (i<=MAX_DISPLAYED_PROPS) {
                        if (comma) {
                            props += ",";
                        } else {
                            comma = true;
                        }
                        if (i == MAX_DISPLAYED_PROPS) {
                            props += "...";
                        } else {
                            props += (i+cd.firstProp);
                        }
                    }
                    break;
                }
            }
        }
        return props;
    }

    public void computeTotalProbaFactorWithCluster(Cluster cl) {
        cd.totalProbabilityFactor = 0;

        //ArrayList al;
        DoubleInteger di;
        double val1, val2;

        for(int i=0; i<cd.sortedPropertyValues.length; i++) {
            for(Object o: cd.sortedPropertyValues[i]) {
                di = (DoubleInteger)o;

                val1 = di.i/(double)(cd.sortedPropertyTotal[i]);
                val2 = cl.getProbability(i+cd.firstProp, di.d);

                if(val1 > (val2 *  cd.probaFactor)) {
                    //TraceManager.addDev("Increasing proba factor");
                    cd.totalProbabilityFactor ++;


                    break;
                }
            }
        }
    }

    public double[] getTheTwoMostFrequentValues(int index) {
        double[] tab = new double[2];

        ComputeAndSortPropertyValue(index);

        ArrayList<DoubleInteger> list = cd.sortedPropertyValues[index-cd.firstProp];

        if (list.size() == 0) {
            return null;
        }

        tab[0] = ((DoubleInteger)(list.get(0))).d;

        if (list.size() == 1) {
            tab[1] = tab[0];
            return tab;
        }

        tab[1] = ((DoubleInteger)(list.get(1))).d;

        return tab;

    }


    public void ComputeAndSortPropertyValue(int index) {
        if (cd.sortedPropertyValues == null) {
            cd.sortedPropertyValues = new ArrayList[cd.nbOfProps];
            cd.sortedPropertyTotal = new int[cd.nbOfProps];
        }

        if (cd.sortedPropertyValues[index-cd.firstProp] != null) {
            return;
        }

        HashMap<Double, Integer> pts = new HashMap<Double, Integer>();
        // (value, nb)
        double value;
        //double nb;
        Integer stored;
        cd.sortedPropertyTotal[index-cd.firstProp] = 0;


        for(ClusterElement ce: elements) {
            if (ce.propertiesV[index]) {
                value = ce.propertiesD[index];
                stored = pts.get(new Double(value));
                if (stored == null) {
                    pts.put(new Double(value), new Integer(1));
                } else {
                    pts.put(new Double(value), new Integer(stored.intValue() + 1));
                }
                cd.sortedPropertyTotal[index-cd.firstProp] += 1;
            }
        }

        cd.sortedPropertyValues[index-cd.firstProp] = new ArrayList<DoubleInteger>();

        for(Double d: pts.keySet()) {
            stored = pts.get(d);
            cd.sortedPropertyValues[index-cd.firstProp].add(new DoubleInteger(d, stored));
        }

        Collections.sort(cd.sortedPropertyValues[index-cd.firstProp]);

        /*DoubleInteger di;
        for(Object o: cd.sortedPropertyValues[index-cd.firstProp]) {
            di = (DoubleInteger)o;
            //TraceManager.addDev("di=" + di.d + ", " + di.i);
	    }*/


        //TraceManager.addDev("Found " + pts.size() + " different values for property " + index);

    }

    public String getStringProbabilities() {
	String ret ="";
	for(int i=cd.firstProp; i<cd.lastProp; i++) {
	    ret += "index:" + i + " ";
	    for(DoubleInteger di: cd.sortedPropertyValues[i-cd.firstProp]) {
		ret+= "(val=" + di.d + ",proba=" + ((double)di.i)/cd.sortedPropertyTotal[i-cd.firstProp] + ") ";
	    }
	    ret += "\n";
	}
	return ret;
    }

    public String getStringProbabilities(Cluster withCluster) {
	String ret ="";
	for(int i=cd.firstProp; i<cd.lastProp; i++) {
	    ret += "index:" + i + " " + "\n\t" + getName() + ":";
	    for(DoubleInteger di: cd.sortedPropertyValues[i-cd.firstProp]) {
		ret+= "(val=" + di.d + ",proba=" + ((double)di.i)/cd.sortedPropertyTotal[i-cd.firstProp] + ") ";
	    }
	    ret +=  "\n\t" + withCluster.getName() + ":";
	    for(DoubleInteger di: withCluster.cd.sortedPropertyValues[i-cd.firstProp]) {
		ret+= "(val=" + di.d + ",proba=" + ((double)di.i)/(withCluster.cd.sortedPropertyTotal[i-cd.firstProp]) + ") ";
	    }
	    ret += "\n";
	}
	return ret;
    }



    public double getProbability(int propertyIndex, double value) {
        DoubleInteger di;
        ArrayList<DoubleInteger> list = cd.sortedPropertyValues[propertyIndex-cd.firstProp];

        for(Object o: list) {
            di = (DoubleInteger)(o);

            if (di.d == value) {
                return di.i/((double)(cd.sortedPropertyTotal[propertyIndex-cd.firstProp]));
            }
        }

        return 0;
    }





    public String getCorrelationString() {
        String ret = "";
        for (int i=0; i<cd.nbOfProps; i++) {
            ret+= "line " + i + ": ";
            for(int j=0; j<cd.nbOfProps; j++) {
                ret += cd.correlations[i][j] + " ";
            }
            ret += "\n";
        }
        return ret;
    }

    public double getCorrelation(int i, int j) {
        return cd.correlations[i][j];
    }

    public String getInterestingCorrelations() {
        String ret = "";
        double val;
        String prop;
        for (int i=0; i<cd.nbOfProps; i++) {
	    //TraceManager.addDev("i=" + i);
            for(int j=0; j<cd.nbOfProps; j++) {
                val = Math.abs(cd.correlations[i][j]);
                if ((val <1.0001) && (val>cd.correlationThreshold)) {
                    ret += "Correlation between property " + (i+cd.firstProp);
                    prop = getPropertyName(i+cd.firstProp);
                    if (prop.length() > 0) {
                        ret +=" (" + prop + ")";
                    }
                    ret += " and "  + j;
                    prop = getPropertyName(j+cd.firstProp);
                    if (prop.length() > 0) {
                        ret +=" (" + prop + ")";
                    }
		    //TraceManager.addDev("Correlation found for (" + i + "," + j + ") val=" + correlations[i][j]);
                    ret += " = " + cd.correlations[i][j] + " mean of property i: " + cd.meanPerProperties[i] + " mean of property j: " + cd.meanPerProperties[j] + "\n";
                }
            }
        }
        return ret;
    }

    public String getShortInterestingCorrelations() {
        String ret = "";
        double val;
        String prop;
        for (int i=0; i<cd.nbOfProps; i++) {
            for(int j=0; j<cd.nbOfProps; j++) {
                val = cd.correlations[i][j];
                if ((val <1.01) && (val>cd.correlationThreshold)) {
                    ret += "Correlation between property " + (i+cd.firstProp);
                    prop = getPropertyName(i+cd.firstProp);
                    if (prop.length() > 0) {
                        ret +=" (" + prop + ")";
                    }
                    ret += " and "  + (j+cd.firstProp);
                    prop = getPropertyName(j+cd.firstProp);
                    if (prop.length() > 0) {
                        ret +=" (" + prop + ")";
                    }
                    ret += " = " + cd.correlations[i][j] + "\n";
                }
            }
        }
        return ret;
    }

    public String getCorrelationsOfCluster(Cluster cl) {
        String prop;
        String ret = "";
        for (PropertiesCorrelation pc: cd.listOfCorrelations) {
            if (pc.cluster == cl) {
                ret += "Correlation between property " + pc.property1Index;
                prop = getPropertyName(pc.property1Index);
                if (prop.length() > 0) {
                    ret +=" (" + prop + ")";
                }
                ret += " and "  + pc.property2Index;
                prop = getPropertyName(pc.property2Index);
                if (prop.length() > 0) {
                    ret +=" (" + prop + ")";
                }
                ret += " = " + pc.correlationValue + "\n";
            }
        }
        return ret;
    }

    public String getAllPropertyValues(int i, int j) {
        String ret = "";
        int diff = 0;

        checkForPropertiesComputed();

        for(ClusterElement ce: elements) {
            if (ce.propertiesV[i] && ce.propertiesV[j]) {
                if (ce.propertiesD[i] != ce.propertiesD[j]) {
                    ret+= "**";
                    diff++;
                }
                ret += "(" + ce.propertiesD[i] + "," + ce.propertiesD[j] + ") ";

            }
        }
        ret += "  (" + diff + " differences)";
        return ret;
    }

    public Vector<Point> getInterestingCorrelationsPoints() {
        Vector<Point> points = new Vector<Point>();
        double val;
        for (int i=0; i<cd.nbOfProps; i++) {
            for(int j=0; j<cd.nbOfProps; j++) {
                val = Math.abs(cd.correlations[i][j]);
                if ((val <1.01) && (val>0.75)) {
                    points.add( new Point(i, j));
                }
            }
        }
        return points;
    }


    /*public String getProbabilitiesString() {
        String ret = "(index, value, prob):\n";
        for(int i=0; i<cd.probabilityOfMostCommonValue.length; i++) {
            ret += "(" + i + "," + cd.mostCommonValue[i] + "," + cd.probabilityOfMostCommonValue[i] + ") ";
        }
        return ret;

	}*/

    /*public boolean areMostCommonValueAndProbabilitiesComputed() {
	return (cd.probabilityOfMostCommonValue != null);
    }


    public void computeMostCommonValueAndProbabilities() {
	if (areMostCommonValueAndProbabilitiesComputed()) {
	    return;
	}

        Vector<DoublePoint> valuesAndNumber = new Vector<DoublePoint>();
        double value;
        boolean found;
        int total;
        cd.probabilityOfMostCommonValue = new double[cd.nbOfProps];
        cd.mostCommonValue = new double[cd.nbOfProps];


        for(int i=cd.firstProp; i<cd.lastProp; i++) {
            valuesAndNumber.clear();
            total = 0;
            for(ClusterElement ce: elements) {
                if (ce.propertiesV[i]) {
                    value = ce.propertiesD[i];
                    total ++;
                    found = false;
                    for(DoublePoint p: valuesAndNumber) {
                        if (p.getX() == value) {
                            p.setLocation(p.getX(), p.getY() + 1);
                            found = true;
                            break;
                        }
                    }
                    if (!found) {
                        DoublePoint p1 = new DoublePoint();
                        p1.setLocation(value, 1.0);
                        valuesAndNumber.add(p1);
                    }
                }
            }

            // Must find the most probable one!
            if (valuesAndNumber.size() > 0) {
                int index = -1;
                int cpt = 0;
                int max = -1;
                for(DoublePoint p: valuesAndNumber) {
                    if (p.getY() > max) {
                        index = cpt;
                        max = (int)(p.getY());
                    }
                    cpt ++;
                }
                cd.probabilityOfMostCommonValue[i-cd.firstProp] = max / ((double)total);
                cd.mostCommonValue[i-cd.firstProp] = valuesAndNumber.get(index).getX();

            } else {
                cd.probabilityOfMostCommonValue[i-cd.firstProp] = -1;
            }
        }


	}*/

    public double getProbabilityOfProperty(double value, int index) {
        int cptOther = 0;
        int cptGood = 0;

        for(ClusterElement ce: elements) {
            if (ce.propertiesV[index]) {
                if (ce.propertiesD[index] == value) {
                    cptGood ++;
                } else {
                    cptOther ++;
                }
            }
        }

        if (cptOther==0 && cptGood == 0) {
            return -1;
        }

        return cptGood/((double)(cptOther+cptGood));
    }



    public void jumpLineInReport() {
        addToReport("", ONLY_TEXT, "\n");
    }

    public void jumpLineInTopReport() {
        addToTopReport("\n");
    }
    public void addToTopReport(String s) {
        if (topReport == null) {
            topReport = "";
        }
        topReport += s + "\n";
    }


    public void addToReport(String header, int type, String footer) {
        addToReport(header, type, footer, null, null);
    }

    public void addToReport(String header, int type, String footer, Cluster cluster1, Cluster cluster2) {
        for(ClusterElement ce: elements) {
            switch(type) {
            case ONLY_TEXT:
                ce.addToReport(header + footer);
                break;
            case NAME:
                ce.addToReport(header + ce.getName() + footer);
                break;
	    case SVM:
                ce.addToReport(header + footer);
                break;
            case DEVIATION:
                ce.addToReport(header + ce.getShortDeviations() + footer);
                break;
	    case INVERSE_DEVIATION:
                ce.addToReport(header + ce.getShortDeviations() + footer);
                break;
            case WEIGHT_DEVIATION:
                ce.addToReport(header + ce.getShortWeightDeviations() + footer);
                break;
	    case INVERSE_WEIGHT_DEVIATION:
                ce.addToReport(header + ce.getShortWeightDeviations() + footer);
                break;
            case DEVIATION_PERCENTAGE:
                ce.addToReport(header + ce.getShortDeviationPercentages(cluster1, cluster2) + footer);
                break;
            case DEVIATION_WEIGHT_PERCENTAGE:
                ce.addToReport(header + ce.getShortDeviationWeightPercentages(cluster1, cluster2) + footer);
                break;
            case CLOSEST:
                ce.addToReport(header + ce.getShortClosestStringByClusterName() + footer);
                break;
	    case FULL_CLOSEST:
                ce.addToReport(header + ce.getClosestStringByClusterName() + footer);
                break;
            case PROBA_DIFFERENCE:
                ce.addToReport(header + ce.getShortProbaDifferencePercentage(cluster1, cluster2) + footer);
                break;
            case CORRELATION:
                ce.addToReport(header + ce.getShortCorrelationSummary(cd.listOfCorrelations) + footer);
                break;
            case PROBA_FACTOR:
                ce.addToReport(header + ce.getProbaFactorSummary(cluster1, cluster2) + footer);
                break;
            case DISTANCE_EPSILON_CLUSTER:
                ce.addToReport(header + ce.getShortSummaryOnEpsilonClusters(cluster1, cluster2) + footer);
                break;
            case VALUE_TO_ADD:
                ce.addToReport(header + footer);
                break;
            case DEGRESSIVE_CLOSEST:
                ce.addToReport(header + ce.getShortClosestStringByClusterName() + footer);
                break;
	    case FULL_DEGRESSIVE_CLOSEST:
                ce.addToReport(header + ce.getShortClosestStringByClusterName() + footer);
                break;
	    case DISTANCE_TO_AVERAGE:
                ce.addToReport(header + ce.getShortDistanceToAverage(cluster1, cluster2) + footer);
                break;
            default:
                ce.addToReport("Unknown element\n");
            }
        }


    }

    public String getAllReports(boolean printIntermediateScore, int nbOfElementsForCluster2, String name1, String name2, String extra) {
        StringBuffer ret= new StringBuffer("");
        if (nbOfElementsForCluster2 > 0) {
            ret.append("*** Overall report of testing mode ***");

        } else {
            ret.append("*** Overall report of " + getName() + " ***\n\n");
        }


        if (extra != null) {
            ret.append(extra);
            ret.append("\n\n ** Overall results **\n\n");
        }

        ret.append(topReport);



        if (printIntermediateScore) {
            int cpt = 0;
            if (nbOfElementsForCluster2 != 0) {
                ret.append("Elements of " + name1 + ":\n---------------\n\n");
            }
            for(ClusterElement ce: elements) {
                //TraceManager.addDev("Appending report of " + cpt);
                if ((nbOfElementsForCluster2 != 0) && (cpt == (elements.size() - nbOfElementsForCluster2))) {
                    ret.append("Elements of " + name2 + ":\n---------------\n\n");
                }

                ret.append("\n" + ce.getReport() + "\n");
                cpt ++;
            }
        }

        TraceManager.addDev("All reports done");

        return ret.toString();
    }

    public String getFullReportForSample(int index) {
        ClusterElement ce = elements.get(index);
        StringBuffer ret= new StringBuffer("Full report of sample " + ce.getName());
        TraceManager.addDev("Fetching report of " + ce.getName());
        ret.append("\n" + ce.getReport() + "\n" + ret.toString() + "\n---end of report");
        TraceManager.addDev("Got full report from " + ce.getName());
        return ret.toString();
    }

    // Epsilon clusters

    public void saveEpsilonClusters(double nb) {
        if (cd.masterEpsilonClusters == null) {
            cd.masterEpsilonClusters  = new Vector<MasterEpsilonClusters>();
        }
        MasterEpsilonClusters mec = new MasterEpsilonClusters(nb);
        mec.listOfClusters = cd.epsilonClusters;
        cd.masterEpsilonClusters.add(mec);

    }


    public void loadEpsilonClusters(double nb) {
        if (cd.masterEpsilonClusters == null) {
            return;
        }

        for(MasterEpsilonClusters mec: cd.masterEpsilonClusters) {
            if (mec.threshold == nb) {
                cd.epsilonClusters = mec.listOfClusters;
                cd.epsilonClusterDistanceThreshold = nb;
                return;
            }
        }
    }

    public void clearEpsilonClusters() {
        if (cd.epsilonClusters != null) {
            cd.epsilonClusters = null;
        }
    }


    public boolean isNullEpsilonCluster() {
        return cd.epsilonClusters == null;
    }

    public void addAllToEpsilonClusters(double dt, Cluster cl, boolean isRegular) {
        if (cd.epsilonClusters == null) {
            cd.epsilonClusters = new Vector<EpsilonCluster>();
        }

        int total = cl.getElements().size();

        int merged = 0;



        cd.epsilonClusterDistanceThreshold = dt;
        //boolean added;
        EpsilonCluster ec;
        LinkedList<EpsilonCluster> vector;
        int cpt = 0;

        for(ClusterElement ce: cl.getElements()) {

            if (cpt % 100 == 0) {
                setInfoOnCurrentComputation("Adding " + total + " elts to " + cd.epsilonClusters.size() + " " + dt + "-epsi clust. (" + merged + " merge)");
            }

            percentageOfCurrentComputation = (100 * cpt) / total;
            cpt ++;

            //added = false;
            vector = getEpsilonClusters(ce);
            if (vector == null) {
                ec = new EpsilonCluster(null, dt);
                cd.epsilonClusters.add(ec);
                ec.elements.add(ce);
                if (isRegular) {
                    ec.nbOfelementsInRegularCluster++;
                } else {
                    ec.nbOfelementsInMalwareCluster++;
                }
                //TraceManager.addDev("New epsilon cluster");
            } else {
                ec = vector.get(0);
                if (vector.size() == 1) {
                    //TraceManager.addDev("One cluster found");
                    //ec = vector.get(0);
                    ec.elements.add(ce);
                    if (isRegular) {
                        ec.nbOfelementsInRegularCluster++;
                    } else {
                        ec.nbOfelementsInMalwareCluster++;
                    }
                } else {
                    //TraceManager.addDev("Epsilon cluster merging");
                    // element is added to the first ec
                    ec.elements.add(ce);
                    if (isRegular) {
                        ec.nbOfelementsInRegularCluster++;
                    } else {
                        ec.nbOfelementsInMalwareCluster++;
                    }

                    // Must merge epsilon clusters together: clusters are merged with the first one
                    for(EpsilonCluster ec0: vector) {
                        if (ec0 != ec) {
                            cd.epsilonClusters.remove(ec0);
                            ec.addEpsilonCluster(ec0);
                            merged ++;
                        }
                    }

                }
            }
            //TraceManager.addDev("Nb of epsilon clusters:" + epsilonClusters.size() + " nb of elements:" + getNbOfElementsInEpsilonClusters());


        }
    }

    public  LinkedList<EpsilonCluster> getEpsilonClusters(ClusterElement ceTo) {
        LinkedList<EpsilonCluster> vector = null;
        for(EpsilonCluster ec: cd.epsilonClusters) {
            for(ClusterElement ce: ec.elements) {
                if (ce.IsSquareDistanceWithLessThan(ceTo, ec.threshold)) {
                    if (vector == null) {
                        vector = new LinkedList<EpsilonCluster>();
                    }
                    vector.add(ec);
                    break;
                }
            }
        }

        return vector;
    }

    public int getNbOfElementsInEpsilonClusters() {
        int nb = 0;
        for(EpsilonCluster ec: cd.epsilonClusters) {
            nb += ec.nbOfelementsInRegularCluster + ec.nbOfelementsInMalwareCluster;
        }

        return nb;
    }


    public static Point nbOfClosestToAveragePerProperty(ClusterElement ce, Cluster cluster1, Cluster cluster2) {

        Point p = new Point(0, 0);
        double d1, d2;

        for(int i=cluster1.cd.firstProp; i<cluster1.cd.lastProp; i++) {
            if (ce.propertiesV[i]) {
                d1 = Math.abs(cluster1.cd.meanPerProperties[i-cluster1.cd.firstProp] - ce.propertiesD[i]);
                d2 = Math.abs(cluster2.cd.meanPerProperties[i-cluster1.cd.firstProp] - ce.propertiesD[i]);
		//TraceManager.addDev("d1 = " + d1 + " d2=" + d2);

                if (d1 > d2) {
                    p.x = p.x + 1;
                } else if (d2 > d1) {
                    p.y = p.y + 1;
                }
            }
        }

        return p;
        //meanPerProperties
    }





    public String getShortSummaryOnEpsilonClusters() {
        String s = cd.epsilonClusters.size() + " epsilon clusters with threshold = " + cd.epsilonClusterDistanceThreshold + ": ";

        return s;
    }

    public String getSummaryOnEpsilonClusters() {
        String s = cd.epsilonClusters.size() + " epsilon clusters with threshold = " + cd.epsilonClusterDistanceThreshold + ": ";
        for(EpsilonCluster ec: cd.epsilonClusters) {
            s += "(" + ec.nbOfelementsInRegularCluster + "," + ec.nbOfelementsInMalwareCluster + ") ";
        }

        return s;
    }

    public double getEpsilonClusterThreshold() {
        return cd.epsilonClusterDistanceThreshold;
    }

    public void purgeProximity() {
        for(ClusterElement ce: elements) {
            ce.closest = null;
        }
	cd.elementsA = null;
    }

    public void purgeDeviationsOfElements() {
        for(ClusterElement ce: elements) {
            ce.deviations = null;
        }
        
    }

    public void purgeWeightDeviationsOfElements() {
        for(ClusterElement ce: elements) {
            ce.weightDeviations = null;
        }
        
    }

    public void purgeAllWeightDeviations() {
	for(ClusterData clda: datas) {
	    purgeWeightDeviations(clda);
	}
    }

    public void purgeWeightDeviations(ClusterData _cd) {
        _cd.weightedMeanPerProperties = null;
	_cd.weightedMeanPerPropertiesV = null;
    }

    public void purgeAllProbabilities() {
	for(ClusterData clda: datas) {
	    clda.sortedPropertyValues = null;
	    clda.sortedPropertyTotal = null;
	}
    }


    public void purgeAllEpsilonClusters() {
	    for(ClusterData clda: datas) {
		clda.epsilonClusters = null;
		clda.masterEpsilonClusters = null;
	    }
    }

    public void purgeAllCorrelations() {
	for(ClusterData clda: datas) {
	    clda.correlations = null;
	    clda.listOfCorrelations = null;
	}
    }

    public void purgeAllSVMModels() {
	for(ClusterData clda: datas) {
	    clda.svmSolutionModel = null;
	}
    }

    public void purgeClusterElements() {
        for(ClusterElement ce: elements) {
            //ce.properties = null;
            ce.propertiesD = null;
            ce.propertiesV = null;
            ce.deviations = null;
	    ce.weightDeviations = null;
            ce.closest = null;
            ce.epsilonClusters = null;
            ce.report = null;
        }
	datas = null;
        /*cd.elementsA = null;
        normalized = false;
        normalizedValues = null;
        cd.sortedPropertyValues = null;
        cd.sortedPropertyTotal = null;
        cd.meanPerProperties = null;
        cd.weightedMeanPerProperties = null;
        cd.meanPerPropertiesV = null;
        cd.meanPerPropertiesComputed = false;
        cd.standardDeviationPerProperties = null;
        //weightedStandardDeviationPerProperties = null;
        cd.standardDeviationPerPropertiesV = null;
        cd.standardDeviationPerPropertiesComputed = false;
        cd.correlations = null;
        cd.listOfCorrelations = null;
        cd.probabilityOfMostCommonValue = null;
        cd.mostCommonValue = null;
        cd.epsilonClusters = null;
        cd.masterEpsilonClusters = null;
        propertiesDComputed = false;
        cd.nbOfClosestComputed = 0;
        normalizedValues = null;
        TraceManager.addDev("Stopping all computation threads");
        if (manager != null) {
            manager.stopAll();
        }
        TraceManager.addDev("All computation threads stopped");
        manager = null;*/
    }


    public void setPercentage(int percentage) {
        percentageOfCurrentComputation = percentage;
        //TraceManager.addDev("Percentage=" + percentage);
    }

    public int getPercentage() {
        return percentageOfCurrentComputation;
    }

    public void resetInfoOnComputation() {
        infoOnCurrentComputation = "unknown";
        percentageOfCurrentComputation = 0;
    }

    public String getComputation() {
        return infoOnCurrentComputation;
    }

    public void setInfoOnCurrentComputation(String info) {
        percentageOfCurrentComputation = 0;
        infoOnCurrentComputation = info;
    }

    public void printMeanValues() {
        String toPrint = "Mean values of properties in " + name + ": ";
        for(int i=0; i<cd.nbOfProps; i++) {
            if (cd.meanPerPropertiesV[i]) {
                toPrint += cd.meanPerProperties[i];
            } else {
                toPrint += "u";
            }
            toPrint += " ";
        }

        TraceManager.addDev(toPrint);
    }


    // SVM
    public void computeSVMModel(Cluster cluster1, Cluster cluster2) {
	if (cd.svmSolutionModel != null) {
	    return;
	}

	if ((cluster1.cd.svmSolutionModel != null) && (cluster1.cd.svmSolutionModel == cluster2.cd.svmSolutionModel)) {
	    cd.svmSolutionModel = cluster1.cd.svmSolutionModel;
	    return;
	}
	

        ImmutableSvmParameterPoint param;
        ImmutableSvmParameterPoint.Builder paramPointBuilder = new ImmutableSvmParameterPoint.Builder();
        paramPointBuilder.nu = 0.5f;
        paramPointBuilder.cache_size = 40;
        paramPointBuilder.C = 2;
        paramPointBuilder.eps = 1e-3f;
        paramPointBuilder.p = 0.1f;
        paramPointBuilder.shrinking = true;
        paramPointBuilder.probability = false;
        paramPointBuilder.kernel = new LinearKernel();

        /*int svm_type = 0;
        int kernel_type = 2;
        int degree = 3;
        float gamma = 0;
        float coef0 = 0;*/

        param = paramPointBuilder.build();

        SVM svm = new C_SVC();
        MutableSvmProblem prob = new MutableBinaryClassificationProblemImpl<Byte, SparseVector>(Byte.class, 3);
        
        // Making vectors
	TraceManager.addDev("Making vectors");
	makeSVMVectors(prob, cluster1, REGULAR);
	makeSVMVectors(prob, cluster2, MALWARE);

        TraceManager.addDev("Computing solution");
        cd.svmSolutionModel = svm.train(prob, param);
	cluster1.cd.svmSolutionModel = cd.svmSolutionModel;
	cluster2.cd.svmSolutionModel = cd.svmSolutionModel;
    }

    public SolutionModel getSVMSolutionModel() {
	return cd.svmSolutionModel;
    }

    private void makeSVMVectors(MutableSvmProblem prob, Cluster cluster, int categorie) {
	for(ClusterElement ce: cluster.getElements()) {
	    SparseVector v = getSparseVector(ce);
	    prob.addExample(v, categorie);
	}
    }

    private SparseVector getSparseVector(ClusterElement ce) {
	SparseVector v = new SparseVector(cd.nbOfProps);
	for(int i=cd.firstProp; i<cd.lastProp; i++) {
	    v.indexes[i-cd.firstProp] = i;
	    if (ce.propertiesV[i]) {
		v.values[i-cd.firstProp] = (float)(ce.propertiesD[i]);
	    } else {
		v.values[i-cd.firstProp] = (float)0.5;
	    }
	}
	return v;
    }

    public int getSVMCategorie(ClusterElement ce) {
	if (cd.svmSolutionModel == null) {
	    return ERROR;
	}

	SparseVector v = getSparseVector(ce);
	int categorie =  (Integer)(((DiscreteModel) cd.svmSolutionModel).predictLabel(v));
	return categorie;
    }


}
