/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

ludovic.apvrille AT telecom-paristech.fr

This software is a computer program whose purpose is to allow malware analysis

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

/**
 * Class ClusterAlgoMultithread
 * Thread to execute  algorithms of cluster
 * Creation: 17/07/2013
 * @version 1.0 17/07/2013
 * @author Ludovic APVRILLE
 * @see
 */

package cluster;

import myutil.*;


public class ClusterAlgoMultithread extends Thread {
	
	public static final int ALGO_CLOSEST = 1;
	public static final int ALGO_CLOSEST_PROPERTIES = 2;
	public static final int ALGO_CORRELATIONS = 3;
	
	private int id;
	
	public  Cluster cluster;
	public Cluster[] withClusters;
	public int algo;
	public int nbOfElements;
	public int nbOfProperties;
	public int startIndex;
	public int endIndex;
	public boolean updatePercentage;
	
	private boolean go;
	private boolean run;
	
	public ClusterAlgoMultithread(int _id, Cluster _cluster) {
		id = _id;
		cluster = _cluster;
		go = true;
		run = false;
	}
	
	public void stopMe() {
		go = false;
		executeAlgo();
		waitForAlgoDone();
	}
	
	// Thread side
	public synchronized void waitForRun() {
		while(run == false) {
			try {
				wait();
			} catch (InterruptedException ie) {
			}
		}
	}
	
	public synchronized void algoDone() {
		run = false;
		notifyAll();
	}
	
	
	// Caller side
	public synchronized void waitForAlgoDone() {
		while (run == true) {
			try {
				wait();
			} catch (InterruptedException ie) {
			}
		}
	}
	
	
	public synchronized void executeAlgo() {
		while (run == true) {
			try {
				wait();
			} catch (InterruptedException ie) {
			}
		}
		run = true;
		notifyAll();
	}
	
	// Main function of thread
	public void run() {
		while(go) {
			waitForRun();
			if (go == true) {
				execute();
				algoDone();
			}
		}
		algoDone();
	}
	
	// Algos called by thread
	public void execute() {
		switch(algo) {
		case ALGO_CLOSEST:
			TraceManager.addDev("Computing closest in thread #" + id + " from element " + startIndex + " to " + endIndex + " (not included)\n");
			cluster.computeClosestElementsRange(nbOfElements, withClusters, startIndex, endIndex, updatePercentage);
			break;
		case ALGO_CLOSEST_PROPERTIES:
			TraceManager.addDev("Computing closest properties in thread #" + id + " from element " + startIndex + " to " + endIndex + " (not included)\n");
			cluster.computeClosestElementsWithPropertiesRange(nbOfElements, nbOfProperties, withClusters, startIndex, endIndex, updatePercentage);
			break;
		case ALGO_CORRELATIONS:
			TraceManager.addDev("Computing correlations in thread #" + id + " from properties " + startIndex + " to " + endIndex + " (not included)\n");
			cluster.computeCorrelations(startIndex, endIndex, updatePercentage);
			break;
		default:
		}
	}
				
}
