/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

   ludovic.apvrille AT telecom-paristech.fr

   This software is a computer program whose purpose is to allow malware analysis

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class ClusterData
   * Notion of data manipulated by clusters
   * Creation: 11/11/2014
   * @version 1.0 11/11/2014
   * @author Ludovic APVRILLE
   * @see
   */

package cluster;

//import myutil.*;
//import java.io.*;
import java.util.*;
//import java.awt.*;

import edu.berkeley.compbio.jlibsvm.SolutionModel;


public class ClusterData  {

    public int firstProp = -1; // included
    public int nbOfProps = -1;
    public int lastProp = -1; // not included: nbOfProps = lastProp - firstProp
    
    public int maxNbOfValues = -1; // Nb Of maximum values for properties
 
    
    // C = Clean or Malware Cluster
    // G = Guess cluster

     // Weight of properties
    public int[] weightOfProperties; // Used by C and G

    // Mean with no prop. weight.
    public double[] meanPerProperties; // Used by C
    public boolean[] meanPerPropertiesV; // Used by C

    // Mean with prop. weight
    public double[] weightedMeanPerProperties; // Used by C
    public boolean[] weightedMeanPerPropertiesV; // Used by C

    // Usual standard&weight  deviations-> stored in ClusterElement of G, computed
    // based on mean Per Properties of C

    // Standard deviation for correlations
    public double[] standardDeviationPerProperties; // Used by C
    public boolean standardDeviationPerPropertiesV []; // Used by C
    // Correlations
    public double[][] correlations; //Use of value -2 when correlation is invalid  // Used by C
    public Vector<PropertiesCorrelation> listOfCorrelations; // Used by C. User also by G for storing non exclusive interesting correlations
    public double correlationThreshold = 0.75;


    // Probabilities (difference, factor)
    // Most common value for each property, and its probability
    // NOT USED ????????????????
    //public double[] probabilityOfMostCommonValue; // -1 not relevant // Used by C
    //public double[] mostCommonValue; // Used by C

    // Classified values DoubleInteger: value, nb of occurences
    public ArrayList<DoubleInteger>[] sortedPropertyValues = null; // Used by C // Used by *correlations*
    public int[] sortedPropertyTotal;  // Total nb of valid properties per property index // Used by C
    public int totalProbabilityDifference; // Total nb of proba diff with another cluster // Used by C
    public int totalProbabilityFactor; // Total nb of proba factor with another cluster // Used by C
    public double probaDifference; // Used by C and G
    public double probaFactor; // Used by C and G

   
    //public Vector<ProbabilityDifference> probaDifferences;
    //public double probaThreshold = 0.5;
    //public double probaRelation = 10;
    //public double lowProbaThreshold = 0.1;
    //public double highProbaThreshold = 1;


    
    // Proximity
    public int nbOfClosestComputed = 0; //Used by G.
    public ClusterElement[] elementsA; // Used by G

    // Epsilon clusters
    public Vector<MasterEpsilonClusters> masterEpsilonClusters; // List of epsilon clusters for various threshold value // Used by G, copied in C. Load from C in null in G, and not in C
    public Vector<EpsilonCluster> epsilonClusters; // Used by G, and copied in C
    public double epsilonClusterDistanceThreshold = -1.0;

    // SVM
    SolutionModel svmSolutionModel; // in G. Copied in C. If not in G, copied from C, if model equal in both C


    public ClusterData() {
    }

    public ClusterData(ClusterData _cd, int _firstProperty, int _nb) {
	firstProp = _firstProperty;
	nbOfProps = _nb;
	lastProp = firstProp + nbOfProps;
	weightOfProperties = new int[_cd.weightOfProperties.length];
	for(int i=0; i<weightOfProperties.length; i++) {
	    weightOfProperties[i] = _cd.weightOfProperties[i];
	}
    }

    public boolean hasSimilarPropertyProfile(ClusterData other) {
	return ((firstProp != other.firstProp) || (lastProp != other.lastProp) || (maxNbOfValues != other.maxNbOfValues)) ? false : true;
	/*if ((firstProp != other.firstProp) || (lastProp != other.lastProp)) {
	    return false;
	    }

	    return true;*/
    }

    public boolean isInvalid() {
	return (firstProp == -1);
    }

     public String toString() {
	return "first=" + firstProp + " last=" + lastProp + " nbOfProps=" + nbOfProps;
    }


}
