/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

   ludovic.apvrille AT telecom-paristech.fr

   This software is a computer program whose purpose is to allow malware analysis
   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.
   /**
   * Class ClusterElement
   * Element of a cluster
   * Creation: 22/06/2012
   * @version 1.0 22/06/2012
   * @author Ludovic APVRILLE
   * @see
   */

package cluster;

import myutil.*;


import java.util.*;
import java.awt.*;


public class ClusterElement  {

    private String name;
    public String[] properties;

    // Only properties between firstProperty and nbOfPropertiesToConsider in mainCluster
    public double[] propertiesD; // double value, if applicable
    public boolean[] propertiesV; // validity

    // Contains the full range of properties
    public double[] originalPropertiesD;
    public boolean[] originalPropertiesV;

    public Vector<ClusterDeviationElement> deviations;
    public Vector<ClusterWeightDeviationElement> weightDeviations;
    public double lastPercentageDeviationWithCluster1;
    public double lastPercentageDeviationWithCluster2;

    // Closest elements. Only in G
    public Vector<ClusterElementDistance> closest;
    //public double distance;
    //public int nbOfConsideredProperties;

    public String report;

    public Cluster mainCluster;

    public Vector<EpsilonCluster> epsilonClusters;

    public Point distanceToAverage;


    public ClusterElement(String _name) {
        name = _name;
        report = "";
    }

    public static ClusterElement makeZeroClusterElement(String _name, int _nbOfProperties) {
	ClusterElement ce = new ClusterElement(_name);
	ce.properties = new String[_nbOfProperties];
	for(int i=0; i<_nbOfProperties; i++) {
	    ce.properties[i] = "0";
	}
	return ce;
    }

    public void propertyWeightChanged() {
	weightDeviations = null;
    }

    public void propertyRangeChanged() {
	deviations = null;
	weightDeviations = null;
	epsilonClusters = null;
	closest = null;
    }

    /*public void setRangeOfProperties(int lowIndex, int nbOfPropertiesToConsider) {
        propertiesD = Arrays.copyOfRange(originalPropertiesD, lowIndex, lowIndex+nbOfPropertiesToConsider);
        propertiesV = Arrays.copyOfRange(originalPropertiesV, lowIndex, lowIndex+nbOfPropertiesToConsider);
	}*/

    public String getDescription() {
        String s = name;
        if ((propertiesV == null) && (properties == null)) {
            return "(no properties)";
        }
        if (properties == null) {
            for(int i=0; i<originalPropertiesD.length; i++) {
                if (originalPropertiesV[i]) {
                    s+= "," + originalPropertiesD[i];
                } else {
                    s+= ", -";
                }
            }
        } else {
            for(int i=0; i<properties.length; i++) {
                s+= "," + properties[i];
            }
        }
        return s;
    }

    public String getName() {
        return name;
    }

    public boolean hasDeviationWith(Cluster cluster) {
        if (deviations == null) {
            return false;
        }

        for(ClusterDeviationElement cde: deviations) {
            if (cde.comparisonCluster == cluster) {
                return true;
            }
        }

        return false;
    }

    public ClusterDeviationElement getDeviationElementWith(Cluster cluster) {
        if (deviations == null) {
            return null;
        }

        for(ClusterDeviationElement cde: deviations) {
            if (cde.comparisonCluster == cluster) {
                return cde;
            }
        }

        return null;
    }

    public boolean hasWeightDeviationWith(Cluster cluster) {
        if (weightDeviations == null) {
            return false;
        }

        for(ClusterWeightDeviationElement cwde: weightDeviations) {
            if (cwde.comparisonCluster == cluster) {
                return true;
            }
        }

        return false;
    }

    public ClusterWeightDeviationElement getWeightDeviationElementWith(Cluster cluster) {
        if (weightDeviations == null) {
            return null;
        }

        for(ClusterWeightDeviationElement cwde: weightDeviations) {
            if (cwde.comparisonCluster == cluster) {
                return cwde;
            }
        }

        return null;
    }


    public DoublePoint getPercentageDeviationsWith(Cluster cluster1, Cluster cluster2) {

        if (deviations == null) {
            return new DoublePoint(0, 0);
        }

        ClusterDeviationElement cde1 = getDeviationElementWith(cluster1);
        ClusterDeviationElement cde2 = getDeviationElementWith(cluster2);

        if ((cde1 == null) && (cde2 == null)) {
            return new DoublePoint(0, 0);
        }

        if (cde1 == null) {
            return new DoublePoint(0, 100);
        }

        if (cde2 == null) {
            return new DoublePoint(100,0);
        }

        int in1 = 0;
        int in2 = 0;
        int totalComputed = 0;

        for(int i=0; i<mainCluster.cd.nbOfProps; i++) {
            if ((cde1.standardDeviationV[i]) || (cde2.standardDeviationV[i])) {

                if (!(cde1.standardDeviationV[i])) {
                    in2 ++;
                } else {
                    if (!(cde2.standardDeviationV[i])) {
                        in1 ++;

                    } else {
                        // Both are valid
                        // We must compare their value
                        if (cde1.standardDeviation[i] < cde2.standardDeviation[i]) {
                            in1 ++;
                        } else {
                            in2 ++;
                        }

                    }
                }


            }
        }

        totalComputed = in1 + in2;
        if (totalComputed == 0) {
            return new DoublePoint(0, 0);
        }


        return new DoublePoint(100 * in1/(double)totalComputed, 100 * in2/(double)totalComputed);
    }

    public DoublePoint getWeightPercentageDeviationsWith(Cluster cluster1, Cluster cluster2) {

        if (deviations == null) {
            return new DoublePoint(0, 0);
        }

        ClusterWeightDeviationElement cde1 = getWeightDeviationElementWith(cluster1);
        ClusterWeightDeviationElement cde2 = getWeightDeviationElementWith(cluster2);

        if ((cde1 == null) && (cde2 == null)) {
            return new DoublePoint(0, 0);
        }

        if (cde1 == null) {
            return new DoublePoint(0, 100);
        }

        if (cde2 == null) {
            return new DoublePoint(100,0);
        }

        int in1 = 0;
        int in2 = 0;
        int totalComputed = 0;

        for(int i=mainCluster.cd.firstProp; i<mainCluster.cd.lastProp; i++) {
            if ((cde1.weightDeviationV[i-mainCluster.cd.firstProp]) || (cde2.weightDeviationV[i-mainCluster.cd.firstProp])) {

                if (!(cde1.weightDeviationV[i-mainCluster.cd.firstProp])) {
                    in2 += mainCluster.cd.weightOfProperties[i];
                } else {
                    if (!(cde2.weightDeviationV[i-mainCluster.cd.firstProp])) {
                        in1 += mainCluster.cd.weightOfProperties[i];

                    } else {
                        // Both are valid
                        // We must compare their value
                        if (cde1.weightDeviation[i-mainCluster.cd.firstProp] < cde2.weightDeviation[i-mainCluster.cd.firstProp]) {
                            in1 += mainCluster.cd.weightOfProperties[i];
                        } else {
                            in2 += mainCluster.cd.weightOfProperties[i];
                        }

                    }
                }


            }
        }

        totalComputed = in1 + in2;
        if (totalComputed == 0) {
            return new DoublePoint(0, 0);
        }


        return new DoublePoint(100 * in1/(double)totalComputed, 100 * in2/(double)totalComputed);
    }


    public boolean hasOnlyInvalidProperties() {
        for(int i=0; i<propertiesV.length; i++) {
            if (propertiesV[i]) {
                return false;
            }
        }
        return true;
    }

    public boolean hasMoreThanInvalidProperties(int maxInvalid) {
        int cpt = 0;

        for(int i=0; i<propertiesV.length; i++) {
            if (!propertiesV[i]) {
                cpt++;
                if (cpt > maxInvalid) {
                    return true;
                }
            }
        }
        return false;
    }

    public void addDeviationElement(ClusterDeviationElement cde) {
        if (deviations == null) {
            deviations = new Vector<ClusterDeviationElement>();
        }
        deviations.add(cde);
    }

    public void addWeightDeviationElement(ClusterWeightDeviationElement cwde) {
        if (weightDeviations == null) {
            weightDeviations = new Vector<ClusterWeightDeviationElement>();
        }
        weightDeviations.add(cwde);
    }

    public String getDeviations() {
        String ret = "deviation of " + name ;
        for(ClusterDeviationElement cde: deviations) {
            if (cde.totalDeviationV) {
                ret += "  with " + cde.comparisonCluster.getName() + " = " + cde.totalDeviation + " ";
            } else {
                ret += cde.comparisonCluster.getName() + " = invalid";
            }
        }
        return ret + "\n";
    }

    public String getWeightDeviations() {
        String ret = "Weight deviation of " + name ;
        for(ClusterWeightDeviationElement cwde: weightDeviations) {
            if (cwde.totalWeightDeviationV) {
                ret += "  with " + cwde.comparisonCluster.getName() + " = " + cwde.totalWeightDeviation + " ";
            } else {
                ret += cwde.comparisonCluster.getName() + " = invalid";
            }
        }
        return ret + "\n";
    }

    public double getTotalDeviation(Cluster _cluster) {
        for(ClusterDeviationElement cde: deviations) {
            if ((cde.totalDeviationV) && (cde.comparisonCluster == _cluster)) {
                return cde.totalDeviation;
            }
        }
        return 0;
    }

    public double getTotalWeightDeviation(Cluster _cluster) {
        for(ClusterWeightDeviationElement cwde: weightDeviations) {
            if ((cwde.totalWeightDeviationV) && (cwde.comparisonCluster == _cluster)) {
                return cwde.totalWeightDeviation;
            }
        }
        return 0;
    }

    public String getShortDeviations() {
        String ret = "";
        for(ClusterDeviationElement cde: deviations) {
            if (cde.totalDeviationV) {
                ret += " with " + cde.comparisonCluster.getName() + " = " + cde.totalDeviation + " ";
            } else {
                ret += " with " + cde.comparisonCluster.getName() + " = invalid ";
            }
        }
        return ret;
    }

    public String getShortDistanceToAverage(Cluster cluster1, Cluster cluster2) {
        String ret = "";
	
	if (distanceToAverage == null) {
	    return "Distance to average not computed...";
	}

        ret += "# of lowest distance to average. With " + cluster1.getName() + ": " + distanceToAverage.x + ". With " + cluster2.getName() + ": " + distanceToAverage.y + ".";
        return ret;
    }

    public String getShortWeightDeviations() {
        String ret = "";
        for(ClusterWeightDeviationElement cwde: weightDeviations) {
            if (cwde.totalWeightDeviationV) {
                ret += " with " + cwde.comparisonCluster.getName() + " = " + cwde.totalWeightDeviation + " ";
            } else {
                ret += cwde.comparisonCluster.getName() + " = invalid";
            }
        }
        return ret;
    }

    public String getShortDeviationPercentages(Cluster cluster1, Cluster cluster2) {
        String ret = " percentage of closest property deviations with " + cluster1.getName() + " = " + lastPercentageDeviationWithCluster1;
        ret+= " percentage of closest property deviations with " + cluster2.getName() + " = " + lastPercentageDeviationWithCluster2;
        return ret;
    }

    public String getShortDeviationWeightPercentages(Cluster cluster1, Cluster cluster2) {
        String ret = " weighted percentage of closest property deviations with " + cluster1.getName() + " = " + lastPercentageDeviationWithCluster1;
        ret+= " weighted percentage of closest property deviations with " + cluster2.getName() + " = " + lastPercentageDeviationWithCluster2;
        return ret;
    }

    public String getDeviationStarComparisonString(int index0, int index1) {
        ClusterDeviationElement cde0 = deviations.get(0);
        ClusterDeviationElement cde1 = deviations.get(1);

        if (!cde0.totalDeviationV || !cde1.totalDeviationV) {
            return name + ": Invalid deviation\n";
        }

        //double total = cde0.totalDeviation + cde1.totalDeviation;
        String ret = cde0.comparisonCluster.getName() + "|";
        int stars = 20;
        for(int i=0; i<(int)(cde0.totalDeviation * stars); i++) {
            ret += "-";
        }
        ret += "*";
        for(int i=0; i<(int)(cde1.totalDeviation * stars); i++) {
            ret += "-";
        }
        ret += "|" + cde1.comparisonCluster.getName() + " for: " + name + "\n";

        return ret;

    }

    public String getMinDeviations() {
        double max = Double.MAX_VALUE;
        ClusterDeviationElement minCde = null;
        for(ClusterDeviationElement cde: deviations) {
            if (cde.totalDeviationV) {
                if (cde.totalDeviation < max) {
                    minCde = cde;
                    max = cde.totalDeviation;
                }
            }
        }

        if (minCde == null) {
            return "Closer group of " + name + ": none found";
        }



        return "Closer group of " + name + ": " + minCde.comparisonCluster.getName() + "\n";
    }

    // Assumes properties have been computed
    public double computeDistanceWith(ClusterElement ce) {
        //TraceManager.addDev("Distance from " + name + " with " + ce.name);
        //int cpt = 0;
        double distance = 0.0;
        double tmp;
        for(int i=mainCluster.cd.firstProp; i<mainCluster.cd.lastProp; i++) {
            if (propertiesV[i] && ce.propertiesV[i]) {
                //cpt += ce.mainCluster.weightOfProperties[i];
                tmp = propertiesD[i] - ce.propertiesD[i];
                distance += tmp * tmp * ce.mainCluster.cd.weightOfProperties[i];
            }
        }

        /*if (cpt == 0) {
          return Double.MAX_VALUE;
          }*/

        //return Math.pow(distance/cpt, 0.5);
        return distance;
    }

    public boolean IsSquareDistanceWithLessThan(ClusterElement ce, double dist) {
        //TraceManager.addDev("Distance from " + name + " with " + ce.name + " dist:" + dist);
        //int cpt = 0;
        double distance = 0.0;
        double tmp;
        for(int i=mainCluster.cd.firstProp; i<mainCluster.cd.lastProp; i++) {
            if (propertiesV[i] && ce.propertiesV[i]) {
                //cpt += ce.mainCluster.weightOfProperties[i];
                tmp = (propertiesD[i] - ce.propertiesD[i]);
                distance = distance + (tmp * tmp * ce.mainCluster.cd.weightOfProperties[i]);
		//if (i % 100 == 0) {
		//TraceManager.addDev("Distance = " + distance +  " i= " + i + " and dist=" + dist + " weight=" + ce.mainCluster.weightOfProperties[i] + " tmp=" + tmp);
		    //}
                if (distance >= dist) {
		    //TraceManager.addDev("Distance = " + distance +  " >  dist=" + dist);
                    return false;
                }
            }
        }

	/*if (distance > 0) {
	TraceManager.addDev("----------------------------------- FOUND Distance = " + distance + " < dist=" + dist);
	TraceManager.addDev("Distance from " + name + " with " + ce.name + " dist:" + dist);
	}*/
        return true;
    }

    public double computeSquareDistanceWith(ClusterElement ce) {
        //TraceManager.addDev("Distance from " + name + " with " + ce.name);
        //int cpt = 0;
        double distance = 0.0;
        double tmp;
	for(int i=mainCluster.cd.firstProp; i<mainCluster.cd.lastProp; i++) {
            if (propertiesV[i] && ce.propertiesV[i]) {
                //cpt += ce.mainCluster.weightOfProperties[i];
                tmp = (propertiesD[i] - ce.propertiesD[i]);
                distance += tmp * tmp * ce.mainCluster.cd.weightOfProperties[i];
            }
        }

        /*if (cpt == 0) {
          return Double.MAX_VALUE;
          }*/

        return distance;
    }

    public double computeSquareDistanceWithLimit(ClusterElement ce, double _maxDistance) {
        //TraceManager.addDev("Distance from " + name + " with " + ce.name);
        //int cpt = 0;
        double distance = 0.0;
        double tmp;
        for(int i=mainCluster.cd.firstProp; i<mainCluster.cd.lastProp; i++) {
            if (propertiesV[i] && ce.propertiesV[i]) {
                //cpt += ce.mainCluster.weightOfProperties[i];
                tmp = (propertiesD[i] - ce.propertiesD[i]);
                distance += tmp * tmp * ce.mainCluster.cd.weightOfProperties[i];
                if (distance >= _maxDistance) {
                    return Double.MAX_VALUE;
                }
            }
        }

        /*if (cpt == 0) {
          return Double.MAX_VALUE;
          }*/

        return distance;
    }

    // Assumes properties have been computed
    public double computeSquareDistanceConsideredPropertiesWith(int nbOfConsideredProperties, ClusterElement ce) {
        int cpt = 0;
        double distance = 0.0;
        double tmp;
        double [] distances = new double[mainCluster.cd.nbOfProps];
        for(int i=mainCluster.cd.firstProp; i<mainCluster.cd.lastProp; i++) {
            if (propertiesV[i] && ce.propertiesV[i]) {

                //TraceManager.addDev("cpt=" + cpt + " i = " + i + " lenghtD =" + propertiesD.length + "  length prop ce = " + ce.propertiesD.length);
                tmp = (propertiesD[i] - ce.propertiesD[i]);
                distances[i-mainCluster.cd.firstProp] = tmp * tmp * ce.mainCluster.cd.weightOfProperties[i];
                cpt += ce.mainCluster.cd.weightOfProperties[i];
            }
        }

        Arrays.sort(distances);

        for(int j=0; j<Math.min(nbOfConsideredProperties, cpt); j++) {
            distance += distances[j];
        }

        return distance;
    }

    // Assumes properties have been computed
    public double computeDistanceConsideredPropertiesWith(int nbOfConsideredProperties, ClusterElement ce) {
        int cpt = 0;
        double distance = 0.0;
        double tmp;
        double [] distances = new double[mainCluster.cd.nbOfProps];
        for(int i=mainCluster.cd.firstProp; i<mainCluster.cd.lastProp; i++)  {
            if (propertiesV[i] && ce.propertiesV[i]) {

                //TraceManager.addDev("cpt=" + cpt + " i = " + i + " lenghtD =" + propertiesD.length + "  length prop ce = " + ce.propertiesD.length);
                tmp = (propertiesD[i] - ce.propertiesD[i]);
                distances[i-mainCluster.cd.firstProp] = tmp * tmp * ce.mainCluster.cd.weightOfProperties[i];
                cpt += ce.mainCluster.cd.weightOfProperties[i];
            }
        }

        Arrays.sort(distances);

        for(int j=0; j<Math.min(nbOfConsideredProperties, cpt); j++) {
            distance += distances[j];
        }

        return Math.sqrt(distance);
    }

    public String getClosestString() {
        String ret = "";
        boolean first = true;
        ret += mainCluster.getName() + "/" + name + ": ";
        for(ClusterElementDistance ced: closest) {
            if (!first) {
                ret += ", ";
            } else {
                first = false;
            }
            ret += ced.ce.mainCluster.getName() + "/" + ced.ce.name + "/" + ced.distance;
        }
        return ret;
    }

    public String getClosestStringByClusterName() {
        String ret = "";
        //boolean first = true;
        //String val;
        //Integer i;
        boolean found;
        ret += mainCluster.getName() + "/" + name + ": ";
        Vector<String> names = new Vector<String>();
        Vector<Integer> values = new Vector<Integer>();
        int index;
        for(ClusterElementDistance ced: closest) {
            found = false;
            index = 0;
            for(String s: names) {
                if (s.compareTo(ced.ce.mainCluster.getName()) == 0) {
                    found = true;
                    break;
                }
                index ++;
            }
            if (found) {
                values.setElementAt(new Integer(values.get(index).intValue()+1), index);
            } else {
                names.add(ced.ce.mainCluster.getName());
                values.add(new Integer(1));
            }

        }
        for(index=0; index<names.size(); index++) {
            if (index > 0) {
                ret += ", ";
            }
            ret += names.get(index) + "/" + values.get(index).intValue();

        }

	ret += "\nClosest: " + getClosestString() + "\n";

        return ret;
    }

    public int getNbOfClosestInCluster(Cluster inCluster) {
        int max = mainCluster.cd.nbOfClosestComputed;
        int cpt = 0;
        int nb = 0;
        for(ClusterElementDistance ced: closest) {
            if (cpt == max) {
                break;
            }
            if (ced.ce.mainCluster == inCluster) {
                nb++;
            }

            cpt ++;
        }

        return nb;
    }

    // y = a*x*x + c
    public double getClosestParabolicInCluster(Cluster inCluster, double a, double c) {
        int max = mainCluster.cd.nbOfClosestComputed;
        int cpt = 0;
        double result = 0;
        for(ClusterElementDistance ced: closest) {
            if (cpt == max) {
                break;
            }
            if (ced.ce.mainCluster == inCluster) {
                result = result + a * cpt * cpt + c;
            }

            cpt ++;
        }

        return result;
    }

    public String getShortClosestStringByClusterName() {
        String ret = "";
        //boolean first = true;
        //String val;
        //Integer i;
        boolean found;
        Vector<String> names = new Vector<String>();
        Vector<Integer> values = new Vector<Integer>();
        //Vector<Integer> nbPerName = new Vector<Integer>();
        int index;
        int max = mainCluster.cd.nbOfClosestComputed;
        int cpt = 0;
        for(ClusterElementDistance ced: closest) {
            found = false;
            index = 0;
            for(String s: names) {
                if (s.compareTo(ced.ce.mainCluster.getName()) == 0) {
                    found = true;
                    break;
                }
                index ++;
            }
            if (found) {
                values.setElementAt(new Integer(values.get(index).intValue()+1), index);
            } else {
                names.add(ced.ce.mainCluster.getName());
                values.add(new Integer(1));
            }
            cpt ++;
            if (cpt == max) {
                break;
            }

        }
        for(index=0; index<names.size(); index++) {
            if (index > 0) {
                ret += ", ";
            }
            ret += names.get(index) + "/" + values.get(index).intValue();

        }
        return ret;
    }

    public void sortClosest() {
        ClusterElementDistance ced;
        Vector<ClusterElementDistance> newV = new Vector<ClusterElementDistance>();
        while(closest.size() > 0) {
            ced = getMinDistance(closest);
            closest.remove(ced);
            newV.add(ced);
            //TraceManager.addDev("Adding " + ced.ce.name + " distance=" + ced.distance);
        }
        //TraceManager.addDev("Done sorting");
        closest = newV;
    }

    private ClusterElementDistance getMinDistance(Vector<ClusterElementDistance> elts) {
        double distance = Double.MAX_VALUE;
        ClusterElementDistance ret = null;
        for(ClusterElementDistance ced: elts) {
            //TraceManager.addDev("current distance=" + distance + " evaluated distance:" +  ced.distance);
            if (ced.distance<distance) {
                ret = ced;
                distance = ced.distance;
            }
        }
        return ret;
    }

    public String getShortProbaDifferencePercentage(Cluster cluster1, Cluster cluster2) {

        double val1, val2;
        String prop1 = "";
        String prop2 = "";
        int cpt1=0, cpt2=0;
        double difference = cluster1.getProbabilityDifference();

        for (int i=cluster1.cd.firstProp; i<cluster1.cd.lastProp; i++) {
            if (propertiesV[i]) {
                val1 = cluster1.getProbability(i, propertiesD[i]);
                val2 = cluster2.getProbability(i, propertiesD[i]);


                if ((val1 >= 0) && (val2 >= 0)) {

                    if (val2 >= (val1  + difference)) {
                        //totalCluster2 ++;
                        if (cpt2 <= Cluster.MAX_DISPLAYED_PROPS) {
                            if (cpt2 == Cluster.MAX_DISPLAYED_PROPS) {
                                prop2 += "...";
                            } else {
                                prop2 += "" + i + " ";
                            }
                        }
                        cpt2 ++;
                        //prop2 += "" + i + " ";
                        //TraceManager.addDev("cluster2++, index=" + i + " val=" +  propertiesD[i] + "val1=" + val1 + " val2 =" + val2);
                    }

                    if (val1 >= (val2 + difference)) {
                        //TraceManager.addDev("cluster1++, index=" + i + " val=" +  propertiesD[i] + "val1=" + val1 + " val2 =" + val2);

                        //totalCluster1 ++;

                        if (cpt1 <= Cluster.MAX_DISPLAYED_PROPS) {
                            if (cpt1 == Cluster.MAX_DISPLAYED_PROPS) {
                                prop1 += "...";
                            } else {
                                prop1 += "" + i + " ";
                            }
                        }
                        cpt1 ++;

                    }
                }

            }
        }

        return " - In " + cluster1.getName() + " sample respects: (" + prop1 + "); In " + cluster2.getName() + " sample respects: (" + prop2 + ")";

    }


    public String getShortCorrelationSummary(Vector<PropertiesCorrelation> listOfCorrelations) {
        String ret = "";

        //TraceManager.addDev("NbOfCorrelations:" + listOfCorrelations.size());

        // Compute different Clusters
        Vector<Cluster> clusters = new Vector<Cluster>();
        for(PropertiesCorrelation pc: listOfCorrelations) {
            if (!(clusters.contains(pc.cluster))) {
                clusters.add(pc.cluster);
            }
        }

        // Computation for different clusters
        int totalInCluster;
        int totalApplicable;
        double shouldBe;
        String prop;
        for(Cluster cluster: clusters) {
            totalInCluster = 0;
            totalApplicable = 0;
            prop = "";
            for(PropertiesCorrelation pc: listOfCorrelations) {
                if (pc.cluster == cluster) {
                    totalInCluster ++;
                    if (propertiesV[pc.property1Index] && propertiesV[pc.property2Index]) {
                        shouldBe = pc.a * propertiesD[pc.property1Index] + pc.b;
                        if (shouldBe - propertiesD[pc.property2Index] < 0.001) {
                            totalApplicable ++;
                            prop += "(" + pc.property1Index + "," +  pc.property2Index + ") ";
                        }
                    }
                }
            }

            ret += "\n  ->  with " + cluster.getName() + ": " + ((totalApplicable / ((double)totalInCluster)) * 100) + "% " + prop + " ";

        }

        if (ret.length() > 0) {
            return ret;
        }

        return " no exclusive correlations in analyzed clusters";

    }


    public double getPercentageOfCorrelations(Cluster _cluster) {
        int totalInCluster;
        int totalApplicable;
        double shouldBe;
        totalInCluster = 0;
        totalApplicable = 0;
        for(PropertiesCorrelation pc: _cluster.getListOfCorrelations()) {
            if (pc.cluster == _cluster) {
                totalInCluster ++;
                if (propertiesV[pc.property1Index+_cluster.cd.firstProp] && propertiesV[pc.property2Index+_cluster.cd.firstProp]) {
                    shouldBe = pc.a * propertiesD[pc.property1Index+_cluster.cd.firstProp] + pc.b;
                    if (shouldBe - propertiesD[pc.property2Index+_cluster.cd.firstProp] < 0.001) {
                        totalApplicable ++;
                    }
                }

            }

        }

        if (totalInCluster == 0) {
            return 0;
        }

        return ((totalApplicable / ((double)totalInCluster)) * 100) ;
    }

    public DoublePoint  getPercentageProbaDifference(Cluster cluster1, Cluster cluster2) {
        int totalCluster1 = 0;
        int totalCluster2 = 0;
        double val1, val2;
        double relation = mainCluster.getProbabilityDifference();
        int totalc1 = cluster1.getTotalProbabilityDifference();
        int totalc2 = cluster2.getTotalProbabilityDifference();

        //TraceManager.addDev("Computing proba diff for diff=" + relation + " totalc1=" + totalc1 + " totalc2=" + totalc2);

        for (int i=cluster1.cd.firstProp; i<cluster1.cd.lastProp; i++) {
            if (propertiesV[i]) {
                val1 = cluster1.getProbability(i, propertiesD[i]);
                val2 = cluster2.getProbability(i, propertiesD[i]);

                if ((val1 >= 0) && (val2 >= 0)) {

                    if (val2 >= (val1  + relation)) {
                        totalCluster2 ++;
                        //TraceManager.addDev("cluster2++, index=" + i + " val=" +  propertiesD[i] + " val1=" + val1 + " val2 =" + val2);
                    }

                    if (val1 >= (val2 + relation)) {
                        //TraceManager.addDev("cluster1++, index=" + i + " val=" +  propertiesD[i] + " val1=" + val1 + " val2 =" + val2);
                        totalCluster1 ++;
                    }
                }

            }
        }

        double total1, total2;
        //DoublePoint dp = new DoublePoint();

        if (totalc1 == 0) {
            total1 = 0;
        } else {
            total1 = totalCluster1 / (double)totalc1;
        }

        if (totalc2 == 0) {
            total2 = 0;
        } else {
            total2 = totalCluster2 / (double)totalc2;
        }

        //TraceManager.addDev("Score = " + (100 * total1) + "," +  (100 * total2));

        return new DoublePoint(100 * total1, 100 * total2);
    }

    public String  getProbaFactorSummary(Cluster cluster1, Cluster cluster2) {
        //int totalCluster1 = 0;
        //int totalCluster2 = 0;
        double val1, val2;
        String prop1 = "";
        String prop2 = "";
        double relation = mainCluster.getProbabilityFactor();
        //int totalc1 = cluster1.getTotalProbabilityFactor();
        //int totalc2 = cluster2.getTotalProbabilityFactor();

        for (int i=cluster1.cd.firstProp; i<cluster1.cd.lastProp; i++) {
            if (propertiesV[i]) {
                val1 = cluster1.getProbability(i, propertiesD[i]);
                val2 = cluster2.getProbability(i, propertiesD[i]);

                if ((val1 >= 0) && (val2 >= 0)) {

                    if (val2 >= (val1  * relation)) {
                        //totalCluster2 ++;
                        prop2 += "" + i + " ";
                        //TraceManager.addDev("cluster2++, index=" + i + " val=" +  propertiesD[i] + "val1=" + val1 + " val2 =" + val2);
                    }

                    if (val1 >= (val2 * relation)) {
                        //TraceManager.addDev("cluster1++, index=" + i + " val=" +  propertiesD[i] + "val1=" + val1 + " val2 =" + val2);

                        //totalCluster1 ++;
                        prop1 += "" + i + " ";
                    }
                }

            }
        }

        return cluster1.getName() + "(" + prop1 + ")  " + cluster2.getName() + "(" + prop2 + ")";
    }

    public DoublePoint  getPercentageProbaFactorSummary(Cluster cluster1, Cluster cluster2) {
        int totalCluster1 = 0;
        int totalCluster2 = 0;
        double val1, val2;
        double relation = mainCluster.getProbabilityFactor();
        int totalc1 = cluster1.getTotalProbabilityFactor();
        int totalc2 = cluster2.getTotalProbabilityFactor();

        for (int i=cluster1.cd.firstProp; i<cluster1.cd.lastProp; i++) {
            if (propertiesV[i]) {
                val1 = cluster1.getProbability(i, propertiesD[i]);
                val2 = cluster2.getProbability(i, propertiesD[i]);

                if ((val1 >= 0) && (val2 >= 0)) {

                    if (val2 >= (val1  * relation)) {
                        totalCluster2 ++;
                        //TraceManager.addDev("cluster2++, index=" + i + " val=" +  propertiesD[i] + "val1=" + val1 + " val2 =" + val2);
                    } else if (val1 >= (val2 * relation)) {
                        //TraceManager.addDev("cluster1++, index=" + i + " val=" +  propertiesD[i] + "val1=" + val1 + " val2 =" + val2);
                        totalCluster1 ++;
                    } else {
                        //total ++;
                    }
                }

            }
        }
        double total1, total2;
        //DoublePoint dp = new DoublePoint();

        if (totalc1 == 0) {
            total1 = 0;
        } else {
            total1 = totalCluster1 / (double)totalc1;
        }

        if (totalc2 == 0) {
            total2 = 0;
        } else {
            total2 = totalCluster2 / (double)totalc2;
        }

        //TraceManager.addDev("Score = " + (100 * total1) + "," +  (100 * total2));

        return new DoublePoint(100 * total1, 100 * total2);
    }

    public String getShortSummaryOnEpsilonClusters(Cluster cluster1, Cluster cluster2) {
        //String ret = "";

        DoublePoint dp = getNbOfNeighboursInEpsilonCluster(cluster1, cluster2);
        double reg = dp.getX();
        double mal = dp.getY();

        double perReg, perMal;

        if (mal + reg == 0) {
            perReg = 0;
            perMal = 0;
        } else {
            perReg = 100*reg/(reg+mal);
            perMal = 100*mal/(reg+mal);
        }


        return "sample has " + reg + "(" + perReg +"%) " + cluster1.getName() + " and " + mal + "(" + perMal + "%) " + cluster2.getName() + " in its epsilon cluster (total of epsilon clusters: " + mainCluster.getEpsilonClusters().size() +")";

    }

    public DoublePoint  getPercentageNeighboursInEpsilonCluster(Cluster cluster1, Cluster cluster2) {
        DoublePoint dp = getNbOfNeighboursInEpsilonCluster(cluster1, cluster2);

        double reg = dp.getX();
        double mal = dp.getY();

        if (reg + mal ==0) {
            return dp;
        }

        return new DoublePoint(100*reg/(reg+mal), 100*mal/(reg+mal));
    }


    public DoublePoint  getNbOfNeighboursInEpsilonCluster(Cluster cluster1, Cluster cluster2) {
        LinkedList<EpsilonCluster> vector = mainCluster.getEpsilonClusters(this);

        int reg = 0;
        int mal = 0;

        if (vector == null) {
            return new DoublePoint(0, 0);
        }

        for(EpsilonCluster ec: vector) {
            reg += ec.nbOfelementsInRegularCluster;
            mal += ec.nbOfelementsInMalwareCluster;
        }

        //TraceManager.addDev("Nb of reg:" + reg + " nb of mal:" + mal);


        return new DoublePoint(reg, mal);

    }


    public void addToReport(String _s) {
        report += _s;
    }

    public String getReport() {
        if (report == null) {
            return "";
        }

        //TraceManager.addDev("Report is:" + report);

        return report;
    }

    public void addSortedInClosest(ClusterElementDistance ced) {
        int cpt = 0;

        for(ClusterElementDistance current: closest) {
            if (ced.distance <= current.distance) {
                break;
            }
            cpt++;
        }
        closest.add(cpt, ced);

    }

    public double getMaxDistanceInClosest() {
        return closest.get(closest.size()-1).distance;
    }


    
}
