/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

ludovic.apvrille AT telecom-paristech.fr

This software is a computer program whose purpose is to allow malware analysis

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

/**
 * Class ProgressionDisplayer.java
 * Displays the progression of algorithms
 * Creation: 22/08/2012
 * @version 1.0 22/08/2012
 * @author Ludovic APVRILLE
 * @see
 */

package myutil;

import java.io.*;
import java.util.*;



public class ProgressionDisplayer extends Thread  {
	
    private StepProvider sp;
	private PercentageProvider pp;
	private boolean monitor = false;
	private boolean go;
	private String[] progressionIcons;
	
	
	public ProgressionDisplayer(PercentageProvider _pp) {
		pp = _pp;
	}
	
	public synchronized void setGo(boolean _go) {
		go = _go;
		notifyAll();
	}
	
	public void setProgressionIcons(String[] icons) {
		progressionIcons = icons;
	}
	
	
	public synchronized void setMonitor(boolean _monitor) {
		monitor = _monitor;
		notifyAll();
	}
	
	public void go() {
		go = true;
		this.start();
	}
	
	
	public synchronized void waitForMonitor() {
		while(!monitor && go) {
			try {
				wait();
			} catch (InterruptedException ie) {
			}
		}
	}
	
	public void setPercentageProvider(PercentageProvider _pp) {
		pp = _pp;
	}
	
	public void setStepProvider(StepProvider _sp) {
		sp = _sp;
	}
	
	
	public void run() {
		String percentageS;
		String displayed;
		int percentageI;
		long maxMemory, allocatedMemory;
		Runtime runtime = Runtime.getRuntime();
		String steps = "";
		
		int iteration = 0;
		
		while(go == true) {
			waitForMonitor();
			
			if (!go) {
				break;
			}
			
			if( sp != null) {
				steps = sp.getCurrentStep() + "/" + sp.getMaxStep() + " ";
			}
			
			percentageI = pp.getPercentage();
			if (percentageI > 100) {
				percentageI = 100;
			}
			percentageS = "" + percentageI;
			if (percentageI<10) {
				percentageS  = " " + percentageS;
			}
			if (percentageI<100) {
				percentageS = " " + percentageS;
			}
			
			
			if (progressionIcons != null) {
				percentageS = progressionIcons[iteration%progressionIcons.length] + "  " + steps + percentageS;
				
			} else {
					percentageS = steps + percentageS;
			}
			
			maxMemory = runtime.maxMemory()/1048576;
			allocatedMemory = runtime.totalMemory()/10488576;
			//freeMemory = runtime.freeMemory()/1024;
			
			displayed =  percentageS + "% " + pp.getComputation() + " | " + allocatedMemory + "MB/" + maxMemory + "MB                                                                          ";
			displayed = displayed.substring(0, 100);
			displayed += "\r";
			
			System.out.print(displayed);
			
			iteration++;
			
			try {
				Thread.currentThread().sleep(250);
			} catch (InterruptedException ie) {
			}
		}
		
		TraceManager.addDev("Progression displayer: All done");
		
		// Then extract the number of steps
	}
	
	
	
}
