/**Copyright or (C) or Copr. GET / ENST, Telecom-Paris, Ludovic Apvrille

   ludovic.apvrille AT enst.fr

   This software is a computer program whose purpose is to allow the
   edition of TURTLE analysis, design and deployment diagrams, to
   allow the generation of RT-LOTOS or Java code from this diagram,
   and at last to allow the analysis of formal validation traces
   obtained from external tools, e.g. RTL from LAAS-CNRS and CADP
   from INRIA Rhone-Alpes.

   This software is governed by the CeCILL  license under French law and
   abiding by the rules of distribution of free software.  You can  use,
   modify and/ or redistribute the software under the terms of the CeCILL
   license as circulated by CEA, CNRS and INRIA at the following URL
   "http://www.cecill.info".

   As a counterpart to the access to the source code and  rights to copy,
   modify and redistribute granted by the license, users are provided only
   with a limited warranty  and the software's author,  the holder of the
   economic rights,  and the successive licensors  have only  limited
   liability.

   In this respect, the user's attention is drawn to the risks associated
   with loading,  using,  modifying and/or developing or reproducing the
   software by the user in light of its specific status of free software,
   that may mean  that it is complicated to manipulate,  and  that  also
   therefore means  that it is reserved for developers  and  experienced
   professionals having in-depth computer knowledge. Users are therefore
   encouraged to load and test the software's suitability as regards their
   requirements in conditions enabling the security of their systems and/or
   data to be ensured and,  more generally, to use and operate it in the
   same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL license and that you accept its terms.

   /**
   * Class AlligatorMalwareScore
   * Creation: 23/07/2012
   * version 1.0 23/07/2012
   * @author Ludovic APVRILLE
   * @see
   */

package analyzer;

import myutil.*;


public class AlligatorMalwareScore extends AlligatorScore  {

    public AlligatorMalwareScore(AlligatorInfoScore _ais, double _intermediateScore) {
        super(_ais, _intermediateScore);
    }

    public void computeComplexExpression() throws AlligatorSyntaxException {
        if (ais.malwareComplexExpression != null) {
            evaluateExpression(ais.malwareComplexExpression);
        }
    }

    public void computeMaxScore() {
        scoreComputed = true;
        lastScore = intermediateScore * ais.malwareMaxMultiplier;
    }

    public void computeAverageScore() {
        scoreComputed = true;
        lastScore = intermediateScore * (ais.malwareMaxMultiplier + ais.malwareMinMultiplier)/2;
    }

    public void computeAllPossibleScores() {

        if (ais.malwareMaxMultiplier == ais.malwareMinMultiplier) {
            allPossibleScores = new double[1];
            allPossibleScores[0] = ais.malwareMaxMultiplier;
	    nbOfScores = 1;
            if (allPossibleScores[0] == Double.POSITIVE_INFINITY) {
                TraceManager.addDev("Infinite score in malware");
                //System.exit(0);
            }
            return;
        }

        if (ais.malwareStep == 0) {
            allPossibleScores = new double[2];
            allPossibleScores[0] = ais.malwareMinMultiplier;
            allPossibleScores[1] = ais.malwareMaxMultiplier;
	    nbOfScores = 2;
            if (allPossibleScores[0] == Double.POSITIVE_INFINITY) {
                TraceManager.addDev("Infinite score in malware 0");
                System.exit(0);
            }
            if (allPossibleScores[1] == Double.POSITIVE_INFINITY) {
                TraceManager.addDev("Infinite score in malware 1");
                //System.exit(0);
            }

            return;
        }

        nbOfScores = (int)((ais.malwareMaxMultiplier - ais.malwareMinMultiplier) / ais.malwareStep);
        nbOfScores++;
	nbOfScores = Math.max(1, nbOfScores);
	nbOfScores = Math.min(nbOfScores, MAX_NB_OF_STEPS);

        if (nbOfScores < MAX_NB_OF_SCORES) {
            allPossibleScores = new double[nbOfScores];

            //int cpt = 0;
            for(int i=0; i<nbOfScores; i++) {
                allPossibleScores[i] = intermediateScore * (ais.malwareMinMultiplier + i*ais.malwareStep);
                if (allPossibleScores[i] == Double.POSITIVE_INFINITY) {
                    TraceManager.addDev("Infinite score in malware i=" + i + " intermediate=" + intermediateScore +
                                        "mult=" + ais.malwareMinMultiplier + "step=" + ais.malwareStep);
                    //System.exit(0);
                }
            }

        } else {
            //TraceManager.addDev("Not creating pre-computed values - malware");  
        }
    }

    public double getScore(int index) {
	//TraceManager.addDev("Index=" + index + " size=" + allPossibleScores.length);
        if (allPossibleScores != null) {
            return allPossibleScores[index];
        } else {
            return intermediateScore * (ais.malwareMinMultiplier + index*ais.malwareStep);
        }

    }

} // Class
